/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 1023 $
 * $Date: 2020-06-04 10:06:55 +0200 (Thu, 04 Jun 2020) $
 */
package at.lcm.bu21.general.dtypes.unit;

import java.util.List;

import at.lcm.bu21.general.dtypes.TBoolean;
import at.lcm.bu21.general.dtypes.TData;
import at.lcm.bu21.general.dtypes.TNull;
import at.lcm.bu21.general.dtypes.TNumeric;
import at.lcm.bu21.general.dtypes.TPeriodic;
import at.lcm.bu21.general.dtypes.TString;

/**
 * Convert unit to another unit
 *
 * @author Werner Koppelsttter
 */
public class UnitConversion {

	/** Unit-string of origin */
	String originString;

	/** List of origin unit */
	List<Unit> origin = null;

	/** Unit string of target */
	String targetString;

	/** List of target unit */
	List<Unit> target = null;

	/** Is conversion possible */
	boolean possible = true;

	/* Is one side an any unit */
	boolean any = false;

	/** Gain of origin unit */
	double originGain = 1.0;

	/** Offset of origin unit */
	double originOffset = 0.0;

	/** Gain of target unit */
	double targetGain = 1.0;

	/** Offset of tartget unit */
	double targetOffset = 0.0;

	/**
	 * Create unit converter
	 * 
	 * @param origin
	 *            String with origin unit
	 * @param target
	 *            String with target unit
	 */
	public UnitConversion(String origin, String target) {
		boolean eq;
		Unit u1, u2;

		this.originString = origin;
		this.targetString = target;

		if (((this.originString != null) && !this.originString.isEmpty())
				|| ((this.targetString != null) && !this.targetString.isEmpty())) {

			try {
				this.origin = new Parser(this.originString).getSyntax().getUnits();
			} catch (Exception e) {
				System.err.println("Evaluate origin unit: " + e.getMessage());
				e.printStackTrace();
			}
			try {
				this.target = new Parser(this.targetString).getSyntax().getUnits();
			} catch (Exception e) {
				System.err.println("Evaluate target unit: " + e.getMessage());
				e.printStackTrace();
			}

			// is possible
			this.possible = this.isPossible();

			// get values for conversion
			if (this.possible && !this.any) {
				// check equal units (no conversion)
				eq = this.origin.size() == this.target.size();
				for (int k = 0; k < this.origin.size(); k++) {
					u1 = this.origin.get(k);
					u2 = this.target.get(k);
					if (!u1.unitString.equals(u2.unitString) || (u1.exponent != u2.exponent) || (u1.gain != u2.gain)
							|| (u1.offset != u2.offset)) {
						eq = false;
						break;
					}
				}

				if (!eq) {

					// origin values
					for (Unit u : this.origin) {
						this.originGain *= u.gain;
						this.originOffset += u.offset;
					}

					// target values
					for (Unit u : this.target) {
						this.targetGain *= u.gain;
						this.targetOffset += u.offset;
					}
				}
			}
		}
	}

	/**
	 * Is conversion possible
	 * 
	 * @return Is possible
	 */
	private boolean isPossible() {
		boolean pos;
		int k1 = 0, k2 = 0;
		Unit u1, u2;

		this.any = (this.origin != null) && (this.origin.size() == 1) && this.origin.get(0).unitString.equals("?");
		this.any |= (this.target != null) && (this.target.size() == 1) && this.target.get(0).unitString.equals("?");
		pos = this.any;

		if (!this.any && (this.origin != null) && (this.target != null)) {
			while ((k1 < this.origin.size()) && this.origin.get(k1).unitString.equals("1")) {
				k1++;
			}
			while ((k2 < this.target.size()) && this.target.get(k2).unitString.equals("1")) {
				k2++;
			}

			pos = ((this.origin.size() - k1 - 1) - (this.target.size() - k2 - 1)) == 0;
			while ((k1 < this.origin.size()) && (k2 < this.target.size())) {
				u1 = this.origin.get(k1);
				u2 = this.target.get(k2);
				pos = u1.unitString.equals(u2.unitString) && (u1.exponent == u2.exponent);
				if (pos) {
					k1++;
					k2++;
				} else {
					break;
				}
			}
		}

		return pos;
	}

	/**
	 * Convert value from origin to target unit
	 * 
	 * @param val
	 *            Value to convert (based on origin)
	 * 
	 * @return Converted value. If conversion it not possible NaN is retruned.
	 * @throws ConversionException
	 */
	public double convert(double val) throws ConversionException {
		if (this.possible) {
			if (!this.any) {
				val *= this.originGain;
				val += this.originOffset;
				val -= this.targetOffset;
				val /= this.targetGain;
			}
		} else {
			throw new ConversionException(
					"Incompatible units (" + this.originString + " -> " + this.targetString + ")");
		}

		return val;
	}

	/**
	 * Convert data object
	 * 
	 * @param data
	 *            Data to convert
	 * @return Converted data
	 * @throws ConversionException
	 */
	public TData convert(TData data) throws ConversionException {
		if (!this.any) {
			if (data instanceof TNumeric) {
				data = this.convert((TNumeric) data);
			} else if (data instanceof TPeriodic) {
				data = this.convert((TPeriodic) data);
			} else if (data instanceof TNull) {
			} else if (data instanceof TBoolean) {
			} else if (data instanceof TString) {
			} else {
				throw new ConversionException("Unit conversion not possible. Incompatible data-type. ("
						+ data.getClass().getSimpleName() + ")");
			}
		}

		return data;
	}

	/**
	 * Convert numeric data
	 * 
	 * @param num
	 *            Numeric to convert
	 * @return Converted numeric
	 * @throws ConversionException
	 */
	public TNumeric convert(TNumeric num) throws ConversionException {
		if (this.possible) {
			if (!this.any) {
				num = num.scale(this.originGain).offset(this.originOffset);
				num = num.offset(-this.targetOffset).scale(1.0 / this.targetGain);
			}
		} else {
			throw new ConversionException(
					"Incompatible units (" + this.originString + " -> " + this.targetString + ")");
		}

		return num;
	}

	/**
	 * Convert periodic data
	 * 
	 * @param per
	 *            Periodic to convert
	 * @return Converted periodic
	 * @throws ConversionException
	 */
	public TPeriodic convert(TPeriodic per) throws ConversionException {
		if (this.possible) {
			if (!this.any) {
				per = per.scale(this.originGain).offset(this.originOffset);
				per = per.offset(-this.targetOffset).scale(1.0 / this.targetGain);
			}
		} else {
			throw new ConversionException(
					"Incompatible units (" + this.originString + " -> " + this.targetString + ")");
		}

		return per;
	}

}
