/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 997 $
 * $Date: 2020-03-19 13:49:34 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.general.math.delaunay;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.Iterator;

/**
 * @author gweidenholzer
 * 
 */
public class Triangulate {

	static double EPSILON = 1e-6;

	/**
	 * @param points
	 * @return triangles
	 */
	public static ArrayList<Triangle> triangulate(ArrayList<Point> points) {

		// sort points in x-direction
		Collections.sort(points, new Comparator<Point>() {
			@Override
			public int compare(Point p1, Point p2) {
				if (p1.x < p2.x) {
					return -1;
				} else if (p1.x > p2.x) {
					return 1;
				} else {
					return 0;
				}
			}
		});

		double xmin = points.get(0).x;
		double ymin = points.get(0).y;
		double xmax = xmin;
		double ymax = ymin;

		Iterator<Point> pIter = points.iterator();
		while (pIter.hasNext()) {
			Point p = pIter.next();
			if (p.x < xmin) {
				xmin = p.x;
			}
			if (p.x > xmax) {
				xmax = p.x;
			}
			if (p.y < ymin) {
				ymin = p.y;
			}
			if (p.y > ymax) {
				ymax = p.y;
			}
		}

		double dx = xmax - xmin;
		double dy = ymax - ymin;
		double dmax = (dx > dy) ? dx : dy;
		double xmid = (xmax + xmin) / 2.0;
		double ymid = (ymax + ymin) / 2.0;

		ArrayList<Triangle> triangles = new ArrayList<Triangle>();
		HashSet<Triangle> complete = new HashSet<Triangle>();
		Triangle boundingTriangle = new Triangle();
		boundingTriangle.a = new Point(xmid - (2.0 * dmax), ymid - dmax, 0.0);
		boundingTriangle.b = new Point(xmid, ymid + (2.0 * dmax), 0.0);
		boundingTriangle.c = new Point(xmid + (2.0 * dmax), ymid - dmax, 0.0);
		triangles.add(boundingTriangle);

		// add points
		ArrayList<Edge> edges = new ArrayList<Edge>();
		pIter = points.iterator();
		while (pIter.hasNext()) {

			Point p = pIter.next();

			edges.clear();

			Point circle = new Point();

			for (int j = triangles.size() - 1; j >= 0; j--) {

				Triangle t = triangles.get(j);
				if (complete.contains(t)) {
					continue;
				}

				boolean inside = Triangulate.circumCircle(p, t, circle);

				if ((circle.x + circle.val) < p.x) {
					complete.add(t);
				}
				if (inside) {
					edges.add(new Edge(t.a, t.b));
					edges.add(new Edge(t.b, t.c));
					edges.add(new Edge(t.c, t.a));
					triangles.remove(j);
				}

			}

			for (int j = 0; j < (edges.size() - 1); j++) {
				Edge e1 = edges.get(j);
				for (int k = j + 1; k < edges.size(); k++) {
					Edge e2 = edges.get(k);
					if ((e1.start == e2.end) && (e1.end == e2.start)) {
						e1.start = null;
						e1.end = null;
						e2.start = null;
						e2.end = null;
					}
					if ((e1.start == e2.start) && (e1.end == e2.end)) {
						e1.start = null;
						e1.end = null;
						e2.start = null;
						e2.end = null;
					}
				}
			}

			for (int j = 0; j < edges.size(); j++) {
				Edge e = edges.get(j);
				if ((e.start == null) || (e.end == null)) {
					continue;
				}
				triangles.add(new Triangle(e.start, e.end, p));
			}

		}

		// Remove edges of bounding triangle
		for (int i = triangles.size() - 1; i >= 0; i--) {
			Triangle t = triangles.get(i);
			if (t.touches(boundingTriangle)) {
				triangles.remove(i);
			}
		}

		return triangles;
	}

	private static boolean circumCircle(Point p, Triangle tri, Point circle) {

		double m1, m2, mx1, mx2, my1, my2;
		double dx, dy, rsqr, drsqr;

		/* Check for coincident points */
		if ((Math.abs(tri.a.y - tri.b.y) < Triangulate.EPSILON)
				&& (Math.abs(tri.b.y - tri.c.y) < Triangulate.EPSILON)) {
			return false;
		}

		if (Math.abs(tri.b.y - tri.a.y) < Triangulate.EPSILON) {
			m2 = -(tri.c.x - tri.b.x) / (tri.c.y - tri.b.y);
			mx2 = (tri.b.x + tri.c.x) / 2.0;
			my2 = (tri.b.y + tri.c.y) / 2.0;
			circle.x = (tri.b.x + tri.a.x) / 2.0;
			circle.y = (m2 * (circle.x - mx2)) + my2;
		} else if (Math.abs(tri.c.y - tri.b.y) < Triangulate.EPSILON) {
			m1 = -(tri.b.x - tri.a.x) / (tri.b.y - tri.a.y);
			mx1 = (tri.a.x + tri.b.x) / 2.0;
			my1 = (tri.a.y + tri.b.y) / 2.0;
			circle.x = (tri.c.x + tri.b.x) / 2.0;
			circle.y = (m1 * (circle.x - mx1)) + my1;
		} else {
			m1 = -(tri.b.x - tri.a.x) / (tri.b.y - tri.a.y);
			m2 = -(tri.c.x - tri.b.x) / (tri.c.y - tri.b.y);
			mx1 = (tri.a.x + tri.b.x) / 2.0;
			mx2 = (tri.b.x + tri.c.x) / 2.0;
			my1 = (tri.a.y + tri.b.y) / 2.0;
			my2 = (tri.b.y + tri.c.y) / 2.0;
			circle.x = ((((m1 * mx1) - (m2 * mx2)) + my2) - my1) / (m1 - m2);
			circle.y = (m1 * (circle.x - mx1)) + my1;
		}

		dx = tri.b.x - circle.x;
		dy = tri.b.y - circle.y;
		rsqr = (dx * dx) + (dy * dy);
		circle.val = Math.sqrt(rsqr);

		dx = p.x - circle.x;
		dy = p.y - circle.y;
		drsqr = (dx * dx) + (dy * dy);

		return drsqr <= rsqr;
	}

}
