/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 997 $
 * $Date: 2020-03-19 13:49:34 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.general.plot;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.Toolkit;
import java.awt.datatransfer.Clipboard;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.image.BufferedImage;
import java.awt.print.PageFormat;
import java.awt.print.Printable;
import java.awt.print.PrinterException;
import java.awt.print.PrinterJob;
import java.io.File;
import java.io.IOException;
import java.util.LinkedList;

import javax.imageio.ImageIO;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JSeparator;
import javax.swing.WindowConstants;
import javax.swing.event.MenuEvent;
import javax.swing.event.MenuListener;
import javax.swing.filechooser.FileNameExtensionFilter;

/**
 * Basic class of plots
 * 
 * @author Christoph Halbmayr
 * @author Werner Koppelsttter
 */
public abstract class Plot implements Printable {

	/** Default order of color */
	private static final Color[] DEFAULT_COLOR_ORDER = new Color[] { Color.blue,
			Color.green.darker(), Color.red, Color.cyan, Color.magenta,
			Color.yellow.darker(), Color.black };

	/** Default style of plot */
	private Style defaultStyle = new Style(0.7f, Color.blue, LineStyle.SOLID,
			true, Marker.CIRLE, 6, false, false, false);

	/** Style of selected points */
	private Style defaultSelectStyle = new Style(0.7f, Color.RED,
			LineStyle.SOLID, false, Marker.CIRLE, 10, true, true, false);

	/** Order of userd colors */
	private Color[] colorOrder = Plot.DEFAULT_COLOR_ORDER;

	/** Order of used markers */
	private Marker[] markerOrder = Marker.values();

	/** List of signals */
	private LinkedList<Signal> signals = new LinkedList<Signal>();

	/** Plot title */
	private String title = null;

	/** Plot JPanel for save */
	private JPanel panel;

	/** Plot JPanel with plot */
	private JPanelRect plotPanel;

	/** Object of linked-plots */
	private LinkPlot link = null;

	/** hold flag */
	private HOLD hold = HOLD.ALL;

	/** x Axis Label */
	private String xLabel = null;

	/** y Axis Label */
	private String yLabel = null;

	/** z Axis Label */
	private String zLabel = null;

	/** Parent sub-plot */
	private SubPlot parent = null;

	/** Popup menu */
	protected JMenu popup = null;

	/** Popup menu entry for suppress selected values */
	private JMenuItem mnuSuppressSelect = null;

	/** Popup menu entry for suppress non selected values */
	private JMenuItem mnuSuppressNonSelect = null;

	/** Popup menu entry for clear suppressed values */
	private JMenuItem mnuSuppressClear = null;

	/** Standard direction of colorbar */
	private boolean colorDirection = true;

	/** Defined color-levels of colorbar */
	private float[] colorLevels = null;

	/** Is signal selectable */
	private boolean selectable = true;

	/**
	 * Is eps-formate possible
	 * 
	 * @return eps possible
	 */
	protected abstract boolean isVectorGraphicPossible();

	/**
	 * Panel to draw mouse-selection rectangle
	 * 
	 * @author Werner Koppelsttter
	 */
	public class JPanelRect extends JPanel {

		/** Serialization version number */
		private static final long serialVersionUID = 1L;

		/** Rectange to draw selection */
		public Rectangle rect = null;

		@Override
		public void paint(Graphics g) {
			super.paint(g);
			if (this.rect != null) {
				g.drawRect(this.rect.x, this.rect.y, this.rect.width,
						this.rect.height);
			}
		}
	}

	/**
	 * Standard constructor
	 */
	protected Plot() {
		this.plotPanel = new JPanelRect();
		this.plotPanel.setLayout(new BorderLayout());
		this.plotPanel.setOpaque(true);
		this.plotPanel.setFocusable(true);
		this.plotPanel.setBackground(Color.WHITE);

		this.panel = new JPanel();
		this.panel.setLayout(new BorderLayout());
		this.panel.add(this.plotPanel, BorderLayout.CENTER);
		this.panel.setOpaque(true);
		this.panel.setBackground(Color.WHITE);

		this.link = new LinkPlot(this);

		this.popup = new JMenu("Popup");
		this.addPopupMenu(this.popup);
	}

	/**
	 * @return the panel
	 */
	public JPanel getPanel() {
		return this.panel;
	}

	/**
	 * @return the panel
	 */
	protected JPanelRect getPlotPanel() {
		return this.plotPanel;
	}

	/**
	 * returns the chart title
	 * 
	 * @return title chart title
	 */
	public String getTitle() {
		return this.title;
	}

	/**
	 * Sets the chart title
	 * 
	 * @param title
	 *            chart title
	 */
	public void setTitle(String title) {
		this.title = title;
		this.redraw();
	}

	/**
	 * Resets the whole panel
	 */
	public abstract void reset();

	/**
	 * returns x-Axis Lable
	 * 
	 * @return xLabel x-Axis Lable
	 */
	public String getXLabel() {
		return this.xLabel;
	}

	/**
	 * sets the X Label
	 * 
	 * @param label
	 *            X Label
	 */
	public void setXLabel(String label) {
		if (label != null) {
			this.xLabel = label;
		}
		this.redraw();
	}

	/**
	 * returns the y Axis Label
	 * 
	 * @return yAxisLabel String
	 */
	public String getYLabel() {
		return this.yLabel;
	}

	/**
	 * sets the Y Label
	 * 
	 * @param label
	 *            Y Label
	 */
	public void setYLabel(String label) {
		if (label != null) {
			this.yLabel = label;
		}
		this.redraw();
	}

	/**
	 * returns the z Axis Label
	 * 
	 * @return zAxisLabel String
	 */
	public String getZLabel() {
		return this.zLabel;
	}

	/**
	 * sets the Z Label
	 * 
	 * @param label
	 *            Z Label
	 */
	public void setZLabel(String label) {
		if (label != null) {
			this.zLabel = label;
		}
		this.redraw();
	}

	/**
	 * Redraw plot
	 */
	abstract void redraw();

	/**
	 * Repaints the whole panel
	 */
	abstract void repaint();

	/**
	 * Exports the whole Plot to a file depending on the file Type
	 * 
	 * @param f
	 *            Destination file
	 * @throws IOException
	 *             File exception
	 */
	public void save(String f) throws IOException {
		this.save(new File(f));
	}

	/**
	 * Exports the whole Plot to a file depending on the file Type
	 * 
	 * @param f
	 *            Destination file
	 * @throws IOException
	 *             File exception
	 */
	public void save(File f) throws IOException {
		JPanel plot;

		plot = this.getPlotPanel();
		this.save(f, plot.getWidth(), plot.getHeight());
	}

	/**
	 * Exports the whole Plot to a file depending on the file Type
	 * 
	 * @param f
	 *            destination File
	 * @param width
	 *            Width of picture
	 * @param height
	 *            Height of picture
	 * @throws IOException
	 *             File exception
	 */
	public void save(String f, int width, int height) throws IOException {
		this.save(new File(f), width, height);
	}

	/**
	 * Exports the whole Plot to a file depending on the file Type
	 * 
	 * @param f
	 *            destination File
	 * @param width
	 *            Width of picture
	 * @param height
	 *            Height of picture
	 * @throws IOException
	 *             File exception
	 */
	public void save(File f, int width, int height) throws IOException {
		String fileType = "";
		int index;

		index = f.getName().lastIndexOf('.');
		if (index > 0) {
			fileType = f.getName().substring(index + 1);
		}

		if (fileType.equalsIgnoreCase("jpg")
				|| fileType.equalsIgnoreCase("jpeg")
				|| fileType.equalsIgnoreCase("png")) {
			ImageIO.write(this.getImage(width, height), fileType, f);
		} else {
			throw new IOException("Undefined file-tpye");
		}

	}

	/**
	 * Get BufferdImage from plot
	 * 
	 * @return BufferdImage
	 */
	protected BufferedImage getBufferedImage() {
		BufferedImage im;

		im = new BufferedImage(this.getPlotPanel().getWidth(),
				this.getPlotPanel().getHeight(), BufferedImage.TYPE_INT_RGB);
		this.getPlotPanel().paint(im.createGraphics());

		return im;
	}

	/**
	 * Get image from plot
	 * 
	 * @return Image
	 */
	public BufferedImage getImage() {
		// Dimension dim;
		//
		// dim = this.getPanel().getSize();
		//
		// return this.getImage(dim.width, dim.height);
		return this.getImage(500, 400);
	}

	/**
	 * Get image from plot
	 * 
	 * @param width
	 *            Width of image
	 * @param height
	 *            Height of image
	 * @return Image
	 */
	public BufferedImage getImage(int width, int height) {
		BufferedImage im;
		JFrame frame;

		// Create frame for display
		frame = new JFrame();
		frame.add(this.getPlotPanel());
		this.getPlotPanel().setPreferredSize(new Dimension(width, height));
		frame.pack();

		// create image
		if (this instanceof Plot3D) {
			((Plot3D) this).getChart().render();
		}
		im = this.getBufferedImage();

		// restore structure
		frame.dispose();
		this.getPanel().add(this.getPlotPanel(), BorderLayout.CENTER);
		this.getPlotPanel().setSize(this.getPanel().getSize());
		this.getPlotPanel().validate();

		// this.getPlotPanel().setDoubleBuffered(false);
		//
		// Dimension dim = this.getPlotPanel().getSize(); // get size of
		// document
		//
		// im = new BufferedImage(this.getPlotPanel().getWidth(), this
		// .getPlotPanel().getHeight(), BufferedImage.TYPE_INT_RGB);
		// Graphics2D g2 = im.createGraphics();
		//
		// // positioning Graphic to line up with beginning of
		// // print-imageable
		// // region
		// // g2.translate(pf.getImageableX(), pf.getImageableY());
		//
		// // positioning Graphic to line up with beginning of next
		// // page to
		// // print
		// // g2.translate(0f, -pageIndex * pageHeight);
		//
		// // positioning the page so the width fits...
		// g2.scale(width / dim.width, height / dim.height);
		//
		// this.getPlotPanel().paintAll(g2);
		//
		// this.getPlotPanel().setDoubleBuffered(true);

		return im;
	}

	/**
	 * Is plot linked to other plots
	 * 
	 * @return Is plot linked
	 */
	public boolean isLink() {
		return this.link != null;
	}

	/**
	 * Get link-object of plot
	 * 
	 * @return Link-object
	 */
	LinkPlot getLink() {
		return this.link;
	}

	/**
	 * Set link-object to plot
	 * 
	 * @param link
	 *            Link-object
	 */
	void setLink(LinkPlot link) {
		this.link = link;
	}

	/**
	 * Adds a XYSeries to the XYSeriesCollection in consideration of hold
	 * 
	 * @param signal
	 *            XYSeries
	 */
	protected void addSignal(Signal signal) {
		Style s;

		s = this.defaultStyle.clone();
		switch (this.hold) {
		case OFF:
			this.signals.clear();
			break;
		case ON:
			break;
		case ALL:
			s.setColor(this.colorOrder[this.getNumSignals()
					% this.colorOrder.length]);
			s.setMarker(this.markerOrder[this.getNumSignals()
					% this.markerOrder.length]);
			break;
		case FADING:
			int len = this.getNumSignals();
			for (int k = 0; k < len; k++) {
				Style style = this.getSignal(k).getStyle();
				Color col;
				float[] hsb;
				float sat;

				sat = (float) (Math
						.tan(0.5 + ((((double) k) / len) * (1.35 - 0.5)))
						/ Math.tan(1.35));
				col = style.getColor();
				hsb = Color.RGBtoHSB(col.getRed(), col.getGreen(),
						col.getBlue(), null);
				style.setColor(Color.getHSBColor(hsb[0], sat, hsb[2]));
			}
			break;
		}
		signal.setStyle(s);
		signal.setSelectStyle(this.getDefaultSelectStyle().clone());
		signal.setEnable(this.selectable);
		this.signals.add(signal);

		this.reset();
	}

	/**
	 * Get number of signals
	 * 
	 * @return Number of signals
	 */
	public int getNumSignals() {
		return this.signals.size();
	}

	/**
	 * Get signal of index
	 * 
	 * @param index
	 *            Index of signal
	 * @return Signal
	 */
	public Signal getSignal(int index) {
		Signal sig = null;

		if ((0 <= index) && (index < this.signals.size())) {
			sig = this.signals.get(index);
		}

		return sig;
	}

	/**
	 * Remove signal of index
	 * 
	 * @param index
	 *            Index of signal
	 * @return Signal
	 */
	public Signal removeSignal(int index) {
		Signal sig = null;

		if ((0 <= index) && (index < this.signals.size())) {
			sig = this.signals.remove(index);
		}
		this.redraw();

		return sig;
	}

	/**
	 * Remove all signals
	 */
	public void clearSignals() {
		this.signals.clear();
		this.redraw();
	}

	/**
	 * Get default style
	 * 
	 * @return Default style
	 */
	public Style getDefaultStyle() {
		return this.defaultStyle;
	}

	/**
	 * Get style of selection
	 * 
	 * @return Style of selection
	 */
	public Style getDefaultSelectStyle() {
		return this.defaultSelectStyle;
	}

	/**
	 * Suppress value of signal
	 * 
	 * @param signal
	 *            Index of signal
	 * @param index
	 *            Index of value
	 * @param suppress
	 *            Suppress state
	 */
	void suppress(int signal, int index, boolean suppress) {
		Signal sig;

		sig = this.getSignal(signal);
		if (sig != null) {
			sig.suppress(index, suppress);
		}

	}

	/**
	 * Suppress selected/non-selected values of signal
	 * 
	 * @param signal
	 *            Index of signal
	 * @param inverse
	 *            Selected or non selected values
	 */
	void suppressSelected(int signal, boolean inverse) {
		Signal sig;

		sig = this.getSignal(signal);
		if (sig != null) {
			sig.suppressSelected(inverse);
		}

	}

	/**
	 * Clear all suppressed
	 */
	void clearSuppress() {
		for (int k = 0; k < this.getNumSignals(); k++) {
			this.getSignal(k).clearSuppress();
		}
	}

	/**
	 * Select point
	 * 
	 * @param signal
	 *            Index of signal
	 * @param index
	 *            Index of point
	 * @param select
	 *            Select/Deselect
	 */
	void select(int signal, int index, boolean select) {
		Signal sig;

		sig = this.getSignal(signal);
		if (sig != null) {
			sig.select(index, select);
		}
	}

	/**
	 * Clear all selections
	 */
	void clearSelect() {
		for (int k = 0; k < this.getNumSignals(); k++) {
			this.getSignal(k).clearSelect();
		}
	}

	/**
	 * returns the hold
	 * 
	 * @return the hold
	 */
	public HOLD getHold() {
		return this.hold;
	}

	/**
	 * sets the hold
	 * 
	 * @param h
	 *            HOLD
	 */
	public void setHold(HOLD h) {
		this.hold = h;
	}

	/**
	 * Set color-order
	 * 
	 * @param colors
	 *            List of colors
	 */
	public void setColorOrder(Color[] colors) {
		this.colorOrder = colors;
	}

	/**
	 * Set marker-order
	 * 
	 * @param markers
	 *            List of markers
	 */
	public void setMarkerOrder(Marker[] markers) {
		this.markerOrder = markers;
	}

	/**
	 * Show plot in a frame
	 * 
	 * @return Frame for plot
	 */
	public JFrame show() {
		return this.show(null);
	}

	/**
	 * Show plot in a frame
	 * 
	 * @param title
	 *            Title of frame
	 * @return Frame for plot
	 */
	public JFrame show(String title) {
		JFrame frame;

		frame = new JFrame(title);
		frame.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
		frame.setLayout(new BorderLayout());
		frame.add(this.getPanel());
		frame.pack();
		frame.setVisible(true);

		return frame;
	}

	/**
	 * Get parent of plot (SubPlot)
	 * 
	 * @return Parent SubPlot
	 */
	protected SubPlot getParent() {
		return this.parent;
	}

	/**
	 * Set parent of plot
	 * 
	 * @param parent
	 *            Parent SubPlot
	 */
	protected void setParent(SubPlot parent) {
		this.parent = parent;
		this.refreshPopupMenu(this.popup);
	}

	/**
	 * Add item of popup-menu
	 * 
	 * @param popup
	 *            Based popup-menu
	 */
	protected void addPopupMenu(JMenu popup) {
		JMenuItem copy, save, saveAdv, print, zoomReset, selectionClear;
		JMenu suppress;

		// copy to clipboard
		copy = new JMenuItem("Copy");
		popup.add(copy);
		copy.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e) {
				Clipboard systemClipboard;

				systemClipboard = Toolkit.getDefaultToolkit()
						.getSystemClipboard();
				systemClipboard.setContents(
						new PlotTransferable(Plot.this.getImage()), null);
			}
		});

		// save in file
		save = new JMenuItem("Save");
		popup.add(save);
		save.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e) {
				File file = null;

				file = Plot.this.selectFilename();
				if (file != null) {
					try {
						Plot.this.save(file);
					} catch (IOException e1) {
						e1.printStackTrace();
					}
				}
			}
		});

		// save advanced in file
		saveAdv = new JMenuItem("Save Advanced");
		popup.add(saveAdv);
		saveAdv.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e) {
				File file = null;
				JPanel panel;
				SaveSettingsDialog set;

				file = Plot.this.selectFilename();
				if (file != null) {
					panel = Plot.this.getPlotPanel();
					set = new SaveSettingsDialog(panel);
					set.setPlotSize(panel.getWidth(), panel.getHeight());

					set.setVisible(true);
					if (set.isValidPlotSize()) {

						try {
							Plot.this.save(file, set.getPlotWidth(),
									set.getPlotHeight());
						} catch (IOException e1) {
							e1.printStackTrace();
						}
					}
				}
			}
		});

		// print plot
		print = new JMenuItem("Print");
		popup.add(print);
		print.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e) {
				PrinterJob printJob;
				PageFormat defaultFormat, printFormat;

				printJob = PrinterJob.getPrinterJob();
				defaultFormat = printJob.defaultPage();
				printFormat = printJob.pageDialog(defaultFormat);
				if (printFormat != defaultFormat) {
					printJob.setPrintable(Plot.this, printFormat);
					if (printJob.printDialog()) {
						try {
							printJob.print();
						} catch (PrinterException ex) {
							ex.printStackTrace();
						}
					}
				}
			}
		});

		// Reset zoom to inital zoom
		popup.addSeparator();
		zoomReset = new JMenuItem("Reset Zoom");
		popup.add(zoomReset);
		zoomReset.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e) {
				Plot.this.reset();
			}
		});

		selectionClear = new JMenuItem("Clear Selection");
		popup.add(selectionClear);
		selectionClear.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e) {
				LinkPlot l;

				l = Plot.this.getLink();
				l.clearSelect();
				l.redraw();
			}
		});

		suppress = new JMenu("Suppress");
		popup.insert(suppress, 7);
		suppress.addMenuListener(new MenuListener() {

			@Override
			public void menuSelected(MenuEvent event) {
				boolean suppress = false;
				boolean select = false;
				Signal sig;

				for (int k = 0; k < Plot.this.getNumSignals(); k++) {
					sig = Plot.this.getSignal(k);
					suppress |= sig.isSuppressed();
					select |= sig.isSelected();
				}

				Plot.this.mnuSuppressSelect.setEnabled(select);
				Plot.this.mnuSuppressNonSelect.setEnabled(select);
				Plot.this.mnuSuppressClear.setEnabled(suppress);
			}

			@Override
			public void menuDeselected(MenuEvent arg0) {
			}

			@Override
			public void menuCanceled(MenuEvent arg0) {
			}
		});

		this.mnuSuppressSelect = new JMenuItem("Selected");
		suppress.add(this.mnuSuppressSelect);
		this.mnuSuppressSelect.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				Signal sig;

				for (int k = 0; k < Plot.this.getNumSignals(); k++) {
					sig = Plot.this.getSignal(k);
					if (sig.isSelected()) {
						Plot.this.getLink().suppressSelected(k, false);
					}
				}
				Plot.this.getLink().reset();
			}
		});

		this.mnuSuppressNonSelect = new JMenuItem("Non selected");
		suppress.add(this.mnuSuppressNonSelect);
		this.mnuSuppressNonSelect.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				Signal sig;

				for (int k = 0; k < Plot.this.getNumSignals(); k++) {
					sig = Plot.this.getSignal(k);
					if (sig.isSelected()) {
						Plot.this.getLink().suppressSelected(k, true);
					}
				}
				Plot.this.getLink().reset();
			}
		});

		this.mnuSuppressClear = new JMenuItem("Clear");
		suppress.add(this.mnuSuppressClear);
		this.mnuSuppressClear.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e) {
				Plot.this.getLink().clearSuppress();
				Plot.this.getLink().reset();
			}
		});
	}

	/**
	 * Refresh popup-menu (parent item)
	 * 
	 * @param popup
	 *            Popup-menu to refresh
	 */
	protected void refreshPopupMenu(JMenu popup) {
		JMenu parentPlot = null;
		int parent = -1, suppress = -1;
		JMenuItem tmp;

		for (int k = 0; k < popup.getItemCount(); k++) {
			tmp = popup.getItem(k);
			if (tmp != null) {
				if (tmp.getText().equals("Parent Plot")) {
					parent = k;
				} else if (tmp.getText().equals("Suppress")) {
					suppress = k;
				}
			}
		}

		// Add/Remove popup-menu of parent subplot
		if (this.getParent() != null) {
			if (parent != -1) {
				parentPlot = (JMenu) popup.getItem(parent);
				parentPlot.removeAll();
			} else {
				popup.add(new JSeparator(), suppress + 1);
				parentPlot = new JMenu("Parent Plot");
				popup.add(parentPlot, suppress + 2);
			}
			this.getParent().addPopupMenu(parentPlot);
			this.getParent().refreshPopupMenu(parentPlot);
		} else {
			if (parent != -1) {
				popup.remove(parent);
				popup.remove(parent - 1);
			}
		}
	}

	@Override
	public int print(Graphics graphics, PageFormat pageFormat, int pageNum) {
		int response;
		Graphics2D g2;
		Dimension dim;
		double scale;

		this.getPlotPanel().setDoubleBuffered(false);

		// calculate scale-factor to print image on page
		dim = this.getPlotPanel().getSize();
		if ((dim.getWidth() / dim.getHeight()) > (pageFormat.getImageableWidth()
				/ pageFormat.getImageableHeight())) {
			scale = pageFormat.getImageableWidth() / dim.getWidth();
		} else {
			scale = pageFormat.getImageableHeight() / dim.getHeight();
		}

		if (pageNum != 0) {
			response = Printable.NO_SUCH_PAGE;
		} else {
			g2 = (Graphics2D) graphics;
			// positioning image to the print area
			g2.translate(pageFormat.getImageableX(),
					pageFormat.getImageableY());
			// scale image to page
			g2.scale(scale, scale);
			// paint image
			this.plotPanel.paintAll(g2);
			this.getPlotPanel().paintAll(g2);

			response = Printable.PAGE_EXISTS;
		}

		this.getPlotPanel().setDoubleBuffered(true);

		return response;

	}

	/**
	 * Select file name
	 * 
	 * @return File-name
	 */
	private File selectFilename() {
		File file = null;
		JFileChooser fileChooser;
		boolean epsPossible;

		fileChooser = new JFileChooser();
		fileChooser.setAcceptAllFileFilterUsed(false);
		epsPossible = Plot.this.isVectorGraphicPossible();
		if (epsPossible) {
			fileChooser.addChoosableFileFilter(new FileNameExtensionFilter(
					"All Image Formats (png, jpeg, jpg, eps, svg, pdf)", "png",
					"jpeg", "jpg", "eps", "svg", "pdf"));
		} else {
			fileChooser.addChoosableFileFilter(new FileNameExtensionFilter(
					"All Image Formats (png, jpeg, jpg)", "png", "jpeg",
					"jpg"));

		}
		fileChooser.addChoosableFileFilter(new FileNameExtensionFilter(
				"Portable Network Graphics (png)", "png"));
		fileChooser.addChoosableFileFilter(new FileNameExtensionFilter(
				"JPEG File Interchange Format (jpeg)", "jpeg", "jpg"));
		if (epsPossible) {
			fileChooser.addChoosableFileFilter(new FileNameExtensionFilter(
					"Encapsulated PostScript (eps)", "eps"));
			fileChooser.addChoosableFileFilter(new FileNameExtensionFilter(
					"Scaleable Vector Graphics (svg)", "svg"));
			fileChooser.addChoosableFileFilter(new FileNameExtensionFilter(
					"Portable Document Format (pdf)", "pdf"));
		}

		if (fileChooser.showSaveDialog(
				Plot.this.getPlotPanel()) == JFileChooser.APPROVE_OPTION) {
			file = fileChooser.getSelectedFile();
			if (!fileChooser.getChoosableFileFilters()[0].accept(file)) {
				file = new File(file.getPath() + "."
						+ ((FileNameExtensionFilter) fileChooser
								.getFileFilter()).getExtensions()[0]);
			}

		}
		return file;
	}

	/**
	 * Get color-direction of colorbar(standard or inverse)
	 * 
	 * @return Is standard color-direction
	 */
	public boolean getColorDirection() {
		return this.colorDirection;
	}

	/**
	 * Set color-direction of colorbar(standard of inverse)
	 * 
	 * @param colorDirection
	 *            Is standard color-direction
	 */
	public void setColorDirection(boolean colorDirection) {
		this.colorDirection = colorDirection;
		this.redraw();
	}

	/**
	 * List color-levels of colorbar
	 * 
	 * @return List of color-levels
	 */
	public float[] getColorLevels() {
		return this.colorLevels;
	}

	/**
	 * Set color-levels of colorbar
	 * 
	 * The ColorPaintScale is a prograssive transition from blue over green to
	 * red. The mix between these 3 colors is expressed by the definition of 2
	 * functions.
	 * 
	 * <code><br>
	 *           blue   green  red         <br>
	 *  1.0 |   /------X------X------\   | <br>
	 *      |  /:     /:\    /:\     :\  | <br>
	 *      | / :    / : \  / : \    : \ | <br>
	 *  0.5 |/  :   /  :  \/  :  \   :  \| <br>
	 *      |   :  /   :  /\  :   \  :   | <br>
	 *      |   : /    : /  \ :    \ :   | <br>
	 *      |   :/     :/    \:     \:   | <br>
	 *  0.0 |-------------|--------------| <br>
	 *      0   :      : 0.5  :      :   1 <br>
	 *         L1     L2     L3     L4     <br>
	 * <br></code>
	 * 
	 * The input is normalized from 0.0 to 1.0.
	 * 
	 * Values outside of boundaries are magenta and NaN where white
	 * 
	 * @param levels
	 *            List of color-levels
	 */
	public void setColorLevels(float[] levels) {
		this.colorLevels = levels;
	}

	/**
	 * Is signal seletable
	 * 
	 * @return Is selectable
	 */
	public boolean isSelectable() {
		return this.selectable;
	}

	/**
	 * Set signals selectable
	 * 
	 * @param select
	 *            Is selectable
	 */
	public void setSelectable(boolean select) {
		this.selectable = select;
	}
}
