/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 997 $
 * $Date: 2020-03-19 13:49:34 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.general.plot;

import java.awt.BasicStroke;
import java.awt.Color;

/**
 * Styles of signals (line, marker, color, ...)
 * 
 * @author Christoph Halbmayr
 * @author Werner Koppelsttter
 */
public class Style implements Cloneable {

	/** Plot style */
	SignalStyle signalStyle = SignalStyle.Line;

	/** Line color */
	private Color lineColor;

	/** Is line visible */
	private boolean lineVisible;

	/** Line width */
	private float lineWidth;

	/** Style of line */
	private LineStyle lineStyle;

	/** Marker type */
	private Marker marker;

	/** Is marker visible */
	private boolean markerVisible;

	/** Is marker Filled */
	private boolean markerFilled;

	/** Is surface */
	private boolean surfaceVisible;

	/** Marker size */
	private int markerSize;

	/**
	 * Create a style-object with standard settings
	 */
	Style() {
		this(0.7f, null, LineStyle.SOLID, true, null, 6, false, false, false);
	}

	/**
	 * Creates a style-object
	 * 
	 * @param lineWidth
	 *            LineWidth
	 * @param lineColor
	 *            LineColor
	 * @param lineStyle
	 *            Style of line
	 * @param lineVisible
	 *            LinesVisible
	 * @param marker
	 *            Type of marker
	 * @param markerSize
	 *            Size of marker
	 * @param markerVisible
	 *            MarkersVisible
	 * @param markerFilled
	 *            MarkersFilles
	 * @param surfaceVisible
	 *            SurfaceVisible
	 */
	Style(float lineWidth, Color lineColor, LineStyle lineStyle,
			boolean lineVisible, Marker marker, int markerSize,
			boolean markerVisible, boolean markerFilled, boolean surfaceVisible) {
		this.lineWidth = lineWidth;
		this.lineColor = lineColor;
		this.lineStyle = lineStyle;
		this.lineVisible = lineVisible;
		this.marker = marker;
		this.markerSize = markerSize;
		this.markerVisible = markerVisible;
		this.markerFilled = markerFilled;
		this.surfaceVisible = surfaceVisible;
	}

	@Override
	public Style clone() {
		Style other = null;
		try {
			other = (Style) super.clone();
		} catch (CloneNotSupportedException e) {
		}
		return other;
	}

	/**
	 * Get style of signal
	 * 
	 * @return Style of signal
	 */
	public SignalStyle getSignalStyle() {
		return this.signalStyle;
	}

	/**
	 * Set style of signal
	 * 
	 * @param signalStyle
	 *            Style of signal
	 */
	public void setSignalStyle(SignalStyle signalStyle) {
		this.signalStyle = signalStyle;
	}

	/**
	 * Get color of line
	 * 
	 * @return Color of line
	 */
	public Color getColor() {
		return this.lineColor;
	}

	/**
	 * Set color of line
	 * 
	 * @param lineColor
	 *            Color of line
	 */
	public void setColor(Color lineColor) {
		this.lineColor = lineColor;
	}

	/**
	 * Get style of line
	 * 
	 * @return Style of line
	 */
	public LineStyle getLineStyle() {
		return this.lineStyle;
	}

	/**
	 * Set style of line
	 * 
	 * @param lineStyle
	 *            Style of line
	 */
	public void setLineStyle(LineStyle lineStyle) {
		this.lineStyle = lineStyle;
	}

	/**
	 * Get width of line
	 * 
	 * @return Width of line
	 */
	public float getLineWidth() {
		return this.lineWidth;
	}

	/**
	 * Set width of line
	 * 
	 * @param lineWidth
	 *            Width of line
	 */
	public void setLineWidth(float lineWidth) {
		this.lineWidth = lineWidth;
	}

	/**
	 * Get stroke of line
	 * 
	 * @return Stoke
	 */
	public BasicStroke getStroke() {
		float[] dash;

		switch (this.lineStyle) {
		case DASHED:
			dash = new float[] { 10.0f, 6.0f };
			break;
		case DOTTED:
			dash = new float[] { 2.0f, 2.0f };
			break;
		case DASHDOT:
			dash = new float[] { 14.0f, 4.0f, 2.0f, 4.0f };
			break;
		default:
			dash = null;
			break;
		}
		return new BasicStroke(this.getLineWidth(), BasicStroke.CAP_SQUARE,
				BasicStroke.JOIN_BEVEL, 0.0f, dash, 0.0f);
	}

	/**
	 * Get type of marker
	 * 
	 * @return Marker-type
	 */
	public Marker getMarker() {
		return this.marker;
	}

	/**
	 * Set type of marker
	 * 
	 * @param marker
	 *            Marker-type
	 */
	public void setMarker(Marker marker) {
		this.marker = marker;
	}

	/**
	 * Get size of marker
	 * 
	 * @return Marker-size
	 */
	public int getMarkerSize() {
		return this.markerSize;
	}

	/**
	 * Set size of marker
	 * 
	 * @param size
	 *            Marker-size
	 */
	public void setMarkerSize(int size) {
		this.markerSize = size;
	}

	/**
	 * Get visible of line
	 * 
	 * @return Visible of line
	 */
	public boolean isLineVisible() {
		return this.lineVisible;
	}

	/**
	 * Set visible of line
	 * 
	 * @param lineVisible
	 *            Visible of line
	 */
	public void setLineVisible(boolean lineVisible) {
		this.lineVisible = lineVisible;
	}

	/**
	 * Get visible of marker
	 * 
	 * @return Visible of marker
	 */
	public boolean isMarkerVisible() {
		return this.markerVisible;
	}

	/**
	 * Set visible of marker
	 * 
	 * @param markerVisible
	 *            Visible of marker
	 */
	public void setMarkerVisible(boolean markerVisible) {
		this.markerVisible = markerVisible;
	}

	/**
	 * Get fill-state of marker
	 * 
	 * @return Fill-state of marker
	 */
	public boolean isMarkerFilled() {
		return this.markerFilled;
	}

	/**
	 * Set fill-state of marker
	 * 
	 * @param markerFilled
	 *            Fill-state of marker
	 */
	public void setMarkerFilled(boolean markerFilled) {
		this.markerFilled = markerFilled;
	}

	/**
	 * Get visible of surface
	 * 
	 * @return Visible of surface
	 */
	public boolean isSurfaceVisible() {
		return this.surfaceVisible;
	}

	/**
	 * Set visible of surface
	 * 
	 * @param surfaceVisible
	 *            Visible of surface
	 */
	public void setSurfaceVisible(boolean surfaceVisible) {
		this.surfaceVisible = surfaceVisible;
	}

}
