/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 997 $
 * $Date: 2020-03-19 13:49:34 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.general.utils;

import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.geom.AffineTransform;

import javax.swing.Icon;
import javax.swing.JLabel;
import javax.swing.SwingConstants;

/**
 * A display area for a short vertical text string or an image, or both
 * 
 * @author Werner Koppelsttter
 */
public class JVLabel extends JLabel {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** Rotate size (needed for paintComponent) */
	private boolean rotateSize = true;

	/**
	 * Creates a JVLabel instance with no image and with an empty string for the
	 * title
	 */
	public JVLabel() {
		this("", null, SwingConstants.LEADING);
	}

	/**
	 * Creates a JVLabel instance with the specified image
	 * 
	 * @param image
	 *            The image to be displayed by the label
	 */
	public JVLabel(Icon image) {
		this(null, image, SwingConstants.CENTER);
	}

	/**
	 * Creates a JVLabel instance with the specified image and horizontal
	 * alignment
	 * 
	 * @param image
	 *            The image to be displayed by the label
	 * @param horizontalAlignment
	 *            One of the following constants defined in SwingConstants:
	 *            LEFT, CENTER, RIGHT, LEADING or TRAILING
	 */
	public JVLabel(Icon image, int horizontalAlignment) {
		this(null, image, horizontalAlignment);
	}

	/**
	 * Creates a JLabel instance with the specified text
	 * 
	 * @param text
	 *            The text to be displayed by the label
	 */
	public JVLabel(String text) {
		this(text, null, SwingConstants.LEADING);
	}

	/**
	 * Creates a JLabel instance with the specified text and horizontal
	 * alignment
	 * 
	 * @param text
	 *            The text to be displayed by the label
	 * @param horizontalAlignment
	 *            One of the following constants defined in SwingConstants:
	 *            LEFT, CENTER, RIGHT, LEADING or TRAILING
	 */
	public JVLabel(String text, int horizontalAlignment) {
		this(text, null, horizontalAlignment);
	}

	/**
	 * Creates a JLabel instance with the specified text, image, and horizontal
	 * alignment
	 * 
	 * @param text
	 *            The text to be displayed by the label
	 * @param icon
	 *            The image to be displayed by the label
	 * @param horizontalAlignment
	 *            One of the following constants defined in SwingConstants:
	 *            LEFT, CENTER, RIGHT, LEADING or TRAILING
	 */
	public JVLabel(String text, Icon icon, int horizontalAlignment) {
		super(text, icon, horizontalAlignment);
	}

	@Override
	public Dimension getPreferredSize() {
		Dimension size;

		size = super.getPreferredSize();
		return new Dimension(size.height, size.width);
	}

	@Override
	public Dimension getSize() {
		Dimension size;

		size = super.getSize();
		if (!this.rotateSize) {
			size = new Dimension(size.height, size.width);
		}

		return size;
	}

	@Override
	public int getHeight() {
		return this.getSize().height;
	}

	@Override
	public int getWidth() {
		return this.getSize().width;
	}

	@Override
	protected void paintComponent(Graphics g) {
		Graphics2D rg;

		// rotate and shift
		AffineTransform transform = new AffineTransform();
		transform.rotate(-Math.PI / 2.0, 0, 0);
		transform.preConcatenate(AffineTransform.getTranslateInstance(0, this.getHeight()));

		rg = (Graphics2D) g.create();
		rg.transform(transform);
		this.rotateSize = false;
		super.paintComponent(rg);
		this.rotateSize = true;
	}

}
