import datetime
from lxml import etree


class Fmi2ModelDescription:
    """
    Class for the fmiModelDescription xml-tag

    fmiVersion
    ----------
    Version of FMI (FMI 2.0 revision is defined as "2.0"; future minor revisions are denoted as 2.0.1, 2.0.2 etc).

    modelName
    ----------
    Class name of FMU, e.g. "A.B.C" (several FMU instances are possible)

    guid
    ----
    Fingerprint of xml-file content to verify that xml-file and C-functions are compatible to each other

    description (Optional)
    ----------------------
    Type: string

    author (Optional)
    -----------------
    Type: string

    version (Optional)
    ------------------
    Version of FMU, e.g., "1.4.1"

    copyright (Optional)
    --------------------
    Information on intellectual property copyright for this FMU, such as “© MyCompany 2011“

    license (Optional)
    -------------------
    Information on intellectual property licensing for this FMU,
    such as “BSD license”, "Proprietary", or "Public Domain"


    generationTool (Optional)
    -------------------------
    Type: string

    generationDateAndTime (Optional)
    --------------------------------
    Type: string

    variableNamingConvention (Optional)
    -----------------------------------
    "flat" or "structured"
    Type: string, Default: "flat"

    numberOfEventIndicators (Optional)
    ----------------------------------
    Type: uint

    """

    def __init__(self, fmi_version, model_name, guid, description, generation_tool):
        """

        Parameters
        ----------
        fmi_version
        model_name
        guid
        description
        generation_tool
        """
        self.element = etree.Element('fmiModelDescription',
                                     fmiVersion=fmi_version,
                                     modelName=model_name,
                                     guid=guid,
                                     description=description,
                                     # author="author1",
                                     # version="version1",
                                     # copyright="copyright1",
                                     # license="license1",
                                     generationTool=generation_tool,
                                     generationDateAndTime=self.timestamp,
                                     variableNamingConvention="structured",
                                     numberOfEventIndicators="0")

    @property
    def timestamp(self):
        """
        Optional date and time when the XML file was generated.
        The format is a subset of “xs:dateTime” and should be: “YYYY-MM-DDThh:mm:ssZ"
        (with one “T” between date and time; “Z” characterizes the Zulu time zone, in other words Greenwich meantime).
        [Example: "2009-12-08T14:33:22Z"].

        Returns
        -------

        """
        # return datetime.datetime.utcnow().strftime('%Y-%m-%d %H:%M:%SZ')
        return datetime.datetime.utcnow().replace(microsecond=0).isoformat() + 'Z'
