#!/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 955 $
$LastChangedDate:: 2019-06-12 15:50:17 +0200#$

This file is part of SyMSpace.

Copyright (c) 2018, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""

import numpy as np
from lcm.math.rbf import kernels
from lcm.math.rbf.core import calc_weigths, eval_z, eval_dz_dx
from lcm.utils.serialization import Serializable


def array3d_to_msgpack(array):
    """
    Create msgpack serializable representation of 3d numpy array

    Parameters
    ----------
    array: (size1, size2, size3) np.ndarray
        Source array to be serialized

    Returns
    -------
    msgpack: dict
        msgpack serializable representation of 3d numpy array

    """

    msgpack = {"rows": array.shape[0],
               "cols": array.shape[1],
               "samples": array.shape[2],
               "data": array.astype('>f8', order='C').tobytes()}

    return msgpack


class StackedMultiRBF:
    """
    Solves a stack of independent, multidimensional rbf problems and
    provide an interpolation function, therefor

    Attributes
    ----------
    xc_t : (xdim, ncenters, nsamples) ndarray
        RBF center coordinates in normalized basis
    t : (x_dim, x_dim, nsamples) ndarray of float
        Transformation matrix for input x
    kernel : lcm.rbf.kernel.AbstractKernel
        Function to calculate the RBF kernel.
        It expects an array_like argument (n_centers, n_samples) of radii
        and returns the RBF values in the same shape
    c : (zdim, ncenters, nsamples) ndarray
        RBF weights
    d : (zdim, xdim+1, nsamples) ndarray
        polynomial coefficients
    """

    @classmethod
    def fromCBOR(cls, cbor_obj):
        """
        Create a deserialized instance from a CBOR serializabled object

        ref: http://cbor.io/

        Parameters
        ----------
        cbor_obj : dict

        Returns
        -------
        StackedMultiRBF
        """
        xc = Serializable.fromCBOR(cbor_obj['xi'])
        t = Serializable.fromCBOR(cbor_obj['t'])
        c = Serializable.fromCBOR(cbor_obj['c'])
        d = Serializable.fromCBOR(cbor_obj['d'])
        kernel_id = Serializable.fromCBOR(cbor_obj['kernel'])
        kernel_params = Serializable.fromCBOR(cbor_obj['params'])
        max_dist = Serializable.fromCBOR(cbor_obj['max_dist'])
        kernel = kernels.AbstractKernel.fromID(kernel_id, kernel_params)

        return cls.fromMatrices(xc, c, d, kernel, t, max_dist=max_dist)

    @classmethod
    def fromMatrices(cls, xc, c, d, kernel, t, max_dist):
        """
        Instantiate StackedMultiRBF from pre-calculated matrices

        Parameters
        ----------
        xc : np.ndarray
        c : np.ndarray
        d : np.ndarray
        kernel : kernels.KernelTPS
        t : np.ndarray
        max_dist : float

        Returns
        -------
        StackedMultiRBF
        """
        return cls(xc, None, kernel, t, max_dist, c=c, d=d)

    def __init__(self, xc, zc, kernel=kernels.KernelTPS(), t=None, gamma=0., max_dist=0., max_cond=0., **kwargs):
        """
        xc : (xdim, ncenters, n_samples) ndarray
            RBF center coordinates
        zc : (z_dim, n_centers, n_samples) ndarray
            RBF center values.
        kernel : lcm.rbf.kernel.AbstractKernel
            Instance of type AbstractKernel to calculate the RBF kernel.
        t : (x_dim, x_dim, n_samples) ndarray of float
            Transformation matrix for input x. Default matrix scales xc coordinates to a range of 1.0
        gamma : float
            smoothing parameter
        max_cond : float
            maximum condition number
        """
        x_dim = xc.shape[0]
        n_samples = xc.shape[2]

        self.kernel = kernel
        self.max_dist = max_dist
        self.xc = xc

        if t is None:
            # ToDo: use normalization like in lcm.math.rbf.rbf -> RBF
            # ToDo: update normalization in lcm.math.rbtransformations module
            t = np.zeros((x_dim, x_dim, n_samples))

            for k in range(0, n_samples):
                pp = xc[:, :, k].max(axis=1) - xc[:, :, k].min(axis=1)
                # pp[pp==0] = np.finfo('d').resolution
                t[:, :, k] = np.diag(1.0 / pp)
        # else:
        #     if t.shape != (x_dim, x_dim, n_samples):
        #         raise Exception('Wrong t.shape:' + str(t.shape) + ' instead of ' + str((x_dim, x_dim, n_samples)))

        self.t = t

        if zc is None:
            xc_t, c, d = self.initFromMatrices(kwargs['c'], kwargs['d'])
        else:
            xc_t, c, d = self.initFromSamples(zc, gamma, max_cond)

        self.c = c
        self.d = d
        self.xc_t = xc_t

    def __call__(self, x):
        """
        Evaluates the RBF model

        Parameters
        ----------
        x : (x_dim, n_interpolations, n_samples) ndarray of float
            periodic signals representing the coordinates where to interpolate

        Returns
        -------
        z : (z_dim, n_interpolations, n_samples) ndarray
            result array
        """
        z = eval_z(self.xc_t, self.c, self.d, self.t, x, self.kernel, self.max_dist)
        return z

    @property
    def xdim(self):
        return self.d.shape[1] - 1

    @property
    def zdim(self):
        return self.d.shape[0]

    @property
    def n_stacks(self):
        return self.d.shape[2]

    def initFromSamples(self, zc, gamma, max_cond):
        """

        Parameters
        ----------
        zc : np.ndarray
        gamma : float
        max_cond : float

        Returns
        -------

        xc_t : np.ndarray
        c : np.ndarray
        d : np.ndarray

        """
        (xc_t, c, d) = calc_weigths(self.xc, self.t, zc, self.kernel, gamma, max_cond)
        return xc_t, c, d

    def initFromMatrices(self, c, d):
        """

        Parameters
        ----------
        c : np.ndarray
        d : np.ndarray
        t : np.ndarray

        Returns
        -------

        xc_t : np.ndarray
        c : np.ndarray
        d : np.ndarray

        """
        # xc_t = np.matmul(self.t.transpose(2, 0, 1), self.xc.transpose(2, 0, 1)).transpose(1, 2, 0)
        xc_t = self.xc
        return xc_t, c, d

    def eval_dz_dx(self, x):
        """
        Helper function: evaluates the jacobian matrix

        Parameters
        ----------
        x : (x_dim, n_interpolations, n_samples) ndarray of float
            periodic signals representing the coordinates where to interpolate

        Returns
        -------
        z : (z_dim, x_dim, n_interpolations, n_samples) ndarray
            result array
        """
        z = eval_dz_dx(self.xc_t, self.c, self.d, self.t, x, self.kernel, self.max_dist)
        return z

    def to_msgpack(self):
        """
        Create msgpack serializable dict representing an instance of this class
        ref: http://msgpack.org/

        Returns
        -------
        msgpack : dict
            msgpack representation

        """

        msgpack = {"kernel": self.kernel.to_msgpack(),
                   "max_dist": self.max_dist,
                   "xc_t": array3d_to_msgpack(self.xc_t),
                   "t": array3d_to_msgpack(self.t),
                   "c": array3d_to_msgpack(self.c),
                   "d": array3d_to_msgpack(self.d)}

        return msgpack

    def write_msgpack(self, file):
        """
        Create msgpack file representing an instance of this class
        ref: http://msgpack.org/


        Parameters
        ----------
        file : String
            path to msgpack file
        """
        import msgpack
        data = self.to_msgpack()
        data["guid"] = "{8c4e810f-3df3-4a00-8276-176fa3c9f000}"

        with open(file, "wb") as fh:
            msgpack.dump(data, fh)

    def toCBOR(self):
        """
        Create CBOR serializable dict representing an instance of this class
        ref: http://cbor.io/

        Returns
        -------
        cbor_obj : dict
            cbor representation

        """

        ''' create cbor serializable dict '''
        cbor_obj = {'type': '__stackedRBF__',
                    'xi': Serializable.toCBOR(self.xc_t),
                    't': Serializable.toCBOR(self.t),
                    'c': Serializable.toCBOR(self.c),
                    'd': Serializable.toCBOR(self.d),
                    'kernel': Serializable.toCBOR(self.kernel.id),
                    'params': Serializable.toCBOR(self.kernel.params),
                    'max_dist': Serializable.toCBOR(self.max_dist)
                    }

        return cbor_obj
