#!/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 944 $
$LastChangedDate:: 2019-02-20 16:44:30 +0100#$

This file is part of SyMSpace.

Copyright (c) 2018, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec


class AnalyticFunction:
    def __init__(self, n_x, n_y):
        self.n_x = n_x
        self.n_y = n_y
        self.n_fncs = 4

    def getXc(self):
        xc_samples = np.empty((2, self.n_x * self.n_y), dtype=np.float)
        for col, y0 in enumerate(np.linspace(-20, 20, self.n_y)):
            for row, x0 in enumerate(np.linspace(-20, 20, self.n_x)):
                idx = col * self.n_x + row
                xc_samples[0, idx] = x0
                xc_samples[1, idx] = y0

        return xc_samples

    def getZc(self):
        xc_samples = self.getXc()

        '''zc_samples'''
        n_samples = xc_samples.shape[1]
        zc_samples = np.empty((self.n_fncs, n_samples), dtype=np.float)
        '''dz/dx[0]'''

        for idx_center in range(0, n_samples):
            zc_samples[:, idx_center] = self.getZi(*xc_samples[:, idx_center])

        return zc_samples

    def getZi(self, x, y):
        """
        Definition der Testfunktion [f1, f2, f3] = fun1(x, y, phi)
        """

        f1 = 2. * x - 3. * y
        f2 = 2. * x - 3. * y + 0.5
        f3 = 2. * x - 3. * y - x * y ** 2
        f4 = 2. * x - 3. * y - x * y ** 2

        return np.array([f1, f2, f3, f4])

    def get_dZi_dx(self, x, y):
        f11 = 2. * np.ones_like(x)
        f21 = 2. * np.ones_like(x)
        f31 = 2. - y ** 2
        f41 = 2. - y ** 2

        f12 = - 3. * np.ones_like(x)
        f22 = - 3. * np.ones_like(x)
        f32 = - 3. - 2 * x * y
        f42 = - 3. - 2 * x * y

        return np.array([[f11, f12], [f21, f22], [f31, f32], [f41, f42]])

    def plotComparison(self, x_interp, x_analytic, atol, rtol):
        index = 1
        nrows = x_interp.shape[0]
        for row in range(nrows):
            a = x_interp[row, :]
            b = x_analytic[row, :]

            error = self.calcError(a, b)
            threshold = atol + rtol * np.abs(b)

            plt.subplot(nrows, 2, index)
            plt.step(a, 'r-.', linestyle='-', label='rbf[{},:]'.format(row))
            plt.step(b, 'g--', linestyle='--', label='analytic[{},:]'.format(row))
            plt.grid(True)
            plt.legend(loc=1)

            plt.subplot(nrows, 2, index + 1)
            plt.step(error, 'r-.', linestyle='-.', label='error[{},:]'.format(row))
            plt.step(threshold, 'g--', linestyle='--', label='threshold[{},:]'.format(row))
            plt.grid(True)
            plt.legend(loc=1)

            index += 2
        plt.tight_layout()
        plt.show()

    def plotJacobiComparison(self, dx_interp, dx_analytic, atol, rtol):

        nrows = dx_interp.shape[0]
        ncols = dx_interp.shape[1]
        fig = plt.figure(figsize=(6 * 2, 4 * 2))
        gs = GridSpec(nrows, ncols * 2, figure=fig)

        for row in range(nrows):
            for col in range(ncols):
                a = dx_interp[row, col, :]
                b = dx_analytic[row, col, :]

                error = self.calcError(a, b)
                threshold = atol + rtol * np.abs(b)

                ax1 = fig.add_subplot(gs[row, col])
                ax1.plot(a, 'r-.', label='dx_dz_rbf[{},{}]'.format(row, col))
                ax1.plot(b, 'g--', label='dx_dz_analytic[{},{}]'.format(row, col))
                ax1.grid(True)
                ax1.legend(loc=1)

                ax2 = fig.add_subplot(gs[row, col + ncols])
                ax2.plot(error, 'r-.', label='error[{},{}]'.format(row, col))
                ax2.plot(threshold, 'g--', label='threshold[{},{}]'.format(row, col))
                ax2.grid(True)
                ax2.legend(loc=1)

                plt.tight_layout()

        plt.show()

    def calcError(self, a, b):
        return np.abs(a - b)
