# !/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 1041 $
$LastChangedDate:: 2021-02-04 13:55:24 +0100#$

This file is part of SyMSpace.

Copyright (c) 2020, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""
import os
import optparse
import pkgutil
import nose
import lcm
from lcm.testing.nose_plugins import LoadCoverageModules


class NoseTestRunner:
    """
    Tests all Python tests found in path and all it's subdirectories using the nose framework

    - writes xunit xml-files with testresults
    - writes coverage xml-files for selected packages

    """

    def __init__(self, start_dir, reports_dir, cov_dir):
        self.start_dir = start_dir
        self.reports_dir = reports_dir
        self.cov_dir = cov_dir
        self.xunit_file = os.path.join(reports_dir, 'nosetests-junit.xml')
        self.runner = nose

    def run(self, exclude_patterns: tuple = None, with_coverage: bool = True, cover_xml: bool = True,
            cover_html: bool = False) -> bool:
        """
        Collect and run tests, returning success or failure.

        Parameters
        ----------
        exclude_patterns: tuple
            exclude patterns as regular expression strings
        with_coverage: bool
            Run with coverage, Default: True
        cover_xml: bool
            Create xml coverage report, if "with_coverage" is also True, Default: True
        cover_html: bool
            Create html coverage report, if "with_coverage" is also True, Default: False

        Examples
        -------
        >>>from lcm.testing.nose_testrunner import NoseTestRunner
        >>>runner = NoseTestRunner.fromCMDParser()
        >>>runner.run(exclude_patterns=('tests_math', 'tests_utils'))

        Returns
        -------
        success: bool
            Tells whether or not this result was a success.
        """
        if not os.path.isdir(self.reports_dir):
            os.makedirs(self.reports_dir, exist_ok=True)

        '''nose runner sys/cmd arguments'''
        arguments = ['', '--where={}'.format(self.start_dir),
                     '--with-xunit', '--xunit-file={}'.format(self.xunit_file),
                     ]

        '''additional plugins'''
        add_plugins = []

        if exclude_patterns is not None:
            for pat in exclude_patterns:
                arguments.append('--exclude={}'.format(pat))
        if with_coverage:
            arguments.append('--with-coverage')
            add_plugins.append(LoadCoverageModules(self.cov_dir))
            if cover_xml:
                cover_xml_file = os.path.join(self.reports_dir, 'coverage.xml')
                print('XML Report: {}'.format(cover_xml_file))
                arguments.extend(['--cover-xml',
                                  '--cover-xml-file={}'.format(cover_xml_file),
                                  ])

            if cover_html:
                cover_html_dir = os.path.join(self.reports_dir, 'html')
                print('HTML Report: {}'.format(os.path.join(cover_html_dir, 'index.html')))
                arguments.extend(['--cover-html',
                                  '--cover-html-dir={}'.format(cover_html_dir)
                                  ])

        cov_pkgs = [pkgname for _, pkgname, is_pkg in pkgutil.iter_modules([self.cov_dir]) if is_pkg]
        for pkgs in cov_pkgs:
            arguments.append('--cover-package={}'.format(pkgs))

        success = self.runner.run(argv=arguments, addplugins=add_plugins)
        return success

    @staticmethod
    def getDefaultDirs() -> tuple:
        py_dir = os.path.dirname(os.path.dirname(os.path.dirname(lcm.__file__)))
        start_dir = os.path.join(py_dir, 'tests')
        reports_dir = os.path.join(py_dir, 'tmp')
        cov_dir = os.path.join(py_dir, 'src')
        return start_dir, reports_dir, cov_dir

    @staticmethod
    def parseCMD(parser: optparse.OptionParser) -> tuple:
        (options, args) = parser.parse_args()
        start_dir = os.path.abspath(options.start_dir)
        reports_dir = os.path.abspath(options.reports_dir)
        cov_dir = os.path.abspath(options.cov_dir)
        return start_dir, reports_dir, cov_dir

    @classmethod
    def fromCMDParser(cls) -> 'NoseTestRunner':
        """
        Commandline parser to create instance of this class

        Commandline Arguments
        ---------------------

        --startDir : str
            The search for tests starts from this directory (Optional).
        --reportsDir : str
            The test reports will be saved to this directory (Optional).
        --coverPkgDir : str
            The packages importable from this directory will be covered  (Optional).

        Returns
        -------
        runner : NoseTestRunner
        """
        start_dir_default, reports_dir_default, cov_dir_default = cls.getDefaultDirs()

        parser = optparse.OptionParser("usage: %prog [options]")
        parser.add_option("-d", "--startDir", dest="start_dir", default=start_dir_default,
                          type="string", help="The search for tests starts from this directory.")
        parser.add_option("-r", "--reportsDir", dest="reports_dir", default=reports_dir_default,
                          type="string", help="The test reports will be saved to this directory.")
        parser.add_option("-c", "--coverPkgDir", dest="cov_dir", default=cov_dir_default,
                          type="string", help="The packages importable from this directory will be covered.")

        start_dir, reports_dir, cov_dir = cls.parseCMD(parser)

        if start_dir == "":
            raise ValueError("--startDir must be given")
        if reports_dir == "":
            raise ValueError("--reportsDir must be given")
        if cov_dir == "":
            raise ValueError("--coverPkgDir must be given")

        print('Start Path: {}'.format(start_dir))
        print('Report Path: {}'.format(reports_dir))
        print('Coverage Path: {}'.format(cov_dir))

        runner = cls(start_dir, reports_dir=reports_dir, cov_dir=cov_dir)
        return runner
