import time
import numpy as np
import lcm.math.rbf as rbf
import lcm.utils.pytperiodic as pm
from lcm.math.rbf import kernels
from lcm.math.rbf import transformations


def run_periodicRBFRandom(n_in=2, n_out=3, n_lp=4):
    """
    Example for periodic RBF Interpolation
    using randomly generated input and output signals with the defined size

    Parameters
    ----------
    n_in : int
        number of inputs
    n_out : int
        number of outputs
    n_lp : int
        number of loadpoints

    """

    '''input matrix'''
    xc_per = pm.PyTPeriodicMatrix(n_in, n_lp, mode='random', maxorder=10) * 100

    '''output matrix'''
    zc_per = pm.PyTPeriodicMatrix(n_out, n_lp, mode='random', maxorder=10)

    '''create rbf'''
    t_per = transformations.normalization(xc_per)  # optional
    rbf1 = rbf.PeriodicMultiRBF(xc_per, zc_per, kernels.KernelTPS(), t_per=t_per, gamma=1e-3, max_dist=5.0)

    '''test'''
    x = np.zeros((n_in, 1), dtype=np.object)
    x[0, 0] = pm.PyTPeriodicMatrix(1, 1., 0.)
    x[1, 0] = pm.PyTPeriodicMatrix(1, 1., 0.)
    x_per = pm.PyTPeriodicMatrix(x, mode='objarray')

    '''evaluate rbf'''
    z_per = rbf1(x_per)

    print('x_per', x_per)
    print('z_per', z_per)

    '''plot with plotly'''
    xc_per.showFullPreviewPLY(mode='matrix')


def run_periodicRBFSine(n_in=2, n_out=3, n_lp=4):
    """
    Example for periodic RBF Interpolation
    using cosine input and sine output signals with the defined size

    Parameters
    ----------
    n_in : int
        number of inputs
    n_out : int
        number of outputs
    n_lp : int
        number of loadpoints

    """

    '''input matrix from 1x1 periodics'''
    xc = np.zeros((n_in, n_lp), dtype=np.object)
    for r in range(n_in):
        for c in range(n_lp):
            magnitude = r + c
            xc[r, c] = pm.PyTPeriodicMatrix(1., magnitude, 0.)  # cosine

    '''output matrix from 1x1 periodics'''
    zc = np.zeros((n_out, n_lp), dtype=np.object)
    for r in range(n_out):
        for c in range(n_lp):
            magnitude = r + c
            zc[r, c] = pm.PyTPeriodicMatrix(1, magnitude, -np.pi / 2)  # sine

    '''numpy to periodic'''
    xc_per = pm.PyTPeriodicMatrix(xc, mode='objarray')
    zc_per = pm.PyTPeriodicMatrix(zc, mode='objarray')

    '''create rbf'''
    rbf1 = rbf.PeriodicMultiRBF(xc_per, zc_per, kernels.KernelTPS(), t_per=None, gamma=1e-3, max_dist=1.0)

    '''test'''
    x = np.zeros((n_in, 1), dtype=np.object)
    x[0, 0] = pm.PyTPeriodicMatrix(1, 1., 0.)
    x[1, 0] = pm.PyTPeriodicMatrix(1, 1., 0.)
    x_per = pm.PyTPeriodicMatrix(x, mode='objarray')

    '''evaluate rbf'''
    z_per = rbf1(x_per)

    print('x_per', x_per)
    print('z_per', z_per)

    '''plot with plotly'''
    xc_per.showFullPreviewPLY(mode='matrix')


def run_plotMaxDist(n_in=3, n_out=1, n_lp=4):
    """
    Example for periodic RBF Interpolation
    using random input and output values with the defined size

    Parameters
    ----------
    n_in : int
        number of inputs
    n_out : int
        number of outputs
    n_lp : int
        number of loadpoints

    """

    import plotly.offline as ply

    '''create reproducible random numbers'''
    np.random.seed(1234)

    '''input matrix'''
    xc_per = pm.PyTPeriodicMatrix(n_in, n_lp, mode='random', maxorder=10) * 100

    '''output matrix'''
    zc_per = pm.PyTPeriodicMatrix(n_out, n_lp, mode='random', maxorder=10)

    '''create rbf'''
    t_per = transformations.normalization(xc_per)  # optional
    rbf1 = rbf.PeriodicMultiRBF(xc_per, zc_per, kernels.KernelTPS(), t_per=t_per, gamma=1e-3, max_dist=5.0)

    if n_in == 2:
        data = rbf1.plotMinRadius2DPLY()
    elif n_in == 3:
        z_min = np.min(xc_per[2, :].min())
        z_max = np.max(xc_per[2, :].max())
        data = rbf1.plotMinRadius3DPLY(z_values=np.linspace(z_min, z_max, 3))
    else:
        raise Exception

    ply.plot(data)


def run_loop(n_times):
    n_in = 2
    n_out = 3
    n_lp = 4

    '''create reproducible random numbers'''
    np.random.seed(1234)

    '''input matrix'''
    xc_per = pm.PyTPeriodicMatrix(n_in, n_lp, mode='random', maxorder=10) * 100

    '''output matrix'''
    zc_per = pm.PyTPeriodicMatrix(n_out, n_lp, mode='random', maxorder=10)

    '''create rbf'''
    t_per = transformations.normalization(xc_per)  # optional
    # rbf1 = rbf.PeriodicMultiRBF(xc_per, zc_per, kernels.KernelTPS(), t_per=t_per, gamma=1e-3, max_dist=5.0)
    rbf1 = rbf.PeriodicMultiRBF(xc_per, zc_per, kernels.KernelCS(), t_per=t_per, gamma=1e-3, max_dist=5.0)

    '''test'''
    # x = np.zeros((n_in, 1), dtype=np.object)
    # x[0, 0] = pm.PyTPeriodicMatrix(1, 1., 0.)
    # x[1, 0] = pm.PyTPeriodicMatrix(1, 1., 0.)
    # x_per = pm.PyTPeriodicMatrix(x, mode='objarray')
    x_per = pm.PyTPeriodicMatrix.random((n_in, 1), maxorder=10)

    dt_sum = 0.0
    for _ in range(n_times):
        t1 = time.time()
        '''evaluate rbf'''
        z_per = rbf1(x_per)
        dt_sum += time.time() - t1

    print('\nrbf: Total={:.2f} s  Avg={:.2f} s'.format(dt_sum, dt_sum / n_times))


if __name__ == '__main__':
    # run_periodicRBFRandom()
    # run_periodicRBFSine()
    # run_plotMaxDist()
    run_loop(n_times=100000)  # for profiling
