/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9715 $
 * $Date: 2020-03-19 10:01:15 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.symspace.center.utils;

import java.awt.BorderLayout;
import java.awt.CardLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.FocusAdapter;
import java.awt.event.FocusEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.io.File;
import java.rmi.RemoteException;

import javax.swing.AbstractListModel;
import javax.swing.GroupLayout;
import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JProgressBar;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;
import javax.swing.JTable;
import javax.swing.JTextArea;
import javax.swing.LayoutStyle.ComponentPlacement;
import javax.swing.ListCellRenderer;
import javax.swing.ListSelectionModel;
import javax.swing.SwingConstants;
import javax.swing.WindowConstants;
import javax.swing.border.EmptyBorder;
import javax.swing.filechooser.FileNameExtensionFilter;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.DefaultTableCellRenderer;

import at.lcm.bu21.magopt.structure.IParameter;
import at.lcm.bu21.magopt.structure.IProject;
import at.lcm.bu21.magopt.structure.Project;
import at.lcm.bu21.magopt.structure.extra.Variants;
import magopt.gui.helpers.JProjectChooser;

/**
 * 
 *
 * @author Werner Koppelsttter
 */
public class VariantsDialog extends JDialog {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** String description for possible combinations */
	private static final String combinationString = "Possible Combinations:    ";
	/** String description for valid combinations */
	private static final String validString = "Valid combinations:     ";
	/** String description for selected combinations */
	private static final String selectedString = "Selected combination:    ";

	/** Settings-panel */
	private JPanel settings;
	/** Layout for settings */
	private CardLayout cards;
	/** Index of layout */
	private int cardIndex = 0;

	/** Panel for unused/used selection */
	private JPanel useSelection;
	/** List for unused-parameters */
	private JList<IParameter> unusedList;
	/** List for used-parameters */
	private JList<IParameter> usedList;
	/** Model for unused-list */
	private VariantsUnusedListModel unusedModel;
	/** Model for used-list */
	private VariantsUsedListModel usedModel;
	/** Scroll-pane for unused list */
	private JScrollPane unusedScroll;
	/** Scroll-pane for used list */
	private JScrollPane usedScroll;
	/** Select parameter to used list */
	private JButton selectUsed;
	/** Deselect parameter from used list */
	private JButton deselectUsed;

	/** Panel for range */
	private JSplitPane range;
	/** Table for range */
	private JTable rangeTable;
	/** Table model for range */
	private RangeTableModel rangeModel;
	/** Label for number of possible combinations */
	private JLabel combinationCount;
	/** Area for validity string */
	private JTextArea validText;

	/** Panel for selection */
	private JPanel selection;
	/** Table for selection */
	private JTable selectTable;
	/** Table model for select table */
	private SelectTableModel selectModel;
	/** Label for number of valid and selectet combinations */
	private JLabel validCount;

	/** Panel with progress-bar */
	private JPanel progress;

	/** State message */
	private JLabel state;

	/** Create variants */
	private JButton create;
	/** Append variations */
	private JButton append;
	/** Cancel dialog */
	private JButton cancel;
	/** Load settings */
	private JButton load;
	/** Save settings */
	private JButton save;
	/** Next settings page */
	private JButton next;
	/** Previous settings page */
	private JButton previous;

	/** Variants-object */
	private Variants variants;

	/** Based project */
	private IProject project;

	/** Used settings-file */
	private File settingFile = null;

	/** Result project */
	private IProject result = null;

	/** Size of dialog (used for resize) */
	private Dimension size;

	/**
	 * List-model for unused parameters
	 * 
	 * @author Werner Koppelsttter
	 */
	private class VariantsUnusedListModel
			extends AbstractListModel<IParameter> {

		/** Serialization version number */
		private static final long serialVersionUID = 1L;

		@Override
		public int getSize() {
			return VariantsDialog.this.variants.getUnusedSize();
		}

		@Override
		public IParameter getElementAt(int index) {
			return VariantsDialog.this.variants.getUnused(index);
		}

		/**
		 * Fire parameter added to the end of list
		 */
		public void fireAdded() {
			super.fireIntervalAdded(this,
					VariantsDialog.this.variants.getUnusedSize() - 1,
					VariantsDialog.this.variants.getUnusedSize() - 1);
		}

		/**
		 * Fire one parameter removed from list
		 */
		public void fireRemoved() {
			super.fireIntervalRemoved(this,
					VariantsDialog.this.variants.getUnusedSize(),
					VariantsDialog.this.variants.getUnusedSize());
			super.fireContentsChanged(this, 0,
					VariantsDialog.this.variants.getUnusedSize() - 1);
		}

		/**
		 * Fire content of list changed
		 */
		public void contentChanged() {
			super.fireContentsChanged(this, 0,
					VariantsDialog.this.variants.getUnusedSize() - 1);
		}

		/**
		 * Is list empty
		 * 
		 * @return Is empty
		 */
		public boolean isEmpty() {
			return VariantsDialog.this.variants.getUnusedSize() == 0;
		}
	}

	/**
	 * List-model for used parameters
	 * 
	 * @author Werner Koppelsttter
	 */
	public class VariantsUsedListModel extends AbstractListModel<IParameter> {

		/** Serialization version number */
		private static final long serialVersionUID = 1L;

		@Override
		public int getSize() {
			return VariantsDialog.this.variants.getUsedSize();
		}

		@Override
		public IParameter getElementAt(int index) {
			return VariantsDialog.this.variants.getUsed(index);
		}

		/**
		 * Fire parameter added to the end of list
		 */
		public void fireAdded() {
			super.fireIntervalAdded(this,
					VariantsDialog.this.variants.getUsedSize() - 1,
					VariantsDialog.this.variants.getUsedSize() - 1);
		}

		/**
		 * Fire one parameter removed from list
		 */
		public void fireRemoved() {
			super.fireIntervalRemoved(this,
					VariantsDialog.this.variants.getUsedSize(),
					VariantsDialog.this.variants.getUsedSize());
			super.fireContentsChanged(this, 0,
					VariantsDialog.this.variants.getUsedSize() - 1);
		}

		/**
		 * Fire content of list changed
		 */
		public void contentChanged() {
			super.fireContentsChanged(this, 0,
					VariantsDialog.this.variants.getUsedSize() - 1);
		}

		/**
		 * Is list empty
		 * 
		 * @return Is empty
		 */
		public boolean isEmpty() {
			return VariantsDialog.this.variants.getUsedSize() == 0;
		}
	}

	/**
	 * Cell render for unused/used-list
	 *
	 * @author Werner Koppelsttter
	 */
	public class IParameterCellRenderer extends JLabel
			implements ListCellRenderer<IParameter> {

		/** Serialization version number */
		private static final long serialVersionUID = 1L;

		@Override
		public Component getListCellRendererComponent(
				JList<? extends IParameter> list, IParameter value, int index,
				boolean isSelected, boolean cellHasFocus) {

			this.setBorder(new EmptyBorder(2, 10, 2, 10));
			try {
				this.setText(value.getName());
				this.setToolTipText(value.getIdentifier());
			} catch (RemoteException e) {
				this.setText("???");
				this.setToolTipText("");
			}
			if (isSelected) {
				this.setBackground(list.getSelectionBackground());
				this.setForeground(list.getSelectionForeground());
			} else {
				this.setBackground(list.getBackground());
				this.setForeground(list.getForeground());
			}
			this.setEnabled(list.isEnabled());
			this.setFont(list.getFont());
			this.setOpaque(true);

			return this;
		}
	}

	/**
	 * Table model for range-table
	 *
	 * @author Werner Koppelsttter
	 */
	public class RangeTableModel extends AbstractTableModel {

		/** Serialization version number */
		private static final long serialVersionUID = 1L;

		@Override
		public int getRowCount() {
			return VariantsDialog.this.variants.getUsedSize();
		}

		@Override
		public int getColumnCount() {
			return 3;
		}

		@Override
		public String getColumnName(int column) {
			String str;

			switch (column) {
			case 0:
				str = "Name";
				break;
			case 1:
				str = "Range/Values";
				break;
			case 2:
				str = "Length";
				break;
			default:
				str = "";
				break;
			}

			return str;
		}

		@Override
		public Class<?> getColumnClass(int column) {
			Class<?> c;

			if (column == 2) {
				c = Integer.class;
			} else {
				c = String.class;
			}

			return c;
		}

		@Override
		public boolean isCellEditable(int rowIndex, int columnIndex) {
			return columnIndex == 1;
		}

		@Override
		public Object getValueAt(int rowIndex, int columnIndex) {
			Object val = null;
			double[] values;

			try {
				switch (columnIndex) {
				case 0:
					val = VariantsDialog.this.variants.getUsed(rowIndex)
							.getName();
					break;
				case 1:
					val = VariantsDialog.this.variants.getRange(rowIndex);
					break;
				case 2:
					values = VariantsDialog.this.variants
							.getRangeValues(rowIndex);
					if (values != null) {
						val = values.length;
					} else {
						val = 0;
					}
					break;
				default:
					val = VariantsDialog.this.variants.getUnused(rowIndex);
				}
			} catch (Exception e) {
				System.err.println(e.getMessage());
				e.printStackTrace();
			}

			return val;
		}

		@Override
		public void setValueAt(Object aValue, int rowIndex, int columnIndex) {
			if (columnIndex == 1) {
				VariantsDialog.this.variants.setRange(rowIndex,
						(String) aValue);
				this.fireTableCellUpdated(rowIndex, columnIndex);
				this.fireTableCellUpdated(rowIndex, 2);
				VariantsDialog.this.refresh();
			}
		}

	}

	/**
	 * Table model for selection table
	 *
	 * @author Werner Koppelsttter
	 */
	public class SelectTableModel extends AbstractTableModel {

		/** Serialization version number */
		private static final long serialVersionUID = 1L;

		@Override
		public int getRowCount() {
			return VariantsDialog.this.variants.getValidSize();
		}

		@Override
		public int getColumnCount() {
			return VariantsDialog.this.variants.getUsedSize() + 2;
		}

		@Override
		public String getColumnName(int column) {
			String str = "";

			if (column == 0) {
				str = "Index";
			} else if (column == (VariantsDialog.this.variants.getUsedSize()
					+ 1)) {
				str = "Selected";
			} else {
				try {
					str = VariantsDialog.this.variants.getUsed(column - 1)
							.getName();
				} catch (RemoteException e) {
					System.err.println(e.getMessage());
					e.printStackTrace();
				}
			}

			return str;
		}

		@Override
		public Class<?> getColumnClass(int column) {
			Class<?> c;

			if (column == 0) {
				c = Integer.class;
			} else if (column == (VariantsDialog.this.variants.getUsedSize()
					+ 1)) {
				c = Boolean.class;
			} else {
				c = Double.class;
			}

			return c;
		}

		@Override
		public boolean isCellEditable(int rowIndex, int columnIndex) {
			return columnIndex == (VariantsDialog.this.variants.getUsedSize()
					+ 1);
		}

		/**
		 * Get combinations index from selection table (only valid combinations)
		 * 
		 * @param validIndex
		 *            Valid index
		 * @return Combination index
		 */
		private int getCombinationIndex(int validIndex) {
			int index = 0;

			while ((index < VariantsDialog.this.variants.getCombinationSize())
					&& (validIndex >= 0)) {
				if (VariantsDialog.this.variants.getValid(index)) {
					validIndex--;
				}
				if (validIndex >= 0) {
					index++;
				}
			}

			if (index >= VariantsDialog.this.variants.getCombinationSize()) {
				index = -1;
			}

			return index;
		}

		@Override
		public Object getValueAt(int rowIndex, int columnIndex) {
			Object val = null;

			if (columnIndex == 0) {
				val = rowIndex;
			} else if (columnIndex == (VariantsDialog.this.variants
					.getUsedSize() + 1)) {
				val = VariantsDialog.this.variants
						.getSelected(this.getCombinationIndex(rowIndex));
			} else {
				val = VariantsDialog.this.variants.getCombination(
						this.getCombinationIndex(rowIndex))[columnIndex - 1];
			}

			return val;
		}

		@Override
		public void setValueAt(Object aValue, int rowIndex, int columnIndex) {
			if (columnIndex == (VariantsDialog.this.variants.getUsedSize()
					+ 1)) {

				VariantsDialog.this.variants.setSelected(
						this.getCombinationIndex(rowIndex), (Boolean) aValue);
				this.fireTableCellUpdated(rowIndex, columnIndex);
				VariantsDialog.this.refresh();
			}
		}
	}

	/**
	 * Renderer for variants tables
	 *
	 * @author Werner Koppelsttter
	 */
	public class VariantsTableRenderer extends DefaultTableCellRenderer {

		/** Serialization version number */
		private static final long serialVersionUID = 1L;

		@Override
		public Component getTableCellRendererComponent(JTable table,
				Object value, boolean isSelected, boolean hasFocus, int row,
				int column) {

			super.getTableCellRendererComponent(table, value, isSelected,
					hasFocus, row, column);

			this.setBorder(new EmptyBorder(2, 10, 2, 10));
			// this.setHorizontalAlignment(SwingConstants.LEFT);

			if (this.getText().isEmpty()) {
				this.setBackground(Color.YELLOW);
			} else {
				this.setBackground(table.getBackground());
			}
			if (isSelected) {
				this.setForeground(table.getSelectionForeground());
				this.setBackground(table.getSelectionBackground());
			} else {
				this.setForeground(table.getForeground());
				this.setBackground(table.getBackground());
			}
			this.setOpaque(true);

			return this;
		}
	}

	/**
	 * Create variants dialog
	 * 
	 * @param owner
	 *            Owner of dialog
	 * @param project
	 *            Base project
	 */
	public VariantsDialog(Component owner, IProject project) {
		this.project = project;
		try {
			this.variants = new Variants(project);
		} catch (RemoteException e1) {
			System.err.println(e1.getMessage());
			e1.printStackTrace();
		}

		this.setModal(true);
		this.init();

		this.pack();
		this.setLocationRelativeTo(owner);

		// resize
		this.size = this.getSize();
		this.addComponentListener(new ComponentAdapter() {

			@Override
			public void componentResized(ComponentEvent e) {
				Dimension newSize, compSize; // , minSize;
				int width, height;

				newSize = e.getComponent().getSize();
				compSize = VariantsDialog.this.unusedScroll.getSize();
				width = compSize.width
						+ (newSize.width - VariantsDialog.this.size.width);
				height = compSize.height
						+ (newSize.height - VariantsDialog.this.size.height);
				VariantsDialog.this.size = newSize;

				VariantsDialog.this.unusedScroll
						.setPreferredSize(new Dimension(width, height));
				VariantsDialog.this.usedScroll
						.setPreferredSize(new Dimension(width, height));
			}
		});

		this.refreshLists();
		this.refresh();

		this.range.setDividerLocation(0.8);

		this.setVisible(true);
	}

	/**
	 * Get result project
	 * 
	 * @return Project with variants
	 */
	public IProject getResultProject() {
		return this.result;
	}

	/**
	 * Init dialog
	 */
	private void init() {
		this.setTitle("Project variants");
		this.setLayout(new BorderLayout());
		this.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);

		GroupLayout layout = new GroupLayout(this.getContentPane());
		this.getContentPane().setLayout(layout);
		layout.setAutoCreateGaps(true);
		layout.setAutoCreateContainerGaps(true);

		this.initButtons();
		this.initSettings();

		layout.setHorizontalGroup(layout.createParallelGroup()
				.addComponent(this.settings).addComponent(this.state)
				.addGroup(layout.createSequentialGroup().addComponent(this.load)
						.addComponent(this.save)
						.addPreferredGap(ComponentPlacement.UNRELATED, 50,
								10000)
						.addComponent(this.previous).addComponent(this.next)
						.addPreferredGap(ComponentPlacement.UNRELATED, 50,
								10000)
						.addComponent(this.append).addComponent(this.create)
						.addComponent(this.cancel)));
		layout.setVerticalGroup(layout.createSequentialGroup()
				.addComponent(this.settings).addComponent(this.state)
				.addGroup(layout.createParallelGroup()
						.addComponent(this.previous).addComponent(this.next)
						.addComponent(this.load).addComponent(this.save)
						.addComponent(this.append).addComponent(this.create)
						.addComponent(this.cancel)));
		layout.linkSize(this.load, this.save, this.create, this.cancel);
		layout.linkSize(this.previous, this.next);
	}

	/**
	 * Initialize settings
	 */
	private void initSettings() {
		this.settings = new JPanel();

		this.initUseSelection();
		this.initRange();
		this.initSelection();
		this.initProgress();

		this.cards = new CardLayout();
		this.settings.setLayout(this.cards);

		this.settings.add(this.useSelection, "0");
		this.settings.add(this.range, "1");
		this.settings.add(this.selection, "2");
		this.settings.add(this.progress, "9");
	}

	/**
	 * Initialize used selection panel
	 */
	private void initUseSelection() {
		GroupLayout layout;

		this.useSelection = new JPanel();
		layout = new GroupLayout(this.useSelection);
		this.useSelection.setLayout(layout);
		this.selectUsed = new JButton(">>");
		this.useSelection.add(this.selectUsed);
		this.selectUsed.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e) {
				int index = VariantsDialog.this.unusedList.getSelectedIndex()
						- 1;

				for (IParameter p : VariantsDialog.this.unusedList
						.getSelectedValuesList()) {
					VariantsDialog.this.selectUsed(p);
				}

				if (index < 0) {
					index = 0;
				}
				VariantsDialog.this.unusedList.setSelectedIndex(index);

				VariantsDialog.this.refresh();
			}
		});
		this.deselectUsed = new JButton("<<");
		this.useSelection.add(this.deselectUsed);
		this.deselectUsed.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e) {
				int index = VariantsDialog.this.usedList.getSelectedIndex() - 1;

				for (IParameter p : VariantsDialog.this.usedList
						.getSelectedValuesList()) {
					VariantsDialog.this.deselectUsed(p);
				}

				if (index < 0) {
					index = 0;
				}
				VariantsDialog.this.usedList.setSelectedIndex(index);

				VariantsDialog.this.refresh();
			}
		});

		// this.nonSelectModel = new DefaultListModel<IParameter>();
		this.unusedModel = new VariantsUnusedListModel();
		this.unusedList = new JList<IParameter>(this.unusedModel);
		this.unusedList.setCellRenderer(new IParameterCellRenderer());
		this.unusedList.setSelectionMode(
				ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
		this.unusedList.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent e) {
				if (e.getClickCount() == 2) {
					VariantsDialog.this.selectUsed(
							VariantsDialog.this.unusedList.getSelectedValue());
				}
			}
		});
		this.unusedScroll = new JScrollPane(this.unusedList);
		this.unusedScroll.setPreferredSize(new Dimension(300, 400));
		this.useSelection.add(this.unusedScroll);

		// this.selectModel = new DefaultListModel<IParameter>();
		this.usedModel = new VariantsUsedListModel();
		this.usedList = new JList<IParameter>(this.usedModel);
		this.usedList.setCellRenderer(new IParameterCellRenderer());
		this.usedList.setSelectionMode(
				ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
		this.usedList.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent e) {
				if (e.getClickCount() == 2) {
					VariantsDialog.this.deselectUsed(
							VariantsDialog.this.usedList.getSelectedValue());
				}
			}
		});
		this.usedScroll = new JScrollPane(this.usedList);
		this.usedScroll.setPreferredSize(new Dimension(300, 400));
		this.useSelection.add(this.usedScroll);

		layout.setHorizontalGroup(
				layout.createSequentialGroup().addComponent(this.unusedScroll)
						.addGroup(layout.createParallelGroup()
								.addComponent(this.selectUsed)
								.addComponent(this.deselectUsed))
						.addComponent(this.usedScroll));
		layout.setVerticalGroup(
				layout.createParallelGroup(GroupLayout.Alignment.CENTER)
						.addComponent(this.unusedScroll)
						.addGroup(layout.createSequentialGroup()
								.addComponent(this.selectUsed)
								.addComponent(this.deselectUsed))
						.addComponent(this.usedScroll));
		layout.linkSize(this.selectUsed, this.deselectUsed);

	}

	/**
	 * Initialize range panel
	 */
	private void initRange() {
		JPanel rangePanel, validPanel;
		JLabel validLabel;
		VariantsTableRenderer renderer;

		this.range = new JSplitPane(JSplitPane.VERTICAL_SPLIT);

		rangePanel = new JPanel();
		this.range.setLeftComponent(rangePanel);
		rangePanel.setLayout(new BorderLayout());

		this.rangeModel = new RangeTableModel();
		this.rangeTable = new JTable(this.rangeModel);
		rangePanel.add(new JScrollPane(this.rangeTable), BorderLayout.CENTER);

		this.rangeTable.getTableHeader().setReorderingAllowed(false);
		this.rangeTable.setCellSelectionEnabled(true);
		renderer = new VariantsTableRenderer();
		this.rangeTable.setDefaultRenderer(String.class, renderer);
		renderer = new VariantsTableRenderer();
		renderer.setHorizontalAlignment(SwingConstants.CENTER);
		this.rangeTable.setDefaultRenderer(Integer.class, renderer);

		this.rangeTable.getColumnModel().getColumn(0).setPreferredWidth(200);
		this.rangeTable.getColumnModel().getColumn(1).setPreferredWidth(50);
		this.rangeTable.getColumnModel().getColumn(2).setPreferredWidth(10);

		this.combinationCount = new JLabel(
				VariantsDialog.combinationString + "0");
		this.combinationCount.setBorder(new EmptyBorder(5, 5, 5, 5));
		rangePanel.add(this.combinationCount, BorderLayout.SOUTH);

		validPanel = new JPanel();
		this.range.setRightComponent(validPanel);
		validPanel.setLayout(new BorderLayout());

		validLabel = new JLabel("Validity Script:");
		validLabel.setBorder(new EmptyBorder(5, 5, 5, 5));
		validPanel.add(validLabel, BorderLayout.NORTH);
		validLabel.setToolTipText(
				"Script should return boolean value (true/false)");

		this.validText = new JTextArea();
		validPanel.add(new JScrollPane(this.validText), BorderLayout.CENTER);
		this.validText.addFocusListener(new FocusAdapter() {

			@Override
			public void focusLost(FocusEvent e) {
				try {
					VariantsDialog.this.variants.setValidityString(
							VariantsDialog.this.validText.getText());
				} catch (Exception e1) {
					System.err.println(e1.getMessage());
					e1.printStackTrace();
				}
				VariantsDialog.this.refresh();
			}
		});
		this.validText.setToolTipText(
				"Script should return boolean value (true/false)");
	}

	/**
	 * Initialize section panel
	 */
	private void initSelection() {
		VariantsTableRenderer renderer;

		this.selection = new JPanel();
		this.selection.setLayout(new BorderLayout());
		this.selectModel = new SelectTableModel();
		this.selectTable = new JTable(this.selectModel);
		this.selection.add(new JScrollPane(this.selectTable),
				BorderLayout.CENTER);

		this.selectTable.getTableHeader().setReorderingAllowed(false);
		this.selectTable.setCellSelectionEnabled(true);
		renderer = new VariantsTableRenderer();
		renderer.setHorizontalAlignment(SwingConstants.CENTER);
		this.selectTable.setDefaultRenderer(Number.class, renderer);
		renderer = new VariantsTableRenderer();
		renderer.setHorizontalAlignment(SwingConstants.RIGHT);
		this.selectTable.setDefaultRenderer(Double.class, renderer);

		this.validCount = new JLabel(VariantsDialog.validString + "0"
				+ VariantsDialog.selectedString + "0");
		this.selection.add(this.validCount, BorderLayout.SOUTH);
		this.validCount.setBorder(new EmptyBorder(10, 5, 5, 5));
	}

	/**
	 * Initialize progress panel
	 */
	private void initProgress() {
		JProgressBar progressBar;
		GroupLayout group;

		this.progress = new JPanel();
		group = new GroupLayout(this.progress);
		this.progress.setLayout(group);

		progressBar = new JProgressBar();
		progressBar.setMinimum(0);
		progressBar.setIndeterminate(true);
		progressBar.setStringPainted(false);
		this.progress.add(progressBar);

		group.setHorizontalGroup(
				group.createSequentialGroup().addComponent(progressBar));
		group.setVerticalGroup(group.createSequentialGroup()
				.addPreferredGap(ComponentPlacement.UNRELATED, 1, 10000)
				.addComponent(progressBar)
				.addPreferredGap(ComponentPlacement.UNRELATED, 1, 10000));
	}

	/**
	 * Select used parameter
	 * 
	 * @param param
	 *            Parameter
	 */
	private void selectUsed(IParameter param) {
		try {
			if (this.variants.selectUsed(param)) {
				this.unusedModel.fireRemoved();
				this.usedModel.fireAdded();
			}
		} catch (RemoteException e1) {
			System.err.println(e1.getMessage());
			e1.printStackTrace();
		}
	}

	/**
	 * Deselect used parameter
	 * 
	 * @param param
	 *            Parameter
	 */
	private void deselectUsed(IParameter param) {
		if (this.variants.deselectUsed(param)) {
			this.unusedModel.fireAdded();
			this.usedModel.fireRemoved();
		}
	}

	/**
	 * Initialize buttons
	 */
	private void initButtons() {

		this.state = new JLabel("Error message");
		this.state.setForeground(Color.RED);
		this.state.setBorder(new EmptyBorder(10, 10, 10, 10));

		this.load = new JButton("Load");
		this.add(this.load);
		this.load.setToolTipText("Load settings");
		this.load.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e) {
				JFileChooser chooser;
				File dir = null;

				chooser = new JFileChooser();
				if (VariantsDialog.this.settingFile != null) {
					dir = VariantsDialog.this.settingFile.getParentFile();
				} else {
					try {
						dir = VariantsDialog.this.project.getMopFile()
								.getParentFile();
					} catch (RemoteException e1) {
						System.err.println(e1.getMessage());
						e1.printStackTrace();
					}
				}
				chooser.setCurrentDirectory(dir);

				// chooser.addChoosableFileFilter(new
				// FileNameExtensionFilter(
				// "Project-Variants-file (*.dpd)", "dpd"));
				chooser.setFileFilter(new FileNameExtensionFilter(
						"Project-Variants-file (*.dpd)", "dpd"));

				chooser.setAcceptAllFileFilterUsed(false);
				chooser.setDialogTitle("Choose Variants-file");
				chooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
				chooser.setMultiSelectionEnabled(false);
				chooser.setFileHidingEnabled(true);

				if (chooser.showOpenDialog(
						VariantsDialog.this) == JFileChooser.APPROVE_OPTION) {

					VariantsDialog.this.settingFile = chooser.getSelectedFile();
					try {
						VariantsDialog.this.variants
								.loadSettings(VariantsDialog.this.settingFile);
						VariantsDialog.this.refreshLists();
						VariantsDialog.this.refresh();
					} catch (Exception e1) {
						System.err.println(e1.getMessage());
						e1.printStackTrace();
					}
				}
			}
		});

		this.save = new JButton("Save");
		this.add(this.save);
		this.save.setToolTipText("Save settings");
		this.save.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e) {
				JFileChooser chooser;
				File dir = null, file;

				chooser = new JFileChooser();
				if (VariantsDialog.this.settingFile != null) {
					dir = VariantsDialog.this.settingFile.getParentFile();
				} else {
					try {
						dir = VariantsDialog.this.project.getMopFile()
								.getParentFile();
					} catch (RemoteException e1) {
						System.err.println(e1.getMessage());
						e1.printStackTrace();
					}

				}
				chooser.setCurrentDirectory(dir);

				// chooser.addChoosableFileFilter(new
				// FileNameExtensionFilter(
				// "Project-Variants-file (*.dpd)", "dpd"));
				chooser.setFileFilter(new FileNameExtensionFilter(
						"Project-Variants-file (*.dpd)", "dpd"));

				chooser.setAcceptAllFileFilterUsed(false);
				chooser.setDialogTitle("Choose Variants-file");
				chooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
				chooser.setMultiSelectionEnabled(false);
				chooser.setFileHidingEnabled(true);

				if (chooser.showSaveDialog(
						VariantsDialog.this) == JFileChooser.APPROVE_OPTION) {

					file = chooser.getSelectedFile();
					if (!file.getName().toLowerCase().endsWith(".dpd")) {
						file = new File(file.getParentFile(),
								file.getName() + ".dpd");
					}
					VariantsDialog.this.settingFile = file;

					try {
						VariantsDialog.this.variants
								.storeSettings(VariantsDialog.this.settingFile);
					} catch (Exception e1) {
						System.err.println(e1.getMessage());
						e1.printStackTrace();
					}
				}

			}
		});

		this.append = new JButton("Append");
		this.add(this.append);
		this.append.setToolTipText("Append variants to project");
		this.append.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e) {
				// show progress bar
				VariantsDialog.this.cardIndex = 9;
				VariantsDialog.this.cards.show(VariantsDialog.this.settings,
						Integer.toString(VariantsDialog.this.cardIndex));
				VariantsDialog.this.refresh();

				new Thread() {
					@Override
					public void run() {
						try {
							VariantsDialog.this.variants.createVariants();
							VariantsDialog.this.result = VariantsDialog.this.project;
							VariantsDialog.this.setVisible(false);
							VariantsDialog.this.dispose();
						} catch (Exception e1) {
							System.err.println(e1.getMessage());
							e1.printStackTrace();
						}
					}
				}.start();
			}
		});

		this.create = new JButton("New");
		this.add(this.create);
		this.create.setToolTipText("Create new project with variants");
		this.create.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e) {
				final JProjectChooser chooser;

				chooser = new JProjectChooser(JProjectChooser.PROJECT_ONLY,
						((Project) VariantsDialog.this.project).getMopFile()
								.getParentFile());

				if (chooser.showSaveDialog(
						VariantsDialog.this) == JFileChooser.APPROVE_OPTION) {
					// show progress bar
					VariantsDialog.this.cardIndex = 9;
					VariantsDialog.this.cards.show(VariantsDialog.this.settings,
							Integer.toString(VariantsDialog.this.cardIndex));
					VariantsDialog.this.refresh();

					new Thread() {
						@Override
						public void run() {
							try {
								VariantsDialog.this.result = VariantsDialog.this.variants
										.createVariants(
												chooser.getSelectedFile());
								VariantsDialog.this.setVisible(false);
								VariantsDialog.this.dispose();
							} catch (Exception e1) {
								System.err.println(e1.getMessage());
								e1.printStackTrace();
							}
						}
					}.start();
				}
			}
		});

		this.cancel = new JButton("Cancel");
		this.add(this.cancel);
		this.cancel.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e) {
				VariantsDialog.this.setVisible(false);
				VariantsDialog.this.dispose();
			}
		});

		this.previous = new JButton("Previous");
		this.add(this.previous);
		this.previous.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e) {
				VariantsDialog.this.cardIndex--;
				VariantsDialog.this.cards.show(VariantsDialog.this.settings,
						Integer.toString(VariantsDialog.this.cardIndex));
				VariantsDialog.this.refresh();
			}
		});

		this.next = new JButton("Next");
		this.add(this.next);
		this.next.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e) {
				VariantsDialog.this.cardIndex++;
				VariantsDialog.this.cards.show(VariantsDialog.this.settings,
						Integer.toString(VariantsDialog.this.cardIndex));
				VariantsDialog.this.refresh();
			}
		});

	}

	/**
	 * Refresh unused/used-list
	 */
	private void refreshLists() {
		VariantsDialog.this.unusedModel.contentChanged();
		VariantsDialog.this.usedModel.contentChanged();
	}

	/**
	 * Refresh dialog
	 */
	private void refresh() {
		// used
		this.selectUsed.setEnabled(!this.unusedModel.isEmpty());
		this.deselectUsed.setEnabled(!this.usedModel.isEmpty());

		// range
		this.rangeModel.fireTableDataChanged();
		this.combinationCount.setText(VariantsDialog.combinationString
				+ this.variants.getCombinationSize() + "   "
				+ VariantsDialog.validString + this.variants.getValidSize());
		// valid
		this.validText.setText(this.variants.getValidityString());

		// selection
		this.selectModel.fireTableStructureChanged();
		this.selectModel.fireTableDataChanged();
		this.selectTable.getColumnModel().getColumn(0).setPreferredWidth(1);
		this.selectTable.getColumnModel()
				.getColumn(this.selectModel.getColumnCount() - 1)
				.setPreferredWidth(2);
		for (int k = 1; k < (this.selectModel.getColumnCount() - 1); k++) {
			this.selectTable.getColumnModel().getColumn(k)
					.setPreferredWidth(100);
		}
		this.validCount.setText(
				VariantsDialog.validString + this.variants.getValidSize()
						+ "    " + VariantsDialog.selectedString
						+ this.variants.getSelectedSize());

		// buttons
		this.previous.setEnabled((this.cardIndex > 0) && (this.cardIndex < 9));
		this.next.setEnabled(this.cardIndex < 2);

		this.create.setEnabled(
				(this.variants.getSelectedSize() > 0) && (this.cardIndex < 9));
		this.append.setEnabled(
				(this.variants.getSelectedSize() > 0) && (this.cardIndex < 9));
		this.cancel.setEnabled(this.cardIndex < 9);
		this.load.setEnabled(this.cardIndex < 9);
		this.save.setEnabled(this.cardIndex < 9);

		// error message
		if (this.variants.getUsedSize() == 0) {
			this.state.setText("No used parameters");
		} else if (this.variants.getCombinationSize() == 0) {
			this.state.setText("No possible combination");
		} else if (this.variants.getValidSize() == 0) {
			this.state.setText("No valid combination");
		} else if (this.variants.getSelectedSize() == 0) {
			this.state.setText("No selected combination");
		} else {
			this.state.setText("");
		}
	}
}
