/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9715 $
 * $Date: 2020-03-19 10:01:15 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.symspace.center.views.indices;

import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;

import javax.swing.BorderFactory;
import javax.swing.GroupLayout;
import javax.swing.ImageIcon;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.SwingConstants;

import at.lcm.bu21.magopt.structure.Project;
import at.lcm.symspace.center.icons.indices.IconIndexButton;
import at.lcm.symspace.center.views.CenterViewManager;
import magopt.gui.dataset.DataSetLayout;

/**
 * Displays a list of all Data Sets in Project.<br>
 * The Status of each Data Set is indicated by one of the colors Red (Error),
 * Orange (Incomplete) or Green (OK).<br>
 * Each Data Set gets an ToolTip String from the Data Set's Index Number &amp;
 * Index Comment.<br>
 * Each Data Set can be selected, which results in changing of the current
 * Index.<br>
 * An IndexListener can be registered to get the Selection Process.
 * 
 * @author Priglinger Markus
 * @author Werner Koppelsttter
 */
public class IndicesViewPanel extends JPanel implements MouseListener {

	private static final long serialVersionUID = 19700717201309102L;

	private CenterViewManager views = null;

	static final int GRIDBASE_X = 0;
	static final int GRIDBASE_Y = 3;
	static final int DATASET_LABEL_WIDTH = 130; // must be adapted
												// experimentally

	/**
	 * Data Set states:<br>
	 * - Error (red background)<br>
	 * - Incomplete (orange background)<br>
	 * - OK (green background)
	 * 
	 * @author Aigner Markus
	 * 
	 */
	enum IndexType {
		ERROR("Error", Color.RED), INCOMPLETE("Incomplete", Color.ORANGE), OK("OK", Color.GREEN);

		private String mLabel;
		private Color mColor;

		private IndexType(String label, Color color) {
			this.mLabel = label;
			this.mColor = color;
		}

		protected String getLabel() {
			return (this.mLabel);
		}

		protected Color getColor() {
			return (this.mColor);
		}
	}

	private Project _project;

	// for the actual index view
	private JPanel _dataSet;
	private int _indicesPerLine = 10;
	private int _insetAfterNumIndices = 5;

	private JLabel lbl_idxText = null;
	private JLabel lbl_commentText = null;

	private DnDButton[] indizes = null;

	public IndicesViewPanel(CenterViewManager views) {
		this();
		this.views = views;
	}

	public IndicesViewPanel() {
		JLabel lError, lIncomplete, lOk, lbl_idx, lbl_comment;
		GroupLayout layout;
		FlowLayout indicesLayout;
		JScrollPane sp;

		layout = new GroupLayout(this);
		layout.setAutoCreateGaps(true);
		this.setLayout(layout);
		this.setBorder(BorderFactory.createEmptyBorder(5, 5, 5, 5));

		this._dataSet = new JPanel();
		indicesLayout = new DataSetLayout(FlowLayout.LEFT, 1, 1);
		this._dataSet.setLayout(indicesLayout);

		sp = new JScrollPane(this._dataSet);

		/* legend */
		lError = new JLabel(IndexType.ERROR.getLabel());
		lError.setOpaque(true);
		lError.setBorder(BorderFactory.createEmptyBorder(1, 1, 1, 1));
		lError.setHorizontalAlignment(SwingConstants.CENTER);
		lError.setBackground(IndexType.ERROR.getColor());

		lIncomplete = new JLabel(IndexType.INCOMPLETE.getLabel());
		lIncomplete.setOpaque(true);
		lIncomplete.setBorder(BorderFactory.createEmptyBorder(1, 1, 1, 1));
		lIncomplete.setHorizontalAlignment(SwingConstants.CENTER);
		lIncomplete.setBackground(IndexType.INCOMPLETE.getColor());

		lOk = new JLabel(IndexType.OK.getLabel());
		lOk.setOpaque(true);
		lOk.setBorder(BorderFactory.createEmptyBorder(1, 1, 1, 1));
		lOk.setHorizontalAlignment(SwingConstants.CENTER);
		lOk.setBackground(IndexType.OK.getColor());

		lbl_idx = new JLabel("Active index: ");
		this.lbl_idxText = new JLabel("-");
		lbl_comment = new JLabel(" Comment: ");
		this.lbl_commentText = new JLabel("-");

		layout.setHorizontalGroup(layout.createSequentialGroup()
				.addGroup(layout.createParallelGroup()
						.addGroup(layout.createSequentialGroup().addComponent(lbl_idx).addComponent(this.lbl_idxText)
								.addComponent(lbl_comment).addComponent(this.lbl_commentText))
						.addComponent(sp).addGroup(layout.createSequentialGroup().addComponent(lOk)
								.addComponent(lIncomplete).addComponent(lError))));

		layout.setVerticalGroup(layout.createSequentialGroup()
				.addGroup(layout.createParallelGroup(GroupLayout.Alignment.BASELINE).addComponent(lbl_idx)
						.addComponent(this.lbl_idxText).addComponent(lbl_comment).addComponent(this.lbl_commentText))
				.addGroup(layout.createParallelGroup(GroupLayout.Alignment.BASELINE).addComponent(sp))
				.addGroup(layout.createParallelGroup(GroupLayout.Alignment.BASELINE).addComponent(lOk)
						.addComponent(lIncomplete).addComponent(lError)));
		layout.linkSize(lOk, lIncomplete, lError);

		this.installResizeListener();
	}

	private void installResizeListener() {

		this.addComponentListener(new ComponentAdapter() {
			@Override
			public void componentResized(ComponentEvent e) {
				// This is only called when the user releases the mouse button.
				Component comp = e.getComponent();

				float rawWidth = (comp.getSize().width - IndicesViewPanel.DATASET_LABEL_WIDTH) / 16.85f;
				// align to
				IndicesViewPanel.this._indicesPerLine = ((int) rawWidth / IndicesViewPanel.this._insetAfterNumIndices)
						* IndicesViewPanel.this._insetAfterNumIndices;

				if (IndicesViewPanel.this._indicesPerLine == 0) {
					IndicesViewPanel.this._indicesPerLine = IndicesViewPanel.this._insetAfterNumIndices;
				}

				IndicesViewPanel.this.refresh();
			}
		});
	}

	public void setProject(Project project) {
		this._project = project;
		this.refresh();
	}

	/**
	 * Refresh description of current dataset
	 */
	private void refreshText() {
		if (this._project == null) {
			this.lbl_idxText.setText("");
			this.lbl_commentText.setText("");
		} else {
			this.lbl_idxText.setText(Integer.toString(this._project.getIndex() + 1));
			this.lbl_commentText.setText(this._project.getDSComment());
		}
	}

	/**
	 * Refresh color and selection of dataset
	 * 
	 * @param index
	 *            Dataset index
	 */
	private void refreshIndex(int index) {
		// set tooltip
		String comment = this._project.getDSComment(index);
		String txt = "Index #" + String.valueOf(index + 1);
		if (!comment.isEmpty()) {
			txt += " (" + comment + ")";
		}
		this.indizes[index].setToolTipText(txt);

		// set color by setting icons
		if (this._project.getError(index)) {
			this.indizes[index].setIcon(new ImageIcon(IconIndexButton.IDXBUTTON_ERROR));
			this.indizes[index].setPressedIcon(new ImageIcon(IconIndexButton.IDXBUTTON_PRESSED_ERROR));
			this.indizes[index].setRolloverIcon(new ImageIcon(IconIndexButton.IDXBUTTON_ROLLOVER_ERROR));

		} else if (this._project.getIncomplete(index)) {
			this.indizes[index].setIcon(new ImageIcon(IconIndexButton.IDXBUTTON_INCOMPLETE));
			this.indizes[index].setPressedIcon(new ImageIcon(IconIndexButton.IDXBUTTON_PRESSED_INCOMPLETE));
			this.indizes[index].setRolloverIcon(new ImageIcon(IconIndexButton.IDXBUTTON_ROLLOVER_INCOMPLETE));
		} else {
			this.indizes[index].setIcon(new ImageIcon(IconIndexButton.IDXBUTTON_OK));
			this.indizes[index].setPressedIcon(new ImageIcon(IconIndexButton.IDXBUTTON_PRESSED_OK));
			this.indizes[index].setRolloverIcon(new ImageIcon(IconIndexButton.IDXBUTTON_ROLLOVER_OK));
		}

		if (index == this._project.getIndex()) {
			this.indizes[index].setBorder(BorderFactory.createLineBorder(Color.BLACK, 2));
		} else {
			this.indizes[index].setBorder(BorderFactory.createEmptyBorder());
		}
	}

	/**
	 * Redraw the all datasets
	 */
	public void refresh() {
		this._dataSet.removeAll();
		if (this._project != null) {

			// Layout definition
			int numOfIndices = this._project.getLength();
			int cols = IndicesViewPanel.GRIDBASE_X + this._indicesPerLine + 1;
			int rows = IndicesViewPanel.GRIDBASE_Y + (numOfIndices / this._indicesPerLine) + 1;

			GridBagLayout gridBagLayout = new GridBagLayout();

			gridBagLayout.columnWidths = new int[cols + 1]; // all 0
			gridBagLayout.rowHeights = new int[rows + 1]; // all 0
			gridBagLayout.columnWeights = new double[cols + 1]; // all 0.0, last
																// is
																// Double.MIN_VALUE
			gridBagLayout.rowWeights = new double[rows + 1]; // all 0.0, last is
																// Double.MIN_VALUE
			for (int i = 0; i < cols; i++) {
				gridBagLayout.columnWidths[i] = 0;
				gridBagLayout.columnWeights[i] = 0.0;
			}
			gridBagLayout.columnWeights[cols] = Double.MIN_VALUE;

			for (int i = 0; i < rows; i++) {
				gridBagLayout.rowHeights[i] = 0;
				gridBagLayout.rowWeights[i] = 0.0;
			}
			gridBagLayout.rowWeights[rows] = Double.MIN_VALUE;

			this._dataSet.setLayout(gridBagLayout);

			GridBagConstraints gbc = null;
			JLabel labelIndexNumbers = null;

			// add dummy label to apply insets
			labelIndexNumbers = new JLabel("");
			gbc = new GridBagConstraints();
			gbc.insets = new Insets(20, 0, 0, 0);
			gbc.gridx = 0;
			gbc.gridy = 0;

			this._dataSet.add(labelIndexNumbers, gbc);

			int insetCount = 0;
			int rowThreshold = this._indicesPerLine;
			int rowCount = -1;
			int colCount = 1;

			// draw indices
			this.indizes = new DnDButton[numOfIndices];
			for (int i = 0; i < numOfIndices; i++) {

				// row label
				if (rowThreshold == this._indicesPerLine) {
					rowCount++;
					labelIndexNumbers = new JLabel();
					String textSeg = "";

					if (i < (numOfIndices - 1)) {
						if ((i + this._indicesPerLine) >= numOfIndices) {
							textSeg = " - " + numOfIndices;
						} else {
							textSeg = " - " + Integer.toString(i + this._indicesPerLine);
						}
					}
					labelIndexNumbers.setText((i + 1) + textSeg);

					gbc = new GridBagConstraints();
					gbc.anchor = GridBagConstraints.EAST;
					gbc.insets = new Insets(0, 20, 0, 20);
					gbc.gridx = IndicesViewPanel.GRIDBASE_X;
					gbc.gridy = IndicesViewPanel.GRIDBASE_Y + rowCount;

					this._dataSet.add(labelIndexNumbers, gbc);
					rowThreshold = 0;
					insetCount = 0;
				}
				rowThreshold++;

				// create button and insert
				this.indizes[i] = new DnDButton();
				this.indizes[i].setIndex(i + 1);
				this.indizes[i].setPreferredSize(new Dimension(15, 15));
				this.indizes[i].addMouseListener(this);
				this.indizes[i].setBackground(null);
				this.indizes[i].setForeground(null);

				if (colCount > this._indicesPerLine) {
					colCount = 1;
				}

				int insetRight = 1;

				insetCount++;
				if (insetCount >= this._insetAfterNumIndices) {
					insetCount = 0;
					insetRight = 5;
				}

				Insets buttonInsets = new Insets(0, 0, 5, insetRight);
				gbc = new GridBagConstraints();
				gbc.anchor = GridBagConstraints.CENTER;
				gbc.insets = buttonInsets;
				gbc.gridx = IndicesViewPanel.GRIDBASE_X + colCount;
				gbc.gridy = IndicesViewPanel.GRIDBASE_Y + rowCount;

				this._dataSet.add(this.indizes[i], gbc);

				colCount++;
			}

			// refresh indices
			for (int k = 0; k < this._project.getLength(); k++) {
				if (k < this.indizes.length) {
					this.refreshIndex(k);
				}
			}
		}

		this.refreshText();

		this.revalidate();
		this.repaint();
	}

	// mouse listener
	@Override
	public void mouseClicked(MouseEvent me) {
		this.views.getController().setIndex(((DnDButton) me.getSource()).getIndex() - 1);
	}

	@Override
	public void mouseEntered(MouseEvent me) {
	}

	@Override
	public void mouseExited(MouseEvent me) {
	}

	@Override
	public void mousePressed(MouseEvent me) {
	}

	@Override
	public void mouseReleased(MouseEvent me) {
	}
}
