/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9715 $
 * $Date: 2020-03-19 10:01:15 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.symspace.gui;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.GraphicsEnvironment;
import java.awt.Image;
import java.awt.Insets;
import java.awt.Rectangle;
import java.awt.Toolkit;
import java.awt.image.BufferedImage;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;

import javax.imageio.ImageIO;
import javax.swing.JDialog;
import javax.swing.JPanel;
import javax.swing.JProgressBar;

/**
 * Splash-screen while loading gui
 *
 * @author Werner Koppelsttter
 */
public class Splash extends JDialog {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** Width of dialog */
	final protected int width = 800;
	/** Height of dialog */
	final protected int height = 600;

	/**
	 * Create splash-screen
	 */
	public Splash() {
		this.init();
		// progress-bar
		JProgressBar bar = new JProgressBar();
		bar.setIndeterminate(true);
		bar.setStringPainted(true);
		bar.setString("Loading ...");
		this.add(bar, BorderLayout.SOUTH);
	}

	/**
	 * Initialize panel
	 */
	protected void init() {
		this.setLayout(new BorderLayout());
		this.setUndecorated(true);

		// screen insets
		Insets insets = Toolkit.getDefaultToolkit().getScreenInsets(
				GraphicsEnvironment.getLocalGraphicsEnvironment().getDefaultScreenDevice().getDefaultConfiguration());
		// screen area
		Rectangle rect = GraphicsEnvironment.getLocalGraphicsEnvironment().getDefaultScreenDevice()
				.getDefaultConfiguration().getBounds();
		// working area
		Rectangle work = new Rectangle(rect.x + insets.left, rect.y + insets.top,
				rect.width - insets.left - insets.right, rect.height - insets.top - insets.bottom);
		// size and location
		this.setLocation(work.x + ((work.width - this.width) / 2), work.y + ((work.height - this.height) / 2));
		this.setSize(this.width, this.height);

		// draw image
		try {
			ImagePanel plot = new ImagePanel(ImageIO.read(this.getClass().getResource(
					"/" + this.getClass().getPackage().getName().replace('.', '/') + "/" + "Splashscreen.jpg")));
			this.add(plot, BorderLayout.CENTER);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	/**
	 * Panel to draw image
	 *
	 * @author Werner Koppelsttter
	 */
	private class ImagePanel extends JPanel {

		/** Serialization version number */
		private static final long serialVersionUID = 1L;

		/** Image to display */
		private BufferedImage image;

		/**
		 * Create image-panel
		 * 
		 * @param image
		 *            Image to draw
		 */
		public ImagePanel(BufferedImage image) {
			this.image = image;
		}

		@Override
		protected void paintComponent(Graphics g) {
			Graphics2D g2d = (Graphics2D) g;

			int width = ImagePanel.this.getWidth();
			int height = ImagePanel.this.getHeight();

			if ((width != 0) && (height != 0)) {

				// scale factor
				double scaleX = ((double) width) / this.image.getWidth();
				double scaleY = ((double) height) / this.image.getHeight();
				double scale = scaleX < scaleY ? scaleX : scaleY;

				// scale image
				BufferedImage scaled = new BufferedImage((int) (this.image.getWidth() * scale),
						(int) (this.image.getHeight() * scale), BufferedImage.TYPE_INT_ARGB);
				Graphics2D g2 = scaled.createGraphics();
				g2.drawImage(this.image.getScaledInstance(scaled.getWidth(), scaled.getHeight(), Image.SCALE_SMOOTH), 0,
						0, null);
				g2.dispose();

				// draw background
				// g2d.setColor(Color.white);
				g2d.setColor(new Color(0, 159, 227));
				g2d.fillRect(0, 0, Splash.this.width, Splash.this.height);

				// draw scaled image
				// g2d.drawImage(scaled, null, (width - scaled.getWidth(null)) /
				// 2,
				// (height - scaled.getHeight(null)) / 2);
				g2d.drawImage(scaled, null, 0, 0);

//				// FIXME remove prerelease
//				Color col = Color.GRAY.brighter();
//				col = new Color(col.getRed(), col.getGreen(), col.getBlue(),
//						200);
//				g2d.setColor(col);
//				// g2d.setColor(Color.RED);
//				Font font = g2d.getFont();
//				font = font.deriveFont(height / 4f);
//				g2d.setFont(font);
//				g2d.translate(width / 8, (height * 8) / 10);
//				g2d.rotate(-Math.atan(
//						((double) scaled.getHeight()) / scaled.getWidth()));
//				g2d.drawString("Prerelease", 0, 0);

				// get version
				String ver = null;
				File version = new File(
						new File(this.getClass().getProtectionDomain().getCodeSource().getLocation().getFile())
								.getParentFile().getParentFile(),
						"Version.txt");
				if (version.exists()) {
					try {
						BufferedReader in = new BufferedReader(new FileReader(version));
						ver = in.readLine();
						in.close();
					} catch (FileNotFoundException e) {
					} catch (NumberFormatException e) {
					} catch (IOException e) {
					}
				}
				if ((ver != null) && !ver.isEmpty()) {
					ver = ver.trim();
				} else {
					ver = "0.0";
				}

				// get revision
				File revision = new File(
						new File(this.getClass().getProtectionDomain().getCodeSource().getLocation().getFile())
								.getParentFile().getParentFile().getParentFile(),
						"Revision.txt");
				String rev = null;
				String build = null;
				if (revision.exists()) {
					try {
						BufferedReader in = new BufferedReader(new FileReader(revision));
						rev = in.readLine();
						build = in.readLine();
						in.close();
					} catch (FileNotFoundException e) {
					} catch (NumberFormatException e) {
					} catch (IOException e) {
					}
				}
				if ((rev != null) && !rev.isEmpty()) {
					rev = rev.trim();
				} else {
					rev = "-";
				}
				if ((build != null) && !build.isEmpty()) {
					rev += " (" + build.trim() + " build)";
				} else {
					rev += " (nightly build)";
				}

				float fontSize = height / 35f;

				Color col = Color.WHITE;
				g2d.setColor(col);
				Font font = g2d.getFont();
				Font normal = font.deriveFont(fontSize);
				Font bold = normal.deriveFont(Font.BOLD);
				g2d.translate(width / 8, (height * 5.8) / 10);
				g2d.setFont(bold);
				g2d.drawString("SyMSpace", 0, 0);
				g2d.setFont(normal);
				g2d.drawString("SyMSpace orchestrates your favorite engineering tools to cover", 0, fontSize * 1.5f);
				g2d.drawString("all necessary modeling and calculation steps in one workflow.", 0, fontSize * 3f);

				g2d.setFont(normal);
				g2d.drawString("Version:", 0, fontSize * 5.5f);
				g2d.setFont(bold);
				g2d.drawString(ver, fontSize * 5f, fontSize * 5.5f);
				g2d.setFont(normal);
				g2d.drawString("Revision:", 0, fontSize * 7f);
				g2d.setFont(bold);
				g2d.drawString(rev, fontSize * 5f, fontSize * 7f);

				g2d.setFont(normal);
				g2d.drawString("Copyright (C) Linz Center of Mechatronics GmbH, Linz, Austria", 0, fontSize * 9.5f);
				g2d.drawString("www.symspace.at", 0, fontSize * 11f);
				g2d.drawString("symspace@lcm.at", 0, fontSize * 12.5f);

			}

		};
	};
}
