/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9767 $
 * $Date: 2020-04-20 20:04:00 +0200 (Mon, 20 Apr 2020) $
 */
package at.lcm.bu21.magopt.docu;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.security.NoSuchAlgorithmException;
import java.util.LinkedList;

import at.lcm.bu21.general.utils.FileUtils;
import at.lcm.bu21.general.utils.TexWriter;
import at.lcm.bu21.magopt.structure.ParamJavaFnc;
import at.lcm.bu21.magopt.structure.PortDesc;
import at.lcm.bu21.magopt.structure.Project;
import at.lcm.bu21.magopt.utils.MatlabControl;

/**
 * Create documentation of functions
 * 
 * @author Werner Koppelsttter
 */
public class ExtractFunctionDocu extends Thread {

	/**
	 * Document-structure for creating a function documentation
	 * 
	 * @author Werner Koppelsttter
	 */
	class DocInfo {
		/** Name of function */
		public String name = "";
		/** Documentation string of function */
		public String[] header = new String[0];
		/** Description of function */
		public String description = "";
		/** Imports of function */
		public PortDesc[] imports = null;
		/** Exports of function */
		public PortDesc[] exports = null;
	}

	/** Search path */
	File file = null;

	/** List of files to generate */
	LinkedList<File> files = new LinkedList<File>();

	/**
	 * Create documentation of functions
	 * 
	 * @param file
	 *            Location for generate. Possible values are function, toolbox or
	 *            null for whole library
	 */
	public ExtractFunctionDocu(String file) {
		if (file != null) {
			this.file = new File(file);
		}
		this.start();
	}

	/**
	 * List files in Library
	 * 
	 * @param file
	 *            File to list funcitons
	 */
	private void listLibraries(File file) {
		String name;

		if (file == null) {
			this.listToolboxes(new File(Project.getLibraryPath()));
		} else if (file.isDirectory()) {
			this.listFiles(file);
		} else {
			name = file.getName().toLowerCase();
			if (name.endsWith(".java")) {
				this.checkJavaFile(file);
			} else if (name.endsWith(".m")) {
				this.checkMatlabFile(file);
			}
		}
	}

	/**
	 * List toolboxes
	 * 
	 * @param file
	 *            Directory to check
	 */
	private void listToolboxes(File file) {
		for (File f : file.listFiles()) {
			if (f.isDirectory()) {
				if (f.getName().startsWith("_")) {
					this.listToolboxes(f);
				} else {
					this.listFiles(f);
				}
			}
		}
	}

	/**
	 * List files of toolboxes
	 * 
	 * @param file
	 */
	private void listFiles(File file) {
		File func, dir;
		func = new File(file, "Functions");
		if (func.exists()) {
			dir = new File(new File(func, "Java"), "src");
			if (dir.exists()) {
				this.listJavaFiles(dir);
			}
			dir = new File(func, "Matlab");
			if (dir.exists()) {
				this.listMatlabFiles(dir);
			}
		}
	}

	/**
	 * List java files and subfiles
	 * 
	 * @param file
	 *            Directories or file to add
	 */
	private void listJavaFiles(File file) {
		for (File f : file.listFiles()) {
			if (f.isDirectory()) {
				this.listJavaFiles(f);
			} else {
				if (f.getName().toLowerCase().endsWith(".java")) {
					this.checkJavaFile(f);
				}
			}
		}
	}

	/**
	 * Check and add ParamJavaFnc
	 * 
	 * @param file
	 *            File to add
	 */
	private void checkJavaFile(File file) {
		Object obj;
		String name;
		File path;

		try {
			// check if class is ParamJavaFnc
			name = file.getName();
			name = name.substring(0, name.length() - 5);
			path = file.getParentFile();
			while (path.getName().compareToIgnoreCase("src") != 0) {
				name = path.getName() + "." + name;
				path = path.getParentFile();
			}
			obj = Class.forName(name).newInstance();
			if (obj instanceof ParamJavaFnc) {
				this.files.add(file);
			}
		} catch (Throwable e) {
		}
	}

	/**
	 * List matlab files and subfiles
	 * 
	 * @param file
	 *            Directories or file to add
	 */
	private void listMatlabFiles(File file) {
		File mf;

		for (File f : file.listFiles()) {
			if (f.isDirectory()) {
				if (f.getName().startsWith("+")) {
					this.listMatlabFiles(f);
				} else if (f.getName().startsWith("@")) {
					mf = new File(f, f.getName().substring(1) + ".m");
					if (mf.exists()) {
						this.checkMatlabFile(mf);
					}
				}
			} else {
				if (f.getName().toLowerCase().endsWith(".m")) {
					this.checkMatlabFile(f);
				}
			}
		}
	}

	/**
	 * Check and add ParamMatlabFnc
	 * 
	 * @param file
	 *            File to add
	 */
	private void checkMatlabFile(File file) {
		String name;
		File path;

		name = file.getName();
		name = name.substring(0, name.length() - 2);
		path = file.getParentFile();
		if (path.getName().startsWith("@")) {
			path = path.getParentFile();
		}
		while (path.getName().startsWith("+")) {
			name = path.getName().substring(1) + "." + name;
			path = path.getParentFile();
		}
		try {
			if ((Boolean) MatlabControl.getProxy(path.getPath()).returningFeval("GetDocInfo", 1,
					new Object[] { name, null })[0]) {
				this.files.add(file);
			}
		} catch (Throwable e) {
		}

	}

	@Override
	public void run() {
		System.out.println("Start list files");
		this.listLibraries(this.file);

		System.out.println("Start extract docu (" + this.files.size() + ")");
		for (File f : this.files) {
			this.extract(f);
		}
		System.out.println("End extract docu");
	}

	/**
	 * Create functions documentation from document-structure
	 * 
	 * @param content
	 *            TexWriter
	 * @param doc
	 *            Document-structure
	 */
	private void addContent(TexWriter content, DocInfo doc) {
		int indexOf[], index;
		String name;

		if ((doc.header != null) && (doc.header.length > 0)) {
			//
			content.add("\n\\section*{");
			content.addString(doc.description);
			content.add(": \\moparam{");
			content.addString(doc.name);
			content.add("}}\n");
			content.add("\\label{");
			content.addName(doc.name);
			content.add(".Sec:heading}\n");
			content.add("\\ifdefined\\usehtml\\else\n");
			content.add("\\addcontentsline{toc}{subsction}{");
			content.addString(doc.name);
			content.add("}");
			content.add("\\fi\n\n");

			for (String str : doc.header) {

				if (str.contains(" *DESCRIPTION* ")) {
					content.add("\n\\subsection*{Description}\n\\label{");
					content.addName(doc.name);
					content.add(".Sec:description}\n\n");
					str = "";
				}

				if (str.contains(" *IMPORTS* ")) {
					content.add("\n\\subsection*{Imports}\n\\label{");
					content.addName(doc.name);
					content.add(".Sec:imports}\n\n");
					str = "";
				}

				if (str.contains(" *EXPORTS* ")) {
					content.add("\n\\subsection*{Exports}\n\\label{");
					content.addName(doc.name);
					content.add(".Sec:exports}\n\n");
					str = "";
				}

				if (str.contains(" *PREVIEW* ")) {
					content.add("\n\\subsection*{Preview}\n\\label{");
					content.addName(doc.name);
					content.add(".Sec:preview}\n\n");
					str = "";
				}

				while (str.contains("\\importtable") || str.contains("\\exporttable") || str.contains("\\input{")) {
					indexOf = new int[3];
					indexOf[0] = str.indexOf("\\importtable");
					indexOf[1] = str.indexOf("\\exporttable");
					indexOf[2] = str.indexOf("\\input{");
					index = -1;
					for (int k = 0; k < indexOf.length; k++) {
						if (indexOf[k] != -1) {
							if ((index == -1) || (indexOf[k] < indexOf[index])) {
								index = k;
							}
						}
					}
					content.add(str.substring(0, indexOf[index]));
					str = str.substring(indexOf[index]);

					switch (index) {
					case 0:
						// importtable
						str = str.substring(12);
						this.addTable(content, doc.imports, "Imports");
						break;
					case 1:
						// exporttable
						str = str.substring(12);
						this.addTable(content, doc.exports, "Exports");
						break;
					case 2:
						// input
						str = str.substring(7);
						index = str.indexOf("}");
						name = str.substring(0, index);
						str = str.substring(index + 1);

						content.add("\\vspace{1ex}");
						content.add("\\input{\\MoHomePath/");
						content.add((FileUtils.getRelativePath(Project.getInstallPath(), content.getDirectory()) + "/"
								+ name).replace('\\', '/'));
						content.add("}");
						content.add("\\vspace{1ex}");
						break;
					}
				}
				content.add(str + "\n");
			}
		}
	}

	/**
	 * Add Table with depends
	 * 
	 * @param content
	 *            TexWriter
	 * @param depends
	 *            Depends to write
	 * @param table
	 *            Name of table
	 */
	private void addTable(TexWriter content, PortDesc[] depends, String table) {

		// begin table
		content.add("\\vspace{1ex}");
		content.add("\\ifdefined\\usehtml");
		content.add(
				"\\begin{tabular}{|>{\\raggedright}p{3cm}|>{\\centering}p{1.5cm}|>{\\raggedright}p{7cm}|p{0.2cm}|}");
		content.add("\\else");
		content.add(
				"\\begin{longtable}{|>{\\raggedright}p{3cm}|>{\\centering}p{1.5cm}|>{\\raggedright}p{7cm}|p{0.2cm}|}");
		content.add("\\fi");
		content.add("\\hline");
		content.add(
				"\\textbf{Variable} & \\textbf{Unit} & \\textbf{Description}& \\hspace*{-0.15cm}\\textbf{(*)}\\tabularnewline");
		content.add("\\hline\\hline");

		// table content
		if (depends != null) {
			for (PortDesc depend : depends) {
				content.add("\\moparam{");
				content.addString(depend.getName());
				content.add("} & ");
				content.addString(depend.getUnit());
				content.add(" & ");
				content.add(depend.getDescription());
				content.add(" & ");
				if (depend.getOptional()) {
					content.add("\\ding{51}");
				}
				content.add(" \\tabularnewline");
				content.add(" \\hline");
				// structure
				if (depend.getStructure() != null) {
					content.add(" \\multicolumn{4}{|p{13.7cm}|}{");
					content.add(" \\vspace{.2ex}");

					this.addStructure(content, depend, 0, false);

					content.add(" \\vspace{.5ex}");
					content.add("} \\tabularnewline \\hline");
				}
			}
		}

		// end table
		content.add(
				"\\multicolumn{4}{r}{\\footnotesize \\textbf{(*)} " + table + " marked with \\ding{51} are optional.}");
		content.add("\\ifdefined\\usehtml");
		content.add("\\end{tabular}");
		content.add("\\else");
		content.add("\\end{longtable}");
		content.add("\\fi");
		content.add("\\vspace{1ex}");

	}

	/**
	 * Add Structure to table
	 * 
	 * @param content
	 *            TexWriter
	 * @param depend
	 *            Dependency to add
	 * @param level
	 *            Level to add
	 * @param optional
	 *            Add optional-flag
	 */
	private void addStructure(TexWriter content, PortDesc depend, int level, boolean optional) {

		content.add(" \\structitem{" + level + "}{");
		content.addString(depend.getName());
		content.add("}{ ");
		if (level != 0) {
			content.add(depend.getDescription());
		}
		content.add(" }");

		if (depend.getStructure() != null) {
			for (PortDesc dep : depend.getStructure()) {
				this.addStructure(content, dep, level + 1, optional);
			}
		}

	}

	/**
	 * Extract documentation from source-file
	 * 
	 * @param file
	 *            File to generate help
	 */
	void extract(File file) {
		DocInfo doc = null;
		String name;
		File path, tex;

		if (file.exists()) {
			name = file.getName();
			path = file.getParentFile();

			if (file.getName().toLowerCase().endsWith(".m")) {
				name = name.substring(0, name.length() - 2);
				if (path.getName().startsWith("@")) {
					path = path.getParentFile();
				}
				while (path.getName().startsWith("+")) {
					name = path.getName().substring(1) + "." + name;
					path = path.getParentFile();
				}
				doc = this.extractMatlabDoc(file, path.getPath(), name);
			} else if (file.getName().toLowerCase().endsWith(".java")) {
				name = name.substring(0, name.length() - 5);
				while (path.getName().compareToIgnoreCase("src") != 0) {
					name = path.getName() + "." + name;
					path = path.getParentFile();
				}
				path = path.getParentFile();
				doc = this.extractJavaDoc(file, name);
			}
			System.out.println("\t" + name + " (" + path.getPath() + ")");

			tex = FileUtils.getFile(path.getPath(), "tex", name.replace('.', File.separatorChar) + "_Doc",
					name.substring(name.lastIndexOf('.') + 1) + ".tex");

			if (doc != null) {
				try {
					if (this.generateTex(tex, doc)) {
						System.out.println("\t\tDocu gernerated");
					}
				} catch (NoSuchAlgorithmException e) {
				} catch (IOException e) {
				}
			}

			// Help
			if (ExtractFunctionDocu.generateHtml(tex)) {
				System.out.println("\t\tHelp generated");
			}
			if (ExtractFunctionDocu
					.generateHtml(new File(tex.getPath().substring(0, tex.getPath().length() - 4) + ".en.tex"))) {
				System.out.println("\t\tHelp (en) generated");
			}
			if (ExtractFunctionDocu
					.generateHtml(new File(tex.getPath().substring(0, tex.getPath().length() - 4) + ".ger.tex"))) {
				System.out.println("\t\tHelp (ger) generaterd");
			}
		} else {
			System.out.println("\t\tFile does not exist");
		}
	}

	/**
	 * Create tex-help from DocInfo
	 * 
	 * @param tex
	 *            Tex-file to export
	 * @param doc
	 *            Info to write
	 * @return Help created
	 * @throws IOException
	 * @throws NoSuchAlgorithmException
	 */
	private boolean generateTex(File tex, DocInfo doc) throws IOException, NoSuchAlgorithmException {
		TexWriter content;
		String fileChecksum, contentChecksum;
		boolean write = false;

		content = new TexWriter(tex);
		this.addContent(content, doc);

		if (content.getFile().exists()) {
			fileChecksum = FileUtils.checkSum(content.getFile());
			contentChecksum = FileUtils.getCheckSum(content.toString());
			if (fileChecksum.compareTo(contentChecksum) != 0) {
				write = true;
			}
		} else {
			if (content.toString().length() > 0) {
				write = true;
			}
		}

		if (write) {
			content.write();
		}

		return write;
	}

	/**
	 * Create html-help from tex-file
	 * 
	 * @param source
	 *            Tex-source
	 * @return Help created
	 */
	public static boolean generateHtml(String source) {
		return ExtractFunctionDocu.generateHtml(new File(source));
	}

	/**
	 * Create html-help from tex-file
	 * 
	 * @param source
	 *            Tex-source
	 * @return Help created
	 */
	public static boolean generateHtml(File source) {
		File html;
		TexWriter content;
		ProcessBuilder pb;
		Process p;
		String name;
		boolean valid = false;

		if (source.exists()) {
			name = source.getName();
			name = name.substring(0, name.length() - 4);

			// write Frame tex-file
			content = new TexWriter(new File(source.getParentFile(), name + "_.tex"));
			content.add("\\documentclass[12pt]{article}\n\n");
			content.add("\\newcommand{\\MoHomePath}{");
			content.add(FileUtils.getRelativePath(source.getParent(), Project.getInstallPath()).replace('\\', '/'));
			content.add("}\n\n");
			content.add("\\newcommand{\\MagOptDoku}{1}\n\n");
			content.add("\\newcommand{\\usehtml}{1}\n\n");
			content.add("\\usepackage{\\MoHomePath/Documentation/General/Styles/generalDoc}\n\n");
			content.add("\\begin{document}\n");
			content.add("\\selectlanguage{american}\n\n");
			content.add("\\input{");
			content.add(name);
			content.add("}\n\n");
			content.add("\\bibliography{\\MoHomePath/Documentation/Bibliography/MagOptLiteratur}\n\n");
			content.add("\\end{document}");

			// convert to html-help
			try {
				content.write();
				FileUtils.copy(FileUtils.getFile(Project.getInstallPath(), "Documentation", "General", "Styles",
						"htlatexconfig.cfg"), new File(source.getParentFile(), "htlatexconfig.cfg"), true);
				FileUtils.copy(
						FileUtils.getFile(Project.getInstallPath(), "Documentation", "General", "Styles", "L.eps"),
						new File(source.getParentFile(), "L.eps"), true);
				html = new File(source.getParentFile(), name + "_.html");

				pb = new ProcessBuilder("cmd", "/c", "start", "/wait", "Generate Html", // name of window
						"cmd", "/c", "htlatex", name + "_.tex", "\"htlatexconfig,html,png,pic-tabular\"");
				pb.directory(source.getParentFile());
				p = pb.start();
				p.waitFor();

				if ((p.exitValue() == 0) && (html.exists())) {
					FileUtils.move(html, new File(source.getParentFile(), name + ".html"), true);
					FileUtils.delete(new File(source.getParentFile(), "htlatexconfig.cfg"));
					FileUtils.delete(new File(source.getParentFile(), "L.eps"));
					for (File f : source.getParentFile().listFiles()) {
						if (f.isFile() && f.getName().startsWith(name + "_")
								&& !f.getName().toLowerCase().endsWith(".css")
								&& !f.getName().toLowerCase().endsWith(".png")) {
							f.delete();
						}
					}
					valid = true;
				}
			} catch (IOException e) {
			} catch (InterruptedException e) {
			}
		}
		return valid;
	}

	/**
	 * Extract doc-info from Matalb-function
	 * 
	 * @param file
	 *            Function to generate doc-info
	 * @param name
	 *            Name of function
	 * @return Doc-info
	 */
	private DocInfo extractMatlabDoc(File file, String path, String name) {
		DocInfo doc = null;
		String[] header;

		header = this.extractHeader(file, '%');

		if (header.length > 0) {
			try {
				doc = new DocInfo();
				doc.name = name;
				doc.header = header;
				if (!(Boolean) MatlabControl.getProxy(path).returningFeval("GetDocInfo", 1,
						new Object[] { name, doc })[0]) {
					doc = null;
				}

			} catch (Exception e) {
			}

		}
		return doc;
	}

	/**
	 * Extract doc-info from Java-function
	 * 
	 * @param file
	 *            Function to generate doc-info
	 * @param name
	 *            Name of function
	 * @return Doc-info
	 */
	private DocInfo extractJavaDoc(File file, String name) {
		ParamJavaFnc func;
		DocInfo doc = null;
		String[] header;

		header = this.extractHeader(file, '*');

		if (header.length > 0) {
			try {
				func = (ParamJavaFnc) Class.forName(name).newInstance();
				doc = new DocInfo();
				doc.name = name;
				doc.description = func.getDefaultDescription();
				doc.imports = func.importVars();
				doc.exports = func.exportVars();
				doc.header = header;
			} catch (Exception e) {
			}
		}

		return doc;
	}

	/**
	 * Extract header from file
	 * 
	 * @param file
	 *            Function with documentation header
	 * @param comment
	 *            Char that indicate comment
	 * @return Extracted header
	 */
	private String[] extractHeader(File file, char comment) {
		BufferedReader in = null;
		LinkedList<String> header;
		String line;
		int index;

		header = new LinkedList<String>();
		try {
			in = new BufferedReader(new FileReader(file));
			// remove line before header
			do {
				line = in.readLine();
			} while ((line != null) && (!line.contains(" *TEXHEADER*")));
			// start with next line
			line = in.readLine();
			// read header till header ends
			while ((line != null) && (!line.contains(" *END TEXHEADER*"))) {
				index = line.indexOf(comment);
				if (index != -1) {
					header.add(line.substring(index + 1));
					line = in.readLine();
				} else {
					line = null;
				}
			}
			// remove lines if header is not complete
			if (line == null) {
				header.clear();
			}
		} catch (Exception e1) {
		} finally {
			if (in != null) {
				try {
					in.close();
				} catch (IOException e) {
				}
			}
		}
		return header.toArray(new String[0]);
	}

}
