/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9928 $
 * $Date: 2020-08-03 12:06:12 +0200 (Mon, 03 Aug 2020) $
 */
package at.lcm.bu21.magopt.structure;

import java.io.File;
import java.io.Serializable;
import java.lang.reflect.InvocationTargetException;
import java.rmi.RemoteException;
import java.util.AbstractList;
import java.util.ArrayList;
import java.util.List;

import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.XMLStreamWriter;

import at.lcm.bu21.general.dtypes.TBoolean;
import at.lcm.bu21.general.dtypes.TData;
import at.lcm.bu21.general.dtypes.TStruct;
import at.lcm.bu21.general.utils.Clone;
import at.lcm.bu21.general.utils.CloneException;
import at.lcm.bu21.general.utils.FileUtils;
import at.lcm.bu21.general.utils.Pair;
import at.lcm.bu21.magopt.docu.DocTexWriter;
import at.lcm.bu21.magopt.structure.client.CParamContainer;
import at.lcm.bu21.magopt.structure.rmi.RParamContainer;

/**
 * Parameter to structure the parameter-list
 * 
 * @author Werner Koppelsttter
 */
public class ParamContainer extends Parameter implements IParamContainer {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** List of child-parameter */
	private ArrayList<Parameter> param = new ArrayList<Parameter>(0);

	/** Access modifier */
	private Access access = Access.PUBLIC;

	/** Enable container */
	private boolean enable = true;

	/** Component properties */
	private Component component = null;

	/** List of public-interfaces */
	private ArrayList<PublicInterface> publicInterface = null;

	/**
	 * Defines a public interface to container
	 * 
	 * @author Werner Koppelsttter
	 */
	private class PublicInterface implements Serializable {

		/** Serialization version number */
		private static final long serialVersionUID = 1L;

		/** Name of interface */
		private String name;

		/** Parameter-string */
		private String param = "";

		/**
		 * Create new interface
		 * 
		 * @param name Name of interface
		 */
		public PublicInterface(String name) {
			this.name = name;
		}

		/**
		 * Clone interface
		 * 
		 * @return Cloned interface
		 */
		protected PublicInterface cloneInterface() {
			PublicInterface other;

			other = new PublicInterface(this.name);
			other.param = this.param;

			return other;
		}

		/**
		 * Get name of interface
		 * 
		 * @return Name
		 */
		public String getName() {
			return this.name;
		}

		/**
		 * Get name of interface
		 * 
		 * @param name New name
		 */
		public void setName(String name) {
			this.name = name;
		}

		/**
		 * Get parameter-string of interface
		 * 
		 * @return Parameter-string
		 */
		public String getParam() {
			return this.param;
		}

		/**
		 * Set parameter-string of interface
		 * 
		 * @param param Parameter-string
		 */
		public void setParam(String param) {
			this.param = param;
		}

		@Override
		public boolean equals(Object obj) {
			if (this == obj) {
				return true;
			}
			if (obj == null) {
				return false;
			}
			if (this.getClass() != obj.getClass()) {
				return false;
			}
			PublicInterface other = (PublicInterface) obj;
			if (!this.getOuterType().equals(other.getOuterType())) {
				return false;
			}
			if (this.name == null) {
				if (other.name != null) {
					return false;
				}
			} else if (!this.name.equals(other.name)) {
				return false;
			}
			if (this.param == null) {
				if (other.param != null) {
					return false;
				}
			} else if (!this.param.equals(other.param)) {
				return false;
			}
			return true;
		}

		/**
		 * Type of outer class
		 * 
		 * @return Outer class
		 */
		private ParamContainer getOuterType() {
			return ParamContainer.this;
		}
	}

	@Override
	public List<String> listTabNames() {
		List<String> list = super.listTabNames();
		list.add(0, "Container Info");
		list.add(1, "Summary");
		// if (this.getComponent() != null) {
		// list.add("Component");
		// }
		return list;
	}

	/**
	 * Constructor for loading a xml-file
	 * 
	 * @throws Exception
	 */
	ParamContainer() throws Exception {
		this("XML_Load", "XML load constructor");
	}

	/**
	 * Constructor with name
	 * 
	 * @param name Name of Parameter
	 * @throws Exception
	 */
	public ParamContainer(String name) throws Exception {
		this(name, name);
	}

	/**
	 * Constructor with name and description
	 * 
	 * @param name        Name of Parameter
	 * @param description Description of Parameter
	 * @throws Exception
	 */
	public ParamContainer(String name, String description) throws Exception {
		super(name, description);
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (!super.equals(obj)) {
			return false;
		}
		if (this.getClass() != obj.getClass()) {
			return false;
		}
		ParamContainer other = (ParamContainer) obj;
		if (this.access != other.access) {
			return false;
		}
		if (this.enable != other.enable) {
			return false;
		}
		if (this.param == null) {
			if (other.param != null) {
				return false;
			}
		} else if (!this.param.equals(other.param)) {
			return false;
		}
		if (this.publicInterface == null) {
			if (other.publicInterface != null) {
				return false;
			}
		} else if (!this.publicInterface.equals(other.publicInterface)) {
			return false;
		}
		return true;
	}

	@Override
	public ParamContainer deepClone() throws CloneException {
		return (ParamContainer) Clone.deepClone(this);
	}

	@Override
	void cloneParam(Parameter clone) throws CloneException {
		Parameter pT, pC;

		try {
			super.cloneParam(clone);

			if (clone instanceof ParamContainer) {
				ParamContainer other = (ParamContainer) clone;

				other.param.clear();
				for (int k = 0; k < this.param.size(); k++) {
					pT = this.param.get(k);
					pC = pT.getClass().getDeclaredConstructor().newInstance();
					pT.cloneParam(pC);
					other.param.add(pC);
					pC.setParent(other);
				}

				other.access = this.access;
				other.enable = this.enable;
				if (this.publicInterface == null) {
					other.publicInterface = null;
				} else {
					other.publicInterface = new ArrayList<ParamContainer.PublicInterface>();
					for (PublicInterface interf : this.publicInterface) {
						other.publicInterface.add(interf.cloneInterface());
					}
				}

				if (this.component == null) {
					other.removeComponent();
				} else {
					other.addComponent();
					this.component.cloneComponent(other.component);
				}
			}
		} catch (IllegalAccessException e) {
			System.err.println(e.getMessage());
			e.printStackTrace();
			throw new CloneException("Unable to clone");
		} catch (InstantiationException e) {
			System.err.println(e.getMessage());
			e.printStackTrace();
			throw new CloneException("Unable to clone");
		} catch (IllegalArgumentException e) {
			System.err.println(e.getMessage());
			e.printStackTrace();
			throw new CloneException("Unable to clone");
		} catch (InvocationTargetException e) {
			System.err.println(e.getMessage());
			e.printStackTrace();
			throw new CloneException("Unable to clone");
		} catch (NoSuchMethodException e) {
			System.err.println(e.getMessage());
			e.printStackTrace();
			throw new CloneException("Unable to clone");
		} catch (SecurityException e) {
			System.err.println(e.getMessage());
			e.printStackTrace();
			throw new CloneException("Unable to clone");
		}

	}

	@Override
	void cloneParam(Parameter clone, Dataset dsThis, Dataset dsClone, boolean docFiles) throws CloneException {
		TData data;

		super.cloneParam(clone, dsThis, dsClone, docFiles);

		if (clone instanceof ParamContainer) {
			ParamContainer other = (ParamContainer) clone;

			data = dsThis.getValue(this.getPId(), ParamSetKey.INCOMPLETE);
			if (data != null) {
				dsClone.setValue(other.getPId(), ParamSetKey.INCOMPLETE, data.deepClone());
			} else {
				dsClone.rmValue(other.getPId(), ParamSetKey.INCOMPLETE);
			}

			for (int k = 0; k < other.param.size(); k++) {
				this.param.get(k).cloneParam(other.param.get(k), dsThis, dsClone, docFiles);
			}
		}

	}

	@Override
	public ParamContainer convertToParamContainer() {
		ParamContainer other = null;

		try {
			other = new ParamContainer("replace");
			this.convertParamContainer(other);
			this.refresh();
			other.refresh();
		} catch (Exception ex) {
			this.error("Unable to convert", ex.getStackTrace());
		}

		return other;
	}

	@Override
	public ParamMaterial convertToParamMaterial() {
		ParamMaterial other = null;

		try {
			other = new ParamMaterial("replace");
			this.convertParamContainer(other);
			this.refresh();
			other.refresh();
		} catch (Exception ex) {
			this.error("Unable to convert", ex.getStackTrace());
		}

		return other;
	}

	@Override
	public ParamCollector convertToParamCollector() {
		ParamCollector other = null;

		try {
			other = new ParamCollector("replace");
			this.convertParamContainer(other);
			this.refresh();
			other.refresh();
		} catch (Exception ex) {
			this.error("Unable to convert", ex.getStackTrace());
		}

		return other;
	}

	@Override
	public ParamFunction convertToParamMatlabFunction(String classPath, String className) {
		ParamMatlabFunction other = null;

		try {
			other = new ParamMatlabFunction("replace", "", classPath, className);
			this.convertParamContainer(other);

			// other.setClassPath(classPath);
			// other.setClassName(className);
			this.refresh();
			other.refresh();

			other.reset();
		} catch (Exception ex) {
			this.error("Unable to convert", ex.getStackTrace());
		}

		return other;
	}

	@Override
	public ParamFunction convertToParamPythonFunction(String classPath, String className) {
		ParamPythonFunction other = null;

		try {
			if ((className != null) && (className != "")) {
				// && PythonConnection.isValidScript(new File(classPath,
				// className + ".py"))) {
				other = new ParamPythonFunction("replace", "", classPath, className);
				this.convertParamContainer(other);

				// other.setClassPath(classPath);
				// other.setClassName(className);
				this.refresh();
				other.refresh();

				other.reset();
			} else {
				this.warning(className + " is not a valid class!");
			}
		} catch (Exception ex) {
			this.error("Unable to convert", ex.getStackTrace());
		}

		return other;
	}

	@Override
	public ParamFunction convertToParamJavaFunction(String classPath, String className) {
		ParamJavaFunction other = null;

		try {
			if ((className != null) && (className != "")) {
//				Class.forName(className);

				other = new ParamJavaFunction("replace", "", classPath, className);
				this.convertParamContainer(other);

				// other.setClassPath(classPath);
				// other.setClassName(className);
				this.refresh();
				other.refresh();

				other.reset();
			} else {
				this.warning(className + " is not a valid class!");
			}
		} catch (ClassNotFoundException ex) {
			this.warning(className + " is not a valid class!");
		} catch (Exception ex) {
			this.error("Unable to convert", ex.getStackTrace());
		}

		return other;
	}

	@Override
	public Component getComponent() {
		return this.component;
	}

	@Override
	public void addComponent() {
		if (this.getComponent() == null) {
			this.component = new Component(this);
		}
	}

	@Override
	public void removeComponent() {
		this.component = null;
	}

	/**
	 * Convert ParamContainer
	 * 
	 * @param other Other ParamContainer
	 * @throws Exception
	 */
	@SuppressWarnings("unchecked")
	protected void convertParamContainer(ParamContainer other) throws Exception {

		super.cloneParam(other);

		other.access = this.access;
		other.enable = this.enable;
		if (this.publicInterface == null) {
			other.publicInterface = null;
		} else {
			other.publicInterface = (ArrayList<PublicInterface>) Clone.deepClone(this.publicInterface);
		}

		if (this.component == null) {
			other.removeComponent();
		} else {
			other.addComponent();
			this.component.cloneComponent(other.component);
		}

		other.param.clear();
		for (Parameter p : this.param) {
			p.setParent(other);
			other.param.add(p);
		}
		this.param.clear();

		if (this.getParent() != null) {
			this.getParent().replaceObj(this, other);
		}

		// this.refresh();
		// other.refresh();
	}

	/**
	 * Replace object with new object
	 * 
	 * @param oldObj Object to replace
	 * @param newObj New Object
	 */
	protected void replaceObj(Parameter oldObj, Parameter newObj) {
		Parameter param;
		for (int i = 0; i < this.param.size(); i++) {
			param = this.param.get(i);
			if (param == oldObj) {
				this.param.set(i, newObj);
				oldObj.setParent(null);
				newObj.setParent(this);
			}
		}
	}

	@Override
	public boolean getUsed() {
		return true;
	}

	/**
	 * Check list for given name
	 * 
	 * @param name Name
	 * @return whether name exist in list
	 */
	boolean existsName(String name) {
		boolean exist = false;

		for (int i = 0; i < this.param.size(); i++) {
			exist |= name.equals(this.param.get(i).getName());
		}
		return exist;
	}

	/**
	 * Get index of parameter in param-list
	 * 
	 * @param parameter Parameter to look for
	 * @return Index of parameter
	 */
	int getIndexOfParam(Parameter parameter) {
		return this.param.indexOf(parameter);
	}

	/**
	 * Get size of param-list
	 * 
	 * @return Size of parameters
	 */
	public int getParamSize() {
		return this.param.size();
	}

	@Override
	public Parameter[] getParam() {
		return this.param.toArray(new Parameter[0]);
	}

	@Override
	public TStruct getDataStructRec() {
		TStruct struct;

//		struct = new TStruct();
		struct = super.getDataStructRec();
		for (Parameter param : this.getParam()) {
			struct.put(param.getName(), param.getDataStructRec());
		}

		return struct;
	}

	@Override
	public boolean isEnable() {
		boolean enable;

		enable = this.getEnable();
		if (this.getParent() != null) {
			enable = enable && this.getParent().isEnable();
		}
		return enable;
	}

	@Override
	public boolean getEnable() {
		return this.enable;
	}

	@Override
	public void setEnable(boolean enable) {
		this.enable = enable;
		this.getRoot().refresh();
	}

	@Override
	public Access getAccess() {
		return this.access;
	}

	@Override
	public void setAccess(Access access) {
		this.access = access;
		this.getRoot().refresh();
	}

	@Override
	public boolean getIncomplete() {
		return this.getIncomplete(this.getIndex());
	}

	/**
	 * Get given incomplete status
	 * 
	 * @param index Index of dataset
	 * @return incomplete status
	 */
	public boolean getIncomplete(int index) {
		Dataset ds;
		TData value;
		boolean bool = true;

		ds = this.getDataset(index);
		if (ds != null) {
			value = ds.getValue(this.getPId(), ParamSetKey.INCOMPLETE);
			if (value != null) {
				bool = ((TBoolean) value).getBoolean();
			}
		}

		return bool;
	}

	@Override
	public boolean[] getAllIncomplete() {
		boolean[] list = new boolean[this.getLength()];
		for (int i = 0; i < this.getLength(); i++) {
			list[i] = this.getIncomplete(i);
		}
		return list;
	}

	/**
	 * Set status of current incomplete
	 * 
	 * @param bool status of incomplete
	 */
	void setIncomplete(boolean bool) {
		this.setIncomplete(this.getIndex(), bool);
	}

	/**
	 * Set status of current incomplete
	 * 
	 * @param index Index of dataset
	 * @param bool  status of incomplete
	 */
	void setIncomplete(int index, boolean bool) {
		Dataset ds;
		ds = this.getDataset(index);
		if (ds != null) {
			ds.setValue(this.getPId(), ParamSetKey.INCOMPLETE, new TBoolean(bool));
		}
	}

	@Override
	public void setConstant(boolean constant) {
		if (constant) {
			this.warning("Set Container to constant is not allowed!");
		}
	}

	@Override
	void execStructure(ExecStructure wrapper) {
		super.execStructure(wrapper);
		for (int k = 0; k < this.param.size(); k++) {
			this.param.get(k).execStructure(wrapper);
		}
	}

	@Override
	AbstractList<Parameter> filterParameter(FilterStructure filter) {
		ArrayList<Parameter> list = new ArrayList<Parameter>(0);
		list.addAll(super.filterParameter(filter));
		for (int k = 0; k < this.param.size(); k++) {
			list.addAll(this.param.get(k).filterParameter(filter));
		}
		return list;
	}

	@Override
	protected boolean refreshIncomplete() {
		boolean incomplete;

		incomplete = super.refreshIncomplete();

		for (int k = 0; k < this.param.size(); k++) {
			incomplete |= this.param.get(k).refreshIncomplete();
		}

		this.setIncomplete(incomplete);
		return incomplete;
	}

	@Override
	public ParameterPath getParamByUniqueIdentifier(String uIdName, boolean allAccess) {
		String name = this.getName();
		ParameterPath path = null;
		String[] keys;
		int k;

		if (uIdName != null) {
			if (uIdName.compareTo(name) == 0) {
				path = new ParameterPath(this);
			} else {
				if (uIdName.startsWith(name + ".")) {
					uIdName = uIdName.substring(name.length() + 1);

					// interface-dependency
					k = 0;
					keys = this.listPublicInterfaceName();
					while ((k < keys.length) && (path == null)) {
						if (keys[k].compareTo(uIdName) == 0) {
							path = this.getPublicInterfacePath(keys[k]);
						}
						k++;
					}

					// direct-dependency
					if (path == null) {
						if (allAccess || (this.getAccess() != Access.PRIVATE)) {
							k = 0;
							while ((k < this.param.size()) && (path == null)) {
								path = this.param.get(k).getParamByUniqueIdentifier(uIdName, allAccess);
								k++;
							}
						}
					}
				}
			}
		}
		return path;
	}

	@Override
	protected AbstractList<ParameterPath> getParamByIdentifier(String idName, boolean thisPublic) {
		String name = this.getName();
		ArrayList<ParameterPath> list = new ArrayList<ParameterPath>(0);

		if (idName.compareTo(name) == 0) {
			list.add(new ParameterPath(this));
			if (thisPublic || (this.getAccess() == Access.PUBLIC)) {
				for (int k = 0; k < this.param.size(); k++) {
					list.addAll(this.param.get(k).getParamByIdentifier(idName, false));
				}
			}
		} else {
			if (idName.startsWith(name + ".")) {
				idName = idName.substring(name.length() + 1);

				// interface-dependencies
				for (String iname : this.listPublicInterfaceName()) {
					if (iname.compareTo(idName) == 0) {
						list.add(this.getPublicInterfacePath(iname));
					}
				}

				// direct-dependencies
				if (thisPublic || (this.getAccess() != Access.PRIVATE)) {
					for (int k = 0; k < this.param.size(); k++) {
						list.addAll(this.param.get(k).getParamByIdentifier(idName, false));
					}
				}
			} else {
				// interface-dependencies
				for (String iname : this.listPublicInterfaceName()) {
					if (iname.compareTo(idName) == 0) {
						list.add(this.getPublicInterfacePath(iname));
					}
				}

				// direct-dependencies
				if (thisPublic || (this.getAccess() == Access.PUBLIC)) {
					for (int k = 0; k < this.param.size(); k++) {
						list.addAll(this.param.get(k).getParamByIdentifier(idName, false));
					}
				}
			}
		}

		return list;
	}

	@Override
	public Parameter getParamByPid(String pid) {
		Parameter p = null;
		int k = 0;

		p = super.getParamByPid(pid);
		while ((p == null) && (k < this.param.size())) {
			p = this.param.get(k).getParamByPid(pid);
			k++;
		}

		return p;
	}

	/**
	 * Add Parameter to structure at the end
	 * 
	 * @param add Parameter to add
	 * @throws Exception
	 */
	public void addParam(Parameter add) throws Exception {
		this.addParam(add, -1);
	}

	/**
	 * Add Parameter to structure
	 * 
	 * @param add Parameter to add
	 * @param pos Position to add (-1 is end of list)
	 * @throws Exception
	 */
	public void addParam(Parameter add, int pos) throws Exception {
		int num = 1;
		String name, pid;
		ArrayList<Pair<String, String>> link;
		Dataset dsAdd, dsThis;
		ParamDataStruct dataAdd, dataThis;
		Project addProject, thisProject;
		File dsAddFile, dsThisFile, addFile, thisFile;
		TData value;
		DocFile[] docfiles;
		DocFile doc;
		String tmp;
		ParamContainer par;

		name = add.getName();

		if (add instanceof Project) {
			throw new Exception("Unable to add project");
		}

		par = add.getParent();
		// check for identical name
		if ((par == null) || (this != par)) {
			while (this.existsName(name)) {
				name = add.getName() + "(" + num++ + ")";
			}
			if (name.compareTo(add.getName()) != 0) {
				this.warning("Parameter '" + add.getName() + "' already exist in (" + this.getIdentifier()
						+ ")! Rename to '" + name + "'");
				add.setName(name);
			}
		} else {
			// change position to add if this is behind current position in
			// container
			if (add.getIndexOfParameter() < pos) {
				pos--;
			}
		}

		// insert parameter
		dataAdd = add.getParameterStruct();
		dataThis = this.getParameterStruct();
		if (dataThis != null) {
			if (dataAdd != null) {
				if (dataThis != dataAdd) {
					thisProject = (Project) this.getRoot();
					addProject = (Project) add.getRoot();

					// add to list (change pid and transfer data)
					this.insertParam(add, pos);
					link = new ArrayList<Pair<String, String>>();
					for (Parameter p : add.getParameters(Parameter.class)) {
						pid = dataThis.getNextPId();
						link.add(new Pair<String, String>(p.getPId(), pid));
						p.setPId(pid);
					}
					// transfer data
					dsAdd = dataAdd.getDataset();
					dsAddFile = addProject.getFilesDirectory(false);
					for (int k = 0; k < dataThis.getLength(); k++) {
						if (dataThis.getLength() == dataAdd.getLength()) {
							dsAdd = dataAdd.getDataset(k);
							dsAddFile = addProject.getFilesDirectory(k, false);
						}
						dsThis = dataThis.getDataset(k);
						dsThisFile = thisProject.getFilesDirectory(k);
						for (Pair<String, String> l : link) {
							// transfer paramdata
							for (ParamSetKey key : ParamSetKey.values()) {
								value = dsAdd.getValue(l.getFirst(), key);
								if (value != null) {
									dsThis.setValue(l.getSecond(), key, value);
								}
							}

							// copy parameter directory
							addFile = FileUtils.getFile(dsAddFile.getPath(), "param", l.getFirst());
							if (addFile.exists()) {
								thisFile = FileUtils.getFile(dsThisFile.getPath(), "param", l.getSecond());
								try {
									FileUtils.copy(addFile, thisFile);
								} catch (Exception ex) {
									this.warning("Unable to copy parameter-directory (" + ex.getMessage() + ")",
											ex.getStackTrace());
								}
							}

							// refresh components
							for (Parameter p : add.getParameters(ParamContainer.class)) {
								Component comp = ((ParamContainer) p).getComponent();
								if (comp != null) {
									comp.refresh();
								}
							}

							// copy DocFiles
							docfiles = dsAdd.getDocFiles(l.getFirst());
							for (DocFile f : docfiles) {
								addFile = new File(dsAddFile, f.getFile());
								if (addFile.exists()) {
									thisFile = this.getUniqueDocFileName(new File(dsThisFile, f.getFile()));
									// add docFile
									doc = new DocFile();
									doc.setName(f.getName());
									tmp = add.getParamByPid(l.getSecond()).getIdentifier();
									if (tmp.indexOf(".") != -1) {
										tmp = tmp.substring(tmp.indexOf(".") + 1);
									}
									doc.setParamName(tmp);
									doc.setFile(FileUtils.getRelativePath(dsThisFile.getPath(), thisFile.getPath()));

									doc.setDescription(f.getDescription());
									// doc.setChecksum(f.getChecksum());
									doc.setChecksum("");

									FileUtils.copy(addFile, thisFile);
									dsThis.setDocFile(l.getSecond(), doc);
								}
							}
						}
					}

					// cleanup
					if ((addProject.getParam().length == 0)
							&& addProject.getFile().startsWith(System.getProperty("java.io.tmpdir"))) {
						Project.delete(addProject.getFile());
					} else {
						addProject.cleanup();
					}
					thisProject.cleanup();
				} else {
					// add to list (change position)
					this.insertParam(add, pos);
				}
			} else {
				// add to list (without data)
				this.insertParam(add, pos);
				for (Parameter p : add.getParameters(Parameter.class)) {
					p.setPId(dataThis.getNextPId());
				}
			}
		} else {
			// add to list (without pid and data
			this.insertParam(add, pos);
		}

		if (par != null) {
			par.refresh();
		}

		// refresh location of directories
		for (Parameter p : add.getParameters(ParamDirectory.class)) {
			((ParamDirectory) p).updateLocation();
		}

		this.refresh();
	}

	/**
	 * Insert parameter at position of list
	 * 
	 * @param add Parameter to insert
	 * @param pos Position to insert
	 */
	private void insertParam(Parameter add, int pos) {
		// remove param
		if (add.getParent() != null) {
			add.getParent().rmParameter(add);
			add.setParent(null);
		}

		// add param
		add.setParent(this);
		if ((0 <= pos) && (pos < this.param.size())) {
			this.param.add(pos, add);
		} else {
			this.param.add(add);
		}
	}

	/**
	 * Remove Parameter from structure
	 * 
	 * @param param Parameter to remove
	 */
	void rmParameter(Parameter param) {
		this.param.remove(param);
	}

	@Override
	public String toString() {
		StringBuffer str = new StringBuffer();
		str.append("ParamContainer\n");
		str.append("\n\tName        : ");
		str.append(this.getName());
		str.append("\n\tDescription : ");
		str.append(this.getDescription());
		str.append("\n\tParams      : ");

		for (int i = 0; i < this.param.size(); i++) {
			str.append("\n\t\tName: " + this.param.get(i).getName());
		}

		return str.toString();
	}

	@Override
	void xml2obj(XMLStreamReader xmlReader) throws Exception {
		String element, tmp;
		Parameter obj = null;
		PublicInterface interf;

		xmlReader.next();
		while (xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT) {
			element = xmlReader.getLocalName();
			if (element.equals("Parameter")) {
				super.xml2obj(xmlReader);
			} else if (element.equals("Param")) {
				xmlReader.next();
				while (xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT) {
					element = xmlReader.getLocalName();
					// read objects
					if (element.equals("ParamContainer")) {
						obj = new ParamContainer();
					} else if (element.equals("ParamField")) {
						obj = new ParamField();
					} else if (element.equals("ParamComboBox")) {
						obj = new ParamComboBox();
					} else if (element.equals("ParamFormula")) {
						obj = new ParamFormula();
					} else if (element.equals("ParamLinkedField")) {
						obj = new ParamLinkedField();
					} else if (element.equals("ParamMaterial")) {
						obj = new ParamMaterial();
					} else if (element.equals("ParamCollector")) {
						obj = new ParamCollector();
					} else if (element.equals("ParamDirectory")) {
						obj = new ParamDirectory();
					} else if (element.equals("ParamFile")) {
						obj = new ParamFile();
					} else if (element.equals("ParamMatlabFunction")) {
						obj = new ParamMatlabFunction();
					} else if (element.equals("ParamPythonFunction")) {
						obj = new ParamPythonFunction();
					} else if (element.equals("ParamPythonFunction_old")) {
						obj = new ParamPythonFunction_old();
					} else if (element.equals("ParamJavaFunction")) {
						obj = new ParamJavaFunction();
					} else {
						this.error("Unknown class: " + element);
					}

					this.param.add(obj);
					obj.setParent(this);
					obj.xml2obj(xmlReader);
				}
				xmlReader.next();

			} else if (element.equals("Access")) {
				xmlReader.next();
				tmp = new String();
				while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
					tmp += xmlReader.getText();
					xmlReader.next();
				}
				this.access = Access.valueOf(tmp);
				xmlReader.next();
			} else if (element.equals("Enable")) {
				xmlReader.next();
				this.enable = Boolean.parseBoolean(xmlReader.getText());
				xmlReader.next();
				xmlReader.next();
			} else if (element.equals("Interface")) {
				xmlReader.next();
				this.publicInterface = new ArrayList<ParamContainer.PublicInterface>();
				while (xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT) {
					xmlReader.next();
					interf = new PublicInterface("");
					this.publicInterface.add(interf);

					while (xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT) {
						element = xmlReader.getLocalName();
						if (element.equals("Name")) {
							xmlReader.next();
							tmp = new String();
							while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
								tmp += xmlReader.getText();
								xmlReader.next();
							}
							interf.setName(tmp);
							xmlReader.next();
						} else if (element.equals("Param")) {
							xmlReader.next();
							tmp = new String();
							while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
								tmp += xmlReader.getText();
								xmlReader.next();
							}
							interf.setParam(tmp);
							xmlReader.next();
						}
					}

					xmlReader.next();
				}
				xmlReader.next();
			} else if (element.equals("Component")) {
				this.component = new Component(this);
				this.component.xml2obj(xmlReader);
			} else {
				throw new Exception("Unrecognized property: " + xmlReader.getLocalName());
			}
		}
		xmlReader.next();
	}

	@Override
	void obj2xml(XMLStreamWriter xmlWriter) throws Exception {

		xmlWriter.writeStartElement("ParamContainer");
		xmlWriter.writeAttribute("object", this.getClass().getSimpleName());
		xmlWriter.writeAttribute("name", this.getName());

		// super class
		super.obj2xml(xmlWriter);

		// Param
		xmlWriter.writeStartElement("Param");
		for (Parameter p : this.param) {
			p.obj2xml(xmlWriter);
		}
		xmlWriter.writeEndElement();

		// Access
		xmlWriter.writeStartElement("Access");
		xmlWriter.writeCharacters(this.access.name());
		xmlWriter.writeEndElement();

		// Enable
		xmlWriter.writeStartElement("Enable");
		xmlWriter.writeCharacters(Boolean.toString(this.enable));
		xmlWriter.writeEndElement();

		// Public interface
		if (this.publicInterface != null) {
			xmlWriter.writeStartElement("Interface");
			for (PublicInterface interf : this.publicInterface) {
				xmlWriter.writeStartElement("entry");

				xmlWriter.writeStartElement("Name");
				xmlWriter.writeCharacters(interf.getName());
				xmlWriter.writeEndElement();

				xmlWriter.writeStartElement("Param");
				xmlWriter.writeCharacters(interf.getParam());
				xmlWriter.writeEndElement();

				xmlWriter.writeEndElement();
			}
			xmlWriter.writeEndElement();
		}

		if (this.component != null) {
			this.component.obj2xml(xmlWriter);
		}

		xmlWriter.writeEndElement();
	}

	@Override
	void reportTable(DocTexWriter content, int indent) {

		content.add("\\rowcolor[RGB]{\\RGBrowcolor}");
		super.reportTable(content, indent);
		content.add(" & &");

		Parameter[] param = this.getParam();
		for (int k = 0; k < param.length; k++) {
			content.add(" \\\\\n\\hline ");
			param[k].reportTable(content, indent + 1);
		}
	}

	@Override
	List<ParamSetKey> listParamSetKey() {
		List<ParamSetKey> list;

		list = super.listParamSetKey();
		list.add(ParamSetKey.INCOMPLETE);

		return list;
	}

	/**
	 * Get public-interface
	 * 
	 * @param name name of interface
	 * @return Public interface
	 */
	private PublicInterface getPublicInterface(String name) {
		PublicInterface interf = null;

		if (this.publicInterface != null) {
			for (PublicInterface i : this.publicInterface) {
				if (i.getName().compareTo(name) == 0) {
					interf = i;
				}
			}
		}

		return interf;
	}

	@Override
	public void addPublicInterface(String name) {
		if ((name != null) && !name.trim().isEmpty()) {
			if (this.getPublicInterface(name) == null) {
				if (this.publicInterface == null) {
					this.publicInterface = new ArrayList<ParamContainer.PublicInterface>(1);
				}
				this.publicInterface.add(new PublicInterface(name));
			}
		}
	}

	@Override
	public void renamePublicInterface(String oldName, String newName) {
		PublicInterface interf;
		ParameterPath path;
		ParamContainer cont;
		String str;
		Parameter[] params;

		interf = this.getPublicInterface(oldName);
		if (interf != null) {

			// change dependency entries
			params = this.getParameters(Parameter.class);
			for (Parameter param : params) {

				Parameter[] usedList = param.getUsedBy();
				for (Parameter used : usedList) {
					this.renameDepends(((Dependency) used).getDepend().getImports(), oldName, newName);
				}

				Parameter[] setList = param.getSetBy();
				for (Parameter set : setList) {
					this.renameDepends(((Dependency) set).getDepend().getExports(), oldName, newName);
				}
			}

			// change interfaces entries of parents
			cont = this.getParent();
			while (cont != null) {
				for (String iname : cont.listPublicInterfaceName()) {
					path = cont.getPublicInterfacePath(iname);
					while (path.getPath().getPath() != null) {
						path = path.getPath();
					}
					if (path.getParameter() == this) {
						str = cont.getPublicInterfaceParam(iname);
						if (str.compareTo(oldName) == 0) {
							cont.setPublicInterfaceParam(iname, newName);
						} else if (str.endsWith("." + oldName)) {
							cont.setPublicInterfaceParam(iname,
									str.substring(0, str.length() - oldName.length()) + newName);
						}
					}
				}
				cont = cont.getParent();
			}

			interf.setName(newName);
			this.refresh();
		}
	}

	/**
	 * Rename depend if interface is in parameter-string
	 * 
	 * @param items   Dependency-items to check of parameter
	 * @param oldName Old name of interface
	 * @param newName New name of interface
	 */
	private void renameDepends(DependItem[] items, String oldName, String newName) {
		ParameterPath path;
		String str;

		for (DependItem item : items) {
			// get parameter of direct link or parameter of first interface
			path = this.getObjectPath(item.getParam());
			if (path != null) {
				while (path.getPath() != null) {
					path = path.getPath();
				}
				if (path.getParameter() == this) {
					str = item.getParam();
					if (str.compareTo(oldName) == 0) {
						item.setParam(newName);
					} else if (str.endsWith("." + oldName)) {
						item.setParam(str.substring(0, str.length() - oldName.length()) + newName);
					}
				}
			}
		}
	}

	@Override
	public String removePublicInterface(String name) {
		PublicInterface interf = null;
		String param = null;

		interf = this.getPublicInterface(name);
		if (interf != null) {
			param = interf.getParam();
			this.publicInterface.remove(interf);
			if (this.publicInterface.isEmpty()) {
				this.publicInterface = null;
			}
		}
		this.refresh();

		return param;
	}

	@Override
	public String getPublicInterfaceParam(String name) {
		PublicInterface interf = null;
		String param = null;

		interf = this.getPublicInterface(name);
		if (interf != null) {
			param = interf.getParam();
		}

		return param;
	}

	@Override
	public void setPublicInterface(String name, String param) {
		this.setPublicInterfaceParam(name, param);
		this.refresh();
	}

	/**
	 * Set parameter-string of public-interface
	 * 
	 * @param name  Name of interface
	 * @param param Parameter-string
	 */
	void setPublicInterfaceParam(String name, String param) {
		PublicInterface interf = null;

		interf = this.getPublicInterface(name);
		if (interf != null) {
			interf.setParam(param);
		}
	}

	@Override
	public String[] listPublicInterfaceName() {
		String[] list = new String[0];

		if (this.publicInterface != null) {
			list = new String[this.publicInterface.size()];
			for (int k = 0; k < list.length; k++) {
				list[k] = this.publicInterface.get(k).getName();
			}
		}

		return list;
	}

	@Override
	public ParameterPath getPublicInterfacePath(String iname) {
		ParameterPath p, path = null;
		AbstractList<ParameterPath> par;
		int k;
		String param;

		param = this.getPublicInterfaceParam(iname);
		if ((param != null) && !param.isEmpty()) {

			// get by unique identifier
			k = 0;
			while ((k < this.param.size()) && (path == null)) {
				path = this.param.get(k).getParamByUniqueIdentifier(this.getPublicInterfaceParam(iname), false);
				k++;
			}

			// get by identifier
			k = 0;
			while ((k < this.param.size()) && (path == null)) {
				par = this.param.get(k).getParamByIdentifier(this.getPublicInterfaceParam(iname), false);
				if ((par != null) && !par.isEmpty()) {
					path = par.get(0);
				}
				k++;
			}
		}

		// unlinked interface
		if (path == null) {
			path = new ParameterPath(null);
		}

		// add this path to end
		p = path;
		while (p.getPath() != null) {
			p = p.getPath();
		}
		p.setPath(new ParameterPath(this));

		return path;
	}

	/**
	 * Get selected parameter from public-interface
	 * 
	 * @param name Name of interface
	 * @return Selected parameter
	 */
	public Parameter getPublicInterfaceParameter(String name) {
		return this.getPublicInterfacePath(name).getParameter();
	}

	@Override
	public IParamContainer getRMIImpl() throws RemoteException {
		return new RParamContainer(this);
	}

	@Override
	public IParamContainer getClientImpl() {
		return new CParamContainer(this);
	}
}
