/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9749 $
 * $Date: 2020-04-10 16:00:50 +0200 (Fri, 10 Apr 2020) $
 */
package at.lcm.bu21.magopt.structure;

import java.rmi.RemoteException;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;

import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.XMLStreamWriter;

import at.lcm.bu21.general.dtypes.TBoolean;
import at.lcm.bu21.general.dtypes.TData;
import at.lcm.bu21.general.utils.Clone;
import at.lcm.bu21.general.utils.CloneException;
import at.lcm.bu21.magopt.docu.DocTexWriter;
import at.lcm.bu21.magopt.structure.client.CParamFormula;
import at.lcm.bu21.magopt.structure.formula.Parser;
import at.lcm.bu21.magopt.structure.formula.ParserException;
import at.lcm.bu21.magopt.structure.formula.ScannerException;
import at.lcm.bu21.magopt.structure.rmi.RParamFormula;
import at.lcm.bu21.magopt.utils.MatlabControl;
import matlabcontrol.MatlabInvocationException;
import matlabcontrol.MatlabProxy;

/**
 * Parameter to calculate formula
 * 
 * @author Werner Koppelsttter
 */
public class ParamFormula extends ParamField implements IParamFormula, Function {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** Expression to calculate */
	private String expression = new String();

	/** Type of formula */
	private FormulaType formulaType = FormulaType.JAVA;

	/** Dependency of Parameter */
	private Depend depend = new Depend(this);

	/**
	 * Constructor for loading a xml-file
	 * 
	 * @throws Exception
	 */
	ParamFormula() throws Exception {
		this("XML_Load", "XML load constructor");
	}

	/**
	 * Constructor with name
	 * 
	 * @param name
	 *            Name of Parameter
	 * @throws Exception
	 */
	public ParamFormula(String name) throws Exception {
		this(name, name);
	}

	/**
	 * Constructor with name and description
	 * 
	 * @param name
	 *            Name of Parameter
	 * @param description
	 *            Description of Parameter
	 * @throws Exception
	 */
	public ParamFormula(String name, String description) throws Exception {
		super(name, description);
		this.addSetBy(this);
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (!super.equals(obj)) {
			return false;
		}
		if (this.getClass() != obj.getClass()) {
			return false;
		}
		ParamFormula other = (ParamFormula) obj;
		if (this.expression == null) {
			if (other.expression != null) {
				return false;
			}
		} else if (!this.expression.equals(other.expression)) {
			return false;
		}
		if (this.depend == null) {
			if (other.depend != null) {
				return false;
			}
		} else if (!this.depend.equals(other.depend)) {
			return false;
		}
		return true;
	}

	@Override
	public ParamFormula deepClone() throws CloneException {
		return (ParamFormula) Clone.deepClone(this);
	}

	@Override
	void cloneParam(Parameter clone) throws CloneException, InstantiationException, IllegalAccessException {

		super.cloneParam(clone);

		if (clone instanceof ParamFormula) {
			ParamFormula other = (ParamFormula) clone;

			other.formulaType = this.formulaType;
			other.expression = this.expression;

			this.depend.cloneDepend(other.depend);
		}
	}

	@Override
	void cloneParam(Parameter clone, Dataset dsThis, Dataset dsClone, boolean docFiles) throws CloneException {
		TData data;

		super.cloneParam(clone, dsThis, dsClone, docFiles);

		if (clone instanceof ParamFormula) {
			ParamFormula other = (ParamFormula) clone;

			data = dsThis.getValue(this.getPId(), ParamSetKey.UPDATEREQUIRED);
			if (data != null) {
				dsClone.setValue(other.getPId(), ParamSetKey.UPDATEREQUIRED, data.deepClone());
			} else {
				dsClone.rmValue(other.getPId(), ParamSetKey.UPDATEREQUIRED);
			}
		}
	}

	@Override
	public Depend getDepend() {
		return this.depend;
	}

	@Override
	public void refreshDependency() {
	}

	/**
	 * Refresh dependencies of formula
	 */
	private void refreshDependency2() {
		String[] list = null;
		MatlabProxy proxy;
		PortDesc[] vars;
		boolean valid = true;

		switch (this.formulaType) {
		case JAVA:
			try {
				list = new Parser(this.expression).getVariables();
			} catch (ParserException ex) {
				this.error(ex.getMessage(), ex.getStackTrace());
			} catch (ScannerException ex) {
				this.error(ex.getMessage(), ex.getStackTrace());
			}
			break;
		case MATLAB:
			if (!this.checkBracket(this.expression, '(', ')')) {
				valid = false;
				this.error("Error with parentheses! '" + this.expression + "'");
			}
			if (!this.checkBracket(this.expression, '[', ']')) {
				valid = false;
				this.error("Error with brackets! '" + this.expression + "'");
			}
			if (!this.checkBracket(this.expression, '{', '}')) {
				valid = false;
				this.error("Error with braces! '" + this.expression + "'");
			}

			if (valid) {
				proxy = MatlabControl.getProxy();
				try {
					list = (String[]) proxy.returningFeval("symvar", 1, this.expression)[0];
				} catch (MatlabInvocationException e) {
					this.error("Unable to parse expression '" + this.expression + "'", e.getStackTrace());
				}
			}
			break;
		case PYTHON:
			// TODO
			this.error("Formula-type 'Python' not yet implemented");
			break;
		default:
			this.error("Undefined formula type");
			break;
		}

		// set dependencies
		if (list != null) {
			vars = new PortDesc[list.length];
			for (int k = 0; k < list.length; k++) {
				vars[k] = new PortDesc(list[k], "", "", false);
			}
			this.getDepend().setImportVars(vars);
		}

		this.refresh();

	}

	@Override
	public void setName(String name) throws Exception {
		super.setName(name);
	}

	@Override
	public FormulaType getFormulaType() {
		return this.formulaType;
	}

	@Override
	public void setFormulaType(FormulaType type) {
		if (this.formulaType != type) {
			this.formulaType = type;
			for (int k = 0; k < this.getLength(); k++) {
				this.setUpdateRequired(k, true);
			}
			this.refreshDependency2();
		}
	}

	@Override
	public String getExpression() {
		return this.expression;
	}

	@Override
	public void setExpression(String exp) {
		if (this.expression.compareTo(exp) != 0) {
			this.expression = exp;
			for (int k = 0; k < this.getLength(); k++) {
				this.setUpdateRequired(k, true);
			}

			this.refreshDependency2();
		}
	}

	/**
	 * Check number and order of bracket
	 * 
	 * @param exp
	 *            Expression to check
	 * @param open
	 *            Open character
	 * @param close
	 *            Close character
	 * @return Valid bracket
	 */
	private boolean checkBracket(String exp, char open, char close) {
		boolean valid = true;
		int indexOpen = -1;
		int indexClose = -1;

		do {
			indexOpen = exp.indexOf(open, ++indexOpen);
			indexClose = exp.indexOf(close, ++indexClose);

			if ((indexOpen == -1) && (indexClose == -1)) {
			} else if ((indexOpen == -1) || (indexClose == -1)) {
				valid = false;
			} else {
				valid &= indexOpen < indexClose;
			}
		} while (valid && (indexOpen != -1) && (indexClose != -1));

		return valid;
	}

	@Override
	public boolean getUpdateRequired() {
		return this.getUpdateRequired(this.getIndex());
	}

	@Override
	public boolean getUpdateRequired(int index) {
		Dataset ds;
		TData value;
		boolean bool = true;

		ds = this.getDataset(index);
		if (ds != null) {
			value = ds.getValue(this.getPId(), ParamSetKey.UPDATEREQUIRED);
			if (value != null) {
				bool = ((TBoolean) value).getBoolean();
			}
		}

		return bool;
	}

	@Override
	public void setUpdateRequired(boolean update) {
		this.setUpdateRequired(this.getIndex(), update);
		// FIXME setUncertain (refresh)
		this.refreshFlags();
	}

	@Override
	public void setUpdateRequired(int index, boolean update) {
		Dataset ds;
		if (update != this.getUpdateRequired(index)) {
			ds = this.getDataset(index);
			if (ds != null) {
				ds.setValue(this.getPId(), ParamSetKey.UPDATEREQUIRED, new TBoolean(update));
			}
			this.resetDocFiles(index);
		}

	}

	@Override
	public void setConstant(boolean constant) {
		if (constant) {
			this.warning("Set Formular to constant is not allowed!");
		}
	}

	@Override
	public boolean isEnable() {
		boolean enable = true;

		if (this.getParent() != null) {
			enable = this.getParent().isEnable();
		}
		return enable;
	}

	@Override
	protected boolean refreshIncomplete() {
		return super.refreshIncomplete() || this.getUpdateRequired();
	}

	@Override
	public void refreshDepend(boolean dirty, boolean load) {
		this.depend.refreshDepend(dirty, load);
	}

	/**
	 * Set unit of Parameter. Default behavior is not convert values.
	 * 
	 * @param unit
	 *            unit
	 * @throws RemoteException
	 * @throws Exception
	 */
	@Override
	public void setUnit(String unit) throws Exception {
		super.setUnit(unit, false);
	}

	@Override
	public String toString() {
		StringBuffer str = new StringBuffer();
		str.append("ParamFormula\n");
		str.append("\n\tName        : ");
		str.append(this.getName());
		str.append("\n\tDescription : ");
		str.append(this.getDescription());
		str.append("\n\tExpression  : ");
		str.append(this.getExpression());
		str.append("\n\tData        : ");
		str.append(this.getData());
		str.append("\n\tUnit        : ");
		str.append(this.getUnit());
		str.append("\n\tTex String  : ");
		str.append(this.getTexString());
		str.append("\n\tType        : ");
		str.append(this.getType());
		return str.toString();
	}

	@Override
	void xml2obj(XMLStreamReader xmlReader) throws Exception {
		String element, tmp;

		xmlReader.next();
		while (xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT) {
			element = xmlReader.getLocalName();
			if (element.equals("ParamField")) {
				super.xml2obj(xmlReader);
			} else if (element.equals("Depend")) {
				this.depend.xml2obj(xmlReader);
			} else if (element.equals("FormulaType")) {
				xmlReader.next();
				tmp = new String();
				while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
					tmp += xmlReader.getText();
					xmlReader.next();
				}
				this.formulaType = FormulaType.valueOf(tmp);
				xmlReader.next();
			} else if (element.equals("Expression")) {
				xmlReader.next();
				while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
					this.expression += xmlReader.getText();
					xmlReader.next();
				}
				xmlReader.next();
			} else {
				throw new Exception("Unrecognized property: " + xmlReader.getLocalName());
			}
		}
		xmlReader.next();
	}

	@Override
	void obj2xml(XMLStreamWriter xmlWriter) throws Exception {

		xmlWriter.writeStartElement("ParamFormula");
		xmlWriter.writeAttribute("object", this.getClass().getSimpleName());
		xmlWriter.writeAttribute("name", this.getName());

		// super class
		super.obj2xml(xmlWriter);

		// Depend
		this.depend.obj2xml(xmlWriter);

		// Type
		xmlWriter.writeStartElement("FormulaType");
		xmlWriter.writeCharacters(this.formulaType.name());
		xmlWriter.writeEndElement();

		// Expression
		xmlWriter.writeStartElement("Expression");
		if (this.expression != null) {
			xmlWriter.writeCharacters(this.expression);
		}
		xmlWriter.writeEndElement();

		xmlWriter.writeEndElement();
	}

	@Override
	public void reset() {
		this.setUpdateRequired(true);
	}

	@Override
	public void reportDepend(DocTexWriter content) {
		this.getDepend().reportImport(content);
	}

	@Override
	void reportInfo(DocTexWriter content) {

		super.reportInfo(content);

		content.add("\tExpression & ");
		content.addString(this.getExpression());
		content.add(" \\\\\n");
	}

	@Override
	List<ParamSetKey> listParamSetKey() {
		List<ParamSetKey> list;

		list = super.listParamSetKey();
		list.add(ParamSetKey.UPDATEREQUIRED);

		return list;
	}

	/**
	 * Get import values
	 * 
	 * @return Map with import values
	 */
	protected Map<String, TData> importValues() {
		Map<String, TData> map = new Hashtable<String, TData>();

		DependItem[] depends = this.getDepend().getImports();
		for (DependItem depend : depends) {
			if (depend.getHandle() != null) {
				map.put(depend.getVarName(), depend.getHandle().getData());
			}
		}
		return map;
	}

	@Override
	public int update(Map<String, Object> args) throws Exception {
		int status = 0;
		Parameter root;
		MatlabProxy proxy;

		switch (this.formulaType) {
		case JAVA:
			this.setData(new Parser(this.expression).eval(this.importValues()));
			break;
		case MATLAB:
			// FIXME Matlab exception
			proxy = MatlabControl.getProxy();
			if (proxy.isRunningInsideMatlab()) {
				proxy.returningFeval("EvalFormula", 0, new Object[] { this });
			} else {
				try {
					proxy.returningFeval("EvalFormula", 0, new Object[] { this.getRMIImpl().getClientImpl() });
				} catch (RemoteException e) {
					e.printStackTrace();
				}
			}
			break;
		case PYTHON:
			// TODO
			throw new Exception("Formula-type 'Python' not yet implemented");
			// break;
		default:
			break;
		}

		root = this.getRoot();
		if (root instanceof Project) {
			((Project) root).check();
		}
		if (status == 0) {
			this.reset();
			this.setUpdateRequired(false);
		}
		return status;
	}

	@Override
	public IParamFormula getRMIImpl() throws RemoteException {
		return new RParamFormula(this);
	}

	@Override
	public IParamFormula getClientImpl() {
		return new CParamFormula(this);
	}
}
