/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9928 $
 * $Date: 2020-08-03 12:06:12 +0200 (Mon, 03 Aug 2020) $
 */
package at.lcm.bu21.magopt.structure;

import java.io.File;
import java.util.List;

import javax.swing.JPanel;
import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.XMLStreamWriter;

import at.lcm.bu21.general.dtypes.TBoolean;
import at.lcm.bu21.general.dtypes.TData;
import at.lcm.bu21.general.dtypes.unit.ConversionException;
import at.lcm.bu21.general.utils.Clone;
import at.lcm.bu21.general.utils.CloneException;
import at.lcm.bu21.magopt.docu.DocTexWriter;
import at.lcm.bu21.magopt.utils.MagOptConfig;

/**
 * Parameter for user-defined functions
 * 
 * @author Werner Koppelsttter
 */
public abstract class ParamFunction extends ParamContainer implements IParamFunction, Function {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** Dependency of Parameter */
	private Depend depend = new Depend(this);

	/** Name of class */
	protected String className;

	/** Path of class */
	protected String classPath;

	/** Location of the class */
	protected Location classLocation = Location.ABSOLUTE;

	@Override
	public List<String> listTabNames() {
		List<String> list = super.listTabNames();
		list.set(0, "Fnc Info");
		list.add(list.size() - 1, "Imports/Exports");
		list.add(list.size() - 1, "Help");

		if (this.hasPreview()) {
			list.add("Preview");
		}

		return list;
	}

	/**
	 * Constructor for loading a xml-file
	 * 
	 * @throws Exception
	 */
	ParamFunction() throws Exception {
		this("XML_Load", "XML load constructor");
	}

	/**
	 * Constructor with name
	 * 
	 * @param name Name of Parameter
	 * @throws Exception
	 */
	public ParamFunction(String name) throws Exception {
		this(name, name);
	}

	/**
	 * Constructor with name and description
	 * 
	 * @param name        Name of Parameter
	 * @param description Description of Parameter
	 * @throws Exception
	 */
	public ParamFunction(String name, String description) throws Exception {
		super(name, description);
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (!super.equals(obj)) {
			return false;
		}
		if (this.getClass() != obj.getClass()) {
			return false;
		}
		ParamFunction other = (ParamFunction) obj;
		if (this.depend == null) {
			if (other.depend != null) {
				return false;
			}
		} else if (!this.depend.equals(other.depend)) {
			return false;
		}
		return true;
	}

	@Override
	public ParamFunction deepClone() throws CloneException {
		return (ParamFunction) Clone.deepClone(this);
	}

	@Override
	void cloneParam(Parameter clone) throws CloneException {

		super.cloneParam(clone);

		if (clone instanceof ParamFunction) {
			ParamFunction other = (ParamFunction) clone;

			this.depend.cloneDepend(other.depend);
		}
	}

	@Override
	void cloneParam(Parameter clone, Dataset dsThis, Dataset dsClone, boolean docFiles) throws CloneException {
		TData data;

		super.cloneParam(clone, dsThis, dsClone, docFiles);

		if (clone instanceof ParamFunction) {
			ParamFunction other = (ParamFunction) clone;

			data = dsThis.getValue(this.getPId(), ParamSetKey.DATA);
			if (data != null) {
				dsClone.setValue(other.getPId(), ParamSetKey.DATA, data.deepClone());
			} else {
				dsClone.rmValue(other.getPId(), ParamSetKey.DATA);
			}

			data = dsThis.getValue(this.getPId(), ParamSetKey.UPDATEREQUIRED);
			if (data != null) {
				dsClone.setValue(other.getPId(), ParamSetKey.UPDATEREQUIRED, data.deepClone());
			} else {
				dsClone.rmValue(other.getPId(), ParamSetKey.UPDATEREQUIRED);
			}

			data = dsThis.getValue(this.getPId(), ParamSetKey.RESET);
			if (data != null) {
				dsClone.setValue(other.getPId(), ParamSetKey.RESET, data.deepClone());
			} else {
				dsClone.rmValue(other.getPId(), ParamSetKey.RESET);
			}
		}

	}

	@Override
	protected void convertParamContainer(ParamContainer other) throws Exception {
		// FIXME usefull?
		super.convertParamContainer(other);
		if (other instanceof ParamFunction) {
			this.depend.cloneDepend(((ParamFunction) other).depend);
		}
	}

	@Override
	public abstract String getClassName();

	/**
	 * Set name of class
	 * 
	 * @param className Name
	 */
	public abstract void setClassName(String className);

	@Override
	public String getClassPath() {
		String loc;
		String cp;

		loc = this.getLocation();
		if ((loc == null) || loc.isEmpty()) {
			cp = this.classPath;
		} else {
			if ((this.classPath == null) || this.classPath.isEmpty()) {
				cp = loc;
			} else {
				cp = this.getLocation() + File.separator + this.classPath;
			}
		}

		return cp;
	}

	/**
	 * Set path of class
	 * 
	 * @param classPath Path
	 */
	public void setClassPath(String classPath) {
		if (classPath == null) {
			classPath = "";
		}
		this.classPath = this.setLocation(classPath);
	}

	/**
	 * Get name of library-package
	 * 
	 * @return library-package
	 */
	public abstract String getLibraryPackage();

	/**
	 * Refresh class-file location
	 */
	abstract void refreshClass();

	/**
	 * Get location of function
	 * 
	 * @return Location
	 */
	public abstract String getLocation();

	/**
	 * Get location from location-code
	 * 
	 * @param location Location-code
	 * @return Location
	 */
	protected String getLocation(Location location) {
		return location.getPath(this);
	}

	/**
	 * Get location-string from location-code
	 * 
	 * @param location Location-code
	 * @return Location
	 */
	protected String getLocationString(Location location) {
		return location.name();
	}

	/**
	 * Set location from path
	 * 
	 * @param path Path
	 * @return updated path
	 */
	protected String setLocation(String path) {
		String location;

		if ((path == null) || path.isEmpty()) {
			this.classLocation = Location.ABSOLUTE;
			path = "";
		} else {
			path = (new File(path)).getPath();
			if (!path.endsWith(File.separator)) {
				path += File.separator;
			}

			for (Location loc : Location.functionValues()) {
				if (loc == Location.ABSOLUTE) {
					this.classLocation = loc;
					break;
				} else {
					location = loc.getPath(this);
					if (location != null) {
						location = location.toLowerCase();
						if (!location.endsWith(File.separator)) {
							location += File.separator;
						}
						if (path.toLowerCase().startsWith(location)) {
							this.classLocation = loc;
							path = path.substring(location.length());
							break;
						}
					}
				}
			}
			if (path.endsWith(File.separator) && !path.endsWith(":" + File.separator)) {
				path = path.substring(0, path.length() - 1);
			}

		}

		return (new File(path)).getPath();
	}

	@Override
	public Depend getDepend() {
		return this.depend;
	}

	/**
	 * Update function
	 * 
	 * @param status Status of function
	 */
	protected void update(int status) {
		Parameter root;

		root = this.getRoot();
		if (root instanceof Project) {
			((Project) root).check();
		}
		if (status == 0) {
			this.reset();
			this.setUpdateRequired(false);
		}
	}

	@Override
	public TData getData(int index) {
		Dataset ds;
		TData data = null;

		ds = this.getDataset(index);
		if (ds != null) {
			data = ds.getValue(this.getPId(), ParamSetKey.DATA);
		}
		try {
			data = data.deepClone();
		} catch (Exception e) {
			data = super.getData(index);
		}

		return data;
	}

	@Override
	public void setData(TData data) throws Exception {
		this.setData(this.getIndex(), data);
	}

	/**
	 * Set data of current dataset
	 * 
	 * @param index Index of dataset
	 * @param data  data to set
	 * @throws Exception
	 */
	public void setData(int index, TData data) throws Exception {
		Dataset ds;
		TData tmp;

		tmp = this.getData(index);

		if (!tmp.equals(data)) {
			ds = this.getDataset(index);
			if (ds != null) {
				ds.setValue(this.getPId(), ParamSetKey.DATA, data.deepClone());
			}
			this.setModified(index, true);
		}
	}

	@Override
	public boolean getUpdateRequired() {
		return this.getUpdateRequired(this.getIndex());
	}

	@Override
	public boolean getUpdateRequired(int index) {
		Dataset ds;
		TData value;
		boolean bool = true;

		ds = this.getDataset(index);
		if (ds != null) {
			value = ds.getValue(this.getPId(), ParamSetKey.UPDATEREQUIRED);
			if (value != null) {
				bool = ((TBoolean) value).getBoolean();
			}
		}

		return bool;
	}

	@Override
	public void setUpdateRequired(boolean update) {
		this.setUpdateRequired(this.getIndex(), update);
		// FIXME setUncertain (refresh)
		this.refreshFlags();
	}

	@Override
	public void setUpdateRequired(int index, boolean update) {
		Dataset ds;

		if (update) {
			this.setResetFlag(index, update);
		}

		if (update != this.getUpdateRequired(index)) {
			ds = this.getDataset(index);
			if (ds != null) {
				ds.setValue(this.getPId(), ParamSetKey.UPDATEREQUIRED, new TBoolean(update));
			}
			this.resetDocFiles(index);
		}
	}

	/**
	 * Get status of reset-flag
	 * 
	 * @return reset-flag
	 */
	public boolean getResetFlag() {
		return this.getResetFlag(this.getIndex());
	}

	/**
	 * Get status of reset-flag
	 * 
	 * @param index Index of flag
	 * @return reset-flag
	 */
	public boolean getResetFlag(int index) {
		Dataset ds;
		TData value;
		boolean bool = false;

		ds = this.getDataset(index);
		if (ds != null) {
			value = ds.getValue(this.getPId(), ParamSetKey.RESET);
			if (value != null) {
				bool = ((TBoolean) value).getBoolean();
			}
		}

		return bool;
	}

	/**
	 * Set status of reset-flag
	 * 
	 * @param reset reset-flag
	 */
	public void setResetFlag(boolean reset) {
		this.setResetFlag(this.getIndex(), reset);
	}

	/**
	 * Set status of reset-flag
	 * 
	 * @param index Index of flag
	 * @param reset reset-flag
	 */
	public void setResetFlag(int index, boolean reset) {
		Dataset ds;

		ds = this.getDataset(index);
		if (ds != null) {
			ds.setValue(this.getPId(), ParamSetKey.RESET, new TBoolean(reset));
		}
	}

	@Override
	void setError(int index, boolean error) {
		if (error) {
			this.setResetFlag(index, true);
		}
		super.setError(index, error);
	}

	@Override
	public void setConstant(boolean constant) {
		if (constant) {
			this.warning("Set Function to constant is not allowed!");
		}
	}

	@Override
	protected boolean refreshIncomplete() {
		super.refreshIncomplete();
		if (this.getUpdateRequired()) {
			this.setIncomplete(true);
		}
		return this.getIncomplete();
	}

	@Override
	public void refreshDepend(boolean dirty, boolean load) {
		this.depend.refreshDepend(dirty, load);
	}

	@Override
	public abstract void show();

	/**
	 * Draw preview to panel
	 * 
	 * @param panel
	 */
	public abstract void preview(JPanel panel);

	@Override
	public String toString() {
		StringBuffer str = new StringBuffer();
		str.append("ParamFunction\n");
		str.append("\n\tName        : ");
		str.append(this.getName());
		str.append("\n\tDescription : ");
		str.append(this.getDescription());
		str.append("\n\tParams:       ");

		for (Parameter p : this.getParam()) {
			str.append("\n\t\tName: " + p.getName());
		}

		return str.toString();
	}

	@Override
	void xml2obj(XMLStreamReader xmlReader) throws Exception {
		String element;

		xmlReader.next();
		while (xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT) {
			element = xmlReader.getLocalName();
			if (element.equals("ParamContainer")) {
				super.xml2obj(xmlReader);
			} else if (element.equals("Depend")) {
				this.depend.xml2obj(xmlReader);
			} else {
				throw new Exception("Unrecognized property: " + xmlReader.getLocalName());
			}
		}
		xmlReader.next();

	}

	@Override
	void obj2xml(XMLStreamWriter xmlWriter) throws Exception {

		xmlWriter.writeStartElement("ParamFunction");
		xmlWriter.writeAttribute("object", this.getClass().getSimpleName());
		xmlWriter.writeAttribute("name", this.getName());

		// super class
		super.obj2xml(xmlWriter);

		// Depend
		this.depend.obj2xml(xmlWriter);

		xmlWriter.writeEndElement();
	}

	@Override
	public void reportDepend(DocTexWriter content) {
		this.getDepend().reportImport(content);
		this.getDepend().reportExport(content);
	}

	@Override
	List<ParamSetKey> listParamSetKey() {
		List<ParamSetKey> list;

		list = super.listParamSetKey();
		list.add(ParamSetKey.DATA);
		list.add(ParamSetKey.UPDATEREQUIRED);
		list.add(ParamSetKey.RESET);

		return list;
	}

	@Override
	public abstract String getFunctionID();

	@Override
	public abstract boolean hasPreview();

	@Override
	public abstract boolean hasShow();

	protected DependItem getImportItem(String name) {
		DependItem imp = null;

		for (DependItem item : this.depend.getImports()) {
			if (item.getVarName().equals(name)) {
				imp = item;
			}
		}

		return imp;
	}

	protected DependItem getExportItem(String name) {
		DependItem exp = null;

		for (DependItem item : this.depend.getExports()) {
			if (item.getVarName().equals(name)) {
				exp = item;
			}
		}

		return exp;
	}

	@Override
	public TData getImportValue(String name) throws Exception {
		DependItem depend;
		Parameter param;
		TData data = null;
		boolean checkOptional = false;
		String tmp;

		tmp = MagOptConfig.getConfig("EnableCheckOptional");
		if (tmp != null) {
			checkOptional = Boolean.parseBoolean(tmp);
		}

		depend = this.getImportItem(name);
		if (depend != null) {
			param = depend.getHandle();
			if (param != null) {
				if (param instanceof ParamField) {
					try {
						data = ((ParamField) param).getData(depend.getVarDesc().getUnit());
					} catch (ConversionException e) {
						Exception ex = new Exception("Error by converting import " + name + ": " + e.getMessage());
						ex.setStackTrace(e.getStackTrace());
						throw ex;
					}
				} else {
					data = param.getData();
				}
			} else {
				if (checkOptional) {
					if (!depend.getVarDesc().getOptional()) {
						throw new Exception("Import '" + name + "' is not optional");
					}
				}
			}
		}

		return data;

	}

	@Override
	public TData getExportValue(String name) throws Exception {
		DependItem depend;
		Parameter param;
		TData data = null;
		boolean checkOptional = false;
		String tmp;

		tmp = MagOptConfig.getConfig("EnableCheckOptional");
		if (tmp != null) {
			checkOptional = Boolean.parseBoolean(tmp);
		}

		depend = this.getExportItem(name);
		if (depend != null) {
			param = depend.getHandle();
			if (param != null) {
				if (param instanceof ParamField) {
					try {
						data = ((ParamField) param).getData(depend.getVarDesc().getUnit());
					} catch (ConversionException e) {
						Exception ex = new Exception("Error by converting export " + name + ": " + e.getMessage());
						ex.setStackTrace(e.getStackTrace());
						throw ex;
					}
				} else {
					data = param.getData();
				}
			} else {
				if (checkOptional) {
					if (!depend.getVarDesc().getOptional()) {
						this.warning("Export '" + name + "' is not optional");
					}
				}
			}
		}

		return data;
	}

	@Override
	public void setExportValue(String name, TData data) throws Exception {
		DependItem depend;
		Parameter param;
		boolean checkOptional = false;
		String tmp;

		tmp = MagOptConfig.getConfig("EnableCheckOptional");
		if (tmp != null) {
			checkOptional = Boolean.parseBoolean(tmp);
		}

		depend = this.getExportItem(name);
		if (depend != null) {
			param = depend.getHandle();
			if (param != null) {
				if (data != null) {
					if (param instanceof ParamField) {
						try {
							((ParamField) param).setData(data, depend.getVarDesc().getUnit());
						} catch (ConversionException e) {
							Exception ex = new Exception("Error by converting export " + name + ": " + e.getMessage());
							ex.setStackTrace(e.getStackTrace());
							throw ex;
						}
					} else {
						param.setData(data);
					}
				} else {
					throw new Exception("Export value for '" + name + "' doesn't exist");
				}
			} else {
				if (checkOptional) {
					if (!depend.getVarDesc().getOptional()) {
						throw new Exception("Export '" + name + "' is not optional");
					}
				}
			}
		}
	}

	@Override
	public String getImportUnit(String name) {
		DependItem depend;
		Parameter param;
		String unit = null;

		depend = this.getImportItem(name);
		if (depend != null) {
			unit = depend.getVarDesc().getUnit();
			if ((unit == null) || unit.isEmpty()) {
				param = depend.getHandle();
				if ((param != null) && (param instanceof ParamField)) {
					unit = ((ParamField) depend.getHandle()).getUnit();
				}
			}
		}

		return unit;
	}

	@Override
	public String getExportUnit(String name) {
		String unit = null;
		DependItem depend;
		Parameter param;

		depend = this.getExportItem(name);
		if (depend != null) {
			unit = depend.getVarDesc().getUnit();
			if ((unit == null) || unit.isEmpty()) {
				param = depend.getHandle();
				if ((param != null) && (param instanceof ParamField)) {
					unit = ((ParamField) depend.getHandle()).getUnit();
				}
			}
		}

		return unit;
	}

}
