/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9713 $
 * $Date: 2020-03-19 09:28:58 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.magopt.structure;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.Serializable;
import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.LinkedList;

import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.XMLStreamWriter;

import at.lcm.bu21.magopt.structure.client.CReportParameter;
import at.lcm.bu21.magopt.structure.rmi.RReportParameter;

/**
 * Report-parameters
 * 
 * @author Werner Koppelsttter
 */
public class ReportParameter implements IReportParameter {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/**
	 * Enumeration of languages
	 * 
	 * @author Werner Koppelsttter
	 */
	public enum Language {
		/** German */
		GERMAN("German", "ngerman"),
		/** English */
		ENGLISH("English", "american");

		/** Name of language */
		private String name;

		/** Tex-code of language */
		private String tex;

		/**
		 * Create a language
		 * 
		 * @param name
		 *            Name of language
		 * @param tex
		 *            Tex-code of language
		 */
		private Language(String name, String tex) {
			this.name = name;
			this.tex = tex;
		}

		/**
		 * Get tex-code of language
		 * 
		 * @return Tex-code
		 */
		public String getTex() {
			return this.tex;
		}

		@Override
		public String toString() {
			return this.name;
		}
	}

	/**
	 * Report-property
	 * 
	 * @author Werner Koppelsttter
	 */
	public class ReportProperty implements Serializable,
			Comparable<ReportProperty> {
		/** Serialization version number */
		private static final long serialVersionUID = 1L;

		/** Key of property */
		private String key;

		/** Value of property */
		private String value;

		/** Property is valid */
		private boolean valid = false;

		/**
		 * Create property for key
		 * 
		 * @param key
		 *            Key of entry
		 */
		public ReportProperty(String key) {
			this(key, "");
		}

		/**
		 * Create property for key and value
		 * 
		 * @param key
		 * @param value
		 */
		public ReportProperty(String key, String value) {
			this.key = key;
			this.value = value;
		}

		@Override
		public boolean equals(Object obj) {
			if (this == obj) {
				return true;
			}
			if (obj == null) {
				return false;
			}
			if (this.getClass() != obj.getClass()) {
				return false;
			}
			ReportProperty other = (ReportProperty) obj;
			if (!this.getOuterType().equals(other.getOuterType())) {
				return false;
			}
			if (this.key == null) {
				if (other.key != null) {
					return false;
				}
			} else if (!this.key.equals(other.key)) {
				return false;
			}
			if (this.value == null) {
				if (other.value != null) {
					return false;
				}
			} else if (!this.value.equals(other.value)) {
				return false;
			}
			if (this.valid != other.valid) {
				return false;
			}
			return true;
		}

		/**
		 * Get key of property
		 * 
		 * @return Key of property
		 */
		public String getKey() {
			return this.key;
		}

		/**
		 * Get value of property
		 * 
		 * @return Value of property
		 */
		public String getValue() {
			return this.value;
		}

		/**
		 * Set value of property
		 * 
		 * @param value
		 *            Value of property
		 */
		public void setValue(String value) {
			if (value != null) {
				this.value = value;
			} else {
				this.value = "";
			}
		}

		/**
		 * Is property valid
		 * 
		 * @return Valid of property
		 */
		public boolean isValid() {
			return this.valid;
		}

		@Override
		public int compareTo(ReportProperty other) {
			return this.key.compareTo(other.key);
		}

		/**
		 * Get outer object
		 * 
		 * @return Outer object
		 */
		private ReportParameter getOuterType() {
			return ReportParameter.this;
		}

	}

	/** Path to style-files */
	private static final String STYLEPATH = "Documentation/General/Styles";

	/** Name of Style file */
	private static final String STYLENAME = "ReportStyle";

	/** Selected Style */
	private String style = "MAGOPT";

	/** Selected Language */
	private Language language = Language.GERMAN;

	/**
	 * List of report-properties
	 */
	private ArrayList<ReportProperty> properties = new ArrayList<ReportProperty>();

	/**
	 * Get list of available styles
	 * 
	 * @return List of styles
	 */
	public static String[] getStyles() {
		File dir;
		LinkedList<String> styles;

		styles = new LinkedList<String>();
		dir = new File(Project.getInstallPath(), ReportParameter.STYLEPATH);
		for (File f : dir.listFiles()) {
			if (f.isDirectory()
					&& new File(f, f.getName() + ReportParameter.STYLENAME
							+ ".sty").exists()) {
				styles.add(f.getName());
			}
		}
		return styles.toArray(new String[0]);
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (obj == null) {
			return false;
		}
		if (this.getClass() != obj.getClass()) {
			return false;
		}
		ReportParameter other = (ReportParameter) obj;
		if (this.language != other.language) {
			return false;
		}
		if (this.properties == null) {
			if (other.properties != null) {
				return false;
			}
		} else if (!this.properties.equals(other.properties)) {
			return false;
		}
		if (this.style == null) {
			if (other.style != null) {
				return false;
			}
		} else if (!this.style.equals(other.style)) {
			return false;
		}
		return true;
	}

	@Override
	public String getStylePath() {
		return ReportParameter.STYLEPATH + "/" + this.style + "/" + this.style
				+ ReportParameter.STYLENAME;
	}

	@Override
	public String getStyle() {
		return this.style;
	}

	@Override
	public void setStyle(String style) {
		LinkedList<String> props;
		ReportProperty prop;
		File file;
		BufferedReader reader = null;
		boolean read = true;
		String line;

		this.style = style;

		// read style properties
		props = new LinkedList<String>();
		file = new File(Project.getInstallPath(), this.getStylePath() + ".sty");
		try {
			reader = new BufferedReader(new FileReader(file));
			while (read && reader.ready()) {
				line = reader.readLine().trim();
				if (line.startsWith("\\newcommand{\\")) {
					props.add(line.substring(line.indexOf('{') + 2,
							line.indexOf('}')).trim());
				} else if (line.startsWith("\\usepackage{")) {
					read = false;
				}
			}
		} catch (FileNotFoundException e) {
		} catch (IOException e) {
		} finally {
			if (reader != null) {
				try {
					reader.close();
				} catch (IOException e) {
				}
			}
		}

		// update properties
		for (ReportProperty p : this.properties) {
			p.valid = false;
		}
		for (String str : props) {
			prop = this.getProperty(str);
			if (prop == null) {
				prop = new ReportProperty(str);
				this.properties.add(prop);
			}
			prop.valid = true;
		}
	}

	@Override
	public Language getLanguage() {
		return this.language;
	}

	@Override
	public void setLanguage(Language lang) {
		this.language = lang;
	}

	@Override
	public ReportProperty[] getProperties() {
		LinkedList<ReportProperty> props;
		ReportProperty[] list;

		props = new LinkedList<ReportProperty>();
		for (ReportProperty entry : this.properties) {
			if (entry.isValid()) {
				props.add(entry);
			}
		}

		list = props.toArray(new ReportProperty[0]);
		Arrays.sort(list);

		return list;
	}

	@Override
	public ReportProperty getProperty(String key) {
		ReportProperty prop = null;
		for (ReportProperty p : this.properties) {
			if (p.key.compareTo(key) == 0) {
				prop = p;
			}
		}
		return prop;
	}

	// public void addProperty(String key) {
	// if (this.getProperty(key) == null) {
	// this.properties.add(new PropertyEntry(key));
	// }
	// }

	// public void rmProperty(String key) {
	// ReportProperty prop;
	//
	// prop = this.getProperty(key);
	// if (prop != null) {
	// this.properties.remove(prop);
	// }
	// }

	@Override
	public void setProperty(String key, String value) {
		ReportProperty prop;

		prop = this.getProperty(key);
		if (prop != null) {
			prop.setValue(value);
		}
	}

	/**
	 * Read object from xml-file
	 * 
	 * @param xmlReader
	 *            XMLReader to read xml-file
	 * @throws Exception
	 */
	public void xml2obj(XMLStreamReader xmlReader) throws Exception {
		String element;
		String str;

		xmlReader.next();
		while (xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT) {
			element = xmlReader.getLocalName();
			if (element.equals("Style")) {
				xmlReader.next();
				this.style = new String();
				while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
					this.style += xmlReader.getText();
					xmlReader.next();
				}
				xmlReader.next();
			} else if (element.equals("Language")) {
				xmlReader.next();
				str = new String();
				while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
					str += xmlReader.getText();
					xmlReader.next();
				}
				this.language = Language.valueOf(str);
				xmlReader.next();
			} else if (element.equals("Properties")) {
				xmlReader.next();
				this.properties.clear();
				while (xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT) {
					ReportProperty prop;
					xmlReader.next();

					// Key
					xmlReader.next();
					str = new String();
					while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
						str += xmlReader.getText();
						xmlReader.next();
					}
					prop = new ReportProperty(str);
					xmlReader.next();

					// Value
					xmlReader.next();
					str = new String();
					while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
						str += xmlReader.getText();
						xmlReader.next();
					}
					prop.setValue(str);
					xmlReader.next();

					prop.valid = true;
					this.properties.add(prop);
					xmlReader.next();
				}
				xmlReader.next();
			} else {
				throw new Exception("Unrecognized property: "
						+ xmlReader.getLocalName());
			}
		}
		xmlReader.next();
	}

	/**
	 * Write object to xml-file
	 * 
	 * @param xmlWriter
	 *            XMLWriter to write xml-file
	 * @throws Exception
	 */
	public void obj2xml(XMLStreamWriter xmlWriter) throws Exception {
		ReportProperty[] props;

		xmlWriter.writeStartElement("ReportParameter");

		// Style
		xmlWriter.writeStartElement("Style");
		xmlWriter.writeCharacters(this.style);
		xmlWriter.writeEndElement();

		// Lang
		xmlWriter.writeStartElement("Language");
		xmlWriter.writeCharacters(this.language.name());
		xmlWriter.writeEndElement();

		// Properties
		xmlWriter.writeStartElement("Properties");
		props = this.getProperties();
		for (ReportProperty p : props) {
			if (p.isValid()) {
				xmlWriter.writeStartElement("prop");

				xmlWriter.writeStartElement("Key");
				if (p.getKey() != null) {
					xmlWriter.writeCharacters(p.getKey());
				}
				xmlWriter.writeEndElement();

				xmlWriter.writeStartElement("Value");
				if (p.getValue() != null) {
					xmlWriter.writeCharacters(p.getValue());
				}
				xmlWriter.writeEndElement();

				xmlWriter.writeEndElement();
			}
		}
		xmlWriter.writeEndElement();

		xmlWriter.writeEndElement();

	}

	/**
	 * Get RMI-adapter of parameter
	 * 
	 * @return RMI-adapter
	 * @throws RemoteException
	 */
	public IReportParameter getRMIImpl() throws RemoteException {
		return new RReportParameter(this);
	}

	@Override
	public IReportParameter getClientImpl() {
		return new CReportParameter(this);
	}

}
