/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9713 $
 * $Date: 2020-03-19 09:28:58 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.magopt.structure.converter;

import java.io.File;
import java.util.ArrayList;
import java.util.List;

import at.lcm.bu21.general.utils.XMLNode;

/**
 * Change structure of parameter specially for directories- and files-parameter
 * and change old function for new strucutre
 * 
 * @author Werner Koppelsttter
 */
class Param {
	/** Parameter-identifier */
	String pid;
	/** Parameter-name */
	String name;
	/** Access-flag for container */
	String access = null;
	/** Location of directory */
	String location = "ABSOLUTE";
	/** Path to directory */
	String dir = null;
	/** File-name of file */
	String file = null;
	/** Checksum of file */
	String checksum = null;
	/** Parent-parameter of tree */
	Param parent = null;
	/** Child-parameters */
	Param[] childs = null;
	/** Reference to xml-structure */
	XMLNode node = null;
	/** Reference to depend directory */
	Param linkedDir = null;
	/** Constant-flag of parameter */
	boolean constant = true;
}

/**
 * Convert to V_0_35.
 * 
 * - Change ParamDirectory and ParamFile structre
 * 
 * @author Werner Koppelsttter
 */
public class Converter_35 {

	/**
	 * Convert the given node
	 * 
	 * @param node
	 *            Data to convert
	 * @return Changed data node
	 */
	static XMLNode convert(XMLNode node) {
		XMLNode dataset;
		Param struct;
		ArrayList<Param> directories;

		// get list of datasets
		dataset = node.getNode(2);

		// remove extension entry
		Converter_35.convertRemoveExtension(dataset);

		// create tree-structure
		directories = new ArrayList<Param>();
		struct = Converter_35.createStructure(node, directories);
		struct.node = node;

		// link depend directories
		for (Param p : directories) {
			Converter_35.linkDirectory(p);
		}

		// convert structure of directory
		Converter_35.convertDirectoryStructure(directories, dataset);

		// update functions
		Converter_35.convertFunctions(node);

		return node;
	}

	/**
	 * Convert directory by remove extension
	 * 
	 * @param node
	 *            Data to convert
	 */
	private static void convertRemoveExtension(XMLNode node) {
		int dir, ext;

		for (XMLNode n : node.getNodes()) {
			if (n.getName().equals("DS")) {
				// convert all datasets
				for (XMLNode item : n.getNodes()) {
					for (XMLNode entry : item.getNodes()) {
						if (entry.getName().equals("ParamData")) {
							for (XMLNode paramSet : entry.getNodes()) {
								// find dataset-value-list
								for (XMLNode data : paramSet.getNodes()) {
									if (data.getName().equals("Data")) {
										dir = -1;
										ext = -1;
										// search dataset-values for directory
										// and extension
										for (int k = 0; k < data.getNodeSize(); k++) {
											if (data.getNode(k).getName()
													.equals("DIRECTORY")) {
												dir = k;
											} else if (data.getNode(k)
													.getName()
													.equals("EXTENSION")) {
												ext = k;
											}
										}
										// if extension exist remove this item
										if (ext != -1) {
											if (dir != -1) {
												data.getNode(dir)
														.getNode(0)
														.setText(
																data.getNode(
																		dir)
																		.getNode(
																				0)
																		.getText()
																		+ data.getNode(
																				ext)
																				.getNode(
																						0)
																				.getText());
											}
											data.rmNode(ext);
										}
									}
								}
							}
						}
					}
				}
			}
		}
	}

	/**
	 * Create tree structure from xml-data
	 * 
	 * @param node
	 *            XMLNode
	 * @param dirs
	 *            List of directories
	 * @return Tree structure
	 */
	private static Param createStructure(XMLNode node, List<Param> dirs) {
		Param param;
		XMLNode list;

		if (node.getName().equals("Parameter")) {
			param = new Param();
			param.pid = node.getNode(0).getText();
			param.name = node.getNode(1).getText();
			param.constant = Boolean.parseBoolean(node.getNode(4).getText());
		} else {
			param = Converter_35.createStructure(node.getNode(0), dirs);
			if (node.getName().equals("ParamDirectory")) {
				dirs.add(param);
			} else if (node.getName().equals("ParamContainer")) {
				param.access = node.getNode(2).getText();

				list = node.getNode(1);
				param.childs = new Param[list.getNodeSize()];

				for (int k = 0; k < param.childs.length; k++) {
					param.childs[k] = Converter_35.createStructure(
							list.getNode(k), dirs);
					param.childs[k].parent = param;
					param.childs[k].node = list.getNode(k);
				}
			}
		}
		return param;
	}

	/**
	 * Create dependency link of directory
	 * 
	 * @param param
	 */
	private static void linkDirectory(Param param) {
		XMLNode node;
		Param p;
		ArrayList<Param> list;
		String ident, paramName;

		// get directory node
		node = param.node;
		if (node.getName().equals("ParamFile")) {
			node = node.getNode(0);
		}
		// get imports
		node = node.getNode(1).getNode(0);
		if (node.getNodeSize() > 0) {
			// get parameter-name of import
			paramName = node.getNode(0).getNode(1).getText();
			if (paramName != null) {
				// list objectives for identifier
				list = new ArrayList<Param>();
				// object by unique identifier
				ident = param.name;
				p = param;
				while (p.parent != null) {
					p = p.parent;
					ident = p.name + "." + ident;
				}
				if (ident.equals(paramName)) {
					list.add(param);
				}
				// object by identifier
				p = param.parent;
				while (p != null) {
					if ((p.parent == null) || (!p.access.equals("PUBLIC"))) {
						Converter_35.listObject(p, paramName, list, true);
					}
					p = p.parent;
				}
				// remove cicles
				while (!list.isEmpty() && (list.get(0) == param)) {
					list.remove(0);
				}
				if (!list.isEmpty()
						&& list.get(0).node.getName().equals("ParamDirectory")) {
					param.linkedDir = list.get(0);
				} else {
					// link to root (project)
					param.location = "PROJECT";
				}
			} else {
				// link to root (project)
				param.location = "PROJECT";
			}
		}
	}

	/**
	 * List available parameters for identifier
	 * 
	 * @param param
	 *            Start parameter
	 * @param ident
	 *            Identifier
	 * @param list
	 *            List of objects
	 * @param sub
	 *            Check sub-parameters
	 */
	private static void listObject(Param param, String ident, List<Param> list,
			boolean sub) {

		// replace . for project
		if ((param.parent == null) && ident.startsWith(".")) {
			if (ident.length() == 1) {
				ident = param.name;
			} else {
				ident = param.name + ident;
			}
		}

		// find parameter
		if (ident.equals(param.name)) {
			// direct match
			list.add(param);
			if ((param.childs != null)
					&& (sub || param.access.equals("PUBLIC"))) {
				for (Param p : param.childs) {
					Converter_35.listObject(p, ident, list, false);
				}
			}
		} else {
			if (ident.startsWith(param.name + ".")) {
				// direct path
				ident = ident.substring(param.name.length() + 1);

				// public interface

				// check sub-parameter
				if ((param.childs != null)
						&& (sub || !param.access.equals("PRIVATE"))) {
					for (Param p : param.childs) {
						Converter_35.listObject(p, ident, list, false);
					}
				}
			} else {
				// public interface

				// check sub-parameter
				if ((param.childs != null)
						&& (sub || param.access.equals("PUBLIC"))) {
					for (Param p : param.childs) {
						Converter_35.listObject(p, ident, list, false);
					}
				}
			}
		}
	}

	/**
	 * Convert directroy stucture from depend-parameter to location entry
	 * 
	 * @param directories
	 *            List of directories
	 * @param datasets
	 *            List of datasets
	 */
	private static void convertDirectoryStructure(List<Param> directories,
			XMLNode datasets) {
		XMLNode node, n, psd;
		boolean dirty = true;
		Param d;
		String strDir, strLoc;

		// get value of const directories
		// change depent to location
		for (Param dir : directories) {
			node = dir.node;
			if (node.getName().equals("ParamFile")) {
				n = node.getNode(1);
				dir.file = n.getText();
				n.setText(null);

				n = node.getNode(2);
				dir.checksum = n.getText();
				n.setText(null);

				node = node.getNode(0);
			}

			n = node.getNode(2);
			dir.dir = n.getText();
			n.setText(null);

			node.rmNode(1);
			n = new XMLNode("Location");
			n.setText("ABSOLUTE");
			node.addNode(1, n);
		}

		// update const flags
		while (dirty) {
			dirty = false;
			for (Param dir : directories) {
				if (dir.linkedDir != null) {
					if (dir.constant && !dir.linkedDir.constant) {
						dirty = true;
						dir.constant = false;
						node = dir.node;
						if (node.getName().equals("ParamFile")) {
							node = node.getNode(0);
						}
						dir.node.getNode(0).getNode(4).setText("false");
					}
				}
			}
		}

		// update const dir/file
		for (Param dir : directories) {
			if (dir.constant) {
				d = dir.linkedDir;
				while (d != null) {
					if (d.dir != null) {
						if (dir.dir != null) {
							dir.dir = d.dir + File.separator + dir.dir;
						} else {
							dir.dir = d.dir;
						}
					}
					dir.location = d.location;
					d = d.linkedDir;
				}
				dir.linkedDir = null;

				// update xml-node
				node = dir.node;
				if (node.getName().equals("ParamFile")) {
					node.getNode(1).setText(dir.file);
					node.getNode(2).setText(dir.checksum);
					node = node.getNode(0);
				}
				node.getNode(1).setText(dir.location);
				node.getNode(2).setText(dir.dir);
			}
		}

		// update non const dir/file
		// update all dataset
		for (XMLNode ds : datasets.getNode(4).getNodes()) {
			// get data from dataset
			for (Param dir : directories) {
				if (!dir.constant) {
					for (XMLNode ps : ds.getNode(2).getNodes()) {
						if (dir.pid.equals(ps.getAttribute("pid"))) {
							psd = ps.getNode(0);
							for (int k = psd.getNodeSize() - 1; k >= 0; k--) {
								if (psd.getNode(k).getName()
										.equals("DIRECTORY")) {
									dir.dir = psd.getNode(k).getNode(0)
											.getText();
									psd.rmNode(k);
								} else if (psd.getNode(k).getName()
										.equals("FILE")) {
									dir.file = psd.getNode(k).getNode(0)
											.getText();
									psd.rmNode(k);
								} else if (psd.getNode(k).getName()
										.equals("CHECKSUM")) {
									dir.checksum = psd.getNode(k).getNode(0)
											.getText();
									psd.rmNode(k);
								}
							}
						}
					}
				}
			}
			// FIXME

			// set data of dataset
			for (Param dir : directories) {
				if (!dir.constant) {
					strLoc = dir.location;
					strDir = dir.dir;

					d = dir.linkedDir;
					while (d != null) {
						if (d.dir != null) {
							if (strDir != null) {
								strDir = d.dir + File.separator + strDir;
							} else {
								strDir = d.dir;
							}
						}
						strLoc = d.location;
						d = d.linkedDir;
					}

					for (XMLNode ps : ds.getNode(2).getNodes()) {
						if (dir.pid.equals(ps.getAttribute("pid"))) {
							psd = ps.getNode(0);

							n = new XMLNode("DIRECTORY");
							n.addNode(new XMLNode("String"));
							n.getNode(0).setText(strDir);
							psd.addNode(n);

							n = new XMLNode("LOCATION");
							n.addNode(new XMLNode("String"));
							n.getNode(0).setText(strLoc);
							psd.addNode(n);

							if (dir.node.getName().equals("ParamFile")) {
								n = new XMLNode("FILE");
								n.addNode(new XMLNode("String"));
								n.getNode(0).setText(dir.file);
								psd.addNode(n);

								n = new XMLNode("CHECKSUM");
								n.addNode(new XMLNode("String"));
								n.getNode(0).setText(dir.checksum);
								psd.addNode(n);
							}
						}
					}
				}
			}
		}
	}

	/**
	 * Convert functions
	 * 
	 * @param node
	 *            Node to convert
	 */
	private static void convertFunctions(XMLNode node) {
		String fnc;

		if (node.getName().equals("ParamContainer")) {
			for (XMLNode n : node.getNode(1).getNodes()) {
				Converter_35.convertFunctions(n);
			}
		} else if (node.getNodeSize() > 0) {
			Converter_35.convertFunctions(node.getNode(0));
		}

		if (node.getName().equals("ParamMatlabFunction")) {
			// update matlab-function in library
			if (node.getNode(3).getText().equals("LIBRARY")) {
				fnc = node.getNode(1).getText();
				if (fnc.equals("FncFemagACMultIXFTrans")
						|| fnc.equals("FncFemagDCSingleSim")
						|| fnc.equals("FncFemagDCSingleSimRot")
						|| fnc.equals("FncFemagDCSingleSimTrans")
						|| fnc.equals("FncFemagPMSM")
						|| fnc.equals("FncFemagGeoSetup")
						|| fnc.equals("FncFemagMultIXRot")
						|| fnc.equals("FncFemagMultIXTrans")
						|| fnc.equals("FncFemagMultIxLSP")
						|| fnc.equals("FncFemagPMSM_curr_harmonic")
						|| fnc.equals("FncFemagParamIdent")
						|| fnc.equals("FncFemagReluktanz")
						|| fnc.equals("FncFemagSoloPMSMTrans")
						|| fnc.equals("FncFemagSoloMultIXTrans")
						|| fnc.equals("FncFemagLdLq")) {
					Converter_35.removeImport(node, "simDir");
					Converter_35.removeImport(node, "zipFile");
				}
			}
		} else if (node.getName().equals("ParamJavaFunction")) {
			// update java-function in library
			if (node.getNode(3).getText().equals("LIBRARY")) {
				if (node.getNode(1).getText().equals("femag.FncSimulate")) {
					Converter_35.removeImport(node, "outputZipFile");
				}
			}
		}
	}

	/**
	 * Remove name from import-list
	 * 
	 * @param node
	 *            Node of function
	 * @param name
	 *            Name of import
	 */
	private static void removeImport(XMLNode node, String name) {
		XMLNode imports;

		imports = node.getNode(0).getNode(1).getNode(0);
		for (int k = imports.getNodeSize() - 1; k >= 0; k--) {
			if (imports.getNode(k).getNode(0).getText().equals(name)) {
				imports.rmNode(k);
			}
		}
	}
}
