/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9713 $
 * $Date: 2020-03-19 09:28:58 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.magopt.structure.formula;

import java.util.Map;

import at.lcm.bu21.general.dtypes.TArray;
import at.lcm.bu21.general.dtypes.TBoolean;
import at.lcm.bu21.general.dtypes.TData;
import at.lcm.bu21.general.dtypes.TDataException;
import at.lcm.bu21.general.dtypes.TNumeric;
import at.lcm.bu21.general.dtypes.TString;

/**
 * 
 * 
 *
 * @author Werner Koppelsttter
 */
public class EvalFunction {

	/**
	 * Eval function token
	 * 
	 * @param node
	 *            Function node to eval
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws EvalException
	 * @throws UserException
	 * @throws TDataException
	 */
	static TData function_(SyntaxNode node, Map<String, TData> vars)
			throws UserException, EvalException, TDataException {
		TData data;
		if ("error".equals(node.tok.str)) {
			data = EvalFunction.fnc_error(node, vars);
		} else if ("abs".equals(node.tok.str)) {
			data = EvalFunction.fnc_abs(node, vars);
		} else if ("angle".equals(node.tok.str)) {
			data = EvalFunction.fnc_angle(node, vars);
		} else if ("isreal".equals(node.tok.str)) {
			data = EvalFunction.fnc_isreal(node, vars);
		} else if ("real".equals(node.tok.str)) {
			data = EvalFunction.fnc_real(node, vars);
		} else if ("imag".equals(node.tok.str)) {
			data = EvalFunction.fnc_imag(node, vars);
		} else if ("size".equals(node.tok.str)) {
			data = EvalFunction.fnc_size(node, vars);
		} else if ("sum".equals(node.tok.str)) {
			data = EvalFunction.fnc_sum(node, vars);
		} else if ("max".equals(node.tok.str)) {
			data = EvalFunction.fnc_max(node, vars);
		} else if ("min".equals(node.tok.str)) {
			data = EvalFunction.fnc_min(node, vars);
		} else if ("mean".equals(node.tok.str)) {
			data = EvalFunction.fnc_mean(node, vars);
		} else if ("round".equals(node.tok.str)) {
			data = EvalFunction.fnc_round(node, vars);
		} else if ("ceil".equals(node.tok.str)) {
			data = EvalFunction.fnc_ceil(node, vars);
		} else if ("floor".equals(node.tok.str)) {
			data = EvalFunction.fnc_floor(node, vars);
		} else if ("sqrt".equals(node.tok.str)) {
			data = EvalFunction.fnc_sqrt(node, vars);
		} else if ("ln".equals(node.tok.str)) {
			data = EvalFunction.fnc_ln(node, vars);
		} else if ("log".equals(node.tok.str)) {
			data = EvalFunction.fnc_log(node, vars);
		} else if ("sin".equals(node.tok.str)) {
			data = EvalFunction.fnc_sin(node, vars);
		} else if ("cos".equals(node.tok.str)) {
			data = EvalFunction.fnc_cos(node, vars);
		} else if ("tan".equals(node.tok.str)) {
			data = EvalFunction.fnc_tan(node, vars);
		} else if ("asin".equals(node.tok.str)) {
			data = EvalFunction.fnc_asin(node, vars);
		} else if ("acos".equals(node.tok.str)) {
			data = EvalFunction.fnc_acos(node, vars);
		} else if ("atan".equals(node.tok.str)) {
			data = EvalFunction.fnc_atan(node, vars);
		} else if ("sinh".equals(node.tok.str)) {
			data = EvalFunction.fnc_sinh(node, vars);
		} else if ("cosh".equals(node.tok.str)) {
			data = EvalFunction.fnc_cosh(node, vars);
		} else if ("tanh".equals(node.tok.str)) {
			data = EvalFunction.fnc_tanh(node, vars);
		} else if ("todeg".equals(node.tok.str)) {
			data = EvalFunction.fnc_todeg(node, vars);
		} else if ("torad".equals(node.tok.str)) {
			data = EvalFunction.fnc_torad(node, vars);
		} else if ("len".equals(node.tok.str)) {
			data = EvalFunction.fnc_len(node, vars);
		} else if ("substring".equals(node.tok.str)) {
			data = EvalFunction.fnc_substring(node, vars);
		} else if ("concat".equals(node.tok.str)) {
			data = EvalFunction.fnc_concat(node, vars);
		} else if ("cmp".equals(node.tok.str)) {
			data = EvalFunction.fnc_cmp(node, vars);
		} else if ("upper".equals(node.tok.str)) {
			data = EvalFunction.fnc_upper(node, vars);
		} else if ("lower".equals(node.tok.str)) {
			data = EvalFunction.fnc_lower(node, vars);
		} else if ("str".equals(node.tok.str)) {
			data = EvalFunction.fnc_str(node, vars);
		} else if ("num".equals(node.tok.str)) {
			data = EvalFunction.fnc_num(node, vars);
		} else if ("rand".equals(node.tok.str)) {
			data = EvalFunction.fnc_rand(node, vars);
		} else if ("ones".equals(node.tok.str)) {
			data = EvalFunction.fnc_ones(node, vars);
		} else if ("zeros".equals(node.tok.str)) {
			data = EvalFunction.fnc_zeros(node, vars);
		} else if ("eye".equals(node.tok.str)) {
			data = EvalFunction.fnc_eye(node, vars);
		} else {
			throw new EvalException("Undefined function", node.tok);
		}

		return data;
	}

	/**
	 * Check minimal number of function parameters
	 * 
	 * @param node
	 *            Syntax node of function
	 * @param min
	 *            Minimal number
	 * 
	 * @throws EvalException
	 */
	private static void check_FncParameterCount(SyntaxNode node, int min)
			throws EvalException {
		int cnt;

		if (node.childs == null) {
			cnt = 0;
		} else {
			cnt = node.childs.length;
		}
		if (cnt < min) {
			throw new EvalException(
					"To less parameter for funtion " + node.tok.str, node.tok);
		}
	}

	/**
	 * Check minimal and maximal number of function parameters
	 * 
	 * @param node
	 *            Syntax node of function
	 * @param min
	 *            Minimal number
	 * @param max
	 *            Maximal number
	 * 
	 * @throws EvalException
	 */
	private static void check_FncParameterCount(SyntaxNode node, int min,
			int max) throws EvalException {
		int cnt;

		EvalFunction.check_FncParameterCount(node, min);
		if (node.childs == null) {
			cnt = 0;
		} else {
			cnt = node.childs.length;
		}
		if (cnt > max) {
			throw new EvalException(
					"To much parameter for funtion " + node.tok.str, node.tok);
		}
	}

	/**
	 * Error function
	 * 
	 * @param node
	 *            Syntax node of function
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws UserException
	 * @throws EvalException
	 */
	private static TData fnc_error(SyntaxNode node, Map<String, TData> vars)
			throws UserException, EvalException {
		EvalFunction.check_FncParameterCount(node, 0, 1);
		if (node.childs == null) {
			throw new UserException(node.tok);
		} else {
			throw new UserException(SyntaxNode.getString(node.childs[0], vars),
					node.tok);
		}
	}

	/**
	 * Abs function
	 * 
	 * @param node
	 *            Syntax node of function
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws UserException
	 * @throws EvalException
	 * @throws TDataException
	 */
	private static TNumeric fnc_abs(SyntaxNode node, Map<String, TData> vars)
			throws EvalException, UserException, TDataException {
		TNumeric num;
		double re, im;
		double[][] res;

		EvalFunction.check_FncParameterCount(node, 1, 1);
		num = SyntaxNode.getNumeric(node.childs[0], vars);
		res = new double[num.size()[0]][num.size()[1]];
		if (num.isReal()) {
			for (int row = 0; row < res.length; row++) {
				for (int col = 0; col < res[0].length; col++) {
					res[row][col] = Math.abs(num.getReal(row, col));
				}
			}
		} else {
			for (int row = 0; row < res.length; row++) {
				for (int col = 0; col < res[0].length; col++) {
					re = num.getReal(row, col);
					im = num.getImag(row, col);
					res[row][col] = Math.sqrt((re * re) + (im * im));
				}
			}
		}

		return new TNumeric(res);
	}

	/**
	 * Angle function
	 * 
	 * @param node
	 *            Syntax node of function
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws UserException
	 * @throws EvalException
	 * @throws TDataException
	 */
	private static TNumeric fnc_angle(SyntaxNode node, Map<String, TData> vars)
			throws EvalException, UserException, TDataException {
		TNumeric num;
		double re, im, angle;
		double[][] res;

		EvalFunction.check_FncParameterCount(node, 1, 1);
		num = SyntaxNode.getNumeric(node.childs[0], vars);
		res = new double[num.size()[0]][num.size()[1]];
		for (int row = 0; row < res.length; row++) {
			for (int col = 0; col < res[0].length; col++) {
				re = num.getReal(row, col);
				im = num.getImag(row, col);
				if (re == 0) {
					if (im > 0.0) {
						angle = Math.PI / 2.0;
					} else if (im < 0.0) {
						angle = -Math.PI / 2.0;
					} else {
						angle = 0.0;
					}
				} else {
					angle = Math.atan(im / re);
					if (re < 0.0) {
						if (im >= 0.0) {
							angle += Math.PI;
						} else {
							angle -= Math.PI;
						}
					}
				}
				res[row][col] = angle;
			}
		}

		return new TNumeric(res);
	}

	/**
	 * Isreal function - Check if matrix is real (no imaginary part)
	 * 
	 * @param node
	 *            Syntax node of function
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws UserException
	 * @throws EvalException
	 */
	private static TBoolean fnc_isreal(SyntaxNode node, Map<String, TData> vars)
			throws EvalException, UserException {
		EvalFunction.check_FncParameterCount(node, 1, 1);
		return new TBoolean(
				SyntaxNode.getNumeric(node.childs[0], vars).isReal());
	}

	/**
	 * Real function - Extract real part of matrix
	 * 
	 * @param node
	 *            Syntax node of function
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws UserException
	 * @throws EvalException
	 * @throws TDataException
	 */
	private static TNumeric fnc_real(SyntaxNode node, Map<String, TData> vars)
			throws EvalException, UserException, TDataException {
		EvalFunction.check_FncParameterCount(node, 1, 1);
		return new TNumeric(
				SyntaxNode.getNumeric(node.childs[0], vars).getRealMatrix());
	}

	/**
	 * Imag function - Extract imaginary part of matrix
	 * 
	 * @param node
	 *            Syntax node of function
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws UserException
	 * @throws EvalException
	 * @throws TDataException
	 */
	private static TNumeric fnc_imag(SyntaxNode node, Map<String, TData> vars)
			throws EvalException, UserException, TDataException {
		EvalFunction.check_FncParameterCount(node, 1, 1);
		return new TNumeric(null,
				SyntaxNode.getNumeric(node.childs[0], vars).getImagMatrix());
	}

	/**
	 * Size function - Get size of numeric or array
	 * 
	 * @param node
	 *            Syntax node of function
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws UserException
	 * @throws EvalException
	 * @throws TDataException
	 */
	private static TNumeric fnc_size(SyntaxNode node, Map<String, TData> vars)
			throws EvalException, UserException, TDataException {
		TData d;
		int[] size;
		EvalFunction.check_FncParameterCount(node, 1, 1);
		d = node.childs[0].eval(vars);
		if (d instanceof TNumeric) {
			((TNumeric) d).size();
			size = ((TNumeric) d).size();
		} else if (d instanceof TArray) {
			size = ((TArray) d).size();
		} else {
			throw new EvalException("Is not numeric or array type",
					node.childs[0].tok);
		}

		return new TNumeric(new double[] { size[0], size[1] });
	}

	/**
	 * Sum function
	 * 
	 * @param node
	 *            Syntax node of function
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws UserException
	 * @throws EvalException
	 * @throws TDataException
	 */
	private static TNumeric fnc_sum(SyntaxNode node, Map<String, TData> vars)
			throws EvalException, UserException, TDataException {
		TNumeric num, dim;
		int mode = 0, end;
		int[] size;
		double[][] sum;

		EvalFunction.check_FncParameterCount(node, 1);

		// check size and mode
		end = node.childs.length;
		num = SyntaxNode.getReal(node.childs[0], vars);
		size = num.size();
		dim = SyntaxNode.getReal(node.childs[node.childs.length - 1], vars);
		if (node.childs.length == 1) {
			mode = 3;
		} else if (!num.isScalar() && dim.isScalar()) {
			mode = SyntaxNode.getInt(dim.getReal(0, 0));
			end--;
		}

		// initialize sum
		switch (mode) {
		case 0:
			sum = new double[size[0]][size[1]];
			break;
		case 1:
			sum = new double[1][size[1]];
			break;
		case 2:
			sum = new double[size[0]][1];
			break;
		case 3:
			sum = new double[1][1];
			break;
		default:
			throw new EvalException("Undefined mode of sum", node.tok);
		}
		// for (int row = 0; row < sum.length; row++) {
		// for (int col = 0; col < sum[0].length; col++) {
		// sum[row][col] = 0.0;
		// }
		// }

		// sum of all children
		for (int k = 0; k < end; k++) {
			if (k != 0) {
				num = SyntaxNode.getReal(node.childs[k], vars);
			}
			// check dimension
			if ((size[0] != num.size()[0]) || (size[1] != num.size()[1])) {
				throw new EvalException("matrix dimension error",
						node.childs[k].tok);
			}
			// sum of matrix
			for (int row = 0; row < size[0]; row++) {
				for (int col = 0; col < size[1]; col++) {
					switch (mode) {
					case 0:
						sum[row][col] += num.getReal(row, col);
						break;
					case 1:
						sum[0][col] += num.getReal(row, col);
						break;
					case 2:
						sum[row][0] += num.getReal(row, col);
						break;
					case 3:
						sum[0][0] += num.getReal(row, col);
						break;
					}
				}
			}
		}

		return new TNumeric(sum);
	}

	/**
	 * Max function
	 * 
	 * @param node
	 *            Syntax node of function
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws UserException
	 * @throws EvalException
	 * @throws TDataException
	 */
	private static TNumeric fnc_max(SyntaxNode node, Map<String, TData> vars)
			throws EvalException, UserException, TDataException {
		TNumeric num, dim;
		int mode = 0, end;
		int[] size;
		double[][] max;

		EvalFunction.check_FncParameterCount(node, 1);

		// check size and mode
		end = node.childs.length;
		num = SyntaxNode.getReal(node.childs[0], vars);
		size = num.size();
		dim = SyntaxNode.getReal(node.childs[node.childs.length - 1], vars);
		if (node.childs.length == 1) {
			mode = 3;
		} else if (!num.isScalar() && dim.isScalar()) {
			mode = SyntaxNode.getInt(dim.getReal(0, 0));
			end--;
		}

		// initialize min
		switch (mode) {
		case 0:
			max = new double[size[0]][size[1]];
			break;
		case 1:
			max = new double[1][size[1]];
			break;
		case 2:
			max = new double[size[0]][1];
			break;
		case 3:
			max = new double[1][1];
			break;
		default:
			throw new EvalException("Undefined mode of max", node.tok);
		}
		for (int row = 0; row < max.length; row++) {
			for (int col = 0; col < max[0].length; col++) {
				max[row][col] = Double.NEGATIVE_INFINITY;
			}
		}

		// max of all children
		for (int k = 0; k < end; k++) {
			if (k != 0) {
				num = SyntaxNode.getReal(node.childs[k], vars);
			}
			// check dimension
			if ((size[0] != num.size()[0]) || (size[1] != num.size()[1])) {
				throw new EvalException("matrix dimension error",
						node.childs[k].tok);
			}
			// max of matrix
			for (int row = 0; row < size[0]; row++) {
				for (int col = 0; col < size[1]; col++) {
					switch (mode) {
					case 0:
						max[row][col] = Math.max(max[row][col],
								num.getReal(row, col));
						break;
					case 1:
						max[0][col] = Math.max(max[0][col],
								num.getReal(row, col));
						break;
					case 2:
						max[row][0] = Math.max(max[row][0],
								num.getReal(row, col));
						break;
					case 3:
						max[0][0] = Math.max(max[0][0], num.getReal(row, col));
						break;
					}
				}
			}
		}

		return new TNumeric(max);
	}

	/**
	 * Min function
	 * 
	 * @param node
	 *            Syntax node of function
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws UserException
	 * @throws EvalException
	 * @throws TDataException
	 */
	private static TNumeric fnc_min(SyntaxNode node, Map<String, TData> vars)
			throws EvalException, UserException, TDataException {
		TNumeric num, dim;
		int mode = 0, end;
		int[] size;
		double[][] min;

		EvalFunction.check_FncParameterCount(node, 1);

		// check size and mode
		end = node.childs.length;
		num = SyntaxNode.getReal(node.childs[0], vars);
		size = num.size();
		dim = SyntaxNode.getReal(node.childs[node.childs.length - 1], vars);
		if (node.childs.length == 1) {
			mode = 3;
		} else if (!num.isScalar() && dim.isScalar()) {
			mode = SyntaxNode.getInt(dim.getReal(0, 0));
			end--;
		}

		// initialize min
		switch (mode) {
		case 0:
			min = new double[size[0]][size[1]];
			break;
		case 1:
			min = new double[1][size[1]];
			break;
		case 2:
			min = new double[size[0]][1];
			break;
		case 3:
			min = new double[1][1];
			break;
		default:
			throw new EvalException("Undefined mode of min", node.tok);
		}
		for (int row = 0; row < min.length; row++) {
			for (int col = 0; col < min[0].length; col++) {
				min[row][col] = Double.POSITIVE_INFINITY;
			}
		}

		// min of all children
		for (int k = 0; k < end; k++) {
			if (k != 0) {
				num = SyntaxNode.getReal(node.childs[k], vars);
			}
			// check dimension
			if ((size[0] != num.size()[0]) || (size[1] != num.size()[1])) {
				throw new EvalException("matrix dimension error",
						node.childs[k].tok);
			}
			// min of matrix
			for (int row = 0; row < size[0]; row++) {
				for (int col = 0; col < size[1]; col++) {
					switch (mode) {
					case 0:
						min[row][col] = Math.min(min[row][col],
								num.getReal(row, col));
						break;
					case 1:
						min[0][col] = Math.min(min[0][col],
								num.getReal(row, col));
						break;
					case 2:
						min[row][0] = Math.min(min[row][0],
								num.getReal(row, col));
						break;
					case 3:
						min[0][0] = Math.min(min[0][0], num.getReal(row, col));
						break;
					}
				}
			}
		}

		return new TNumeric(min);
	}

	/**
	 * Meanx function
	 * 
	 * @param node
	 *            Syntax node of function
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws UserException
	 * @throws EvalException
	 * @throws TDataException
	 */
	private static TNumeric fnc_mean(SyntaxNode node, Map<String, TData> vars)
			throws EvalException, UserException, TDataException {
		TNumeric num, dim;
		int mode = 0, end;
		int[] size;
		double[][] mean;

		EvalFunction.check_FncParameterCount(node, 1);

		// check size and mode
		end = node.childs.length;
		num = SyntaxNode.getReal(node.childs[0], vars);
		size = num.size();
		dim = SyntaxNode.getReal(node.childs[node.childs.length - 1], vars);
		if (node.childs.length == 1) {
			mode = 3;
		} else if (!num.isScalar() && dim.isScalar()) {
			mode = SyntaxNode.getInt(dim.getReal(0, 0));
			end--;
		}

		// initialize min
		switch (mode) {
		case 0:
			mean = new double[size[0]][size[1]];
			break;
		case 1:
			mean = new double[1][size[1]];
			break;
		case 2:
			mean = new double[size[0]][1];
			break;
		case 3:
			mean = new double[1][1];
			break;
		default:
			throw new EvalException("Undefined mode of mean", node.tok);
		}
		// for (int row = 0; row < mean.length; row++) {
		// for (int col = 0; col < mean[0].length; col++) {
		// mean[row][col] = 0;
		// }
		// }

		// sum of all children
		for (int k = 0; k < end; k++) {
			if (k != 0) {
				num = SyntaxNode.getReal(node.childs[k], vars);
			}
			// check dimension
			if ((size[0] != num.size()[0]) || (size[1] != num.size()[1])) {
				throw new EvalException("matrix dimension error",
						node.childs[k].tok);
			}
			// sum of matrix
			for (int row = 0; row < size[0]; row++) {
				for (int col = 0; col < size[1]; col++) {
					switch (mode) {
					case 0:
						mean[row][col] += num.getReal(row, col);
						break;
					case 1:
						mean[0][col] += num.getReal(row, col);
						break;
					case 2:
						mean[row][0] += num.getReal(row, col);
						break;
					case 3:
						mean[0][0] += num.getReal(row, col);
						break;
					}
				}
			}
		}

		for (int row = 0; row < mean.length; row++) {
			for (int col = 0; col < mean[0].length; col++) {
				switch (mode) {
				case 0:
					mean[row][col] /= end;
					break;
				case 1:
					mean[row][col] /= end * size[0];
					break;
				case 2:
					mean[row][col] /= end * size[1];
					break;
				case 3:
					mean[row][col] /= end * size[0] * size[1];
					break;
				}
			}
		}
		return new TNumeric(mean);
	}

	/**
	 * Round function
	 * 
	 * @param node
	 *            Syntax node of function
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws UserException
	 * @throws EvalException
	 * @throws TDataException
	 */
	private static TNumeric fnc_round(SyntaxNode node, Map<String, TData> vars)
			throws EvalException, UserException, TDataException {
		TNumeric num;
		double[][] res;

		EvalFunction.check_FncParameterCount(node, 1, 1);
		num = SyntaxNode.getReal(node.childs[0], vars);
		res = new double[num.size()[0]][num.size()[1]];
		for (int row = 0; row < res.length; row++) {
			for (int col = 0; col < res[0].length; col++) {
				res[row][col] = Math.round(num.getReal(row, col));
			}
		}

		return new TNumeric(res);
	}

	/**
	 * Ceil function
	 * 
	 * @param node
	 *            Syntax node of function
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws UserException
	 * @throws EvalException
	 * @throws TDataException
	 */
	private static TNumeric fnc_ceil(SyntaxNode node, Map<String, TData> vars)
			throws EvalException, UserException, TDataException {
		TNumeric num;
		double[][] res;

		EvalFunction.check_FncParameterCount(node, 1, 1);
		num = SyntaxNode.getReal(node.childs[0], vars);
		res = new double[num.size()[0]][num.size()[1]];
		for (int row = 0; row < res.length; row++) {
			for (int col = 0; col < res[0].length; col++) {
				res[row][col] = Math.ceil(num.getReal(row, col));
			}
		}

		return new TNumeric(res);
	}

	/**
	 * Floor function
	 * 
	 * @param node
	 *            Syntax node of function
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws UserException
	 * @throws EvalException
	 * @throws TDataException
	 */
	private static TNumeric fnc_floor(SyntaxNode node, Map<String, TData> vars)
			throws EvalException, UserException, TDataException {
		TNumeric num;
		double[][] res;

		EvalFunction.check_FncParameterCount(node, 1, 1);
		num = SyntaxNode.getReal(node.childs[0], vars);
		res = new double[num.size()[0]][num.size()[1]];
		for (int row = 0; row < res.length; row++) {
			for (int col = 0; col < res[0].length; col++) {
				res[row][col] = Math.floor(num.getReal(row, col));
			}
		}

		return new TNumeric(res);
	}

	/**
	 * Sqrt function
	 * 
	 * @param node
	 *            Syntax node of function
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws UserException
	 * @throws EvalException
	 * @throws TDataException
	 */
	private static TNumeric fnc_sqrt(SyntaxNode node, Map<String, TData> vars)
			throws EvalException, UserException, TDataException {
		TNumeric num;
		double[][] res;

		EvalFunction.check_FncParameterCount(node, 1, 1);
		num = SyntaxNode.getReal(node.childs[0], vars);
		res = new double[num.size()[0]][num.size()[1]];
		for (int row = 0; row < res.length; row++) {
			for (int col = 0; col < res[0].length; col++) {
				res[row][col] = Math.sqrt(num.getReal(row, col));
			}
		}

		return new TNumeric(res);
	}

	/**
	 * Ln function
	 * 
	 * @param node
	 *            Syntax node of function
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws UserException
	 * @throws EvalException
	 * @throws TDataException
	 */

	private static TNumeric fnc_ln(SyntaxNode node, Map<String, TData> vars)
			throws EvalException, UserException, TDataException {
		TNumeric num;
		double[][] res;

		EvalFunction.check_FncParameterCount(node, 1, 1);
		num = SyntaxNode.getReal(node.childs[0], vars);
		res = new double[num.size()[0]][num.size()[1]];
		for (int row = 0; row < res.length; row++) {
			for (int col = 0; col < res[0].length; col++) {
				res[row][col] = Math.log(num.getReal(row, col));
			}
		}

		return new TNumeric(res);
	}

	/**
	 * Log function
	 * 
	 * @param node
	 *            Syntax node of function
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws UserException
	 * @throws EvalException
	 * @throws TDataException
	 */
	private static TNumeric fnc_log(SyntaxNode node, Map<String, TData> vars)
			throws EvalException, UserException, TDataException {
		TNumeric num;
		double[][] res;

		EvalFunction.check_FncParameterCount(node, 1, 1);
		num = SyntaxNode.getReal(node.childs[0], vars);
		res = new double[num.size()[0]][num.size()[1]];
		for (int row = 0; row < res.length; row++) {
			for (int col = 0; col < res[0].length; col++) {
				res[row][col] = Math.log10(num.getReal(row, col));
			}
		}

		return new TNumeric(res);
	}

	/**
	 * Sin function
	 * 
	 * @param node
	 *            Syntax node of function
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws UserException
	 * @throws EvalException
	 * @throws TDataException
	 */
	private static TNumeric fnc_sin(SyntaxNode node, Map<String, TData> vars)
			throws EvalException, UserException, TDataException {
		TNumeric num;
		double[][] res;

		EvalFunction.check_FncParameterCount(node, 1, 1);
		num = SyntaxNode.getReal(node.childs[0], vars);
		res = new double[num.size()[0]][num.size()[1]];
		for (int row = 0; row < res.length; row++) {
			for (int col = 0; col < res[0].length; col++) {
				res[row][col] = Math.sin(num.getReal(row, col));
			}
		}

		return new TNumeric(res);
	}

	/**
	 * Cos function
	 * 
	 * @param node
	 *            Syntax node of function
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws UserException
	 * @throws EvalException
	 * @throws TDataException
	 */
	private static TNumeric fnc_cos(SyntaxNode node, Map<String, TData> vars)
			throws EvalException, UserException, TDataException {
		TNumeric num;
		double[][] res;

		EvalFunction.check_FncParameterCount(node, 1, 1);
		num = SyntaxNode.getReal(node.childs[0], vars);
		res = new double[num.size()[0]][num.size()[1]];
		for (int row = 0; row < res.length; row++) {
			for (int col = 0; col < res[0].length; col++) {
				res[row][col] = Math.cos(num.getReal(row, col));
			}
		}

		return new TNumeric(res);
	}

	/**
	 * Tan function
	 * 
	 * @param node
	 *            Syntax node of function
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws UserException
	 * @throws EvalException
	 * @throws TDataException
	 */
	private static TNumeric fnc_tan(SyntaxNode node, Map<String, TData> vars)
			throws EvalException, UserException, TDataException {
		TNumeric num;
		double[][] res;

		EvalFunction.check_FncParameterCount(node, 1, 1);
		num = SyntaxNode.getReal(node.childs[0], vars);
		res = new double[num.size()[0]][num.size()[1]];
		for (int row = 0; row < res.length; row++) {
			for (int col = 0; col < res[0].length; col++) {
				res[row][col] = Math.tan(num.getReal(row, col));
			}
		}

		return new TNumeric(res);
	}

	/**
	 * ASin function
	 * 
	 * @param node
	 *            Syntax node of function
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws UserException
	 * @throws EvalException
	 * @throws TDataException
	 */
	private static TNumeric fnc_asin(SyntaxNode node, Map<String, TData> vars)
			throws EvalException, UserException, TDataException {
		TNumeric num;
		double[][] res;

		EvalFunction.check_FncParameterCount(node, 1, 1);
		num = SyntaxNode.getReal(node.childs[0], vars);
		res = new double[num.size()[0]][num.size()[1]];
		for (int row = 0; row < res.length; row++) {
			for (int col = 0; col < res[0].length; col++) {
				res[row][col] = Math.asin(num.getReal(row, col));
			}
		}

		return new TNumeric(res);
	}

	/**
	 * ACos function
	 * 
	 * @param node
	 *            Syntax node of function
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws UserException
	 * @throws EvalException
	 * @throws TDataException
	 */
	private static TNumeric fnc_acos(SyntaxNode node, Map<String, TData> vars)
			throws EvalException, UserException, TDataException {
		TNumeric num;
		double[][] res;

		EvalFunction.check_FncParameterCount(node, 1, 1);
		num = SyntaxNode.getReal(node.childs[0], vars);
		res = new double[num.size()[0]][num.size()[1]];
		for (int row = 0; row < res.length; row++) {
			for (int col = 0; col < res[0].length; col++) {
				res[row][col] = Math.acos(num.getReal(row, col));
			}
		}

		return new TNumeric(res);
	}

	/**
	 * ATan function
	 * 
	 * @param node
	 *            Syntax node of function
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws UserException
	 * @throws EvalException
	 * @throws TDataException
	 */
	private static TNumeric fnc_atan(SyntaxNode node, Map<String, TData> vars)
			throws EvalException, UserException, TDataException {
		TNumeric num;
		double[][] res;

		EvalFunction.check_FncParameterCount(node, 1, 1);
		num = SyntaxNode.getReal(node.childs[0], vars);
		res = new double[num.size()[0]][num.size()[1]];
		for (int row = 0; row < res.length; row++) {
			for (int col = 0; col < res[0].length; col++) {
				res[row][col] = Math.atan(num.getReal(row, col));
			}
		}

		return new TNumeric(res);
	}

	/**
	 * Sinh function
	 * 
	 * @param node
	 *            Syntax node of function
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws UserException
	 * @throws EvalException
	 * @throws TDataException
	 */
	private static TNumeric fnc_sinh(SyntaxNode node, Map<String, TData> vars)
			throws EvalException, UserException, TDataException {
		TNumeric num;
		double[][] res;

		EvalFunction.check_FncParameterCount(node, 1, 1);
		num = SyntaxNode.getReal(node.childs[0], vars);
		res = new double[num.size()[0]][num.size()[1]];
		for (int row = 0; row < res.length; row++) {
			for (int col = 0; col < res[0].length; col++) {
				res[row][col] = Math.sinh(num.getReal(row, col));
			}
		}

		return new TNumeric(res);
	}

	/**
	 * Cosh function
	 * 
	 * @param node
	 *            Syntax node of function
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws UserException
	 * @throws EvalException
	 * @throws TDataException
	 */
	private static TNumeric fnc_cosh(SyntaxNode node, Map<String, TData> vars)
			throws EvalException, UserException, TDataException {
		TNumeric num;
		double[][] res;

		EvalFunction.check_FncParameterCount(node, 1, 1);
		num = SyntaxNode.getReal(node.childs[0], vars);
		res = new double[num.size()[0]][num.size()[1]];
		for (int row = 0; row < res.length; row++) {
			for (int col = 0; col < res[0].length; col++) {
				res[row][col] = Math.cosh(num.getReal(row, col));
			}
		}

		return new TNumeric(res);
	}

	/**
	 * Tanh function - Tangente hyperbolicus
	 * 
	 * @param node
	 *            Syntax node of function
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws UserException
	 * @throws EvalException
	 * @throws TDataException
	 */
	private static TNumeric fnc_tanh(SyntaxNode node, Map<String, TData> vars)
			throws EvalException, UserException, TDataException {
		TNumeric num;
		double[][] res;

		EvalFunction.check_FncParameterCount(node, 1, 1);
		num = SyntaxNode.getReal(node.childs[0], vars);
		res = new double[num.size()[0]][num.size()[1]];
		for (int row = 0; row < res.length; row++) {
			for (int col = 0; col < res[0].length; col++) {
				res[row][col] = Math.tanh(num.getReal(row, col));
			}
		}

		return new TNumeric(res);
	}

	/**
	 * Todeg function - convert to degree
	 * 
	 * @param node
	 *            Syntax node of function
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws UserException
	 * @throws EvalException
	 * @throws TDataException
	 */
	private static TNumeric fnc_todeg(SyntaxNode node, Map<String, TData> vars)
			throws EvalException, UserException, TDataException {
		TNumeric num;
		double[][] res;

		EvalFunction.check_FncParameterCount(node, 1, 1);
		num = SyntaxNode.getReal(node.childs[0], vars);
		res = new double[num.size()[0]][num.size()[1]];
		for (int row = 0; row < res.length; row++) {
			for (int col = 0; col < res[0].length; col++) {
				res[row][col] = Math.toDegrees(num.getReal(row, col));
			}
		}

		return new TNumeric(res);
	}

	/**
	 * Torad function - convert to radiant
	 * 
	 * @param node
	 *            Syntax node of function
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws UserException
	 * @throws EvalException
	 * @throws TDataException
	 */
	private static TNumeric fnc_torad(SyntaxNode node, Map<String, TData> vars)
			throws EvalException, UserException, TDataException {
		TNumeric num;
		double[][] res;

		EvalFunction.check_FncParameterCount(node, 1, 1);
		num = SyntaxNode.getReal(node.childs[0], vars);
		res = new double[num.size()[0]][num.size()[1]];
		for (int row = 0; row < res.length; row++) {
			for (int col = 0; col < res[0].length; col++) {
				res[row][col] = Math.toRadians(num.getReal(row, col));
			}
		}

		return new TNumeric(res);
	}

	/**
	 * Len function - Get length of string
	 * 
	 * @param node
	 *            Syntax node of function
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws UserException
	 * @throws EvalException
	 * @throws TDataException
	 */
	private static TNumeric fnc_len(SyntaxNode node, Map<String, TData> vars)
			throws EvalException, UserException, TDataException {
		EvalFunction.check_FncParameterCount(node, 1, 1);
		return new TNumeric(
				SyntaxNode.getString(node.childs[0], vars).length());
	}

	/**
	 * Substring function - Get a substring (end is optional)
	 * 
	 * @param node
	 *            Syntax node of function
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws UserException
	 * @throws EvalException
	 */
	private static TString fnc_substring(SyntaxNode node,
			Map<String, TData> vars) throws EvalException, UserException {
		String str;
		int begin, end;

		EvalFunction.check_FncParameterCount(node, 2, 3);
		str = SyntaxNode.getString(node.childs[0], vars);
		begin = SyntaxNode
				.getInt(SyntaxNode.getScalarReal(node.childs[1], vars));
		SyntaxNode.checkBounds(begin, str.length(), node.childs[1].tok);
		if (node.childs.length == 2) {
			str = str.substring(begin);
		} else {
			end = SyntaxNode
					.getInt(SyntaxNode.getScalarReal(node.childs[2], vars));
			SyntaxNode.checkBounds(end, str.length() + 1, node.childs[2].tok);
			if (end < begin) {
				throw new EvalException("End index is lower than begin index",
						node.childs[2].tok);
			}
			str = str.substring(begin, end);
		}

		return new TString(str);
	}

	/**
	 * Concat function - Concat strings (list of strings)
	 * 
	 * @param node
	 *            Syntax node of function
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws UserException
	 * @throws EvalException
	 */
	private static TString fnc_concat(SyntaxNode node, Map<String, TData> vars)
			throws EvalException, UserException {
		StringBuffer str;

		EvalFunction.check_FncParameterCount(node, 1);
		str = new StringBuffer(SyntaxNode.getString(node.childs[0], vars));
		for (int k = 1; k < node.childs.length; k++) {
			str.append(SyntaxNode.getString(node.childs[k], vars));
		}

		return new TString(str.toString());
	}

	/**
	 * Cmp function - Compare strings
	 * 
	 * @param node
	 *            Syntax node of function
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws UserException
	 * @throws EvalException
	 * @throws TDataException
	 */
	private static TNumeric fnc_cmp(SyntaxNode node, Map<String, TData> vars)
			throws EvalException, UserException, TDataException {
		EvalFunction.check_FncParameterCount(node, 2, 2);
		return new TNumeric(SyntaxNode.getString(node.childs[0], vars)
				.compareTo(SyntaxNode.getString(node.childs[1], vars)));
	}

	/**
	 * Upper function - Convert string to upper case
	 * 
	 * @param node
	 *            Syntax node of function
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws UserException
	 * @throws EvalException
	 */
	private static TString fnc_upper(SyntaxNode node, Map<String, TData> vars)
			throws EvalException, UserException {
		EvalFunction.check_FncParameterCount(node, 1, 1);
		return new TString(
				SyntaxNode.getString(node.childs[0], vars).toUpperCase());
	}

	/**
	 * Lower function - Convert string to lower case
	 * 
	 * @param node
	 *            Syntax node of function
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws UserException
	 * @throws EvalException
	 */
	private static TString fnc_lower(SyntaxNode node, Map<String, TData> vars)
			throws EvalException, UserException {
		EvalFunction.check_FncParameterCount(node, 1, 1);
		return new TString(
				SyntaxNode.getString(node.childs[0], vars).toLowerCase());
	}

	/**
	 * Str function - Get string representation of real number
	 * 
	 * @param node
	 *            Syntax node of function
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws UserException
	 * @throws EvalException
	 */
	private static TString fnc_str(SyntaxNode node, Map<String, TData> vars)
			throws EvalException, UserException {
		EvalFunction.check_FncParameterCount(node, 1);
		if (node.childs.length == 1) {
			return new TString(Double
					.toString(SyntaxNode.getScalarReal(node.childs[0], vars)));
		} else {
			Object[] list = new Double[node.childs.length];
			for (int k = 1; k < node.childs.length; k++) {
				list[k - 1] = SyntaxNode.getScalarReal(node.childs[k], vars);
			}
			return new TString(String
					.format(SyntaxNode.getString(node.childs[0], vars), list));
		}
	}

	/**
	 * Num function - Parse string to numeric value
	 * 
	 * @param node
	 *            Syntax node of function
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws UserException
	 * @throws EvalException
	 */
	private static TNumeric fnc_num(SyntaxNode node, Map<String, TData> vars)
			throws EvalException, UserException {
		EvalFunction.check_FncParameterCount(node, 1, 1);
		try {
			return new TNumeric(Double
					.parseDouble(SyntaxNode.getString(node.childs[0], vars)));
		} catch (Exception ex) {
			throw new EvalException(ex.getMessage(), node.childs[0].tok);
		}
	}

	/**
	 * Rand function - Get random number
	 * 
	 * @param node
	 *            Syntax node of function
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws UserException
	 * @throws EvalException
	 * @throws TDataException
	 */
	private static TNumeric fnc_rand(SyntaxNode node, Map<String, TData> vars)
			throws EvalException, UserException, TDataException {
		EvalFunction.check_FncParameterCount(node, 0, 0);
		return new TNumeric(Math.random());
	}

	/**
	 * Ones function - Get ones matrix
	 * 
	 * @param node
	 *            Syntax node of function
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws UserException
	 * @throws EvalException
	 * @throws TDataException
	 */
	private static TNumeric fnc_ones(SyntaxNode node, Map<String, TData> vars)
			throws EvalException, UserException, TDataException {
		int row = 1, col = 1;
		double[][] values;

		EvalFunction.check_FncParameterCount(node, 0, 2);
		if (node.childs != null) {
			if (node.childs.length >= 1) {
				col = row = SyntaxNode
						.getInt(SyntaxNode.getScalarReal(node.childs[0], vars));
			}
			if (node.childs.length == 2) {
				col = SyntaxNode
						.getInt(SyntaxNode.getScalarReal(node.childs[1], vars));
			}
		}

		values = new double[row][col];
		for (int r = 0; r < row; r++) {
			for (int c = 0; c < col; c++) {
				values[r][c] = 1.0;
			}
		}

		return new TNumeric(values);
	}

	/**
	 * Ones function - Get zeros matrix
	 * 
	 * @param node
	 *            Syntax node of function
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws UserException
	 * @throws EvalException
	 * @throws TDataException
	 */
	private static TNumeric fnc_zeros(SyntaxNode node, Map<String, TData> vars)
			throws EvalException, UserException, TDataException {
		int row = 1, col = 1;
		double[][] values;

		EvalFunction.check_FncParameterCount(node, 0, 2);
		if (node.childs != null) {
			if (node.childs.length >= 1) {
				col = row = SyntaxNode
						.getInt(SyntaxNode.getScalarReal(node.childs[0], vars));
			}
			if (node.childs.length == 2) {
				col = SyntaxNode
						.getInt(SyntaxNode.getScalarReal(node.childs[1], vars));
			}
		}

		values = new double[row][col];
		// for (int r = 0; r < row; r++) {
		// for (int c = 0; c < col; c++) {
		// values[r][c] = 0.0;
		// }
		// }

		return new TNumeric(values);
	}

	/**
	 * Ones function - Get eye matrix (identiy)
	 * 
	 * @param node
	 *            Syntax node of function
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws UserException
	 * @throws EvalException
	 * @throws TDataException
	 */
	private static TNumeric fnc_eye(SyntaxNode node, Map<String, TData> vars)
			throws EvalException, UserException, TDataException {
		int row = 1, col = 1;
		double[][] values;

		EvalFunction.check_FncParameterCount(node, 0, 2);
		if (node.childs != null) {
			if (node.childs.length >= 1) {
				col = row = SyntaxNode
						.getInt(SyntaxNode.getScalarReal(node.childs[0], vars));
			}
			if (node.childs.length == 2) {
				col = SyntaxNode
						.getInt(SyntaxNode.getScalarReal(node.childs[1], vars));
			}
		}

		values = new double[row][col];
		for (int r = 0; r < row; r++) {
			for (int c = 0; c < col; c++) {
				if (r == c) {
					values[r][c] = 1.0;
					// } else {
					// values[r][c] = 0.0;
				}
			}
		}

		return new TNumeric(values);
	}

}
