/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9713 $
 * $Date: 2020-03-19 09:28:58 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.magopt.structure.xsd;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URL;

import javax.xml.XMLConstants;
import javax.xml.transform.stream.StreamSource;
import javax.xml.validation.SchemaFactory;
import javax.xml.validation.Validator;

import org.w3c.dom.ls.LSInput;
import org.w3c.dom.ls.LSResourceResolver;
import org.xml.sax.SAXException;

import at.lcm.bu21.general.dtypes.xsd.LSDataInput;
import at.lcm.bu21.general.dtypes.xsd.XMLDataValidator;
import magopt.utils.MopAccess;
import magopt.utils.MopFile;

/**
 * Validate file
 * 
 * @author Werner Koppelsttter
 */
public class XMLValidator {

	/** Schema for validation */
	String schema;

	/**
	 * Validate xml-file
	 * 
	 * @param file
	 *            XML-file
	 * @param schema
	 *            Name of Schema
	 * @throws SAXException
	 * @throws IOException
	 */
	public void validate(File file, String schema) throws SAXException, IOException {
		this.validate(new FileInputStream(file), schema);
	}

	/**
	 * Validate xml-file
	 * 
	 * @param is
	 *            Inputstream of xml-file
	 * @param schema
	 *            Name of Schema
	 * @throws SAXException
	 * @throws IOException
	 */
	public void validate(InputStream is, String schema) throws SAXException, IOException {
		SchemaFactory schemaFactory;
		Validator vali;
		URL schemaUrl;

		schemaFactory = SchemaFactory.newInstance(XMLConstants.W3C_XML_SCHEMA_NS_URI);
		// ResourceResolver to locate data-types
		schemaFactory.setResourceResolver(new LSResourceResolver() {
			@Override
			public LSInput resolveResource(String type, String namespaceURI, String publicId, String systemId,
					String baseURI) {
				LSInput input = null;

				if (systemId.compareTo("TData.xsd") == 0) {
					input = new LSDataInput(systemId);
				}

				return input;
			}
		});

		// find rexource
		schemaUrl = XMLValidator.class.getResource(schema);
		if (schemaUrl == null) {
			schemaUrl = XMLDataValidator.class.getResource(schema);
		}

		vali = schemaFactory.newSchema(schemaUrl).newValidator();
		vali.validate(new StreamSource(is));
	}

	/**
	 * Validate mop-file
	 * 
	 * @param mop
	 *            Mop-file
	 * @param xml
	 *            xml-file to validate
	 * @param schema
	 *            Schema for validation
	 * @throws Exception
	 */
	public void validateMop(MopFile mop, String xml, String schema) throws Exception {

		this.schema = schema;
		mop.readStream(new MopAccess() {
			@Override
			public void readStream(InputStream is) throws Exception {
				XMLValidator.this.validate(is, XMLValidator.this.schema);
			}

			@Override
			public void writeStream(OutputStream os) throws Exception {
			}
		}, xml);
	}

}
