/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9714 $
 * $Date: 2020-03-19 09:37:13 +0100 (Thu, 19 Mar 2020) $
 */
package magopt.gui.helpers;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.util.ArrayList;

import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.DefaultListCellRenderer;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.KeyStroke;

import at.lcm.bu21.magopt.structure.ipython.ConnectionFile;

/**
 * Select connection and name for variable in IPython
 *
 * @author Werner Koppelsttter
 */
public class SendToIPythonDialog {

	/** Dialog to select material */
	private JDialog dialog;

	/** ComboBox for connection */
	private JComboBox<ConnectionFile> connection;

	/** Field for variable-name */
	private JTextField variable;

	/** Label for validate-string */
	private JLabel validate;

	/** Ok */
	private JButton ok;

	/** Cancel */
	private JButton cancel;

	/** Selected connection */
	private ConnectionFile conn = null;

	/** Selected variable-name */
	private String varName = null;

	/**
	 * Create dialog for IPython connection
	 * 
	 * @param parent
	 *            Parent frame
	 */
	public SendToIPythonDialog(Frame parent) {
		this.init(parent);
		this.dialog.setLocationRelativeTo(parent);
		this.dialog.setSize(new Dimension(400, 200));
	}

	/**
	 * Get selected connection
	 * 
	 * @return Connection-file
	 */
	public ConnectionFile getConnection() {
		return this.conn;
	}

	/**
	 * Get name of variable
	 * 
	 * @return Variable name
	 */
	public String getVariableName() {
		return this.varName;
	}

	/**
	 * Initialize dialog
	 * 
	 * @param parent
	 *            Parent frame
	 */
	protected void init(Frame parent) {
		JPanel mainPanel, panel;
		JLabel label;
		GridBagLayout layout;
		GridBagConstraints c;
		// Font font;

		// create dialog
		this.dialog = new JDialog(parent, "Sent data to IPython", true);
		this.dialog.addWindowListener(new WindowAdapter() {
			@Override
			public void windowClosing(WindowEvent e) {
				SendToIPythonDialog.this.dialog.setVisible(false);
				SendToIPythonDialog.this.dialog.dispose();
			}
		});
		this.dialog.setLayout(new BorderLayout());

		// key events
		this.dialog.getRootPane().registerKeyboardAction(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e) {
				if (SendToIPythonDialog.this.ok.isEnabled()) {
					SendToIPythonDialog.this.finish(true);
				}
			}
		}, KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, 0),
				JComponent.WHEN_IN_FOCUSED_WINDOW);
		this.dialog.getRootPane().registerKeyboardAction(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e) {
				SendToIPythonDialog.this.finish(false);
			}
		}, KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0),
				JComponent.WHEN_IN_FOCUSED_WINDOW);

		// main panel
		mainPanel = new JPanel();
		mainPanel.setBorder(BorderFactory.createEmptyBorder(5, 5, 5, 5));
		mainPanel.setLayout(new BorderLayout());
		this.dialog.add(mainPanel, BorderLayout.CENTER);

		// center area
		panel = new JPanel();
		layout = new GridBagLayout();
		c = new GridBagConstraints();
		c.fill = GridBagConstraints.BOTH;
		c.insets = new Insets(5, 5, 5, 5);
		panel.setLayout(layout);
		mainPanel.add(panel, BorderLayout.CENTER);

		// list of connections
		label = new JLabel("Connection:");
		c.gridwidth = 1;
		c.weightx = 0.0;
		layout.setConstraints(label, c);
		panel.add(label);
		this.connection = new JComboBox<ConnectionFile>();
		c.gridwidth = GridBagConstraints.REMAINDER;
		c.weightx = 1.0;
		layout.setConstraints(this.connection, c);
		panel.add(this.connection);

		// variable name
		label = new JLabel("Variable:");
		c.gridwidth = 1;
		c.weightx = 0.0;
		layout.setConstraints(label, c);
		panel.add(label);
		this.variable = new JTextField();
		c.gridwidth = GridBagConstraints.REMAINDER;
		c.weightx = 1.0;
		layout.setConstraints(this.variable, c);
		panel.add(this.variable);
		this.variable.addKeyListener(new KeyAdapter() {
			@Override
			public void keyReleased(KeyEvent e) {
				SendToIPythonDialog.this.validate();
			}
		});

		// Validate string
		label = new JLabel(" ");
		c.gridwidth = 1;
		c.weightx = 0.0;
		layout.setConstraints(label, c);
		panel.add(label);
		this.validate = new JLabel();
		c.gridwidth = GridBagConstraints.REMAINDER;
		c.weightx = 1.0;
		layout.setConstraints(this.validate, c);
		panel.add(this.validate);
		this.validate.setForeground(Color.RED);

		// add buttons
		panel = new JPanel();
		panel.setBorder(BorderFactory.createEmptyBorder(5, 5, 5, 5));
		panel.setLayout(new BoxLayout(panel, BoxLayout.X_AXIS));
		mainPanel.add(panel, BorderLayout.SOUTH);

		panel.add(Box.createHorizontalGlue());

		this.ok = new JButton("OK");
		this.ok.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent arg0) {
				SendToIPythonDialog.this.finish(true);
			}
		});
		panel.add(this.ok);

		panel.add(Box.createHorizontalGlue());

		this.cancel = new JButton("Cancel");
		this.cancel.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent arg0) {
				SendToIPythonDialog.this.finish(false);
			}
		});
		panel.add(this.cancel);

		panel.add(Box.createHorizontalGlue());

		this.dialog.pack();
	}

	/**
	 * Show dialog
	 * 
	 * @param varName
	 *            Default value of variable name
	 * @throws Exception
	 */
	public void show(String varName) throws Exception {
		ArrayList<ConnectionFile> list;

		// connection-file
		list = ConnectionFile.findConnectionFiles();
		for (ConnectionFile item : list) {
			this.connection.addItem(item);
		}
		// this.connection.setSelectedItem(ConnectionFile.getLatest());
		this.connection.setRenderer(new DefaultListCellRenderer() {

			/**	 */
			private static final long serialVersionUID = 1L;

			@Override
			public Component getListCellRendererComponent(JList<?> list,
					Object value, int index, boolean isSelected,
					boolean cellHasFocus) {
				JLabel label;

				label = (JLabel) super.getListCellRendererComponent(list, value,
						index, isSelected, cellHasFocus);
				if (value != null) {
					label.setText(((ConnectionFile) value).getID());
				}

				return label;
			}
		});

		if (varName == null) {
			varName = "CurData";
		}
		this.variable.setText(varName);

		this.validate();
		this.dialog.setVisible(true);

	}

	/**
	 * Finish dialog
	 * 
	 * @param ok
	 *            Selected of cancel
	 */
	private void finish(boolean ok) {
		SendToIPythonDialog.this.dialog.setVisible(false);
		if (ok) {
			SendToIPythonDialog.this.conn = (ConnectionFile) SendToIPythonDialog.this.connection
					.getSelectedItem();
			SendToIPythonDialog.this.varName = SendToIPythonDialog.this.variable
					.getText();
		}
		SendToIPythonDialog.this.dialog.dispose();
	}

	/**
	 * Validate variable name
	 */
	private void validate() {
		boolean valid;

		if (this.connection.getSelectedItem() == null) {
			this.validate.setText("No kernel found!");
			valid = false;
		} else {
			valid = this.variable.getText().matches("[a-zA-Z]\\w*");
			if (valid) {
				this.validate.setText("");
			} else {
				this.validate.setText("Invalid variable name!");
			}
		}
		this.ok.setEnabled(valid);
	}
}
