/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9767 $
 * $Date: 2020-04-20 20:04:00 +0200 (Mon, 20 Apr 2020) $
 */
package magopt.utils;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.net.URISyntaxException;
import java.util.Arrays;
import java.util.Hashtable;
import java.util.LinkedList;

import com.jezhumble.javasysmon.JavaSysMon;

/**
 * This class execute external command with watchdog-observation
 * 
 * @author Werner Koppelsttter
 */
public class Execute {

	/**
	 * States process execution
	 * 
	 * @author Werner Koppelsttter
	 */
	public enum State {
		/** Process is not started */
		IDLE,
		/** Process is started */
		ALIVE,
		/** Process finished */
		FINISHED,
		/** Process stopped by watchdog */
		TIMEOUT,
		/** Process stopped by user */
		KILLED
	};

	/** State of executor */
	State state = State.IDLE;

	/** Command to execute */
	protected String[] command;

	/** Directory to execute command */
	protected File directory;

	/** List of environments variables */
	protected Hashtable<String, String> environment = null;

	/** Process identifier of started process */
	protected long pid = 0;

	/** Process handle of external program */
	protected Process proc = null;

	/** Timeout for watchdog */
	protected long timeout = 0;

	/** Execution-thread */
	protected Thread thread = null;

	/** Watchdog-thread */
	protected Thread watchdog = null;

	/** Runtime of external process */
	protected long runtime = 0;

	/** Redirect streams */
	protected boolean redirectStream = false;

	/** Discard streams */
	protected boolean discardStream = false;

	/**
	 * Construct a process with specified command
	 * 
	 * @param command
	 *            Command to execute
	 */
	public Execute(String... command) {
		this(null, command);
	}

	/**
	 * Construct a process with specified command in specified directory
	 * 
	 * @param directory
	 *            Working directory
	 * @param command
	 *            Command to execute
	 */
	public Execute(File directory, String... command) {
		this(directory, 0, command);
	}

	/**
	 * Construct a process with specified command in specified directory and a
	 * watchdog
	 * 
	 * @param directory
	 *            Working directory
	 * @param timeout
	 *            Timeout for Watchdog in millisecounds (0 if no watchdog needed)
	 * @param command
	 *            Command to execute
	 */
	public Execute(File directory, long timeout, String... command) {
		this.directory = directory;
		this.timeout = timeout;
		this.command = command;
	}

	/**
	 * Start execution of the external process
	 */
	public void start() {

		Execute.this.state = Execute.State.ALIVE;
		// main execution thread
		this.thread = new Thread() {

			@Override
			public void run() {

				// add watchdog
				if (Execute.this.timeout > 0) {
					// watchdog thread
					Execute.this.watchdog = new Thread() {
						@Override
						public void run() {
							try {
								Thread.sleep(Execute.this.timeout);
							} catch (InterruptedException e) {
							}
							if ((Execute.this.proc != null) && (Execute.this.state == Execute.State.ALIVE)) {
								Execute.this.state = Execute.State.TIMEOUT;
								new JavaSysMon().killProcessTree((int) Execute.this.pid, false);
							}
						}
					};
					Execute.this.watchdog.start();
				}

				// start execution
				Execute.this.runtime = -System.currentTimeMillis();
				try {
					Execute.this.proc = Execute.this.execute();

					if (Execute.this.redirectStream) {
						Execute.this.redirectStreams();
					}

					Execute.this.proc.waitFor();
				} catch (Exception e) {
					System.err.println(e.getMessage());
					e.printStackTrace();
				}
				if (Execute.this.state == Execute.State.ALIVE) {
					Execute.this.runtime += System.currentTimeMillis();
					Execute.this.state = Execute.State.FINISHED;
				}
				// kill watchdog
				if ((Execute.this.watchdog != null) && Execute.this.watchdog.isAlive()) {
					Execute.this.watchdog.interrupt();
				}
			}
		};
		this.thread.start();

		try {
			while (this.pid == 0) {
				Thread.sleep(100);
			}
		} catch (InterruptedException e) {
		}
	}

	protected Process execute() throws IOException, URISyntaxException {
		Process p;
		ProcessBuilder builder;
		LinkedList<String> cmd = new LinkedList<String>();

		cmd.add(new File(new File(System.getProperty("java.home"), "bin"), "java").getPath());
		cmd.add("-Xms16M");
		cmd.add("-Xmx64M");
		cmd.add("-cp");
		cmd.add(new File(Execute.class.getProtectionDomain().getCodeSource().getLocation().toURI()).getPath());
		cmd.add("magopt.utils.ExecuteStarter");
		if (Execute.this.directory != null) {
			cmd.add(Execute.this.directory.getAbsolutePath());
		} else {
			cmd.add("@");
		}
		for (String str : Execute.this.command) {
			cmd.add(str);
		}

		builder = new ProcessBuilder(cmd);
		builder.directory(Execute.this.directory);
		if (Execute.this.environment != null) {
			for (String key : Execute.this.environment.keySet()) {
				builder.environment().put(key, Execute.this.environment.get(key));
			}
		}
		p = builder.start();

		// read process-id
		InputStream in = p.getInputStream();
		byte[] b = new byte[1];
		byte[] str = new byte[0];
		boolean eof = false;
		while (!eof) {
			if (in.available() > 0) {
				in.read(b);
				if ((b[0] == '\n') || (b[0] == '\r')) {
					eof = true;
					while ((in.available() > 0) && ((b[0] == '\n') || (b[0] == '\r'))) {
						in.mark(2);
						in.read(b);
					}
					if ((b[0] != '\n') && (b[0] != '\r')) {
						in.reset();
					}
				} else {
					str = Arrays.copyOf(str, str.length + 1);
					str[str.length - 1] = b[0];
				}
			}
		}
		Execute.this.pid = Long.parseLong(new String(str));

		return p;
	}

	/**
	 * Stop external process
	 */
	public void kill() {
		if (this.state == State.ALIVE) {
			if (this.pid != 0) {
				new JavaSysMon().killProcessTree((int) this.pid, false);
			}
			this.state = State.KILLED;
		}
	}

	/**
	 * Get return-code of external process
	 * 
	 * @return Return-code
	 */
	public int exitValue() {
		int exitValue = -1;
		try {
			while (this.thread.isAlive()) {
				Thread.sleep(100);
			}
			if (this.proc != null) {
				exitValue = this.proc.exitValue();
			}
		} catch (InterruptedException e) {
		}
		return exitValue;
	}

	/**
	 * Get state of Executor
	 * 
	 * @return State
	 */
	public State getState() {
		return this.state;
	}

	/**
	 * Tests if the Executor is alive
	 * 
	 * @return true if this thread is alive; false otherwise.
	 */
	public boolean isAlive() {
		return (this.thread == null) || this.thread.isAlive();
	}

	/**
	 * Get the runtime of the finished process in milliseconds
	 * 
	 * @return Runtime
	 */
	public long getRuntime() {
		return (this.runtime > 0) ? this.runtime : 0;
	}

	/**
	 * Return process identifier
	 * 
	 * @return pid
	 */
	public long getPID() {
		return this.pid;
	}

	/**
	 * Get defined timeout in milliseconds
	 * 
	 * @return Timeout
	 */
	public long getTimeout() {
		return this.timeout;
	}

	/**
	 * Set timeout in milliseconds
	 * 
	 * @param timeout
	 */
	public void setTimeout(long timeout) {
		this.timeout = timeout;
	}

	/**
	 * Get defined command
	 * 
	 * @return Defined command
	 */
	public String[] getCommand() {
		return this.command;
	}

	/**
	 * Set command to execute
	 * 
	 * @param command
	 *            Command to execute
	 */
	public void setCommand(String... command) {
		this.command = command;
	}

	/**
	 * Get defined working-directory
	 * 
	 * @return Working-directory
	 */
	public File getDirectory() {
		return this.directory;
	}

	/**
	 * Set working-directory
	 * 
	 * @param direcotry
	 *            Working-directory
	 */
	public void setDirectory(File direcotry) {
		this.directory = direcotry;
	}

	/**
	 * Get list of environment variables
	 * 
	 * @return List
	 */
	public Hashtable<String, String> getEnvironments() {
		return this.environment;
	}

	/**
	 * Set environment variable
	 * 
	 * @param key
	 *            Key
	 * @param value
	 *            Value
	 */
	public void setEnvironment(String key, String value) {
		if (this.environment == null) {
			this.environment = new Hashtable<String, String>();
		}

		this.environment.put(key, value);
	}

	/**
	 * Get executed process
	 * 
	 * @return Process
	 */
	public Process getProcess() {
		return this.proc;
	}

	/**
	 * Gets the error stream of the subprocess. If process is not started it return
	 * null.
	 * 
	 * @return InputStream connected to the error stream of the process
	 */
	public InputStream getErrorStream() {
		InputStream in = null;

		if ((this.proc != null) && !this.discardStream) {
			in = this.proc.getErrorStream();
		}

		return in;
	}

	/**
	 * Gets the input stream of the subprocess. If process is not started it return
	 * null.
	 * 
	 * @return InputStream connected to the normal input of the process
	 */
	public InputStream getInputStream() {
		InputStream in = null;

		if ((this.proc != null) && !this.discardStream) {
			in = this.proc.getInputStream();
		}

		return in;
	}

	/**
	 * Gets the output stream of the subprocess. If process is not started it return
	 * null.
	 * 
	 * @return OutputStream connected to the normal output of the process
	 */
	public OutputStream getOutputStream() {
		OutputStream out = null;

		if ((this.proc != null) && !this.discardStream) {
			out = this.proc.getOutputStream();
		}

		return out;
	}

	/**
	 * Should streams discard (output und error)
	 * 
	 * @return Discard streams
	 */
	public boolean getDiscardStreams() {
		return this.discardStream;
	}

	/**
	 * Set to discard streams (output and error). This has to be called befor
	 * start()
	 *
	 * @param discard
	 *            Discard streams
	 */
	public void setDiscardStreams(boolean discard) {
		this.discardStream = discard;
		if (discard) {
			this.redirectStream = true;
		}
	}

	/**
	 * Redirect streams (output und error)
	 * 
	 * @return Redirect streams
	 */
	public boolean getRedirectStreams() {
		return this.redirectStream;
	}

	/**
	 * Set redirect streams (output and error). This has to be called before start()
	 *
	 * @param redirect
	 *            Redirect streams
	 */
	public void setRedirectStreams(boolean redirect) {
		this.redirectStream = redirect;
	}

	/**
	 * Discard output and error stream
	 */
	protected void redirectStreams() {
		// output-stream
		new Thread() {
			@Override
			public void run() {
				String line;
				BufferedReader in;

				in = new BufferedReader(new InputStreamReader(Execute.this.proc.getInputStream()));
				try {
					while ((line = in.readLine()) != null) {
						if (!Execute.this.discardStream) {
							System.out.println(line);
						}
					}
				} catch (IOException e) {
					System.err.println(e.getMessage());
					e.printStackTrace();
				} finally {
					try {
						in.close();
					} catch (IOException e) {
					}
				}
			}
		}.start();

		// error-stream
		new Thread() {
			@Override
			public void run() {
				String line;
				BufferedReader in;

				in = new BufferedReader(new InputStreamReader(Execute.this.proc.getErrorStream()));
				try {
					while ((line = in.readLine()) != null) {
						if (!Execute.this.discardStream) {
							System.err.println(line);
						}
					}
				} catch (IOException e) {
					System.err.println(e.getMessage());
					e.printStackTrace();
				} finally {
					try {
						in.close();
					} catch (IOException e) {
					}
				}

			}
		}.start();
	}
}
