/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9714 $
 * $Date: 2020-03-19 09:37:13 +0100 (Thu, 19 Mar 2020) $
 */
package magopt.utils;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.sql.SQLException;
import java.util.Hashtable;
import java.util.Iterator;

import magopt.structure.database.DatabaseConnection;

import org.apache.poi.ss.usermodel.Cell;
import org.apache.poi.ss.usermodel.CellStyle;
import org.apache.poi.ss.usermodel.IndexedColors;
import org.apache.poi.ss.usermodel.Name;
import org.apache.poi.ss.usermodel.Row;
import org.apache.poi.ss.usermodel.Sheet;
import org.apache.poi.ss.usermodel.Workbook;
import org.apache.poi.ss.usermodel.WorkbookFactory;

import at.lcm.bu21.magopt.structure.ParamField;
import at.lcm.bu21.magopt.structure.Parameter;
import at.lcm.bu21.magopt.structure.Project;

/**
 * Create and update the parameter-database link
 * 
 * @author Werner Koppelsttter
 */
public class ParamDatabase {

	/**
	 * Connection to the database
	 */
	protected DatabaseConnection conn = null;

	/**
	 * Create object for communication with database
	 * 
	 * @throws Exception
	 */
	public ParamDatabase() throws Exception {
		this.conn = new DatabaseConnection();
	}

	/**
	 * Link file with database
	 * 
	 * @param file
	 *            File to link
	 * @throws Exception
	 */
	public void linkToDatabase(String file) throws Exception {
		this.linkToDatabase(new File(file));
	}

	/**
	 * Link file with database
	 * 
	 * @param file
	 *            File to link
	 * @throws Exception
	 */
	public void linkToDatabase(File file) throws Exception {
		Project proj;
		Workbook wb;
		FileInputStream in;
		FileOutputStream out;

		if (file.getName().toLowerCase().endsWith(".mop")
				|| file.getName().toLowerCase().endsWith(".mopt")) {
			proj = Project.load(file);
			this.linkProjectToDatabase(proj);
			proj.save();
		} else if (file.getName().toLowerCase().endsWith(".xls")
				|| file.getName().toLowerCase().endsWith(".xlsx")) {
			in = new FileInputStream(file);
			wb = WorkbookFactory.create(in);
			this.linkWorkbookToDatabase(wb);
			out = new FileOutputStream(file);
			wb.write(out);
			out.close();
		} else {
			throw new Exception("Unknown fileformat");
		}
	}

	/**
	 * Update file with database
	 * 
	 * @param file
	 *            File to update
	 * @throws Exception
	 */
	public void updateFromDatabase(String file) throws Exception {
		this.updateFromDatabase(new File(file));
	}

	/**
	 * Update file with database
	 * 
	 * @param file
	 *            File to update
	 * @throws Exception
	 */
	public void updateFromDatabase(File file) throws Exception {
		Project proj;
		Workbook wb;
		FileInputStream in;
		FileOutputStream out;

		if (file.getName().toLowerCase().endsWith(".mop")
				|| file.getName().toLowerCase().endsWith(".mopt")) {
			proj = Project.load(file);
			this.updateProjectFromDatabase(proj);
			proj.save();
		} else if (file.getName().toLowerCase().endsWith(".xls")
				|| file.getName().toLowerCase().endsWith(".xlsx")) {
			in = new FileInputStream(file);
			wb = WorkbookFactory.create(in);
			this.updateWorkbookFromDatabase(wb);
			out = new FileOutputStream(file);
			wb.write(out);
			out.close();
		} else {
			throw new Exception("Unknown fileformat");
		}

	}

	/**
	 * Link Project with the database
	 * 
	 * @param proj
	 *            Project to link
	 * @throws SQLException
	 */
	public void linkProjectToDatabase(Project proj) throws SQLException {
		Parameter[] params;
		ParamField field;

		params = proj.getParameters(ParamField.class);

		// link parameter with database
		for (Parameter p : params) {
			field = (ParamField) p;
			// check unlinked parameter
			if (field.getDbId() < 0) {
				field.setDbId(this.conn.getDBParameterId(field.getName()));
			}
		}

	}

	/**
	 * Update Project from database
	 * 
	 * @param proj
	 *            Project to update
	 * @throws Exception
	 */
	public void updateProjectFromDatabase(Project proj) throws Exception {
		Parameter[] params;
		ParamField field;
		Hashtable<String, String> table;

		params = proj.getParameters(ParamField.class);

		// update with database
		for (Parameter p : params) {
			field = (ParamField) p;
			if (field.getDbId() >= 0) {
				table = this.conn.getDBParameter(field.getDbId());

				field.setName(table.get("Name"));
				field.setDescription(table.get("Description"));
				field.setTexString(table.get("Tex"));
				field.setUnit(table.get("Unit"));

			}
		}
	}

	/**
	 * Create a 'Database' register with links to the database
	 * 
	 * @param wb
	 *            Workbook
	 * @throws SQLException
	 */
	protected void linkWorkbookToDatabase(Workbook wb) throws SQLException {
		Sheet sheet;
		Row row;
		long id;
		Cell cell;
		CellStyle style;

		// get database-sheet
		sheet = wb.getSheet("Database");
		if (sheet == null) {
			sheet = wb.createSheet("Database");

			style = wb.createCellStyle();
			style.setFillForegroundColor(IndexedColors.GREY_25_PERCENT
					.getIndex());
			style.setFillPattern(CellStyle.SOLID_FOREGROUND);

			// header
			row = sheet.createRow(0);

			cell = row.createCell(0);
			cell.setCellValue("Reference");
			cell.setCellStyle(style);

			cell = row.createCell(1);
			cell.setCellValue("Name");
			cell.setCellStyle(style);

			cell = row.createCell(2);
			cell.setCellValue("Identifier");
			cell.setCellStyle(style);
		}

		// update database-identifier
		this.updateReferenceList(sheet);
		for (int i = 1; i <= sheet.getLastRowNum(); i++) {
			row = sheet.getRow(i);
			if (row != null) {
				cell = row.getCell(1);
				if (cell != null) {
					if (row.getCell(2) == null) {
						id = this.conn.getDBParameterId(this
								.getParameterName(cell.getStringCellValue()));
						if (id >= 0) {
							cell = row.createCell(2);
							cell.setCellValue(id);
						}
					}
				}

			}
		}

		// adjust the size of the column
		sheet.autoSizeColumn(0);
		sheet.autoSizeColumn(1);
		sheet.autoSizeColumn(2);
	}

	/**
	 * Update list of named references
	 * 
	 * @param sheet
	 *            Sheet to write references
	 */
	private void updateReferenceList(Sheet sheet) {
		Workbook wb;
		Row row;
		Cell cell;
		Name name;
		boolean found;

		wb = sheet.getWorkbook();

		// add rows to list
		for (int i = 0; i < wb.getNumberOfNames(); i++) {
			name = wb.getNameAt(i);
			found = false;
			for (int j = 1; j <= sheet.getLastRowNum(); j++) {
				row = sheet.getRow(j);
				if (row != null) {
					cell = row.getCell(1);
					if (cell != null) {
						if (cell.getStringCellValue().compareTo(
								name.getNameName()) == 0) {
							found = true;
							cell = row.getCell(0);
							if (cell == null) {
								cell = row.createCell(0);
							}
							cell.setCellValue(name.getRefersToFormula());
						}
					}
				}
			}
			// new row
			if (!found) {
				row = sheet.createRow(sheet.getLastRowNum() + 1);
				row.createCell(0).setCellValue(name.getRefersToFormula());
				row.createCell(1).setCellValue(name.getNameName());
			}
		}

		// remove rows from list
		for (int i = sheet.getLastRowNum(); i >= 1; i--) {
			row = sheet.getRow(i);
			if (row != null) {
				cell = row.getCell(1);
				if (cell != null) {
					name = wb.getName(cell.getStringCellValue());
					if (name == null) {
						// remove and shift
						sheet.removeRow(row);
					}
				}
			}
		}
	}

	/**
	 * Update workbook with database
	 * 
	 * @param wb
	 *            Workbook
	 * @throws SQLException
	 */
	protected void updateWorkbookFromDatabase(Workbook wb) throws SQLException {
		Sheet sheet;
		Row row;
		Cell cellRef, cellName, cellId;
		Name name;
		long id;
		String str, symbol = null;

		// get database-sheet
		sheet = wb.getSheet("Database");
		if (sheet != null) {
			Iterator<Row> iter = sheet.rowIterator();
			// header
			if (iter.hasNext()) {
				iter.next();
			}

			// read all rows
			while (iter.hasNext()) {
				row = iter.next();
				// read name-cell
				cellName = row.getCell(1);
				if (cellName != null) {
					// get name-object
					str = cellName.getStringCellValue();
					name = wb.getName(str);
					if (name != null) {
						// get database id
						cellId = row.getCell(2);
						if (cellId != null) {
							id = (long) cellId.getNumericCellValue();

							// get parameter-name from database
							symbol = this.conn.getDBParameterName(id);
							if (symbol != null) {
								name.setNameName(this.replaceParameterName(
										name.getNameName(), symbol));
								cellName.setCellValue(name.getNameName());
							}
						}

						// update reference
						cellRef = row.getCell(0);
						if (cellRef == null) {
							cellRef = row.createCell(0);
						}
						cellRef.setCellValue(name.getRefersToFormula());
					}
				}
			}

			// adjust the size of the column
			sheet.autoSizeColumn(0);
			sheet.autoSizeColumn(1);
			sheet.autoSizeColumn(2);
		}
	}

	/**
	 * Get name of parameter. Name of parameter is the name between the first
	 * and the second point
	 * 
	 * @param tag
	 *            Tag with name of parameter
	 * @return Name of parameter
	 */
	private String getParameterName(String tag) {
		// remove pre- and post-string
		if (tag.indexOf(".") >= 0) {
			tag = tag.substring(tag.indexOf(".") + 1);
			if (tag.indexOf(".") >= 0) {
				tag = tag.substring(0, tag.indexOf("."));
			}
		}
		return tag;
	}

	/**
	 * Replace the name of parameter with the given name Name of parameter is
	 * the name between the first and the second point
	 * 
	 * @param tag
	 *            Tag with name of parameter
	 * @param name
	 *            New parameter-name
	 * @return Tag with the new parameter-name
	 */
	private String replaceParameterName(String tag, String name) {
		String str = "";
		if (tag.indexOf(".") >= 0) {
			str += tag.substring(0, tag.indexOf(".") + 1);
			tag = tag.substring(tag.indexOf(".") + 1);
		}
		str += name;
		if (tag.indexOf(".") >= 0) {
			str += tag.substring(tag.indexOf("."));
		}
		return str;
	}

}
