/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 247 $
 * $Date: 2020-03-19 10:13:52 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.optim.algorithm;

import at.lcm.bu21.general.dtypes.TStruct;
import at.lcm.bu21.optim.core.Individual;
import at.lcm.bu21.optim.core.Problem;
import at.lcm.bu21.optim.core.ResultDesc;
import at.lcm.bu21.optim.core.VariableDesc;

/**
 * Solver algorithm
 * 
 * @author Gntehr Weidenholzer
 * @author Werner Koppelsttter
 */
public abstract class Solver {

	/**
	 * Problem on with the Optimizer based (Description of variables and results)
	 */
	private Problem problem = null;

	/**
	 * Create optimizer
	 * 
	 * @param problem
	 *            Based problem with description of variables and results
	 */
	public Solver(Problem problem) {
		this.problem = problem;
	}

	/**
	 * Get based problem
	 * 
	 * @return Problem
	 */
	public Problem getProblem() {
		return this.problem;
	}

	/**
	 * Get name of optimizer
	 * 
	 * @return Name of optimizer
	 */
	public abstract String getOptimizerName();

	/**
	 * Get description of optimizer
	 * 
	 * @return Description of optimizer
	 */
	public abstract String getDescription();

	/**
	 * Get setting-panel of optimizer
	 * 
	 * @return panel Panel with settings
	 */
	public abstract JPanelSolver getPanel();

	/**
	 * Get next individual to calculate
	 * 
	 * @return Next individual
	 * @throws Exception
	 */
	public abstract Individual getNextIndividual() throws Exception;

	/**
	 * Set result of calculated individual
	 * 
	 * @param individual
	 *            Individual with results
	 * @throws Exception
	 */
	public abstract void setResolvedIndividual(Individual individual) throws Exception;

	/**
	 * Has next individual
	 * 
	 * @return If exist next individual
	 */
	public abstract boolean hasNext();

	/**
	 * Get the settings value of algorithm
	 * 
	 * @return Optimizer settings
	 * @throws Exception
	 */
	public abstract TStruct getSettings() throws Exception;

	/**
	 * Set settings of algorithm
	 * 
	 * @param resumeData
	 *            Optimizer settings
	 * @throws Exception
	 */
	public abstract void setSettings(TStruct resumeData) throws Exception;

	/**
	 * Get state of algorithm
	 * 
	 * @return Optimizer state
	 * @throws Exception
	 */
	public abstract TStruct getState() throws Exception;

	/**
	 * Set state of algorithm
	 * 
	 * @param resumeData
	 *            Optimizer state
	 * @throws Exception
	 */
	public abstract void setState(TStruct resumeData) throws Exception;

	/**
	 * Get archive of solver
	 * 
	 * @return Individuals of archive
	 * @throws Exception
	 */
	public abstract Individual[] getArchive() throws Exception;

	/**
	 * Set archive of solver
	 * 
	 * @param ind
	 *            Individuals of archive
	 * @throws Exception
	 */
	public abstract void setArchive(Individual[] ind) throws Exception;

	/**
	 * Get pareto of solver archive
	 * 
	 * @return Individuals of pareto
	 * @throws Exception
	 */
	public abstract Individual[] getPareto() throws Exception;

	@Override
	public String toString() {
		VariableDesc[] varDesc;
		ResultDesc[] resDesc;
		StringBuffer str;

		str = new StringBuffer(this.getDescription());
		int k;

		str.append("\n  Variables: ");
		varDesc = this.getProblem().getVariables();
		for (k = 0; k < varDesc.length; k++) {
			if (varDesc[k].isVaried()) {
				str.append(varDesc[k].getName());
				str.append(", ");
			}
		}
		str.setLength(str.length() - 2);

		str.append("\n  Objectives: ");
		resDesc = this.getProblem().getResults();
		for (k = 0; k < resDesc.length; k++) {
			if (resDesc[k].isObjective()) {
				str.append(resDesc[k].getName());
				str.append(", ");
			}
		}
		str.setLength(str.length() - 2);

		return str.toString();
	}

	/**
	 * Check if problem is valid for solver
	 * 
	 * @return Message if problem is not valid
	 */
	public String validateProblem() {
		String msg = null;

		if (this.getProblem().getNumVariedVariables() < 1) {
			msg = "To less variables";
		} else if (this.getProblem().getNumObjectives() < 1) {
			msg = "To less objectives";
		} else {
			for (VariableDesc var : this.getProblem().getVariables()) {
				if (!var.isValid()) {
					msg = "At least one variable is not valid";
				}
			}

			for (ResultDesc res : this.getProblem().getResults()) {
				if (!res.isValid()) {
					msg = "At least one result is not valid";
				}
			}
		}

		return msg;
	}
}
