/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 247 $
 * $Date: 2020-03-19 10:13:52 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.optim.algorithm.decmo;

import java.util.ArrayList;
import java.util.List;

import at.lcm.bu21.general.dtypes.TArray;
import at.lcm.bu21.general.dtypes.TNumeric;
import at.lcm.bu21.general.dtypes.TStruct;
import at.lcm.bu21.optim.algorithm.JMSolver;
import at.lcm.bu21.optim.core.Archive;
import at.lcm.bu21.optim.core.Individual;
import at.lcm.bu21.optim.core.Problem;
import at.lcm.bu21.optim.core.ResultDesc;
import jmetal.core.SolutionSet;
import jmetal.metaheuristics.decmo.DecmoReal;
import jmetal.util.JMException;

/**
 * DECMO implementation
 * 
 * @author Ciprian Zavoianu
 */
public abstract class Decmo extends JMSolver {

	/** Implementation of DECMO */
	protected DecmoReal solver;

	/** Crossover-probability */
	private double crossoverProbability = 0.9;

	/** Crossover-distribution-index */
	private double crossoverDistributionIndex = 20;

	/** Mutation-probability */
	private double mutationProbability = 0.4;

	/** Mutation-distribution-index */
	private double mutationDistributionIndex = 5;

	/** DE crossover-rate */
	private double deCR = 0.2;

	/** DE mutation-factor */
	private double deF = 0.5;

	/** Archive-size */
	private int archiveSize = 200;

	/** Number of consecutive weak sharing stages */
	private int weakSharingStagesCount = 0;

	/** The pool ID of the best performing pool */
	private int bestPerformingPoolID = -1;

	/** Current number of sub-population 1 insertions */
	private int pool1Additions = 0;

	/** Current number of sub-population 2 insertions */
	private int pool2Additions = 0;

	/** Array that stores the pool domination history over the runtime */
	List<Integer> dominationHistory = new ArrayList<Integer>();

	/** Array that stores the runtime minimum values of each objective */
	private double[] minObjValues;

	/** Array that stores the runtime maximum values of each objective */
	private double[] maxObjValues;

	private void initializeExtremeValues(Problem problem) {

		int size = 2;

		if (problem != null) {
			size = problem.getNumObjectives();
		}

		this.minObjValues = new double[size];
		this.maxObjValues = new double[size];
		for (int i = 0; i < size; i++) {
			this.minObjValues[i] = Double.POSITIVE_INFINITY;
			this.maxObjValues[i] = Double.NEGATIVE_INFINITY;
		}

		if (problem != null) {
			Individual[] validIndividuals = problem.getArchive().getMemory()
					.getValidIndividual(problem);

			for (Individual ind : validIndividuals) {
				int i = 0;
				for (ResultDesc rDesc : problem.getResults()) {
					if (rDesc.isObjective()) {
						Double val = ind.getResult(rDesc.getName());
						if (this.minObjValues[i] > val) {
							this.minObjValues[i] = val;
						}
						if (this.maxObjValues[i] < val) {
							this.maxObjValues[i] = val;
						}
						i++;
					}
				}
			}
		}
	}

	public Decmo(Problem problem) throws JMException {
		super(problem);

		this.initializeExtremeValues(problem);

		this.solver = new DecmoReal(this.jmProblem, this.archiveSize,
				this.crossoverProbability, this.crossoverDistributionIndex,
				this.mutationProbability, this.mutationDistributionIndex,
				this.deCR, this.deF, this.weakSharingStagesCount,
				this.bestPerformingPoolID, this.pool1Additions,
				this.pool2Additions, this.minObjValues, this.maxObjValues);
	}

	@Override
	public String getOptimizerName() {
		return "DECMO";
	}

	@Override
	public String getDescription() {
		return "DECMO Algorithm, pool1 = (Crossover: SBX / Mutation: polynomial / Selection: BinaryTournament), pool2 = (DE/rand/1/bin)";
	}

	/**
	 * Get size of archive
	 * 
	 * @return Size of archive
	 */
	public int getArchiveSize() {
		return this.archiveSize;
	}

	/**
	 * Set size of archive
	 * 
	 * @param archiveSize
	 *            Size of archive
	 */
	public void setArchiveSize(int archiveSize) {
		this.archiveSize = archiveSize;
	}

	/**
	 * Get crossover-probability
	 * 
	 * @return Value
	 */
	public double getCrossoverProbability() {
		return this.crossoverProbability;
	}

	/**
	 * Set crossover-probability
	 * 
	 * @param crossoverProbability
	 *            Value
	 */
	public void setCrossoverProbability(double crossoverProbability) {
		this.crossoverProbability = crossoverProbability;
	}

	/**
	 * Get crossover-distribution-index
	 * 
	 * @return Value
	 */
	public double getCrossoverDistributionIndex() {
		return this.crossoverDistributionIndex;
	}

	/**
	 * Set crossover-distribution-index
	 * 
	 * @param crossoverDistributionIndex
	 *            Value
	 */
	public void setCrossoverDistributionIndex(
			double crossoverDistributionIndex) {
		this.crossoverDistributionIndex = crossoverDistributionIndex;
	}

	/**
	 * Get mutation-probability
	 * 
	 * @return Value
	 */
	public double getMutationProbability() {
		return this.mutationProbability;
	}

	/**
	 * Set mutation-probability
	 * 
	 * @param mutationProbability
	 *            Value
	 */
	public void setMutationProbability(double mutationProbability) {
		this.mutationProbability = mutationProbability;
	}

	/**
	 * Get mutation-distribution-index
	 * 
	 * @return Value
	 */
	public double getMutationDistributionIndex() {
		return this.mutationDistributionIndex;
	}

	/**
	 * Set mutation-distribution-index
	 * 
	 * @param mutationDistributionIndex
	 *            Value
	 */
	public void setMutationDistributionIndex(double mutationDistributionIndex) {
		this.mutationDistributionIndex = mutationDistributionIndex;
	}

	/**
	 * Get DE crossover-rate
	 * 
	 * @return Value
	 */
	public double getDeCR() {
		return this.deCR;
	}

	/**
	 * Set DE crossover-rate
	 */
	public void setDeCR(double deCR) {
		this.deCR = deCR;
	}

	/**
	 * Get DE mutation-factor
	 * 
	 * @return Value
	 */
	public double getDeF() {
		return this.deF;
	}

	/**
	 * Set weak sharing stages count
	 */
	public void setDeF(double deF) {
		this.deF = deF;
	}

	/**
	 * Get DE mutation-factor
	 * 
	 * @return Value
	 */
	public int getWeakSharingStagesCount() {
		this.weakSharingStagesCount = this.solver.getWeakSharingStagesCount();
		return this.weakSharingStagesCount;
	}

	/**
	 * Set weak sharing stages count
	 */
	public void setWeakSharingStagesCount(int weakSharingStagesCount) {
		this.weakSharingStagesCount = weakSharingStagesCount;
		this.solver.setWeakSharingStagesCount(weakSharingStagesCount);
	}

	/**
	 * Get best performing pool ID
	 * 
	 * @return Value
	 */
	public int getBestPerformingPoolID() {
		this.bestPerformingPoolID = this.solver.getBestPerformingPoolID();
		return this.bestPerformingPoolID;
	}

	/**
	 * Set best performing pool ID
	 */
	public void setBestPerformingPoolID(int bestPerformingPoolID) {
		this.bestPerformingPoolID = bestPerformingPoolID;
		this.solver.setBestPerformingPoolID(bestPerformingPoolID);
	}

	/**
	 * Get minimum objective values
	 * 
	 * @return Value
	 */
	public double[] getMinObjValues() {
		return this.minObjValues;
	}

	/**
	 * Set minimum objective values
	 */
	public void setMinObjValues(double[] minObjValues) {
		this.minObjValues = minObjValues;
	}

	/**
	 * Get maximum objective values
	 * 
	 * @return Value
	 */
	public double[] getMaxObjValues() {
		return this.maxObjValues;
	}

	/**
	 * Set maximum objective values
	 */
	public void setMaxObjValues(double[] maxObjValues) {
		this.maxObjValues = maxObjValues;
	}

	/**
	 * Get pool1 additions
	 * 
	 * @return Value
	 */
	public int getPool1Additions() {
		this.pool1Additions = this.solver.getPool1Additions();
		return this.pool1Additions;
	}

	/**
	 * Set pool1 additions
	 */
	public void setPool1Additions(int pool1Additions) {
		this.pool1Additions = pool1Additions;
		this.solver.setPool1Additions(pool1Additions);
	}

	/**
	 * Get pool2 additions
	 * 
	 * @return Value
	 */
	public int getPool2Additions() {
		this.pool2Additions = this.solver.getPool2Additions();
		return this.pool2Additions;
	}

	/**
	 * Set pool2 additions
	 */
	public void setPool2Additions(int pool2Additions) {
		this.pool2Additions = pool2Additions;
		this.solver.setPool2Additions(pool2Additions);
	}

	/** ----------------------------------------------------------------- */
	@Override
	public TStruct getSettings() throws Exception {
		TStruct resumeData;

		resumeData = new TStruct();

		// archive size
		resumeData.put("archiveSize", new TNumeric(this.getArchiveSize()));

		// crossover probability
		resumeData.put("crossoverProbability",
				new TNumeric(this.getCrossoverProbability()));

		// crossover distribution index
		resumeData.put("crossoverDistributionIndex",
				new TNumeric(this.getCrossoverDistributionIndex()));

		// mutation probybility
		resumeData.put("mutationProbability",
				new TNumeric(this.getMutationProbability()));

		// mutation distribution index
		resumeData.put("mutationDistributionIndex",
				new TNumeric(this.getMutationDistributionIndex()));

		// DE crossover rate
		resumeData.put("deCR", new TNumeric(this.getDeCR()));

		// DE mutation factor
		resumeData.put("deF", new TNumeric(this.getDeF()));

		// weak sharing stages count
		resumeData.put("weakSharingStagesCount",
				new TNumeric(this.getWeakSharingStagesCount()));

		// best performing pool ID
		resumeData.put("bestPerformingPoolID",
				new TNumeric(this.getBestPerformingPoolID()));

		// pool1 additions
		resumeData.put("pool1Additions",
				new TNumeric(this.getPool1Additions()));

		// pool2 additions
		resumeData.put("pool2Additions",
				new TNumeric(this.getPool2Additions()));

		// domination history
		this.dominationHistory = this.solver.getDominationHistory();
		if (this.dominationHistory.size() > 0) {
			TArray array = new TArray(1, this.dominationHistory.size());
			resumeData.put("dominationHistory", array);
			for (int i = 0; i < this.dominationHistory.size(); i++) {
				array.set(0, i, new TNumeric(this.dominationHistory.get(i)));
			}
		}

		return resumeData;
	}

	@Override
	public void setSettings(TStruct resumeData) throws Exception {
		TNumeric num;

		// archive size
		num = (TNumeric) resumeData.get("archiveSize");
		if (num != null) {
			this.setArchiveSize((int) num.getReal(0, 0));
		}

		// crossover probability
		num = (TNumeric) resumeData.get("crossoverProbability");
		if (num != null) {
			this.setCrossoverProbability(num.getReal(0, 0));
		}

		// crossover distribution index
		num = (TNumeric) resumeData.get("crossoverDistributionIndex");
		if (num != null) {
			this.setCrossoverDistributionIndex(num.getReal(0, 0));
		}

		// mutation probybility
		num = (TNumeric) resumeData.get("mutationProbability");
		if (num != null) {
			this.setMutationProbability(num.getReal(0, 0));
		}

		// mutation distribution index
		num = (TNumeric) resumeData.get("mutationDistributionIndex");
		if (num != null) {
			this.setMutationDistributionIndex(num.getReal(0, 0));
		}

		// DE crossover rate
		num = (TNumeric) resumeData.get("deCR");
		if (num != null) {
			this.setDeCR(num.getReal(0, 0));
		}

		// DE mutation factor
		num = (TNumeric) resumeData.get("deF");
		if (num != null) {
			this.setDeF(num.getReal(0, 0));
		}

		// weak sharing stages count
		num = (TNumeric) resumeData.get("weakSharingStagesCount");
		if (num != null) {
			this.setWeakSharingStagesCount((int) num.getReal(0, 0));
		}

		// best performing pool ID
		num = (TNumeric) resumeData.get("bestPerformingPoolID");
		if (num != null) {
			this.setBestPerformingPoolID((int) num.getReal(0, 0));
		}

		// weak sharing stages count
		num = (TNumeric) resumeData.get("pool1Additions");
		if (num != null) {
			this.setPool1Additions((int) num.getReal(0, 0));
		}

		// best performing pool ID
		num = (TNumeric) resumeData.get("pool2Additions");
		if (num != null) {
			this.setPool2Additions((int) num.getReal(0, 0));
		}

		// domination history
		TArray array = (TArray) resumeData.get("dominationHistory");
		if (array != null) {
			for (int i = 0; i < array.size()[1]; i++) {
				this.dominationHistory
						.add((int) ((TNumeric) array.get(0, i)).getReal(0, 0));
			}
			this.solver.setDominationHistory(this.dominationHistory);
		}

	}

	@Override
	public TStruct getState() throws Exception {
		TStruct resumeData;
		SolutionSet sol;

		resumeData = new TStruct();

		// archive
		sol = this.solver.getArchive();
		if (sol.size() != 0) {
			resumeData
					.put("archive",
							new TNumeric(Archive.int2double(this.getProblem()
									.getArchive().getMemory().getIndexOf(
											this.getIndividuals(sol)))));
		}

		// pool1
		sol = this.solver.getPool1();
		if (sol.size() != 0) {
			resumeData
					.put("pool1",
							new TNumeric(Archive.int2double(this.getProblem()
									.getArchive().getMemory().getIndexOf(
											this.getIndividuals(sol)))));
		}

		// pool2
		sol = this.solver.getPool2();
		if (sol.size() != 0) {
			resumeData
					.put("pool2",
							new TNumeric(Archive.int2double(this.getProblem()
									.getArchive().getMemory().getIndexOf(
											this.getIndividuals(sol)))));
		}

		return resumeData;
	}

	@Override
	public void setState(TStruct resumeData) throws Exception {
		TNumeric num;

		// archive
		num = (TNumeric) resumeData.get("archive");
		if (num != null) {
			this.solver.setArchive(this.getSolutionSet(this.getProblem()
					.getArchive().getMemory().getIndividual(Archive
							.double2int(num.getRealMatrix().getRow(0)))));
		}

		// pool1
		num = (TNumeric) resumeData.get("pool1");
		if (num != null) {
			this.solver.setPool1(this.getSolutionSet(this.getProblem()
					.getArchive().getMemory().getIndividual(Archive
							.double2int(num.getRealMatrix().getRow(0)))));
		}

		// pool2
		num = (TNumeric) resumeData.get("pool2");
		if (num != null) {
			this.solver.setPool2(this.getSolutionSet(this.getProblem()
					.getArchive().getMemory().getIndividual(Archive
							.double2int(num.getRealMatrix().getRow(0)))));
		}
	}

	/** ----------------------------------------------------------------- */

	@Override
	public Individual[] getArchive() throws Exception {
		return this.getIndividuals(this.solver.getArchive());
	}

	@Override
	public void setArchive(Individual[] ind) throws Exception {
		int size = ind.length / 2;

		Individual[] pool1Ind = new Individual[size];
		Individual[] pool2Ind = new Individual[ind.length - size];
		System.arraycopy(ind, 0, pool1Ind, 0, pool1Ind.length);
		System.arraycopy(ind, pool1Ind.length, pool2Ind, 0, pool2Ind.length);

		this.solver.setPool1(this.getSolutionSet(pool1Ind));
		this.solver.setPool2(this.getSolutionSet(pool2Ind));

		this.solver.setArchive(this.getSolutionSet(ind));
	}

	@Override
	public Individual[] getPareto() throws Exception {
		return this.getIndividuals(this.solver.getPareto());
	}

}
