/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 254 $
 * $Date: 2021-02-10 09:34:25 +0100 (Wed, 10 Feb 2021) $
 */
package at.lcm.bu21.optim.core.project;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Arrays;
import java.util.Date;
import java.util.Hashtable;
import java.util.LinkedList;
import java.util.Map;
import java.util.UUID;
import java.util.logging.Level;
import java.util.logging.Logger;

import at.lcm.bu21.general.dtypes.TData;
import at.lcm.bu21.general.dtypes.TNumeric;
import at.lcm.bu21.general.utils.CloneException;
import at.lcm.bu21.general.utils.FileUtils;
import at.lcm.bu21.magopt.structure.Parameter;
import at.lcm.bu21.magopt.structure.Project;
import at.lcm.bu21.magopt.structure.event.Event;
import at.lcm.bu21.magopt.structure.event.EventLogEvent;
import at.lcm.bu21.magopt.structure.event.EventLogListener;
import at.lcm.bu21.magopt.structure.jep.JepProxy;
import at.lcm.bu21.magopt.utils.MagOptConfig;
import at.lcm.bu21.optim.algorithm.GenerationBased;
import at.lcm.bu21.optim.algorithm.grid.Grid;
import at.lcm.bu21.optim.algorithm.me_nsga2.ME_Nsga2;
import at.lcm.bu21.optim.core.Individual;
import at.lcm.bu21.optim.core.Optimization;
import at.lcm.bu21.optim.core.Problem;
import at.lcm.bu21.optim.core.ResultDesc;
import at.lcm.bu21.optim.core.SavePoint.SavePointType;
import at.lcm.bu21.optim.core.VariableDesc;
import at.lcm.bu21.optim.core.event.OptimizerEvent;
import at.lcm.bu21.optim.gui.OptimizerGui;
import jobmanager.JobDesc;
import jobmanager.JobStatus;

/**
 * Compute magopt project
 * 
 * @author Werner Koppelsttter
 */
public class ComputeProject {

	/**
	 * Run optimization
	 * 
	 * @author Gnther Weidenholzer
	 * @author Werner Koppelsttter
	 */
	private class Calculate extends Thread {

		/** Start time of generation */
		private long genTime = -1;

		/** Time for next save-point */
		private long saveTime = -1;

		/** Wait after pool is traversed */
		private boolean wait;

		@Override
		public void run() {
			boolean empty;
			Individual ind;
			Project project;

			ComputeProject.this.run = true;
			// refresh
			ComputeProject.this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.STATUS));

			// duplicate dataset
			ComputeProject.this.optim.addLog("Initialize project");
			while (!ComputeProject.this.stop
					&& (ComputeProject.this.proj.getLength() < (ComputeProject.this.problem.getPoolSize()))) {
				try {
					ComputeProject.this.proj.duplicate();
					ComputeProject.this.proj.getEventLog().addEventLogListener(ComputeProject.this.listener);
				} catch (CloneException e) {
					e.printStackTrace();
				} catch (IOException e) {
					e.printStackTrace();
				}
				ComputeProject.this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.POOL));
			}

			// time for next save-point
			this.saveTime = new Date().getTime() + (ComputeProject.this.problem.getAutosaveTime() * 60 * 1000);

			// Initialization of buffer
			try {
				if (!(ComputeProject.this.problem.getSolver() instanceof Grid)) {
					empty = ComputeProject.this.buffer.isEmpty();

					project = ComputeProject.this.problem.getArchive().getProject();
					// init with existing datasets
					for (int k = 0; k < project.getLength(); k++) {
						try {
							project.setIndex(k);
							ind = new Individual(ComputeProject.this.problem.getArchive());
							for (VariableDesc var : ComputeProject.this.problem.getVariables()) {
								ind.setVariable(var.getName(),
										((TNumeric) project.getObject(var.getName()).getData()).getReal(0, 0));
							}
							if (ind.isValidVar(ComputeProject.this.problem)) {
								ComputeProject.this.buffer.add(ind);
							}
						} catch (Exception ex) {
						}
					}

					// Initialization if no valid solutions exists
					if (empty && (ComputeProject.this.problem.getSolver().getArchive().length < 2)) {
						int init;
						if (ComputeProject.this.problem.getSolver() instanceof GenerationBased) {
							init = ((GenerationBased) ComputeProject.this.problem.getSolver()).getPopulationSize();
						} else {
							init = ComputeProject.this.problem.getAutosaveInterval();
						}
						ComputeProject.this.initBuffer(init);
						// refresh
						ComputeProject.this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.STATUS));
					}
				}
			} catch (Exception e1) {
			}

			while (!ComputeProject.this.stop) {
				this.wait = true;

				try {

					// update all datasets
					if (ComputeProject.this.getNumUsed() > 0) {
						for (int k = 0; k < ComputeProject.this.problem.getPoolSize(); k++) {
							// update
							if (!ComputeProject.this.stop) {
								this.update(k);
							}
						}
					}

					// if opt is finished create new generation
					if (!ComputeProject.this.stop && this.finished()) {
						if (ComputeProject.this.problem.getSolver() instanceof GenerationBased) {
							ComputeProject.this.optim.addLog("Init Generation");
							if (ComputeProject.this.nSolved > 0) {
								this.save();
							}
							// create generation
							this.genTime = new Date().getTime();
							((GenerationBased) ComputeProject.this.problem.getSolver()).createPopulationIndividuals();
						}
					}

					// add Datasets
					if (!ComputeProject.this.stop) {
						if (ComputeProject.this.add || !ComputeProject.this.buffer.isEmpty()
								|| (ComputeProject.this.problem.getArchive().getRecomputeMemory()
										.getNextRecomputeIndividual() != null)) {
							this.add();
						}
					}
				} catch (Exception ex) {
					ComputeProject.this.stop = true;

					System.err.println(ex.getMessage());
					ex.printStackTrace();

					Logger.getLogger(OptimizerGui.class.getName()).log(Level.SEVERE, ex.getMessage(), (Throwable) null);

					// XXX DEBUGGING print error
					// ComputeProject.this.optim.addLog(ex.toString());
					// ComputeProject.this.optim.addLog(ex.getMessage());
					// if (ex.getCause() != null) {
					// ComputeProject.this.optim.addLog(ex.getCause()
					// .toString());
					// }
					// ex.printStackTrace();
					// ComputeProject.this.optim.addLog("\n");
				}

				// sleep if loop is to fast (relax communication)
				try {
					if (this.wait) {
						System.out.println("Wait");
						Thread.sleep(5000);
					} else {
						Thread.sleep(100);
					}
				} catch (InterruptedException e) {
				}
			}

			if (!ComputeProject.this.optim.isTestMode()) {
				// save
				try {
					if (ComputeProject.this.nSolved > 0) {
						if (ComputeProject.this.problem.getSolver() instanceof GenerationBased) {
							// assume generation
							((GenerationBased) ComputeProject.this.problem.getSolver()).takePopulationResults();
						}
						this.save();
					}
				} catch (Exception e) {
				}

				// save solved
				if (ComputeProject.this.nSolved > 0) {
					try {
						if (ComputeProject.this.problem.getSolver() instanceof GenerationBased) {
							// take results
							((GenerationBased) ComputeProject.this.problem.getSolver()).takePopulationResults();
						}
						this.saveSavepoint(SavePointType.USER);
						// refresh
						ComputeProject.this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.STATUS));
						ComputeProject.this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.PROBLEM));
					} catch (Exception e) {
					}

				}
			}

			ComputeProject.this.run = false;
			// refresh
			ComputeProject.this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.STATUS));
		}

		/**
		 * Update index in pool
		 * 
		 * @param k Index of pool
		 * @throws Exception
		 */
		private void update(int k) throws Exception {
			String[] parameters;
			boolean regen;

			if ((ComputeProject.this.ind[k] != null) && ComputeProject.this.used[k]) {

				if (ComputeProject.this.optim.isTestMode() && (ComputeProject.this.poolRuntime[k][0] == 0l)) {
					// start time
					ComputeProject.this.poolRuntime[k][0] = System.currentTimeMillis();
				}

				// timeout
				if ((ComputeProject.this.problem.getSolver() instanceof GenerationBased) && (this.genTime != -1)
						&& ((new Date().getTime() - this.genTime) > (60000
								* ((GenerationBased) ComputeProject.this.problem.getSolver())
										.getGenerationTimeout()))) {
					ComputeProject.this.used[k] = false;
					// remove old jobs from JobManager
					Hashtable<String, JobStatus> list = JobDesc.getStatusList(ComputeProject.this.proj.getFile(),
							ComputeProject.this.proj.getDsId(k), null);
					if (list != null) {
						for (String jobId : list.keySet()) {
							JobDesc.remove(jobId);
						}
					}
					// refresh
					ComputeProject.this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.STATUS));
				} else {
					parameters = JobDesc.getParameterList(ComputeProject.this.proj.getFile(),
							ComputeProject.this.proj.getDsId(k));
					if ((parameters == null) || (parameters.length == 0)) {
						regen = true;
					} else {
						regen = false;
						for (String param : parameters) {
							regen |= JobDesc.allFinished(ComputeProject.this.proj.getFile(),
									ComputeProject.this.proj.getDsId(k), param);
						}
					}

					if (regen) {
						this.wait = false;
						// regenerate
						this.regenerate(k);
						// refresh
						ComputeProject.this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.STATUS));
					}
				}

				if (ComputeProject.this.optim.isTestMode()) {
					// start time of first job
					if (ComputeProject.this.poolRuntime[k][1] == 0l) {
						ComputeProject.this.poolRuntime[k][1] = System.currentTimeMillis();
					}
					// end time
					ComputeProject.this.poolRuntime[k][2] = System.currentTimeMillis();
					// save runtime if index finished
					if (!ComputeProject.this.used[k]) {
						ComputeProject.this.runtime[ComputeProject.this.nSolved
								- 1][0] = ComputeProject.this.poolRuntime[k][1] - ComputeProject.this.poolRuntime[k][0];
						ComputeProject.this.runtime[ComputeProject.this.nSolved
								- 1][1] = ComputeProject.this.poolRuntime[k][2] - ComputeProject.this.poolRuntime[k][1];
					}
				}
			}

			// save
			this.save();

			// stop condition
			if (ComputeProject.this.add) {
				// savepoint
				if (ComputeProject.this.problem.getNumRealSavePoints() >= ComputeProject.this.stopSavePoint) {
					ComputeProject.this.finishOptimizer();
					ComputeProject.this.optim.addLog("Finish by savepoint stop-condition");
				}

				// time
				if (System.currentTimeMillis() >= ComputeProject.this.stopTime) {
					ComputeProject.this.finishOptimizer();
					ComputeProject.this.optim.addLog("Finish by time stop-condition");
				}
			}
		}

		/**
		 * Regenerate index of pool
		 * 
		 * @param k
		 * @throws Exception
		 */
		private void regenerate(int k) throws Exception {
			ComputeProject.this.optim.addLog("Regenerating #" + k);
			ComputeProject.this.proj.setIndex(k);

			// refresh
			ComputeProject.this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.STATUS));

			Map<String, Object> args = new Hashtable<String, Object>();
			args.put("RunSim", true);
			args.put("RemoteCompute", true);
			args.put("Optimize", true);
			args.put("Timeout", 60L * ComputeProject.this.problem.getFunctionTimeout());
			args.put("Aims", ComputeProject.this.problem.getArchive().getResults());

			int status = ComputeProject.this.proj.regenerate(args);

			if (ComputeProject.this.proj.getError()) {
				// remove old jobs from JobManager
				Hashtable<String, JobStatus> list = JobDesc.getStatusList(ComputeProject.this.proj.getFile(),
						ComputeProject.this.proj.getDsId(k), null);
				if (list != null) {
					for (String jobId : list.keySet()) {
						JobDesc.remove(jobId);
					}
				}

				// set objectives
				this.setResults(k);
			} else if (status == 0) {
				// set objectives
				this.setResults(k);
			}
		}

		/**
		 * Set results
		 * 
		 * @param k Index of dataset
		 * @throws Exception
		 */
		private void setResults(int k) throws Exception {
			String varname;
			Parameter param;
			double val;
			TData data;
			TNumeric num;
			String conf;
			File script = null;
			JepProxy jep;

			// set objectives
			for (int r = 0; r < ComputeProject.this.problem.getNumResults(); r++) {
				varname = ComputeProject.this.problem.getResults()[r].getName();
				param = ComputeProject.this.proj.getObject(varname);
				if (param == null) {
					throw new Exception("Result '" + varname + "' doesn't exist!");
				}
				val = Double.NaN;
				if (!param.getUncertain()) {
					data = param.getData();
					if (data instanceof TNumeric) {
						num = (TNumeric) data;
						if (!num.isScalar()) {
							param.error(param.getIdentifier(), "Only scalar values are allowed for optimization",
									new Exception().getStackTrace());
						} else if (!num.isReal()) {
							param.error(param.getIdentifier(), "Only real values are allowed for optimization",
									new Exception().getStackTrace());
						} else {
							val = num.getReal(0, 0);
						}
					} else {
						param.error(param.getIdentifier(), "Only numeric values are allowed for optimization",
								new Exception().getStackTrace());
					}
				}
				ComputeProject.this.ind[k].setResult(r, val);
			}

			if (ComputeProject.this.ind[k].isRecompute()) {
				ComputeProject.this.ind[k].setFromMapping(false);
			} else {

				// add to archive
				ComputeProject.this.problem.getArchive().getMemory().addIndividual(ComputeProject.this.ind[k]);
				if (ComputeProject.this.ind[k].isError()) {
					ComputeProject.this.nError++;
				} else if (ComputeProject.this.ind[k].isValid(ComputeProject.this.problem)) {
					// set results
					ComputeProject.this.problem.getSolver().setResolvedIndividual(ComputeProject.this.ind[k]);
					ComputeProject.this.nValid++;
				} else {
					ComputeProject.this.nInvalid++;
				}
				ComputeProject.this.nSolved++;
			}

			// run post processing script
			conf = MagOptConfig.getConfig("OptimizerPostSript");
			if (conf != null) {
				script = new File(conf);
				if (script.exists()) {
					try {
						jep = ComputeProject.this.proj.getJep();
						jep.eval("from symspace.wrapper.symwrapper import SymParameter");
						jep.set("__project__", ComputeProject.this.proj);
						jep.eval("__project__ = SymParameter._convertObject(__project__)");
						jep.set("__problem__", ComputeProject.this.problem);
						jep.eval("del SymParameter");

						jep.eval("exec(open(r'" + script.getAbsolutePath() + "').read())");

						jep.eval("del __project__");
						jep.eval("del __problem__");
					} catch (Exception ex) {
						ComputeProject.this.proj.warning(
								"Error by executing optimizer post-script (" + ex.getMessage() + ")",
								ex.getStackTrace());
					}
				} else {
					ComputeProject.this.proj
							.warning("Optimizer post-script not found (" + script.getAbsolutePath() + ")");
				}
			}

			ComputeProject.this.used[k] = false;
		}

		/**
		 * Check if save-point is reached
		 * 
		 * @throws Exception
		 */
		private void save() throws Exception {
			// check save for autosave/generation
			if (ComputeProject.this.problem.getSolver() instanceof GenerationBased) {
				if (this.finished()) {
					// take generation results
					((GenerationBased) ComputeProject.this.problem.getSolver()).takePopulationResults();

					if (ComputeProject.this.nSolved > 0) {
						// save
						this.saveSavepoint(SavePointType.INTERVAL);

						// reset Individual
						for (int p = 0; p < ComputeProject.this.ind.length; p++) {
							ComputeProject.this.used[p] = false;
							ComputeProject.this.ind[p] = null;
						}

						// remove old jobs from JobManager
						Hashtable<String, JobStatus> list = JobDesc.getStatusList(ComputeProject.this.proj.getFile(),
								null, null);
						if (list != null) {
							for (String jobId : list.keySet()) {
								JobDesc.remove(jobId);
							}
						}
					}

					// refresh
					ComputeProject.this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.STATUS));
				}
			} else {
				if (ComputeProject.this.nSolved >= ComputeProject.this.problem.getAutosaveInterval()) {
					this.saveSavepoint(SavePointType.INTERVAL);

					// refresh
					ComputeProject.this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.STATUS));
				}
			}

			if (!ComputeProject.this.optim.isTestMode()) {
				if (ComputeProject.this.problem.getAutosaveTimeEnabled() && (new Date().getTime() > this.saveTime)) {
					this.saveSavepoint(SavePointType.TIME);

					// time for next save-point
					this.saveTime = new Date().getTime() + (ComputeProject.this.problem.getAutosaveTime() * 60 * 1000);
					// refresh
					ComputeProject.this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.STATUS));
				}
			}

			if (ComputeProject.this.nRecompute > ComputeProject.this.problem.getAutosaveInterval()) {
				ComputeProject.this.problem.getArchive().save();
			}
		}

		/**
		 * Create save-point and write to file
		 * 
		 * @param type Type of save-point
		 * @throws Exception
		 */
		private void saveSavepoint(SavePointType type) throws Exception {
			ComputeProject.this.optim
					.addLog("Save #" + (ComputeProject.this.problem.getNumSavePoints() + 1) + " (" + type.name() + ")");

			if (type != SavePointType.TIME) {
				ComputeProject.this.nSolved = 0;
			}
			ComputeProject.this.nRecompute = 0;

			// Append history and save results
			ComputeProject.this.problem.createSavePoint(type, false, ComputeProject.this.nValid,
					ComputeProject.this.nInvalid, ComputeProject.this.nError);
			ComputeProject.this.problem.getArchive().save();

			ComputeProject.this.nValid = 0;
			ComputeProject.this.nInvalid = 0;
			ComputeProject.this.nError = 0;

			// Refresh Gui
			ComputeProject.this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.PROBLEM));

			// XXX DEBUGGING save project
			// File file = PanelOptimize.this.proj.getMopFile();
			// PanelOptimize.this.proj.saveAs(new File(file
			// .getParentFile(), "Gen_"
			// + ((TArray) PanelOptimize.this.wizardData
			// .get("paretoHistory")).size()[1] + ".mop"));
			// PanelOptimize.this.proj.saveAs(file);

			if (ComputeProject.this.optim.isTestMode()) {
				String filename = "d:\\\\Results\\" + ComputeProject.this.getProject().getName() + "\\"
						+ ComputeProject.this.getProblem().getName();
				this.storeSavepointInformationOnDisk(filename);
				// ComputeProject.this.problem.getSolver().getArchive();
				// ComputeProject.this.runtime;
			}
		}

		private void storeSavepointInformationOnDisk(String directoryName) throws Exception {
			int currentSavePointNr = ComputeProject.this.getProblem().getNumSavePoints();

			String fileName1 = directoryName + "\\" + currentSavePointNr + ".runtimes";
			String fileName2 = directoryName + "\\" + currentSavePointNr + ".archive";

			File f = new File(fileName1);
			File dir = new File(f.getParent());
			if (!dir.exists() && !dir.mkdirs()) {
				throw new IOException("Could not create directory path: " + f.getParent());
			}

			if (!f.exists()) {
				f.createNewFile();
			}
			BufferedWriter bw = new BufferedWriter(new FileWriter(fileName1, true));
			for (int i = 0; i < ComputeProject.this.problem.getAutosaveInterval(); i++) {
				String line = "";
				for (int j = 0; j < ComputeProject.this.runtime[0].length; j++) {
					line += " " + ComputeProject.this.runtime[i][j];
					ComputeProject.this.runtime[i][j] = 0L;
				}
				bw.write(line + "\n");
			}
			bw.close();

			File g = new File(fileName2);
			if (!g.exists()) {
				g.createNewFile();
			}
			bw = new BufferedWriter(new FileWriter(fileName2, true));
			for (Individual ind : ComputeProject.this.problem.getSolver().getArchive()) {
				String line = "";
				for (ResultDesc rDesc : ComputeProject.this.problem.getResults()) {
					if (rDesc.isObjective()) {
						Individual archivedInd = ComputeProject.this.problem.getArchive().getMemory()
								.getIndividual(ind);
						Double value = archivedInd.getResult(rDesc.getName());
						line += " " + String.format("%.6f", value);
					}
				}
				bw.write(line + "\n");
			}
			bw.close();

			ComputeProject.this.memoryArchiveHits = 0;
		}

		/**
		 * Check if all individual of pool are finished and no next individual exists
		 * 
		 * @return Finished pool
		 */
		private boolean finished() {
			boolean finish;

			finish = ComputeProject.this.buffer.isEmpty();
			finish &= ComputeProject.this.problem.getArchive().getRecomputeMemory()
					.getNextRecomputeIndividual() == null;
			finish &= !ComputeProject.this.problem.getSolver().hasNext();
			finish &= ComputeProject.this.getNumUsed() == 0;

			return finish;
		}

		/**
		 * Add new individual to pool if free index exists
		 * 
		 * @throws Exception
		 */
		private void add() throws Exception {
			if (!ComputeProject.this.buffer.isEmpty()
					|| (ComputeProject.this.problem.getArchive().getRecomputeMemory()
							.getNextRecomputeIndividual() != null)
					|| ComputeProject.this.problem.getSolver().hasNext()) {
				int k;
				Individual ind;
				String varname;
				Parameter param;

				k = this.getNextIndex();
				if (k != -1) {
					// get Next Individual
					ind = this.getNextIndividual();
					if (ind != null) {
						this.wait = false;
						// init dataset
						ComputeProject.this.optim.addLog("Init #" + k);
						ComputeProject.this.ind[k] = ind;
						ComputeProject.this.used[k] = true;
						if (ComputeProject.this.optim.isTestMode()) {
							ComputeProject.this.poolRuntime[k][0] = 0l;
							ComputeProject.this.poolRuntime[k][1] = 0l;
							ComputeProject.this.poolRuntime[k][2] = 0l;
						}

						ComputeProject.this.proj.setIndex(k);
						ComputeProject.this.proj.setError(false);

						// set variable values to dataset
						for (int q = 0; q < ComputeProject.this.problem.getNumVariables(); q++) {
							varname = ComputeProject.this.problem.getVariables()[q].getName();
							param = ComputeProject.this.proj.getObject(varname);
							if (param == null) {
								throw new Exception("Variable '" + varname + "' doesn't exist!");
							}
							param.setData(new TNumeric(ComputeProject.this.ind[k].getVariable(q)));

						}
						ComputeProject.this.proj.refresh();
						// refresh
						ComputeProject.this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.STATUS));
					}
				}
			}

		}

		/**
		 * Get next free index in pool
		 * 
		 * @return Free index
		 */
		private int getNextIndex() {
			int k, index = -1;

			// find free index
			k = 0;
			while ((index == -1) && (k < ComputeProject.this.ind.length)) {
				if (ComputeProject.this.ind[k] == null) {
					index = k;
				}
				k++;
			}

			// no index found -> clean
			if (index == -1) {
				// clean pool
				k = 0;
				while (k < ComputeProject.this.used.length) {
					if (!ComputeProject.this.used[k]) {
						ComputeProject.this.ind[k] = null;
					}
					k++;
				}

				// find free index
				k = 0;
				while ((index == -1) && (k < ComputeProject.this.ind.length)) {
					if (ComputeProject.this.ind[k] == null) {
						index = k;
					} else {
						k++;
					}
				}
			}

			return index;
		}

		/**
		 * Get next individual to calculate
		 * 
		 * @return Next individual
		 * @throws Exception
		 */
		private Individual getNextIndividual() throws Exception {
			Individual ind = null, res;
			int k = 0;

			// true to find new individual
			while (((ind == null) && (k < 100)) && (!ComputeProject.this.buffer.isEmpty()
					|| (ComputeProject.this.problem.getArchive().getRecomputeMemory()
							.getNextRecomputeIndividual() != null)
					|| ComputeProject.this.problem.getSolver().hasNext())) {
				k++;
				if (!ComputeProject.this.buffer.isEmpty()) {
					ind = ComputeProject.this.buffer.remove();
				}
				if (ind == null) {
					ind = ComputeProject.this.problem.getArchive().getRecomputeMemory().getNextRecomputeIndividual();
					if (ind != null) {
						ind.setRecompute(true);
					}
				}
				if (ind == null) {
					ind = ComputeProject.this.problem.getSolver().getNextIndividual();
				}
				// check if ind is in archive
				if ((!ComputeProject.this.optim.isTestMode())
						|| (ComputeProject.this.problem.getSolver() instanceof GenerationBased)) {
					res = ComputeProject.this.problem.getArchive().getMemory().getIndividual(ind);
					if (res != null) {
						if (!res.isError() || !ComputeProject.this.problem.getReattemptError()) {
							if (ind.isRecompute()) {
								// transfer results
								for (int n = 0; n < ComputeProject.this.problem.getNumResults(); n++) {
									ind.setResult(n, res.getResult(n));
								}
								ind.setRecompute(false);
								ind.setFromMapping(false);
							} else {
								ComputeProject.this.memoryArchiveHits++;
								System.out.println(
										"There are: " + ComputeProject.this.memoryArchiveHits + " memory hits!");
								// set to optimizer if valid
								if (res.isValid(ComputeProject.this.problem)) {
									ComputeProject.this.problem.getSolver().setResolvedIndividual(res);
								}
							}
							ind = null;
						}
					}
				} else { // end isTestMode if
					System.out.println("Arhive memory hits check disabled in test mode.");
				}
			}

			if ((ind == null) && ComputeProject.this.problem.getSolver().hasNext()) {
				this.wait = false;
			}

			return ind;
		}
	}

	/** Stop optimization */
	private boolean stop = true;

	/** Add new to optimization */
	private boolean add = false;

	/** Thread is running */
	private boolean run = false;

	/** Recompute mapping based (Sensitivity) */
	private boolean recompute = true;

	/** Optimization */
	private Optimization optim;

	/** Problem of optimizer */
	private Problem problem;

	/** Project for optimizer */
	private Project proj;

	/** Number of solved individuals of save-point */
	private int nSolved;

	/** Number of valid individuals of save-point */
	private int nValid;

	/** Number of invalid individuals of save-point */
	private int nInvalid;

	/** Number of individuals with errors of save-point */
	private int nError;

	/** Number of recomputed individuals */
	private int nRecompute;

	/** Array of individuals */
	private Individual[] ind;

	/** Array which individual of pool is used */
	private boolean[] used;

	/** Buffer of individuals */
	private LinkedList<Individual> buffer = new LinkedList<Individual>();

	/** Calculation thread */
	private Calculate calcThread = null;

	/** Current runtime of each job for test-mode */
	private long[][] poolRuntime = null;

	/** Runtime of each job for test-mode */
	private long[][] runtime = null;

	/** Event-log listener to collect event */
	private EventLogListener listener;

	/** Stop-condition of savepoints */
	private int stopSavePoint = Integer.MAX_VALUE;

	/** Stop-condition of time */
	private long stopTime = Long.MAX_VALUE;

	/**
	 * The number of memory archive hits during the last #nrIndividuals, where
	 * #nrIndividuals is the size of the save interval
	 */
	private int memoryArchiveHits;

	/**
	 * Create settings for magopt-project
	 * 
	 * @param optim Optimization
	 * @throws Exception
	 */
	public ComputeProject(Optimization optim) throws Exception {
		this.optim = optim;
		this.problem = optim.getProblem();

		// ////////////////////////////////////////////////////////////
		// obtain data
		this.problem.getArchive().getProject().setIndex(this.problem.getArchive().getBasedIndex());

		File projFile;
		do {
			String str = UUID.randomUUID().toString() + ".mop";
			projFile = new File(this.problem.getArchive().getMop().getParentFile(), str);

		} while (projFile.exists());

		this.proj = new Project("Optimization", "", projFile);
		this.problem.getArchive().getProject().cloneParam(this.proj, this.problem.getArchive().getBasedIndex(), false);
		this.proj.refresh(true);
		this.listener = new EventLogListener() {
			@Override
			public void eventLogChanged(EventLogEvent event) {
				switch (event.getType()) {
				case EVENT_ADDED:
					for (Event ev : event.getEvents()) {
						ComputeProject.this.optim.getEventLog().add(ev);
					}
					break;
				default:
					break;
				}
			}
		};
		this.proj.getEventLog().addEventLogListener(this.listener);

		this.used = new boolean[this.problem.getPoolSize()];
		this.ind = new Individual[this.problem.getPoolSize()];
		if (this.optim.isTestMode()) {
			// runtimes
			this.poolRuntime = new long[this.problem.getPoolSize()][3];
			this.runtime = new long[(this.problem.getAutosaveInterval())
					+ this.problem.getArchive().getProject().getLength()][2];
		}
	}

	/**
	 * Initial buffer
	 * 
	 * @param num Number of individual for initialize
	 */
	public void initBuffer(int num) {
		this.optim.addLog("Create init individuals to buffer");
		if (!(ComputeProject.this.problem.getSolver() instanceof ME_Nsga2)) {
			this.buffer.addAll(this.problem.createInitDistribution(num));
		} else {
			this.buffer.addAll(this.problem.createBinaryInitDistribution(num));
		}
		this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.STATUS));
	}

	/**
	 * Clear individuals of buffer
	 */
	public void clearBuffer() {
		this.optim.addLog("Clear individuals from buffer");
		this.buffer.clear();
		this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.STATUS));
	}

	/**
	 * Run optimizer
	 */
	public void runOptimizer() {
		this.add = true;
		this.stop = false;
		if ((this.calcThread == null) || !this.calcThread.isAlive()) {
			this.calcThread = new Calculate();
			this.calcThread.start();
		}
		this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.STATUS));
	}

	/**
	 * Run optimizer
	 */
	public void finishOptimizer() {
		this.add = false;
		this.stop = false;
		if ((this.calcThread == null) || !this.calcThread.isAlive()) {
			this.calcThread = new Calculate();
			this.calcThread.start();
		}
		this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.STATUS));
	}

	/**
	 * Stop optimizer
	 */
	public void stopOptimizer() {
		this.stop = true;
		this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.STATUS));
	}

	/**
	 * Stop optimization and shutdown optimizer gui
	 */
	public void shutdown() {
		Hashtable<String, JobStatus> list;

		this.stop = true;

		while ((this.calcThread != null) && this.calcThread.isAlive()) {
			try {
				Thread.sleep(100);
			} catch (InterruptedException e) {
			}
		}

		// delete jobs from JobManager
		if (JobDesc.isRunning()) {
			list = JobDesc.getStatusList(this.proj.getFile(), null, null);

			if (list != null) {
				for (String jobId : list.keySet()) {
					JobDesc.remove(jobId);
				}
			}
		}

		// delete Files-directory
		try {
			FileUtils.delete(this.proj.getFilesDirectory().getParent());
		} catch (Exception ex) {
		}
	}

	/**
	 * Get used problem
	 * 
	 * @return Problem
	 */
	public Problem getProblem() {
		return this.problem;
	}

	/**
	 * Get used project for optimize
	 * 
	 * @return Optimize project
	 */
	public Project getProject() {
		return this.proj;
	}

	/**
	 * Get number of solved individuals since last save-point
	 * 
	 * @return Number of solved
	 */
	public int getNumberSolved() {
		return this.nSolved;
	}

	/**
	 * Get number of valid individuals since last save-point
	 * 
	 * @return Number of valid
	 */
	public int getNumberValid() {
		return this.nValid;
	}

	/**
	 * Get number of invalid individuals since last save-point
	 * 
	 * @return Number of invalid
	 */
	public int getNumberInvalid() {
		return this.nInvalid;
	}

	/**
	 * Get number of error individuals since last save-point
	 * 
	 * @return Number of error
	 */
	public int getNumberError() {
		return this.nError;
	}

	/**
	 * Get number of individuals of init-buffer
	 * 
	 * @return Size of buffer
	 */
	public int getBufferSize() {
		return this.buffer.size();
	}

	/**
	 * Get number of used of pool
	 * 
	 * @return Number of used
	 */
	public int getNumUsed() {
		int num = 0;

		for (boolean b : this.used) {
			if (b) {
				num++;
			}
		}

		return num;
	}

	/**
	 * Get array of used-state of pool
	 * 
	 * @return Used array
	 */
	public boolean[] getUsed() {
		return this.used;
	}

	/**
	 * Get array of individuals in pool
	 * 
	 * @return Individual array
	 */
	public Individual[] getInd() {
		return this.ind;
	}

	/**
	 * Get status of run (status)
	 * 
	 * @return State of flag
	 */
	public boolean isRun() {
		return this.run;
	}

	/**
	 * Get status of add (control)
	 * 
	 * @return State of flag
	 */
	public boolean isAdd() {
		return this.add;
	}

	/**
	 * Get status of stop (control)
	 * 
	 * @return State of flag
	 */
	public boolean isStop() {
		return this.stop;
	}

	/**
	 * Set status of recompute mapping based
	 * 
	 * @param recomp Recompute
	 */
	public void setRecompute(boolean recomp) {
		this.recompute = recomp;
	}

	/**
	 * Get status of recompute
	 * 
	 * @return Recompute
	 */
	public boolean isRecompute() {
		return this.recompute;
	}

	/**
	 * Add list of individual to init-buffer
	 * 
	 * @param inds Individuals
	 */
	public void add2Buffer(Individual[] inds) {
		this.buffer.addAll(Arrays.asList(inds));
		this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.STATUS));
	}

	/**
	 * Get number of savepoint for stop-condition
	 * 
	 * @return Number of savepoint to stop
	 */
	public int getStopConditionSavepoint() {
		return this.stopSavePoint;
	}

	/**
	 * Set number of savepoint for stop-condition
	 * 
	 * 
	 * @param sp Number of savepoint to stop
	 */
	public void setStopConditionSavepoint(int sp) {
		this.stopSavePoint = sp;
	}

	/**
	 * Set time for stop-condition
	 * 
	 * @return Time to stop
	 */
	public long getStopConditionTime() {
		return this.stopTime;
	}

	/**
	 * Get time for stop-condition
	 * 
	 * @param time Time to stop
	 */
	public void setStopConditionTime(long time) {
		this.stopTime = time;
	}

	/**
	 * Clear stop conditions
	 */
	public void clearStopCondition() {
		this.setStopConditionSavepoint(Integer.MAX_VALUE);
		this.setStopConditionTime(Long.MAX_VALUE);
	}

	/**
	 * Get status of computing
	 * 
	 * @return Status
	 */
	public String getStatus() {
		String status;

		if (this.isStop()) {
			if (this.isRun()) {
				status = "Stopping";
			} else {
				status = "Stopped";
			}
		} else {
			if (this.isAdd()) {
				status = "Running";
			} else {
				status = "Finishing";
			}
		}

		return status;
	}

	/**
	 * Is individual (variables) in progress (in computation or buffer)
	 * 
	 * @param ind Individual to check (variables)
	 * @return Is in progress
	 */
	public boolean isInProgress(Individual ind) {
		int index;

		index = Arrays.asList(this.ind).indexOf(ind);
		return ((index != -1) && this.used[index]) || this.buffer.contains(ind);
	}
}
