/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 247 $
 * $Date: 2020-03-19 10:13:52 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.optim.gui;

import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.File;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.swing.UIManager;
import javax.swing.UnsupportedLookAndFeelException;

import at.lcm.bu21.magopt.structure.Project;
import at.lcm.bu21.magopt.utils.MatlabControl;
import at.lcm.bu21.optim.gui.setting.SettingsDialog;

/**
 * Optimizer gui
 * 
 * @author Gnther Weidenholzer
 */
public class OptimizerGuiMain extends Thread {

	/** Init-project */
	private Project project;
	/** Test-mode */
	private boolean testmode;
	/** Init-path */
	private String path;

	/**
	 * Start optimizer gui
	 * 
	 * @param args
	 *            Optimizer gui arguments
	 * @throws Exception
	 */
	public static void main(String[] args) throws Exception {
		String path;
		if (args.length == 0) {
			path = System.getProperty("user.dir");
		} else {
			path = args[0];
		}
		new OptimizerGuiMain(path);
	}

	/**
	 * Create optimizer gui with initial path
	 * 
	 * @param path
	 *            Initial path
	 * @throws ClassNotFoundException
	 * @throws InstantiationException
	 * @throws IllegalAccessException
	 * @throws UnsupportedLookAndFeelException
	 */
	public OptimizerGuiMain(String path) throws ClassNotFoundException, InstantiationException, IllegalAccessException,
			UnsupportedLookAndFeelException {
		this(path, false);
	}

	/**
	 * Create optimizer gui with initial path
	 * 
	 * @param path
	 *            Initial path
	 * @param testMode
	 *            Is test-mode
	 * @throws ClassNotFoundException
	 * @throws InstantiationException
	 * @throws IllegalAccessException
	 * @throws UnsupportedLookAndFeelException
	 */
	public OptimizerGuiMain(String path, boolean testMode) throws ClassNotFoundException, InstantiationException,
			IllegalAccessException, UnsupportedLookAndFeelException {
		this.path = path;
		this.project = null;
		this.testmode = testMode;
		this.start();
		// this.run(path, null, testMode);
	}

	/**
	 * Create optimizer gui with initial project
	 * 
	 * @param project
	 *            Initial project
	 * @throws ClassNotFoundException
	 * @throws InstantiationException
	 * @throws IllegalAccessException
	 * @throws UnsupportedLookAndFeelException
	 */
	public OptimizerGuiMain(Project project) throws ClassNotFoundException, InstantiationException,
			IllegalAccessException, UnsupportedLookAndFeelException {
		this(project, false);
	}

	/**
	 * Create optimizer gui with initial project
	 * 
	 * @param project
	 *            Initial project
	 * @param testMode
	 *            Is test-mode
	 * @throws ClassNotFoundException
	 * @throws InstantiationException
	 * @throws IllegalAccessException
	 * @throws UnsupportedLookAndFeelException
	 */
	public OptimizerGuiMain(Project project, boolean testMode) throws ClassNotFoundException, InstantiationException,
			IllegalAccessException, UnsupportedLookAndFeelException {
		this.path = project.getMopFile().getParent();
		this.project = project;
		this.testmode = testMode;
		this.run();
		// this.run(project.getMopFile().getParent(), project, testMode);
	}

	@Override
	public void run() {
		try {
			this.run(this.path, this.project, this.testmode);
		} catch (ClassNotFoundException e) {
			e.printStackTrace();
		} catch (InstantiationException e) {
			e.printStackTrace();
		} catch (IllegalAccessException e) {
			e.printStackTrace();
		} catch (UnsupportedLookAndFeelException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Run wizard and open optimizer-frame
	 * 
	 * @param path
	 *            Initial path
	 * @param project
	 *            Optional initial project
	 * @param testMode
	 *            Is test-mode
	 * @throws ClassNotFoundException
	 * @throws InstantiationException
	 * @throws IllegalAccessException
	 * @throws UnsupportedLookAndFeelException
	 */
	public void run(String path, Project project, boolean testMode) throws ClassNotFoundException,
			InstantiationException, IllegalAccessException, UnsupportedLookAndFeelException {

		// Use native L&F
		UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());

		final SettingsDialog dia;
		dia = new SettingsDialog(null, new File(path));

		dia.addWindowListener(new WindowAdapter() {
			@Override
			public void windowClosed(WindowEvent event) {
				if (dia.getProblem() != null) {
					try {
						new OptimizerGui(dia.getProblem().getOptimization());
					} catch (Exception ex) {
						Logger.getLogger(OptimizerGui.class.getName()).log(Level.SEVERE, null, ex);
					}
				} else {
					if (!MatlabControl.isInsideMatlab()) {
						System.exit(0);
					}
				}
			}
		});
	}
}
