#!/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 9880 $
$LastChangedDate:: 2020-06-17 19:08:35 +0200#$

This file is part of SyMSpace.

Copyright (c) 2020, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""
import os
import webbrowser
import tempfile
import shutil
import plotly.offline as ply


class FigurePLY:
    """Helper class to preview plotly graphs in SyMSpace"""

    @staticmethod
    def __getTempDir() -> str:
        tmp_dir = os.path.join(tempfile.gettempdir(), 'tmp_SyMSpace_Preview')
        return tmp_dir

    @classmethod
    def __makeTempDir(cls) -> str:
        tmp_dir = cls.__getTempDir()
        if not os.path.isdir(tmp_dir):
            os.makedirs(tmp_dir)
        return tmp_dir

    @classmethod
    def removeTempDir(cls):
        # Todo: probably this method should be called at some point to free space on the drive,
        #  not so important if data handover is not file based -> pickled base64 obj
        tmp_dir = cls.__getTempDir()
        if os.path.isdir(tmp_dir):
            shutil.rmtree(tmp_dir)

    @classmethod
    def openHTML(cls, html: str, suffix: str = '.html'):
        """
        Open a html string in a webbrowser

        Parameters
        ----------
        html : str
            html string
        suffix : str
            html file suffix
        """
        with tempfile.NamedTemporaryFile(dir=cls.__makeTempDir(), mode='w+t', suffix=suffix, delete=False) as temp:
            temp.write(html)
            webbrowser.open('file://' + temp.name, new=2)

    def __init__(self, figure, b64_png=None):
        self.__figure = figure
        self.__b64_png = b64_png

    def toBase64PNG(self, dpi: int = None) -> str:
        """
        Convert Figure to a base64 PNG string

        Parameters
        ----------
        dpi : int
            image resolution in dots per inch

        Returns
        -------
        data : str
            base64 encoded PNG
        """
        # Todo: no implementation for plotly version < 4
        if self.__b64_png is None:
            raise NotImplementedError('no implementation for plotly version < 4')
        return self.__b64_png

    def toHTML(self) -> str:
        """
        Convert Figure to a HTML string

        Returns
        -------
        html : str
            HTML string
        """
        html = ply.plot(self.__figure, auto_open=False, output_type='div')
        return html

    def show(self):
        """Shortcut for showHTML"""
        ply.plot(self.__figure, auto_open=True)

    def showHTML(self):
        """
        Show Figure in the webbrowser

        Returns
        -------
        html : str
            HTML string
        """
        html = self.toHTML()
        self.openHTML(html)

    def toSyMPreview(self, title: str) -> tuple:
        """
        Return a PNG preview for SyMSpace, calls the SyMSpace function method "doubleClickPLY" on doubleclick

        Parameters
        ----------
        title : str
            preview tab title

        Returns
        -------
        tuple
        """
        ret = (title, 'image/png', self.toBase64PNG(), 'doubleClickPLY', self.toHTML())
        return ret

    @classmethod
    def doubleClick(cls, html: str):
        """Specific double click function for a preview tab
        """
        cls.openHTML(html)


if __name__ == '__main__':
    import plotly.graph_objs as go
    import numpy as np

    np.random.seed(1)

    N = 100
    x = np.random.rand(N)
    y = np.random.rand(N)
    colors = np.random.rand(N)
    sz = np.random.rand(N) * 30

    fig = go.Figure()
    fig.add_trace(go.Scatter(
        x=x,
        y=y,
        mode="markers",
        marker=go.scatter.Marker(
            size=sz,
            color=colors,
            opacity=0.6,
            colorscale="Viridis"
        )
    ))

