# !/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 10184 $
$LastChangedDate:: 2021-05-05 14:10:10 +0200#$

This file is part of SyMSpace.

Copyright (c) 2020, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""
import os
import sys
import glob
import warnings
import importlib.util
import importlib.machinery
from modulefinder import ModuleFinder
import symspace.osutils.processutils
from symspace.pyutils.distpkgutil import DistPkgUtil


class PkgUtil:
    """Helper class for Python modules and packages"""

    @staticmethod
    def isPkg(path: str, include_source: bool = True, include_dist: bool = True) -> bool:
        """
        Check if the path points to a python package

        Parameters
        ----------
        path: str
            File system path to check
        include_source: bool
            Include source packages
        include_dist: bool
            Include distribution packages supported by DistPkgUtil

        Returns
        -------
        bool
        """

        if os.path.isdir(path) and include_source:
            all_suffixes = importlib.machinery.all_suffixes()
            for suffix in all_suffixes:
                init_file = os.path.join(path, '__init__{}'.format(suffix))
                if os.path.isfile(init_file):
                    return True

        elif os.path.isfile(path) and include_dist:
            if DistPkgUtil.isDistPkg(path):
                return True

        return False

    @staticmethod
    def getSrcPackageVersion(pkg_name: str, pkg_path: str) -> str:
        """
        Get the version of a source code package, in the form "Major.Minor.Revison".
        If possible and intended the revison is automatically retrieved from the version control system, e.g. SVN

        Parameters
        ----------
        pkg_name: str
            Name of the package
        pkg_path: str
            Absolute path to the package directory

        Examples
        --------
        >>> pkg_version = PkgUtil.getSrcPackageVersion('my_pkg', 'x:\\my_pkg')

        Returns
        -------
        version : str
            Package version in the form "Major.Minor.Revison"
        """
        version = None
        sys_path_added = False
        parent_dir = os.path.dirname(pkg_path)
        if parent_dir not in sys.path:
            sys.path.insert(0, parent_dir)
            sys_path_added = True
        pkg = importlib.import_module(pkg_name)
        if hasattr(pkg, '__version__'):
            version = pkg.__version__
        if sys_path_added:
            sys.path.remove(parent_dir)

        if version is not None and '__revision__' in version:
            try:
                svn_revision, error = symspace.osutils.processutils.getSVNRevision(pkg_path)
                if error == '' and svn_revision != '':
                    version = version.replace('__revision__', svn_revision)
                elif error == '' and svn_revision == '':
                    raise Exception('Empty revision number!')
                else:
                    raise Exception(error)
            except Exception as ex:
                raise Exception('Unable to get svn revision: {}'.format(ex))

        return version

    @staticmethod
    def findSrcPackages(directories: list) -> dict:
        """
        Search for source code packages in these directories.
        Only the top-level directories will be searched.

        Parameters
        ----------
        directories : list
            List of directories to search (absolute path)

        Examples
        --------
        >>> found_pkgs = PkgUtil.findSrcPackages(['x:\\my_pkg1', 'x:\\my_pkg1'])

        Returns
        -------
        pkgs : dict
            Collection of package and their file system location
        """
        pkgs = dict()
        for directory in set(directories):
            pkg_pattern = os.path.join(directory, '*', '__init__.py')
            for pkg_init in glob.iglob(pkg_pattern, recursive=False):
                pkg_dir = os.path.dirname(pkg_init)
                pkg_name = os.path.basename(pkg_dir)
                pkgs[pkg_name] = pkg_dir
        return pkgs

    @staticmethod
    def findModules(directories: list, recursive: bool) -> list:
        """
        Search for python modules (*.py) in these directories.

        Parameters
        ----------
        directories: list
            List of directories to search (absolute path)
        recursive: bool
            If recursive is True also sub-folders will be searched.

        Examples
        --------
        >>> found_modules = PkgUtil.findModules(['x:\\project1\\src', 'x:\\project2\\src'])

        Returns
        -------
        unique_modules : list
            List of found modules
        """
        modules = []
        for directory in set(directories):
            if recursive:
                files = glob.iglob(os.path.join(directory, '**', '*.py'), recursive=True)
            else:
                files = glob.iglob(os.path.join(directory, '*.py'), recursive=False)
            modules.extend(files)
        unique_modules = list(set(modules))
        return unique_modules

    @staticmethod
    def getDependencies(modules: list, importable_paths: list = None, find_packages=True, find_modules=True) -> dict:
        """
        Collect import dependencies in these modules.

        Parameters
        ----------
        modules : list
            Python modules to analyze
        importable_paths : list
            List of paths (directories)  which can be searched for dependencies,
            if not specified (None), sys.path is used.
        find_packages : bool
            Add packages to dependencies, True (Default)
        find_modules : bool
            Add modules to dependencies, True (Default)

        Examples
        --------
        >>> depend = PkgUtil.getDependencies(['x:\\project1\\src\\module1.py', 'x:\\project1\\src\\module2.py'])

        Returns
        -------
        pkgs: dict
            Collection of package and their file system location
        """
        if importable_paths is None:
            importable_paths = sys.path

        dependencies = dict()
        unique_importable_paths = list(set(importable_paths))
        dist_pkgs = DistPkgUtil.findLatestPackages(unique_importable_paths)
        for py_module in set(modules):
            finder = ModuleFinder(unique_importable_paths)
            try:
                finder.run_script(py_module)

                for module_path, mod in finder.modules.items():
                    if isinstance(module_path, str) and '.' not in module_path:
                        '''only search toplevel paths'''
                        if mod.__path__ is not None and find_packages:
                            '''this seams to be a package'''
                            pkg_dir = mod.__path__[0]
                            pkg_name = module_path
                            if pkg_name not in dependencies and os.path.isdir(pkg_dir):
                                dependencies[pkg_name] = pkg_dir
                        if mod.__path__ is None and mod.__file__ is not None and find_modules:
                            '''this seems to be a standalone module'''
                            module_file = mod.__file__
                            if os.path.isfile(module_file):
                                if module_path == '__main__':
                                    module_name = os.path.splitext(os.path.basename(module_file))[0]
                                else:
                                    module_name = module_path

                                if module_name not in dependencies:
                                    dependencies[module_name] = module_file

                if find_packages:
                    for key, value in finder.badmodules.items():
                        '''check if the "badmodules" (here: packages) can be found in distributions'''
                        if key in dist_pkgs:
                            dependencies[key] = dist_pkgs[key]['file']
            except SyntaxError:
                warnings.warn('SyntaxError while analyzing file: {}'.format(py_module))

        return dependencies
