#!/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.5

$LastChangedRevision$
$LastChangedDate::                           $

This file is part of SyMSpace.

Copyright (c) 2018, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""

import time
import itertools as it
import numpy as np
from symspace.wrapper.symwrapper import *


try:
    from magopt.utils import MopFile
except Exception as e:
    print(e)


class SymTolerance:
    """
    Tolerance analysis - only for temporary use

    Examples:


    from symspace.symtolerance import *

    tol = [('Geometry.delta', [2.4, 2.5, 2.6]),
           ('Geometry.Rotor.bm_tol', [0.0, 0.08, 0.1]),
           ('Geometry.Rotor.hm_tol', [0.0, 0.1, 0.2]),
           ('Material_Magnet.Br', [1.28, 1.33]),
           ('Material_Magnet.Theta_m', [20, 100])]

    t = SymTolerance(CurProject)
    t.setTolerance(tol)
    lst = t.expandTolerance()
    t.appendProject(lst, 30)
    """

    def __init__(self, sym_prj):
        self.sym_prj = sym_prj
        self.tol = []
        self.ident = []
        self.result = {}

    def __str__(self):
        res = ''
        for t in tol:
            res = res + str(t[0]) + ': '+ str(t[1]) + ' \n '
        return res

    def expandTolerance(self):
        """
        Expands tolerance to list

        Returns
        -------
        lst : list
            expanded list
        """
        param, values = list(zip(*self.tol))
        for par in param:
            if self.sym_prj.getObject(par) is not None:
                self.ident.append(self.sym_prj.getObject(par).getIdentifier())
            else:
                raise ValueError('Parameter "' + par + '" not found!')
        return list(it.product(*values))

    def getData(self, param):
        """
        Collect data from all tolerance indices

        Parameters
        ----------
        param : str
            Parameter of project

        Returns
        -------
        value : list
        """
        if param not in self.result:
            value = []
            for fh in self.getTolFiles():
                prj = SymProject.load(fh)
                value.extend(SymProject.getDataOverIndices(prj.getObject(param))[1])
            self.result[param] = value
        else:
            value = self.result[param]
        return value

    def getTolFiles(self):
        """
        Returns list of tolerance files

        Returns
        -------
        lst : list
            Tolerance files
        """
        lst = []
        fname = self.sym_prj.getFile().split('.')[0] + '.tol'
        for fh in os.listdir(self.sym_prj.getDirectory()):
            if os.path.isfile(os.path.join(self.sym_prj.getDirectory(), fh)) and fh.startswith(fname) and fh.endswith('.mop'):
                lst.append(os.path.join(self.sym_prj.getDirectory(), fh))
        return lst

    def setTolerance(self, tol):
        """
        Sets tolerance
        Parameters
        ----------
        tol : list
            Configuration : [('symparameter', [0,1]), ...]

        """
        self.tol = tol

    def appendProject(self, tol_lst, max_idx = None):
        """
        Writes SyMSpace project with name .tolXX.mop

        Parameters
        ----------
        tol_lst : list
            Tolerance values

        max_idx : int
            Number of indexes allowed for tolerance project

        """
        # Delete tolerance files first
        files_tol = self.getTolFiles()
        if len(files_tol) > 0:
            raise ValueError('Tolerance files already exist')
        # for fh in self.getTolFiles():
        #     MopFile(fh).delete()

        cnt_file = 1
        nlist = self.sym_prj.getFile().split('.')
        nlist.insert(1, 'tol{:02d}'.format(cnt_file))
        fname = '.'.join(nlist)
        prj = SymProject.load(os.path.join(self.sym_prj.getDirectory(), self.sym_prj.getFile()))
        mopfile = os.path.join(self.sym_prj.getDirectory(), fname)
        prj.saveAs(mopfile)

        k = 1
        for tol in tol_lst:
            if max_idx is not None and k > max_idx:
                prj.save()
                cnt_file += 1
                # Create new file
                nlist = self.sym_prj.getFile().split('.')
                nlist.insert(1, 'tol{:02d}'.format(cnt_file))
                fname = '.'.join(nlist)
                prj = SymProject.load(os.path.join(self.sym_prj.getDirectory(), self.sym_prj.getFile()))
                mopfile = os.path.join(self.sym_prj.getDirectory(), fname)
                prj.saveAs(mopfile)
                k = 1

            if prj.getLength() < k:
                prj.duplicate()
            prj.setIndex(k)

            # Set tolerance for index
            cnt = 0
            for par in self.ident:
                prj.getObject(par).setData(tol[cnt])
                cnt += 1

            k += 1
        prj.save()

    def simulateAll(self, timeout=100):
        """
        Performs simulation of all tolerance files

        Parameters
        ----------
        timeout : int
            Timeout in seconds

        """
        for fh in self.getTolFiles():
            print('\nSimulate ' + os.path.basename(fh))
            prj = SymProject.load(fh)
            sim = SymSimulateAll(prj)
            sim.setRunSimulation(True)
            sim.setRemoteCompute(True)
            sim.start()
            finished = False
            cnt = 0
            while not finished:
                print('    Number of indices to be solved: {:2d}'.format(sim.getNumIncomplete()))
                finished = sim.isStop()
                if not finished:
                    time.sleep(2)

                cnt += 1
                if cnt*2 > timeout:
                    raise ValueError('Timeout')
                if sim.getNumError() >= 0:
                    raise ValueError('Error in Simulation ')
            print('    Simulation finished!')
            prj.save()

    def plotHistogram(self, param):
        """
        Plot histogram of parameter over all tolerance indices

        Parameters
        ----------
        param : str
            ID of parameter
        """
        nr_bins = 10

        values = self.getData(param)
        for value in values:
            if value is None:
                raise ValueError('Parameter contains data of None type')
            elif not isinstance(value, numbers.Number):
                raise ValueError('Compare requires numeric data')
        """
        y, bin_edges = np.histogram(values, bins=nr_bins)
        y_scaled = y / len(values)
        bin_center = (bin_edges[0:-1]+bin_edges[1:])/2
        bin_span = bin_edges[1]-bin_edges[0]
        plot = SymPlot2Bar()
        
        plot.addSignal('{}'.format(pyParam.getName()), np.unique(values), y_scaled.tolist())
        """
        pyParam = self.sym_prj.getObject(param)
        plot = SymPlot2Hist()
        plot.addSignal('{}'.format(pyParam.getName()), values)
        # plot.setTitle('Histogramm of {}'.format(pyParam.getName()))
        plot.setXLabel('{} in {}'.format(pyParam.getName(), pyParam.getUnit()))
        plot.setYLabel('density')
        return plot
