# -*- coding: utf-8 -*-
"""
Created on Thu Jan 11 12:39:02 2018

@author: mkitzberger

Testrunner module for collecting and
running symspace testprojects.
"""

import os
import unittest
import ast
import datetime
import time
import numpy as np
from symspace.osutils.pathutils import Paths

try:
    from symspace.wrapper.symwrapper import SymProject, SymSimulateAll, SymFunction, SymContainer
    from symspace.wrapper.symwrapper import SymField
except Exception as e:
    print(e)

try:
    # Start of java imports
    # noinspection PyUnresolvedReferences
    from javax.swing import JPanel
except Exception as e:
    print('Unable to load Java Libraries!\n{}'.format(e))


class FileCollector:
    """
    Helper class for recursively collecting files inside directories.
    """

    @staticmethod
    def discover(dirpath: str, extension: str, startswith: str):
        """
        Discover files in directory.

        Parameters
        ----------
        dirpath : str
            Path of starting directory
        extension : str
            Extension of files to search for
        startswith: str
            Substring searched files should start with

        Returns
        -------
        paths : list
            List of full path names

        """
        if not os.path.isdir(dirpath):
            raise ValueError('dirpath needs to be a directory')

        if not extension.startswith('.'):
            extension = '.' + extension

        paths = []
        for root, dirs, files in os.walk(dirpath):
            for filename in files:
                fullpath = os.path.join(root, filename)
                if os.path.splitext(fullpath)[1] == extension and os.path.basename(fullpath).lower().startswith(
                        startswith):
                    paths.append(fullpath)

        return paths


class SymProjectTestCase(unittest.TestCase):
    """
    A testcase class for |sym|'s testprojects.
    This class allows usage of unittest.TestCase assert methods by inheritance from TestCase class.
    Furthermore it allows usage of numpy.testing assert methods by delegation to numpy.testing.

    If custom assert methods are required they can be implemented as members of the class
    and then will be available in the comment section of |sym|'s test parameters.
    """

    longMessage = True
    run_args = dict(RunSim=False, RemoteCompute=False, Optimize=False)
    test_preview = True
    generate_docfiles = True
    save_finished = False
    fail_incomplete = False

    @staticmethod
    def assert_almost_equal(*args, **kwargs):
        """
        Raises an AssertionError if two items are not equal up to desired precision.
        """
        np.testing.assert_almost_equal(*args, **kwargs)

    @staticmethod
    def assert_approx_equal(*args, **kwargs):
        """
        Raises an AssertionError if two items are not equal up to significant digits.
        """
        np.testing.assert_approx_equal(*args, **kwargs)

    @staticmethod
    def assert_array_almost_equal(*args, **kwargs):
        """
        Raises an AssertionError if two objects are not equal up to desired precision.
        """
        np.testing.assert_array_almost_equal(*args, **kwargs)

    @staticmethod
    def assert_allclose(*args, **kwargs):
        """
        Raises an AssertionError if two objects are not equal up to desired tolerance.
        """
        np.testing.assert_allclose(*args, **kwargs)

    @staticmethod
    def assert_array_almost_equal_nulp(*args, **kwargs):
        """
        Compare two arrays relatively to their spacing.
        """
        np.testing.assert_array_almost_equal_nulp(*args, **kwargs)

    @staticmethod
    def assert_array_max_ulp(*args, **kwargs):
        """
        Check that all items of arrays differ in at most N Units in the Last Place.
        """
        np.testing.assert_array_max_ulp(*args, **kwargs)

    @staticmethod
    def assert_array_equal(*args, **kwargs):
        """
        Raises an AssertionError if two array_like objects are not equal.
        """
        np.testing.assert_array_equal(*args, **kwargs)

    @staticmethod
    def assert_array_less(*args, **kwargs):
        """
        Raises an AssertionError if two array_like objects are not ordered by less than.
        """
        np.testing.assert_array_less(*args, **kwargs)

    @staticmethod
    def assert_equal(*args, **kwargs):
        """
        Raises an AssertionError if two objects are not equal.
        """
        np.testing.assert_equal(*args, **kwargs)

    @staticmethod
    def assert_raises(*args, **kwargs):
        """
        Fail unless an exception of class exception_class is thrown by callable when invoked with arguments
        args and keyword arguments kwargs.
        """
        np.testing.assert_raises(*args, **kwargs)

    @staticmethod
    def assertProjectOK(error, eventlog):
        if error:
            raise Exception('Project index finished with errors\n{}'.format(eventlog))


class SymProjectTestCaseGenerator:
    """
    Static generator class
    """

    cs_home_replacement = '$ComponentSpace'
    symspace_home_replacement = '$MagOpt_Home'
    dot_replacement = '$dot'

    @classmethod
    def getTestID(cls, file_path: str) -> str:
        """
        Return an id of the test based on the filepath

        SyMSpace and ComponentSpace patch will be replaced to make the ID shorter
        also dots in the filepath are replaced with a placeholder, since dots are a source of errors.

        e.g.:
        C:/opt/jenkins/jobs/SyMSpace/workspace/SyMSpace -> $MagOpt_Home

        Parameters
        ----------
        file_path : str

        Returns
        -------
        test_id :str
        """
        test_id = file_path.replace('.mop', '')
        test_id = test_id.replace(Paths.getSyMSpaceRoot(), cls.symspace_home_replacement)
        try:
            test_id = test_id.replace(Paths.getComponentSpaceRoot(), cls.cs_home_replacement)
        except Exception as e:
            '''ComponentSpace not be defined'''
            pass
        test_id = test_id.replace('.', cls.dot_replacement)
        return test_id

    @staticmethod
    def generate(assertionMethodName: str, *args, **kwargs):
        """
        A generator for assertion methods.

        For a documentation of available assertion
        methods of unittest.TestCase class see unittest documentation.
        https://docs.python.org/3/library/unittest.html#unittest.TestCase

        Parameters
        ----------
        assertionMethodName : str
            Name of the assertion method to use eg. 'assertEqual', 'assertLess', etc.
        args : arguments
            Arguments of the corresponding assertion method
        kwargs: (optional) keyword arguments
            Keyword arguments of the assertion method

        Returns
        -------
        test
        """

        def test(self):
            '''retrieve bound method'''
            fun = getattr(self, assertionMethodName)
            '''call the method with given arguments'''
            fun(*args, **kwargs)

        return test

    @staticmethod
    def generateMain(filename: str, index: int):
        """
        A generator an assertion method, which resets and regenerates the current project index and checks for errors.

        For a documentation of available assertion
        methods of unittest.TestCase class see unittest documentation.
        https://docs.python.org/3/library/unittest.html#unittest.TestCase

        Parameters
        ----------
        filename : str
            Name of the file
        index: int
            project index to evaluate

        Returns
        -------
        test
        """

        def checkPreview(param: SymProject):
            for sub_param in param.getParameters('SymFunction'):
                if sub_param.hasPreview():
                    panel = JPanel()
                    sub_param.preview(panel)

        def test(self):
            """

            Parameters
            ----------
            self : SymProjectTestCase

            """
            '''retrieve bound method'''
            fun = getattr(self, 'assertProjectOK')

            project = SymProject.load(filename)
            project.setIndex(index)
            sfilename = os.path.basename(filename)

            '''reset'''
            timestamp = str(datetime.datetime.now().strftime('%H:%M:%S'))
            print('{} Reset project: {}, index: {:d}'.format(timestamp, sfilename, index))
            project.resetAllFunctions()

            '''regenerate'''
            timestamp = str(datetime.datetime.now().strftime('%H:%M:%S'))
            print('{} Regenerate project: {}, index: {:d}'.format(timestamp, sfilename, index))

            status = 1
            if not self.run_args['RunSim']:
                status = project.regenerate(**self.run_args)
            else:
                print('\t\t|-- Regenerating with "RunSim={}" this may take a while'.format(self.run_args['RunSim']))
                while status:
                    status = project.regenerate(**self.run_args)
                    time.sleep(1)

            timestamp = str(datetime.datetime.now().strftime('%H:%M:%S'))
            if status == 0:
                print('{} Global Update completed successfully: {}, index: {:d}'.format(timestamp, sfilename, index))
            else:
                print('{} Warning: Global Update incomplete: {}, index: {:d}'.format(timestamp, sfilename, index))

            error = project.getError()

            if not error:
                timestamp = str(datetime.datetime.now().strftime('%H:%M:%S'))
                if status == 0 and self.generate_docfiles:
                    print('{} Generate documentation files: {}, index: {:d}'.format(timestamp, sfilename, index))
                    project.generateDocumentationFiles()
                elif status != 0 and self.generate_docfiles:
                    print(
                        '{} Skipped generate documentation files: {}, index: {:d}'.format(timestamp, sfilename, index))

            error = project.getError()

            if not error:
                if status == 0 and self.test_preview:
                    print('{} Testing preview: {}, index: {:d}'.format(timestamp, sfilename, index))
                    checkPreview(project)
                elif status != 0 and self.test_preview:
                    print('{} Skipped testing preview: {}, index: {:d}'.format(timestamp, sfilename, index))

            error = project.getError()
            event_log = project.getEventLog()

            if not error and status != 0 and self.fail_incomplete:
                error = True
                print('{} Error: Global Update incomplete: {}, index: {:d}'.format(timestamp, sfilename, index))

            if self.save_finished:
                project.save()
                print('Saved file: {}'.format(filename))

            project.close()

            '''call the method with given arguments'''
            fun(error, event_log)

        return test

    @staticmethod
    def generateFailed(msg: str):
        """
        A generator for assertion methods, which is bound to fail.

        Parameters
        ----------
        msg : str
            Error message

        Returns
        -------
        test
        """

        def test(self):
            '''retrieve bound method'''
            fun = getattr(self, 'fail')
            '''call the method with given arguments'''
            fun(msg)

        return test

    @staticmethod
    def parseFunction(string: str):
        """
        Method for parsing and evaluating function calls provided as strings

        Parameters
        ----------
        string : str
            Full signature string of the function

        Returns
        -------
        methodName : str
            Name of the function as string

        args : tuple
            Arguments of the function

        kwargs : dict
            Keyword arguments of the function

        """

        '''Check syntax'''
        if '(' in string:
            methodName = string.split('(')[0]
            if ')' not in string.split('(')[1]:
                raise ValueError('Missing closing parenthesis')
            else:
                argStr = string[(string.find('(') + 1): string.find(')')].split(',')
        else:
            raise ValueError('Missing parenthesis')

        '''Read args and kwargs of function'''
        args = ()
        kwargs = {}
        for arg in argStr:
            if arg != '':
                if '=' in arg:
                    name = arg.split('=')[0].rstrip().lstrip()
                    valueStr = arg.split('=')[1].rstrip().lstrip()
                    kwargs[name] = ast.literal_eval(valueStr)
                else:
                    valueStr = arg.rstrip().rstrip().lstrip()
                    args = args + (ast.literal_eval(valueStr),)

        return methodName, args, kwargs

    @classmethod
    def failedCaseFromMop(cls, filename: str, msg: str):
        """
        Method for creating a unittest TestCase object, which fails on purpose.

        Parameters
        ----------
        filename : str
            Filename of Symspace test project.
        msg : str
            Error message
        """
        test_id = cls.getTestID(filename)
        '''Create the failed test method'''
        testMethod = SymProjectTestCaseGenerator.generateFailed(msg)

        addedMethodName = 'test: {}, __generate__'.format(test_id)
        '''Bind assertion function to the TestCaseClass'''
        setattr(SymProjectTestCase, addedMethodName, testMethod)

    @classmethod
    def caseFromMop(cls, filename: str):
        """
        Method for creating a unittest TestCase object from a SymSpace test project given in filename.

        Parameters
        ----------
        filename : str
            Filename of Symspace test project.
        """
        '''do not close this project, this may result in an JRE Access Violation'''
        project = SymProject.load(filename)

        '''automatically change absolute classpaths in ComponentSpace'''
        # FIXME: Doesn't work with if one of the ComponentSpace roots is in e.g. "ComponentSpace\trunk"
        # Todo: remove in future, since it may only be required for SymFunction test projects
        #  and at the moment it seems as if they are not used
        # sym_functions = project.getParameters('SymFunction')
        # for i, item in enumerate(sym_functions):
        #     cp_old = item.getClassPath()
        #     if 'ComponentSpace' in cp_old:
        #         tail = cp_old.split('ComponentSpace')[-1]
        #         cp_new = os.path.join(cs_home, tail.lstrip(os.path.sep))
        #         if cp_new != cp_old:
        #             '''replace function classpath relative to componentspace'''
        #             item.setClassPath(cp_new)
        #             print('Changed classpath {} -> {}'.format(cp_old, cp_new))

        '''
        Empty lists which get filled with params to test (assertie)
        and corresponding test parameter (assertieTest)
        '''
        asserties = []
        assertiesTest = []

        test_id = cls.getTestID(filename)

        '''Check if all indices regenerate without error'''
        for index in range(project.getLength()):
            '''Create check if the project regenerates successfully'''
            testMethod = SymProjectTestCaseGenerator.generateMain(filename, index)

            addedMethodName = 'test: {}, index: {}, __main__'.format(test_id, index)
            '''Bind assertion function to the TestCaseClass'''
            setattr(SymProjectTestCase, addedMethodName, testMethod)

        '''Find testcases in project'''
        symParams = project.getParameters('SymField')
        for i, item in enumerate(symParams):
            if item.getName().startswith('_test'):
                '''check if there is an assertie in the same container'''
                childs = item.getParent().getParam()

                '''filter SymFields'''
                childs = [child for child in childs if isinstance(child, SymField)]
                childsNames = [child.getName() for child in childs]
                assertieName = (item.getName().split('_test', 1)[1]).lstrip('_')

                if assertieName not in childsNames:
                    '''Fail if assertie (parameter) does not exist'''
                    raise Exception('No assertion found for test parameter {} in same container'.format(item.getName()))
                else:
                    assertieIndex = childsNames.index(assertieName)
                    asserties.append(childs[assertieIndex])
                    assertiesTest.append(item)

        '''Dynamically add attributes (testMethods) to the test case class'''
        for assertie, assertieTest in zip(asserties, assertiesTest):
            '''Get comment'''
            comment = assertieTest.getComment().rstrip().lstrip()
            if comment == '':
                raise ValueError('{}: Missing assertion in comment section of parameter {} in container {}'.format(
                    filename, assertieTest.getName(), assertieTest.getParent().getName()))
            '''Parse function signature and arguments'''
            methodName, args, kwargs = SymProjectTestCaseGenerator.parseFunction(comment)
            '''Check if it is a valid assertion method'''
            if methodName not in dir(SymProjectTestCase):
                raise ValueError(
                    'Unsupported assertion statement in comment section of parameter {}, Supported assertions '
                    'are listed in documentation of unittest.TestCase and numpy.testing'.format(assertieTest.getName()))

            test_id = cls.getTestID(filename)
            '''Bind assertion method'''
            for index in range(project.getLength()):
                '''Create assertion function based on the parsed comment'''
                testMethod = SymProjectTestCaseGenerator.generate(methodName, assertie.getData(index),
                                                                  assertieTest.getData(index), **kwargs)

                addedMethodName = 'test: {}, index: {}, container: {}, param: {}'.format(test_id,
                                                                                         index,
                                                                                         assertie.getParent().getName(),
                                                                                         assertie.getName())
                '''Bind assertion function to the TestCaseClass'''
                setattr(SymProjectTestCase, addedMethodName, testMethod)


class SymProjectTestsuite:
    """
    A testsuite class for |sym|'s testprojects.
    A suite collects test cases into a single runable instance.
    Class delegates to a unittest.TestSuite instance for running tests.
    """

    def __init__(self, path: str, run_args=None):
        """
        Create test suite. If path is a .mop file with tests the tests are collected from the file.
        If path is a directory then the directory is searched recursively to find .mop files starting
        with 'test'. Afterwards tests are collected from the respective .mop files.

        Parameters
        ----------
        path : str
            path to a file containing tests or a directory containing test .mop files
        run_args : dict
            SyMSpace run arguments default =  {RunSim=False, RemoteCompute=False, Optimize=False}
        """
        if run_args is None or not bool(run_args):
            '''set default if run_args is None or empty'''
            run_args = dict(RunSim=False, RemoteCompute=False, Optimize=False)

        self.suite = unittest.TestSuite()
        self.loader = unittest.TestLoader()
        self.runner = unittest.TextTestRunner(verbosity=3)

        if os.path.isdir(path):
            testfiles = self.discover(path)
        else:
            testfiles = [path]

        for testfile in testfiles:
            try:
                SymProjectTestCaseGenerator.caseFromMop(testfile)
            except Exception as ex:
                '''Create a failed test case, if the generation of testcases fails somehow.
                   This way the failed test is listed in the unittest results (e.g. junit.xml)'''
                SymProjectTestCaseGenerator.failedCaseFromMop(testfile, str(ex))

        testCase = SymProjectTestCase()
        SymProjectTestCase.run_args = run_args
        self.suite.addTests(self.loader.loadTestsFromModule(testCase))

    @staticmethod
    def discover(dirpath: str) -> list:
        """
        Discover testable .mop files in dirpath.

        Parameters
        ----------
        dirpath : str
            Path to directory

        Returns
        -------
        ret : list
            List of full path names

        """
        return FileCollector.discover(dirpath, '.mop', 'test')

    def run(self) -> unittest.TestResult:
        """
        Function for running the created testsuite
        """
        return self.runner.run(self.suite)


if __name__ == "__main__":
    '''This script needs to be run in a SyMSpace console!'''

    # start_path = os.path.join(os.getenv('ComponentSpace'), 'Part', 'tests')
    start_path = os.path.join(Paths.getSyMSpaceRoot(), 'System', 'Python', 'tests', 'test_symspace')

    suite = SymProjectTestsuite(start_path)
    result = suite.run()
