/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9720 $
 * $Date: 2020-03-19 15:59:49 +0100 (Thu, 19 Mar 2020) $
 */
package jobmanager.logging;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.Date;

/**
 * Write simple log-file for jobs-errors
 *
 * @author Werner Koppelsttter
 */
public class Logger {

	/** Size limit of log-file */
	static long fileLimit = 10L * 1024 * 1024;

	/** Date-formater */
	static SimpleDateFormat sdf = new SimpleDateFormat("yyyy.MM.dd E HH:mm:ss X : ");

	String newLine = "\n";

	/** Log-file */
	File log;
	/** Old log-file (rotating) */
	File old;
	/** Out-writer */
	FileWriter out;

	/**
	 * Open logger for path
	 * 
	 * @param file
	 *            Path for log-file
	 * @throws IOException
	 */
	public Logger(String file) throws IOException {
		this(new File(file));
	}

	/**
	 * Open logger for file
	 * 
	 * @param file
	 *            File of log-file
	 * @throws IOException
	 */
	public Logger(File file) throws IOException {
		File tmp;
		BufferedReader in;
		String line;

		this.log = file.getAbsoluteFile();
		this.old = new File(this.log.getAbsoluteFile() + ".1");
		// transfer logs
		if (this.log.exists()) {
			tmp = new File(this.log.getAbsoluteFile() + ".tmp");
			this.log.renameTo(tmp);

			in = new BufferedReader(new FileReader(tmp));
			this.out = new FileWriter(this.log);
			line = in.readLine();
			while (line != null) {
				this.out.write(line + "\n");
				line = in.readLine();
			}
			in.close();

			tmp.delete();
		} else {
			this.out = new FileWriter(this.log);
		}
	}

	/**
	 * Check size of log-file
	 * 
	 * @throws IOException
	 */
	private void checkLog() throws IOException {
		if (this.log.length() > Logger.fileLimit) {
			this.out.close();
			this.switchLog();
			this.out = new FileWriter(this.log);
		}
	}

	/**
	 * Switch to new log-file
	 */
	private void switchLog() {
		if (this.log.exists()) {
			if (this.old.exists()) {
				this.old.delete();
			}
			this.log.renameTo(this.old);
		}

	}

	/**
	 * Write message to log
	 * 
	 * @param msg
	 * @throws IOException
	 */
	public void write(String msg) throws IOException {
		this.checkLog();
		this.out.write(Logger.sdf.format(new Date()) + "\t" + msg + System.lineSeparator());

		this.out.flush();
	}
}
