/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 10113 $
 * $Date: 2021-02-17 10:54:57 +0100 (Wed, 17 Feb 2021) $
 */
package jobmanager.server;

import java.io.File;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.Socket;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Hashtable;

import jobmanager.Commands;
import jobmanager.JobStatus;
import jobmanager.JobType;
import jobmanager.JobUniverse;

/**
 * Create a JobListener which read command from given connection and return the
 * result.
 * 
 * @author Werner Koppelsttter
 */
class JobListener extends Thread {

	/** Number of JobListener Threads */
	static int count = 0;

	/** Connection to client */
	Socket socket;

	/** OutputStream of connection */
	ObjectOutputStream out = null;

	/** InputStream of connection */
	ObjectInputStream in = null;

	/**
	 * Create a JobListener for the given connection
	 * 
	 * @param socket Connection to client
	 */
	public JobListener(Socket socket) {
		this.socket = socket;

		JobListener.count++;
		this.start();
	}

	/**
	 * Read the command from connection and run suited method
	 */
	@Override
	public void run() {

		try {
			this.out = new ObjectOutputStream(this.socket.getOutputStream());
			this.in = new ObjectInputStream(this.socket.getInputStream());

			switch ((Commands) this.in.readObject()) {
			case VIEW:
				this.setVisible();
				break;
			case CREATE_JAVA_JOB:
				this.createJavaJob();
				break;
			case CREATE_DOCKER_JOB:
				this.createDockerJob();
				break;
			case CREATE_DEFAULT_JOB:
				this.createDefaultJob();
				break;
			case SET_REQUIREMENT:
				this.setRequirement();
				break;
			case SET_TIMEOUT:
				this.setTimeout();
				break;
			case SUBMIT:
				this.submit();
				break;
			case STOP:
				this.stopJob();
				break;
			case RESTART:
				this.restart();
			case RESTART_TYPE:
				this.restartType();
			case REQUEST_JOBSTATUS:
				this.getStatus();
				break;
			case REQUEST_JOBSTATUS_LIST:
				this.getStatusList();
				break;
			case REQUEST_PARAMETER_LIST:
				this.getParameterList();
				break;
			case GET_PATH:
				this.getPath();
				break;
			case LINK_FILE:
				this.linkFile();
				break;
			case REMOVE_JOB:
				this.removeJob();
				break;
			default:
				break;
			}
		} catch (Exception e1) {
		} finally {
			if (this.out != null) {
				try {
					this.out.close();
				} catch (IOException e) {
				}
			}
			if (this.in != null) {
				try {
					this.in.close();
				} catch (IOException e) {
				}
			}
			try {
				this.socket.close();
			} catch (IOException e) {
			}
		}
		JobListener.count--;
	}

	/**
	 * Create job object on JobManager
	 * 
	 * @throws IOException
	 * @throws ClassNotFoundException
	 */
	private void createJavaJob() throws IOException, ClassNotFoundException {
		String project, dataset, parameter, name;
		String[] args;
		Job job = null;

		project = (String) this.in.readObject();
		dataset = (String) this.in.readObject();
		parameter = (String) this.in.readObject();
		name = (String) this.in.readObject();
		args = (String[]) this.in.readObject();
		job = new Job(project, dataset, parameter, name, JobUniverse.JAVA, args);
		job.setCommand("java");

		this.out.writeObject(job.getIdentifier());
	}

	/**
	 * Create job object on JobManager
	 * 
	 * @throws IOException
	 * @throws ClassNotFoundException
	 */
	private void createDockerJob() throws IOException, ClassNotFoundException {
		String project, dataset, parameter, name, dockerImage, command;
		String[] args;
		Job job = null;

		project = (String) this.in.readObject();
		dataset = (String) this.in.readObject();
		parameter = (String) this.in.readObject();
		name = (String) this.in.readObject();
		dockerImage = (String) this.in.readObject();
		command = (String) this.in.readObject();
		args = (String[]) this.in.readObject();
		job = new Job(project, dataset, parameter, name, JobUniverse.DOCKER, args);
		job.setDockerImage(dockerImage);
		job.setCommand(command);

		this.out.writeObject(job.getIdentifier());
	}

	/**
	 * Create job object on JobManager
	 * 
	 * @throws IOException
	 * @throws ClassNotFoundException
	 */
	private void createDefaultJob() throws IOException, ClassNotFoundException {
		String project, dataset, parameter, name, command;
		String[] args;
		Job job = null;

		project = (String) this.in.readObject();
		dataset = (String) this.in.readObject();
		parameter = (String) this.in.readObject();
		name = (String) this.in.readObject();
		command = (String) this.in.readObject();
		args = (String[]) this.in.readObject();
		job = new Job(project, dataset, parameter, name, JobUniverse.DEFAULT, args);
		job.setCommand(command);

		this.out.writeObject(job.getIdentifier());
	}

	/**
	 * Set requirement for job
	 * 
	 * @throws IOException
	 * @throws ClassNotFoundException
	 */
	private void setRequirement() throws IOException, ClassNotFoundException {
		String ident, requirement;
		Job job;

		ident = (String) this.in.readObject();

		job = Job.joblist.get(ident);
		if (job != null) {
			requirement = (String) this.in.readObject();
			job.setRequirement(requirement);
		}

		this.out.writeObject(true);
	}

	/**
	 * Set timeout value of job
	 * 
	 * @throws IOException
	 * @throws ClassNotFoundException
	 */
	private void setTimeout() throws IOException, ClassNotFoundException {
		String ident;
		long timeout;
		Job job;

		ident = (String) this.in.readObject();

		job = Job.joblist.get(ident);
		if (job != null) {
			timeout = (Long) this.in.readObject();
			job.setTimeout(timeout);
		}

		this.out.writeObject(true);
	}

	/**
	 * Submit job to execution queue
	 * 
	 * @throws InstantiationException
	 * @throws IllegalAccessException
	 * @throws Exception
	 */
	private void submit() throws InstantiationException, IllegalAccessException, Exception {
		String ident;
		Job job;
		JobType type;

		ident = (String) this.in.readObject();
		type = (JobType) this.in.readObject();

		job = Job.joblist.get(ident);
		if ((job != null) && (job.getAdapter() == null)) {
			job.setType(type);
			job.setStatus(JobStatus.SUBMIT);
		}

		this.out.writeObject(true);
	}

	/**
	 * Remove job from execution-queue
	 * 
	 * @throws IOException
	 * @throws ClassNotFoundException
	 */
	private void stopJob() throws IOException, ClassNotFoundException {
		String ident;
		Job job;

		ident = (String) this.in.readObject();

		job = Job.joblist.get(ident);
		if (job != null) {
			job.stop();
		}

		this.out.writeObject(true);

	}

	/**
	 * Restart job
	 * 
	 * @throws IOException
	 * @throws ClassNotFoundException
	 */
	private void restart() throws IOException, ClassNotFoundException {
		String ident;
		Job job;

		ident = (String) this.in.readObject();

		job = Job.joblist.get(ident);
		if (job != null) {
			job.restart();
		}

		this.out.writeObject(true);

	}

	/**
	 * Restart job
	 * 
	 * @throws IOException
	 * @throws ClassNotFoundException
	 */
	private void restartType() throws IOException, ClassNotFoundException {
		String ident;
		Job job;
		JobType type;

		ident = (String) this.in.readObject();
		type = (JobType) this.in.readObject();

		job = Job.joblist.get(ident);
		if (job != null) {
			job.setType(type);
			job.restart();
		}

		this.out.writeObject(true);

	}

	/**
	 * Get status of job
	 * 
	 * @throws IOException
	 * @throws ClassNotFoundException
	 */
	private void getStatus() throws IOException, ClassNotFoundException {
		String ident;
		JobStatus status = JobStatus.UNDEFINED;
		Job job;

		ident = (String) this.in.readObject();

		job = Job.joblist.get(ident);
		if (job != null) {
			status = job.getStatus();
		}

		this.out.writeObject(status);
	}

	/**
	 * Get status-list of jobs
	 * 
	 * @throws IOException
	 * @throws ClassNotFoundException
	 */
	private void getStatusList() throws IOException, ClassNotFoundException {
		String project, dataset, parameter;
		Job[] jobs;
		Hashtable<String, JobStatus> list;

		list = new Hashtable<String, JobStatus>();

		project = (String) this.in.readObject();
		dataset = (String) this.in.readObject();
		parameter = (String) this.in.readObject();

		jobs = Job.joblist.values().toArray(new Job[0]);
		Arrays.sort(jobs);
		for (Job job : jobs) {
			if ((project == null) || (project.compareTo(job.getProject()) == 0)) {
				if ((dataset == null) || (dataset.compareTo(job.getDataset()) == 0)) {
					if ((parameter == null) || (parameter.compareTo(job.getParameter()) == 0)) {
						list.put(job.getIdentifier(), job.getStatus());
					}
				}
			}
		}

		this.out.writeObject(list);
	}

	/**
	 * Get list of parameter categories
	 * 
	 * @throws IOException
	 * @throws ClassNotFoundException
	 */
	private void getParameterList() throws IOException, ClassNotFoundException {
		String project, dataset;
		Job[] jobs;
		HashSet<String> set;

		set = new HashSet<String>();

		project = (String) this.in.readObject();
		dataset = (String) this.in.readObject();

		jobs = Job.joblist.values().toArray(new Job[0]);
		Arrays.sort(jobs);
		for (Job job : jobs) {
			if ((project == null) || (project.compareTo(job.getProject()) == 0)) {
				if ((dataset == null) || (dataset.compareTo(job.getDataset()) == 0)) {
					set.add(job.getParameter());
				}
			}
		}

		this.out.writeObject(set.toArray(new String[0]));
	}

	/**
	 * Get the path of local directory
	 * 
	 * @throws Exception
	 */
	private void getPath() throws Exception {
		String ident;
		Job job;
		File path = null;

		ident = (String) this.in.readObject();

		job = Job.joblist.get(ident);
		if (job != null) {
			path = job.getJobDirectory();
		}
		this.out.writeObject(path);
	}

	/**
	 * Link file
	 * 
	 * @throws Exception
	 */
	private void linkFile() throws Exception {
		String ident;
		Job job;
		File file;

		ident = (String) this.in.readObject();
		file = (File) this.in.readObject();

		job = Job.joblist.get(ident);
		if (job != null) {
			job.linkFile(file);
		}

		this.out.writeObject(true);
	}

	/**
	 * Remove job from JobManager
	 * 
	 * @throws Exception
	 */
	private void removeJob() throws Exception {
		String ident;
		Job job;

		ident = (String) this.in.readObject();

		job = Job.joblist.get(ident);
		if (job != null) {
			job.remove();
		}

		this.out.writeObject(true);
	}

	/**
	 * Show or hide JobManager Gui
	 * 
	 * @throws IOException
	 */
	private void setVisible() throws IOException {
		boolean visible;

		visible = this.in.readBoolean();
		JobManager.setVisible(visible);
	}

}
